/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.checkpoint;

import beehive.util.Config;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class FailureDetectorServer extends UnicastRemoteObject implements FailureDetector {

    private String hostFile;
    private Map<String, ClusterInfo> clusters;
    private Set<String> expectedHosts;

    // Maximum number of times a cluster can miss a heartbeat before the server declares it "dead"
    private int graces = 3;
    private boolean recoveryInProgress = false;

    private int pollingFrequency = 5000;


    public FailureDetectorServer(String hostFile) throws IOException {
        this.hostFile = hostFile;
        expectedHosts = new HashSet<>();

        BufferedReader reader = new BufferedReader(new FileReader(hostFile));
        String hostName;

        while ((hostName = reader.readLine()) != null) {
            expectedHosts.add(hostName);
        }

        clusters = new HashMap<>();
        Config configParams = Config.getInstance();
        if (configParams != null) {
            graces = configParams.getIntValue("FailureDetection.graces");
            pollingFrequency = configParams.getIntValue("FailureDetection.pollingFrequency");
            int heartbeatFrequency = configParams.getIntValue("FailureDetection.heartbeatFrequency");
            if (heartbeatFrequency >= pollingFrequency) {
                throw new IOException("Polling frequency must be greater than the heartbeat frequency");
            }
        }
    }

    @Override
    public void reportHeartbeat(String serverHost, boolean alive, boolean master) throws RemoteException {
        if (expectedHosts.contains(serverHost)) {
            System.out.println(serverHost + " is alive");
            clusters.put(serverHost, new ClusterInfo());
        } else {
            throw new RemoteException("Unexpected host reported a heartbeat: " + serverHost);
        }
    }

    public void start() {
        System.out.println("Failure Detector running... ");
        long currentTime = System.currentTimeMillis();
        while (true) {
            try {
                Thread.sleep(pollingFrequency);

                boolean allResponsive = true;
                for (String clusterId : clusters.keySet()) {

                    ClusterInfo cInfo = clusters.get(clusterId);
                    long lastCheckTime = cInfo.time;
                    if (currentTime <= lastCheckTime) {
                        cInfo.unresponsiveCount = 0;
                    } else {
                        System.out.println(clusterId + " has died");
                        cInfo.unresponsiveCount++;
                        allResponsive = false;

                        if (cInfo.unresponsiveCount == graces) {
                            if (!recoveryInProgress) {
                                recoveryInProgress = true;
                                recoverClusters();
                            }
                        }
                    }
                }

                if (allResponsive) {
                    // Once every cluster is back up from recoverClusters and reporting, recoverClusters has finished
                    // If you set it to false at the end of the call to recoverClusters, then the system will
                    // recoverClusters n times (n = number of cluster nodes) since the system takes time
                    // booting up.
                    recoveryInProgress = false;
                }
                currentTime = System.currentTimeMillis();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    private void recoverClusters() {
        String commandFilePath = System.getProperty("user.dir");
        try {
            Runtime rt = Runtime.getRuntime();

            System.out.println("FailureDetectionServer calling clearCluster");
            Process clearCluster = rt.exec("sh " + commandFilePath + "/clearCluster " + hostFile);
            clearCluster.waitFor();
            System.out.println("FailureDetectionServer completed calling clearCluster");

            // Allow the cluster nodes to rest for a bit and in case some cluster nodes are lagging behind
            // in ending their processes
            Thread.sleep(3000);

            System.out.println("FailureDetectionServer calling recover-cluster.sh");
            Process recoverCluster = rt.exec("sh " + commandFilePath + "/recover-cluster.sh " + hostFile);
            recoverCluster.waitFor();
            System.out.println("FailureDetectionServer completed calling recover-cluster");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static class ClusterInfo {
        public long time;
        int unresponsiveCount;

        ClusterInfo() {
            time = System.currentTimeMillis();
            unresponsiveCount = 0;
        }
    }

    public static void main(String args[]) throws Exception {
        Registry localRegistry;
        System.setSecurityManager(new RMISecurityManager());

        if (args.length != 2) {
            System.out.println("Usage: java beehive.checkpoint.FailureDetectionServer <configParamFile> <host file>");
            System.exit(1);
        }

        String hostFile = args[1];

        System.out.println("Starting Failure Detector!");
        FailureDetector fds = new FailureDetectorServer(hostFile);

        boolean bound = true;
        try {
            localRegistry = LocateRegistry.getRegistry(1099);
            localRegistry.bind("FailureDetectionServer", fds);
        } catch (RemoteException e) {
            try {
                localRegistry = LocateRegistry.createRegistry(1099);
                localRegistry.bind("FailureDetectionServer", fds);
            } catch (RemoteException e2) {
                System.out.println("Couldn't create Registry");
                bound = false;
            }
        }
        if (bound) {
            ((FailureDetectorServer) fds).start();
        } else {
            System.exit(1);
        }
    }
}
