/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.server;

import beehive.checkpoint.CheckpointPathBuilder;
import beehive.util.*;
import beehive.thrift.*;

import java.io.*;
import java.rmi.NotBoundException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.rmi.registry.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.ReentrantLock;
import java.lang.reflect.Method;
import java.net.*;

import org.apache.thrift.TException;

public class HashtableServerImpl implements HashtableServer, HashtableServer_thrift.Iface {

    private String myHost;
    private String myURL;
    int myId;
    //private Hashtable table;
    private ConcurrentHashMap table;
    private ConcurrentHashMap<String, String> locationTable;
    // Cache and metadata table
    ConcurrentHashMap<String, Object> cacheTable;
    ConcurrentHashMap<String, NodeMetadata> metaDataTable;
    private HashSet<String> lockedObjects;

    //Stat measures for caching
    private StatContainer statCachedCopyNumTimesUsed;
    private StatContainer statCountInvalidateCallSize;
    private StatContainer statInvalidateTime;
    private StatContainer statGetCallTime;
    private StatContainer statPutCallTime;
    private StatContainer statGetNodeForCachingTime;
    private StatContainer statGetNodesForCachingCallSize;
    private StatContainer invalidateCacheCopiesCallSize;
    private StatContainer statInvalidateCacheCopiesTime;

    //Counts for caching instrumentation
    private int statCountNumCacheStarts = 0;
    private int statCountNumCachePuts = 0;
    private int statCountNumCacheInvalidations = 0;

    private boolean useCaching = false;
    private Logger log;
    private boolean recordStats;
    private int cacheInvalidationScheme = 2;

    //a queue set for keeping track of cache usage
    LinkedBlockingQueue<String> cacheUsageQueue;

    int CACHE_TABLE_SIZE = 10000;
    int CACHE_WINDOW = 10000;
    float CACHE_CLEANUP_TARGET = 0.8f;
    private boolean USE_NODEDATA_REFLECTION = true;

    //number of servers in the cluster
    int numServers;

    //handles to other servers
    ConcurrentHashMap<Integer, HashtableServer> servers;

    private static HashMap<Byte, String> nodeClassNameMap = new HashMap<Byte, String>();
    private static HashMap<Byte, Class> nodeClassMap = new HashMap<Byte, Class>();
    private static HashMap<String, Byte> nodeClassNametoIDMap = new HashMap<String, Byte>();

    private boolean globalValidationEnable = true;

    // Validation Lock fields
    private List<ReentrantLock> cTableLockList;
    private int nLocks = 256;
    private ConcurrentHashMap<String, NodeRWValidationLock> nodeRWLockMap = new ConcurrentHashMap<String, NodeRWValidationLock>();

    private boolean checkpoint = false;
    private String checkpointSaveDir;
    private String checkpointRestoreDir;
    private String checkpointBakDir;

    /**
     * server-id unaware constructor
     *
     * @param url
     * @throws RemoteException
     */
    public HashtableServerImpl(String url) throws RemoteException {
        table = new ConcurrentHashMap();
        myURL = url;
        locationTable = new ConcurrentHashMap<String, String>();
        lockedObjects = new HashSet<String>();
        log = new Logger();

        statGetCallTime = new StatContainer();
        statPutCallTime = new StatContainer();
    }

    /**
     * Constructor for the caching case
     *
     * @param url
     * @param serverId
     * @param configParams
     * @throws RemoteException
     */
    public HashtableServerImpl(String url, int serverId, Config configParams, boolean caching, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws RemoteException {
        this(url);
        this.myId = serverId;

        HashtableServerImpl.nodeClassNameMap = nodeClassNameMap;
        HashtableServerImpl.nodeClassMap = nodeClassMap;

        for (Byte key : nodeClassNameMap.keySet()) {
            String value = nodeClassNameMap.get(key);
            nodeClassNametoIDMap.put(value, key);
        }

        if (caching) {
            this.useCaching = true;
            cacheTable = new ConcurrentHashMap<String, Object>();
            metaDataTable = new ConcurrentHashMap<String, NodeMetadata>();
            cacheUsageQueue = new LinkedBlockingQueue<String>();

            statCachedCopyNumTimesUsed = new StatContainer();
            statCountInvalidateCallSize = new StatContainer();
            statInvalidateTime = new StatContainer();
            statGetNodeForCachingTime = new StatContainer();
            statGetNodesForCachingCallSize = new StatContainer();
            invalidateCacheCopiesCallSize = new StatContainer();
            statInvalidateCacheCopiesTime = new StatContainer();

            //start the cache purge thread
            if (configParams.has("Server.CacheTableSize")) {
                CACHE_TABLE_SIZE = configParams.getIntValue("Server.CacheTableSize");
            }
            if (configParams.has("Server.CacheCleanupTarget")) {
                CACHE_CLEANUP_TARGET = configParams.getFloatValue("Server.CacheCleanupTarget");
            }
            if (configParams.has("Server.CacheWindow")) {
                CACHE_WINDOW = configParams.getIntValue("Server.CacheWindow");
            }
            if (configParams.has("Server.CacheInvalidationScheme")) {
                this.cacheInvalidationScheme = configParams.getIntValue("Server.CacheInvalidationScheme");
            }

            CachePurgeThread purgeThread = new CachePurgeThread(this);
            purgeThread.start();
        }

        if (configParams.has("System.checkpoint")) {
            this.checkpoint = configParams.getBoolValue("System.checkpoint");
        }

        if (configParams.has("Server.useNodeDataReflection")) {
            this.USE_NODEDATA_REFLECTION = configParams.getBoolValue("Server.useNodeDataReflection");
        }

        if (configParams.has("System.globalValidationEnable")) {
            this.globalValidationEnable = configParams.getBoolValue("System.globalValidationEnable");
            System.out.println("Global Validation enabled : " + globalValidationEnable);
            if (!globalValidationEnable) {
                cTableLockList = new ArrayList<ReentrantLock>();
                for (int i = 0; i < nLocks; i++) {
                    cTableLockList.add(new ReentrantLock());
                }
            }
        }
        try {
            myHost = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        }
        if (checkpoint) {
            initCheckpointingDir();
        }
    }

    public HashtableServerImpl(String url, int serverId, Config configParams, boolean caching, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap, boolean recover) throws RemoteException {
        this(url, serverId, configParams, caching, nodeClassNameMap, nodeClassMap);
        if (recover) {
            // load checkpoint
            try {
                FileInputStream hashtableServerCPFile;
                if (new File(checkpointRestoreDir).exists()) {
                    hashtableServerCPFile = new FileInputStream(checkpointRestoreDir + "/HashtableServer.ser");
                } else {
                    if (new File(checkpointBakDir).exists()) {
                        hashtableServerCPFile = new FileInputStream(checkpointBakDir + "/HashtableServer.ser");
                    } else {
                        hashtableServerCPFile = new FileInputStream(checkpointSaveDir + "/HashtableServer.ser");
                    }
                }
                ObjectInputStream hashtableServerCPStream = new ObjectInputStream(hashtableServerCPFile);

                table = (ConcurrentHashMap) hashtableServerCPStream.readObject();
                locationTable = (ConcurrentHashMap<String, String>) hashtableServerCPStream.readObject();

                hashtableServerCPStream.close();
                hashtableServerCPFile.close();

            } catch (IOException i) {
                i.printStackTrace();
            } catch (ClassNotFoundException c) {
                System.out.println("class not found, recovery failure");
                c.printStackTrace();
            }

            //System.out.println("Deserialized HashtableServer ...");
        }
    }

    private void initCheckpointingDir() {
        checkpointSaveDir = CheckpointPathBuilder.buildCheckpointSavePath(myHost);
        checkpointRestoreDir = CheckpointPathBuilder.buildCheckpointRecoverPath(myHost);
        checkpointBakDir = CheckpointPathBuilder.buildCheckpointBakPath(myHost);

        new File(checkpointSaveDir).mkdirs();
        new File(checkpointRestoreDir).mkdirs();
        new File(checkpointBakDir).mkdirs();
    }


    /**
     * Configure server to hold rmi handles of other servers
     * Allows for inter-server communication
     *
     * @param hostList
     */
    public void configureServer(Vector<String> hostList) {
        //populate the handles to the other servers
        this.servers = new ConcurrentHashMap<Integer, HashtableServer>();
        //first set the number of hosts
        this.numServers = hostList.size();
        int serverId = 0;
        Boolean reached;
        for (String host : hostList) {
            if (myId != serverId) {//don't store your own rmi handle
                reached = false;
                while (!reached) {
                    try {
                        String url = "//" + host + "/HashtableServer";
                        HashtableServer server = (HashtableServer) Naming.lookup(url);
                        this.servers.put(serverId, server);
                        reached = true;
                    } catch (NotBoundException ignored) {
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
            }
            serverId++;
        }
    }

    /**
     * Turn on stat recording
     */
    public void enableStats() {
        this.recordStats = true;
    }

    /**
     * Test if the hashtable server is up
     *
     * @return Date
     */
    public Date ping() {
        return new Date();
    }

	/*Single node get calls*/

    /**
     * Given a node id, returns a node to the storage system (application)
     * Forks off to getUseCache in the useCaching = true case
     *
     * @param nodeId
     * @param storageSystemThreadId
     * @return
     */
    public ResultObject get(String nodeId, long storageSystemThreadId) {
        ResultObject result;
        long start = System.currentTimeMillis();
        if (useCaching) {
            result = getUseCache(nodeId, storageSystemThreadId);
        } else {
            if (locationTable.containsKey(nodeId)) {
                // node is relocated
                String newLocation = locationTable.get(nodeId);
                result = new ResultObject(nodeId, newLocation);
            } else if (table.containsKey(nodeId)) {
                // I have the node object
                Node node = (Node) table.get(nodeId);
                result = new ResultObject(node, node.nodeId);
            } else {
                // I don't have the node object and I don't know its location.
                result = new ResultObject(nodeId);
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statGetCallTime, end - start);
        return result;
    }

    /**
     * get() call for the caching case. Not an API, called internally through get()
     *
     * @param nodeId
     * @param storageSystemThreadId
     * @return
     */
    private ResultObject getUseCache(String nodeId, long storageSystemThreadId) {
        log.logDebug("***** HashtableServerImpl : get : reading from server " + myId + " the node " + nodeId + " *****");
        Node node;
        if (table.containsKey(nodeId)) {
            // I have the node object
            node = (Node) table.get(nodeId);
            return new ResultObject(node, nodeId, Constants.FROM_MASTER);
        } else if ((node = (Node) cacheTable.get(nodeId)) != null) {
            NodeMetadata metadata = metaDataTable.get(nodeId);
            //recording how many times we use a cached copy
            metadata.numTimesUsed = metadata.numTimesUsed + 1;
            metadata.lastAccessedTS = System.currentTimeMillis();
            return new ResultObject(node, nodeId, Constants.FROM_CACHE);
        } else {
            // Start caching
            NodeMetadata metadata = null;
            if (!metaDataTable.containsKey(nodeId)) {
                synchronized (metaDataTable) {
                    while (!metaDataTable.containsKey(nodeId)) {
                        metaDataTable.put(nodeId, new NodeMetadata());//num times used is set to 0 here.
                        break;
                    }
                }
            }
            metadata = metaDataTable.get(nodeId);
            boolean status = metadata.testAndSet(true, storageSystemThreadId);
            //we initiate caching in either case (some cases might be overriding the previous ones)
            statCountNumCacheStarts++;
            Node n = null;
            if (!status) {
                return new ResultObject(n, nodeId, Constants.INIT_CACHING);
            } else {
                return new ResultObject(n, nodeId, Constants.INIT_CACHING);
            }
        }
    }

	/*Single node put calls*/

    /**
     * put the node into the table, given a node-id
     * Forks off to putUseCache in the useCaching = true case
     * Returns status as FAILED if the object is put for the first time.
     *
     * @param nodeId
     * @param value
     * @return
     * @throws RemoteException
     */
    public ResultObject put(String nodeId, Object value) throws RemoteException {
        ResultObject result;
        long start = System.currentTimeMillis();
        if (useCaching) {
            result = putUseCache(nodeId, value);
        } else {
            if (locationTable.containsKey(nodeId)) {
                // node is relocated
                String newLocation = locationTable.get(nodeId);
                result = new ResultObject(nodeId, newLocation);
            } else {
                if (globalValidationEnable) {
                    // Global Validator is enabled
                    Node node = (Node) table.put(nodeId, value);
                    int status = Constants.SUCCESS;
                    if (node == null) {
                        node = (Node) value;
                        status = Constants.FAILED;
                    }
                    result = new ResultObject(node.nodeId, status);
                } else {
                    ReentrantLock lock = cTableLockList.get(Math.abs(nodeId.hashCode()) % nLocks);
                    lock.lock();

                    if (!nodeRWLockMap.containsKey(nodeId)) {
                        nodeRWLockMap.put(nodeId, new NodeRWValidationLock());
                    }

                    NodeRWValidationLock rwLock = nodeRWLockMap.get(nodeId);
                    int lockStatus = rwLock.getLock(Constants.VALIDATION_LOCK_WRITE_MODE);

                    if (lockStatus == Constants.LOCK_ACQUIRED) {
                        Node oldNode = (Node) table.get(nodeId);

                        if (oldNode == null) {
                            // ADDED ON Feb. 9 2018
                            // Bypass for graph loading because nodes do not exist in the server yet
                            table.put(nodeId, value);
                            result = new ResultObject(nodeId, Constants.SUCCESS);
                        } else {
                            if (oldNode.TS + 1 == ((Node) value).TS) {
                                Node node = (Node) table.put(nodeId, value);
                                result = new ResultObject(node.nodeId, Constants.SUCCESS);
                            } else {
                                // Timestamps don't match
                                int status = Constants.VALIDATION_ABORT;
                                result = new ResultObject(nodeId, status);
                            }
                        }
                        rwLock.releaseLock(Constants.VALIDATION_LOCK_WRITE_MODE);
                    } else {
                        // Write lock was not acquired
                        result = new ResultObject(nodeId, Constants.VALIDATION_WRITE_LOCK_FAILED);
                    }
                    lock.unlock();
                }
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statPutCallTime, end - start);
        return result;
    }

    /**
     * put for the write through cache case
     *
     * @param nodeId
     * @param value
     * @param callerServerId
     * @return
     * @throws RemoteException
     */
    public ResultObject put(String nodeId, Object value, int callerServerId) throws RemoteException {
        ResultObject result;
        long start = System.currentTimeMillis();
        if (useCaching) {
            result = putUseCache(nodeId, value, callerServerId);
        } else {
            if (locationTable.containsKey(nodeId)) {
                // node is relocated
                String newLocation = locationTable.get(nodeId);
                result = new ResultObject(nodeId, newLocation);
            } else {
                Node node = (Node) table.put(nodeId, value);
                int status = Constants.SUCCESS;
                if (node == null) {
                    status = Constants.FAILED;
                    result = new ResultObject(nodeId, status);
                } else {
                    result = new ResultObject(node.nodeId, status);
                }
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statPutCallTime, end - start);
        return result;
    }

    /**
     * put() call for caching case.
     * Similar to getUseCache, this is not an API. Called internally from put()
     *
     * @param nodeId
     * @param value
     * @return
     * @throws RemoteException
     */
    private ResultObject putUseCache(String nodeId, Object value) throws RemoteException {
        log.logDebug("***** HashtableServerImpl : putUseCache : " + nodeId + " in server " + myId + " *****");
        //write this to the local table
        Node node = (Node) value;
        NodeMetadata origMeta = null;
        if (!metaDataTable.containsKey(nodeId)) {
            synchronized (metaDataTable) {
                while (!metaDataTable.containsKey(nodeId)) {
                    metaDataTable.put(nodeId, new NodeMetadata());
                    break;
                }
            }
        }
        origMeta = metaDataTable.get(nodeId);
        //record stat for the number of invalidation calls
        recordStat(statCountInvalidateCallSize, origMeta.cacheCopyNodes.size());

        long start = System.currentTimeMillis();
        synchronized (origMeta) {
            //invalidate the cache copies
            invalidateCacheCopies(nodeId, origMeta.cacheCopyNodes);
            origMeta.setCacheCopyNodes(new ArrayList<Integer>());
            //flushing the cacheCopyList since it has been invalidated elsewhere
            metaDataTable.put(nodeId, origMeta);
            table.put(nodeId, value);
        }
        long end = System.currentTimeMillis();
        recordStat(statInvalidateTime, end - start);
        return new ResultObject(node, node.nodeId);
    }

    /**
     * putUseCache for the write through cache case
     * also takes in a calling server id, to keep track of cache invaldations (write through cache = no invalidation on calling server)
     *
     * @param key
     * @param value
     * @param callingServerId
     * @return
     * @throws RemoteException
     */
    private ResultObject putUseCache(String key, Object value, int callingServerId) throws RemoteException {
        log.logDebug("***** HashtableServerImpl : putUseCache : " + key + " in server " + myId + " *****");
        //write this to the local table
        Node node = (Node) value;
        NodeMetadata origMeta = null;
        if (!metaDataTable.containsKey(key)) {
            synchronized (metaDataTable) {
                while (!metaDataTable.containsKey(key)) {
                    metaDataTable.put(key, new NodeMetadata());
                    break;
                }
            }
        }
        origMeta = metaDataTable.get(key);
        //record stat for the number of invalidation calls
        recordStat(statCountInvalidateCallSize, origMeta.cacheCopyNodes.size());

        long start = System.currentTimeMillis();
        synchronized (origMeta) {
            //invalidate the cache copies
            //check if the callingSiteId is present in the list
            boolean isCallerPresent = origMeta.cacheCopyNodes.contains(new Integer(callingServerId));
            if (isCallerPresent) {
                origMeta.cacheCopyNodes.remove(new Integer(callingServerId));
                invalidateCacheCopies(key, origMeta.cacheCopyNodes);
                origMeta.setCacheCopyNodes(new ArrayList<Integer>());
                origMeta.cacheCopyNodes.add(new Integer(callingServerId));
            } else {
                invalidateCacheCopies(key, origMeta.cacheCopyNodes);
                origMeta.setCacheCopyNodes(new ArrayList<Integer>());
            }
            //flushing the cacheCopyList since it has been invalidated elsewhere
            metaDataTable.put(key, origMeta);
            table.put(key, value);
        }
        long end = System.currentTimeMillis();
        recordStat(statInvalidateTime, end - start);
        ResultObject result = new ResultObject(node, node.nodeId);
        return result;
    }

	/*get/update node data APIs*/

    /**
     * retrieve the node data for a node, given its node ids and the required properties
     *
     * @param n
     * @return
     * @throws RemoteException
     */
    public ResultObject getNodeProperties(NodeData n) throws RemoteException {
        ResultObject result;
        long start = System.currentTimeMillis();
        String nodeId = n.nodeId;
        if (locationTable.containsKey(nodeId)) {
            // node is relocated
            String newLocation = locationTable.get(nodeId);
            result = new ResultObject(nodeId, newLocation, n);
        } else if (table.containsKey(nodeId)) {
            // I have the node object
            Node node = (Node) table.get(nodeId);
            if (node.TS >= n.TS) {
                if (USE_NODEDATA_REFLECTION) {
                    NodeData nd = readNodePropertiesByReflection(nodeId, node, n.properties, node.TS);
                    result = new ResultObject(nd, nd.nodeId);
                } else {
                    NodeData nd = readNodeProperties(node, n.properties);
                    result = new ResultObject(nd, nd.nodeId);
                }
            } else {
                NodeData nd = null;
                result = new ResultObject(nd, node.nodeId);
            }
        } else {
            // I don't have the node object and I don't know its location.
            result = new ResultObject(nodeId);
        }
        long end = System.currentTimeMillis();
        recordStat(statGetCallTime, end - start);
        return result;
    }

    // TODO DOCUMENTATION
    private NodeData readNodePropertiesByReflection(String nodeId, Object n, Map<String, Object> properties, long TS) {
        NodeData retNode = new NodeData(nodeId, TS, nodeClassNametoIDMap.get(n.getClass().getName()));
        //create node properties from the fields
        Class c = n.getClass();
        Method m;
        Object argValue;
        Object v;
        for (String key : properties.keySet()) {//for now not dealing with function params
            try {
                argValue = properties.get(key);

                StringBuilder buff = new StringBuilder();
                buff.append("get");
                buff.append(key.substring(0, 1).toUpperCase());
                buff.append(key.substring(1));

                if (argValue == null) {
                    m = c.getMethod(buff.toString(), null);
                    v = m.invoke(n);
                } else {
                    m = c.getMethod(buff.toString(), argValue.getClass());
                    v = m.invoke(n, argValue);
                }

                retNode.addProperty(key, v);
            } catch (Exception e) {
                e.printStackTrace();
                retNode.addProperty(key, e);
            }
        }

        return retNode;
    }

    // TODO DOCUMENTATION
    private NodeData readNodeProperties(Object n, Map<String, Object> properties) {
        Node node = (Node) n;
        return node.toNodeData(properties, nodeClassNametoIDMap.get(n.getClass().getName()));
    }

    /**
     * update the properties of a node, given the node id and the updated property map
     *
     * @param value
     * @return
     * @throws RemoteException
     */
    public ResultObject updateNodeProperties(NodeData value) throws RemoteException {
        ResultObject result;
        String key = value.nodeId;

        long start = System.currentTimeMillis();
        if (locationTable.containsKey(key)) {
            // node is relocated
            String newLocation = locationTable.get(key);
//            result = new ResultObject(key, newLocation);
            return new ResultObject(key, newLocation);
        } else {
            //update the node properties using reflection
            Node node = (Node) table.get(key);

            int status = Constants.SUCCESS;
            if (node == null) {
                status = Constants.FAILED;
                result = new ResultObject(key, status);
            } else {
                if (globalValidationEnable) {
                    // Global Validator is enabled
                    node.TS = value.TS;
                    HashMap<String, Object> failedUpdates = updateNodePropertiesByReflection(node, value.properties);
                    table.put(key, node);
                    //insert node back to the table
                    if (failedUpdates != null) {
                        status = Constants.FAILED;
                    }
                    result = new ResultObject(node.nodeId, status);
                } else {
                    ReentrantLock lock = cTableLockList.get(Math.abs(key.hashCode()) % nLocks);
                    lock.lock();

                    if (!nodeRWLockMap.containsKey(key)) {
                        nodeRWLockMap.put(key, new NodeRWValidationLock());
                    }

                    NodeRWValidationLock rwLock = nodeRWLockMap.get(key);
                    int lockStatus = rwLock.getLock(Constants.VALIDATION_LOCK_WRITE_MODE);

                    if (lockStatus == Constants.LOCK_ACQUIRED) {
                        node = (Node) table.get(key);
                        if (node.TS + 1 == value.TS) {
                            node.TS = value.TS;
                            HashMap<String, Object> failedUpdates = updateNodePropertiesByReflection(node, value.properties);
                            table.put(key, node);
                            //insert node back to the table
                            if (failedUpdates != null) {
                                status = Constants.FAILED;
                            }
                            result = new ResultObject(node.nodeId, status);
                        } else {
                            // Timestamps don't match
                            status = Constants.VALIDATION_ABORT;
                            result = new ResultObject(key, status);
                        }
                        rwLock.releaseLock(Constants.VALIDATION_LOCK_WRITE_MODE);
                    } else {
                        // Write lock was not acquired
                        result = new ResultObject(key, Constants.VALIDATION_ABORT);
                    }
                    lock.unlock();
                }
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statPutCallTime, end - start);
        return result;
    }

    // TODO DOCUMENTATION
    private HashMap<String, Object> updateNodePropertiesByReflection(Object node, Map<String, Object> propertyMap) {
        Class c = node.getClass();
        Object v;
        Method m;
        HashMap<String, Object> failedUpdates = null;
        for (String key : propertyMap.keySet()) {
            try {
                v = propertyMap.get(key);
                StringBuffer buff = new StringBuffer();
                buff.append("set");
                buff.append(key.substring(0, 1).toUpperCase());
                buff.append(key.substring(1));
                m = c.getMethod(buff.toString(), v.getClass());
                m.invoke(node, v);
            } catch (Exception e) {
                e.printStackTrace();
                if (failedUpdates == null) {
                    failedUpdates = new HashMap<String, Object>();
                }
                failedUpdates.put(key, e);
            }
        }

        return failedUpdates;
    }

    // TODO DOCUMENTATION
    public ResultObject exec(NodeData n) throws RemoteException {
        ResultObject result;
        long start = System.currentTimeMillis();
        String key = n.nodeId;

        if (locationTable.containsKey(key)) {
            // node is relocated
            String newLocation = locationTable.get(key);
            result = new ResultObject(key, newLocation, n);
        } else if (table.containsKey(key)) {
            // I have the node object
            Object node = table.get(key);
            NodeData nd = execOnNode(n.nodeId, node, n.properties);
            result = new ResultObject(nd, nd.nodeId);
        } else {
            // I dont have the node object and I dont know its location.
            result = new ResultObject(key);
        }
        long end = System.currentTimeMillis();
        recordStat(statGetCallTime, end - start);
        return result;
    }

    // TODO DOCUMENTATION
    private NodeData execOnNode(String nodeId, Object n, Map<String, Object> methodMap) throws RemoteException {
        Class c = n.getClass();
        Method m;
        Object argValue;
        Object v = null;
        HashMap<String, Object> retMap = new HashMap<String, Object>();
        for (String key : methodMap.keySet()) {//for now not dealing with function params
            try {
                argValue = methodMap.get(key);
                if (argValue == null) {
                    m = c.getMethod(key, null);
                    if (m.getReturnType() == Void.class) {
                        m.invoke(n);
                        v = null;
                    } else {
                        v = m.invoke(n);
                    }
                } else {
                    m = c.getMethod(key, argValue.getClass());
                    if (m.getReturnType() == Void.class) {
                        m.invoke(n, argValue);
                        v = null;
                    } else {
                        v = m.invoke(n, argValue);
                    }
                }

                retMap.put(key, v);
            } catch (Exception e) {
                e.printStackTrace();
                retMap.put(key, e);
            }
        }

        NodeData nd = new NodeData(nodeId, nodeClassNametoIDMap.get(n.getClass().getName()));
        nd.properties = retMap;

        return nd;
    }

    // TODO DOCUMENTATION
    public Vector<ResultObject> execBulk(Vector<NodeData> keyList) throws RemoteException {
        Vector<ResultObject> resultList = new Vector<ResultObject>();
        for (NodeData key : keyList) {
            resultList.add(exec(key));
        }
        return resultList;
    }

	/*Bulk get/put/update calls*/

    /**
     * getObjects from the hash table. Called only in non-caching case
     *
     * @param keyList
     * @return
     */
    public Vector<ResultObject> getObjects(Vector<String> keyList) {
        if (useCaching)
            System.out.println("HashtableServerImpl : getObjects : ALERT!!! This function should not be called for caching case");
        Vector<ResultObject> resultList = new Vector<ResultObject>();
        for (String key : keyList) {
            ResultObject result = get(key, -1);
            resultList.add(result);
        }
        return resultList;
    }

    /**
     * putObjects back into the hash table
     *
     * @param objList
     * @return
     * @throws RemoteException
     */
    public Vector<ResultObject> putObjects(ConcurrentHashMap<String, Node> objList) throws RemoteException {
        Vector<ResultObject> resultList = new Vector<ResultObject>();

        for (String nodeId : objList.keySet()) {
            Object val = objList.get(nodeId);
            ResultObject result = put(nodeId, val);
            resultList.add(result);
        }

        return resultList;
    }

    /**
     * put objects back into the hash table for the write through cache case
     *
     * @param objList
     * @param callingServerId
     * @return
     * @throws RemoteException
     */
    public Vector<ResultObject> putObjects(ConcurrentHashMap<String, Node> objList, int callingServerId) throws RemoteException {
        Vector<ResultObject> resultList = new Vector<ResultObject>();

        Enumeration<String> keyList = objList.keys();

        while (keyList.hasMoreElements()) {
            String key = keyList.nextElement();
            Object value = objList.get(key);
            ResultObject result = put(key, value, callingServerId);
            resultList.add(result);
        }

        return resultList;
    }

    /**
     * Bulk call to get object properties for a set of nodes
     *
     * @param keyList
     * @return
     * @throws RemoteException
     */
    public Vector<ResultObject> getObjectsProperties(Vector<NodeData> keyList) throws RemoteException {
        Vector<ResultObject> resultList = new Vector<ResultObject>();
        for (NodeData key : keyList) {
            resultList.add(getNodeProperties(key));
        }
        return (resultList);
    }

    /**
     * In the list provided, get the objects that are available in the cache
     *
     * @param nodeIds
     * @return Cached Nodes
     * @throws RemoteException
     */
    public Vector<Node> getCachedObjects(Vector<String> nodeIds) throws RemoteException {
        Vector<Node> retList = new Vector<Node>();
        Node node;
        for (String nodeId : nodeIds) {
            if (cacheTable.containsKey(nodeId)) {
                node = (Node) cacheTable.get(nodeId);
                if (node != null) {
                    retList.add(node);
                }
            }
        }
        return retList;
    }

    /**
     * This is the reflection counterpart of putObjects, used for bulk update of node properties
     * repeatedly calls the updateNodeProperties (can be done in parallel)
     *
     * @param objList - hashtable of nodeId/Node data object
     * @return result obects for reflection
     */
    public Vector<ResultObject> updateNodePropertiesParallel(Vector<NodeData> objList) throws RemoteException {
        Vector<ResultObject> res = new Vector<ResultObject>();
        for (NodeData n : objList) {
            res.add(updateNodeProperties(n));
        }
        return res;
    }

	/*Cache maintenance related functions*/

    /**
     * Master node uses this function to call other servers and ask them to invalidate cache copy
     * 4-modes of invaldation
     * CACHE_INVALIDATION_ASYNCHRONOUS - master invokes n threads to n servers but does so asynchronously. Spoils correctness
     * CACHE_INVALIDATION_SYNCHRONOUS_PARALLEL - invoke n threds to n servers and wait for their response
     * CACHE_INVALIDATION_SYNCHRONOUS_SERIAL - serially call invalidation on n servers
     * CACHE_INVALIDATION_ASYNC_SERIAL - asynchronously invaldate cache copies on n servers, but not in a multi-threaded fashion
     *
     * @param nodeID    - node for which cache copy has to be invalidated
     * @param cacheList
     */
    private void invalidateCacheCopies(String nodeID, List<Integer> cacheList) throws RemoteException {

        if (cacheList.size() <= 0) return;

        recordStat(invalidateCacheCopiesCallSize, cacheList.size());
        long start = System.currentTimeMillis();

        //Using the specific remote cache invalidator
        if (this.cacheInvalidationScheme == Constants.CACHE_INVALIDATION_ASYNCHRONOUS) {
            Barrier barrier = new Barrier(cacheList.size());
            Vector<RemoteCacheInvalidator> handlerThreads = new Vector<RemoteCacheInvalidator>();
            for (int i = 0; i < cacheList.size(); i++) {
                int id = cacheList.get(i);
                HashtableServer serverHandle = this.servers.get(id);
                RemoteCacheInvalidator remoteRequestHandler = new RemoteCacheInvalidator(serverHandle, nodeID, barrier);
                remoteRequestHandler.setPriority(Thread.MAX_PRIORITY);
                handlerThreads.add(remoteRequestHandler);
                remoteRequestHandler.start();
            }
            //no barrier wait in case of async invalidation
        } else if (this.cacheInvalidationScheme == Constants.CACHE_INVALIDATION_SYNCHRONOUS_PARALLEL) {
            Barrier barrier = new Barrier(cacheList.size());
            Vector<RemoteCacheInvalidator> handlerThreads = new Vector<RemoteCacheInvalidator>();
            for (int i = 0; i < cacheList.size(); i++) {
                int id = cacheList.get(i);
                HashtableServer serverHandle = this.servers.get(id);
                RemoteCacheInvalidator remoteRequestHandler = new RemoteCacheInvalidator(serverHandle, nodeID, barrier);
                remoteRequestHandler.setPriority(Thread.MAX_PRIORITY);
                handlerThreads.add(remoteRequestHandler);
                remoteRequestHandler.start();
            }
            try {
                barrier.barrierWait();
            } catch (Exception localException) {
                localException.printStackTrace();
            }
        } else if (this.cacheInvalidationScheme == Constants.CACHE_INVALIDATION_SYNCHRONOUS_SERIAL) {
            for (int i : cacheList) {
                HashtableServer server = servers.get(i);
                server.cacheInvalidate(nodeID);
            }
        } else if (this.cacheInvalidationScheme == Constants.CACHE_INVALIDATION_ASYNC_SERIAL) {
            RemoteSerialCacheInvalidator handler = new RemoteSerialCacheInvalidator(nodeID, cacheList, this.servers);
            handler.start();
        }

        long end = System.currentTimeMillis();
        recordStat(statInvalidateCacheCopiesTime, end - start);
    }

    /**
     * get node for nodeId from the cache table
     * returns null if node not found
     *
     * @param nodeId
     * @return
     * @throws RemoteException
     */
    public Node getNodeFromCacheTable(String nodeId) throws RemoteException {
        if (cacheTable.containsKey(nodeId)) {
            return (Node) this.cacheTable.get(nodeId);
        }
        return null;
    }

    /**
     * Once a node has been fetched for caching, insert into cache table.
     *
     * @param nodeId
     * @param node
     * @param masterCopySite
     * @param storageSystemThreadId
     * @return
     */
    public ResultObject cachePut(String nodeId, Node node, int masterCopySite, long storageSystemThreadId) {
        log.logDebug("***** HashtableServerImpl : cachePut : putting to cache of server " + myId + " the node " + nodeId + " with master site : " + masterCopySite + " *****");
        NodeMetadata metadata = metaDataTable.get(nodeId);
        metadata = metaDataTable.get(nodeId);
        synchronized (metadata) {
            if (metadata.cachingInProgressFlag == storageSystemThreadId) {
                cacheTable.put(nodeId, node);
                metadata.cachingInProgressFlag = -1;
                //set the last accessed ts to now
                metadata.lastAccessedTS = System.currentTimeMillis();
                metaDataTable.put(nodeId, metadata);
                statCountNumCachePuts++;
                Node obj = (Node) cacheTable.get(nodeId);
                //add this to the cache usage set
                cacheUsageQueue.add(nodeId);
                return new ResultObject(obj, obj.nodeId, metadata, Constants.FROM_CACHE);
            } else {
                log.logWarning("***** HashtableServerImpl : cachePut : Concurrent invalidation since caching is in progress!!! *****");
                Node n = null;
                return new ResultObject(n, Integer.toString(-1), metadata, Constants.CONCURRENT_INVALIDATION);// -1 for the node id signifies non-existent node and is not used
                //the thread which did the concurrent invalidation is responsible for notifyAll
            }
        }
    }


    /**
     * Bulk call for cachePut
     *
     * @param nodes                 - Map of master site v/s list of nodes obtained from that master site
     * @param storageSystemThreadId
     * @return
     */
    public Vector<ResultObject> cachePut(Map<Integer, Vector<Node>> nodes, long storageSystemThreadId) {
        log.logDebug("***** HashtableServerImpl : cachePut : putting to cache of server " + myId + " " + nodes.toString() + " nodes *****");
        Vector<ResultObject> res = new Vector<ResultObject>();
        for (int serverId : nodes.keySet()) {
            for (Node node : nodes.get(serverId)) {
                String nodeId = new Integer(node.nodeId).toString();
                res.add(cachePut(nodeId, node, serverId, storageSystemThreadId));
            }
        }
        return res;
    }

    /**
     * Master server invokes this function on a remote server when asking it to invalidate its cache copy
     *
     * @param nodeId - node for which invalidation has to take place
     */
    public void cacheInvalidate(String nodeId) {
        log.logDebug("***** HashtableServerImpl : cacheInvalidate : invalidate cache on server " + myId + " for node " + nodeId + " *****");
        NodeMetadata metadata = metaDataTable.get(nodeId);
        synchronized (metadata) {
            if (metadata.cachingInProgressFlag >= 0) {
                metadata.cachingInProgressFlag = -1;
            }
            metadata.lastAccessedTS = 0;
            cacheTable.remove(nodeId);
            statCountNumCacheInvalidations++;
            recordStat(statCachedCopyNumTimesUsed, metadata.numTimesUsed);
            metadata.numTimesUsed = 0;
        }
    }

    /**
     * cache eviction thread invokes this method when removing a copy from the cache table (cache replacement)
     *
     * @param nodeId
     * @return
     */
    public Node cacheEvict(String nodeId) {
        NodeMetadata metadata = metaDataTable.get(nodeId);
        Node retNode = null;
        synchronized (metadata) {
            if (metadata.cachingInProgressFlag >= 0) {
                metadata.cachingInProgressFlag = -1;
            }
            retNode = (Node) cacheTable.remove(nodeId);
            metadata.lastAccessedTS = 0;
            statCountNumCacheInvalidations++;
            recordStat(statCachedCopyNumTimesUsed, metadata.numTimesUsed);
            metadata.numTimesUsed = 0;
        }
        return retNode;
    }

    /**
     * cache eviction thread invokes this method on the master site informing it of the cache eviction
     *
     * @param cachingServerId
     * @param evictedNodes
     * @throws RemoteException
     */
    public void cacheEvictMasterNotify(String cachingServerId, Hashtable<String, Long> evictedNodes) throws RemoteException {
        NodeMetadata metadata;
        Node nodeFromTable = null;
        Integer cachingServer = new Integer(cachingServerId);
        for (String nodeId : evictedNodes.keySet()) {
            metadata = metaDataTable.get(nodeId);
            synchronized (metadata) {
                //first get the node from the table
                nodeFromTable = (Node) table.get(nodeId);
                //compare if the version evicted was the same as the one which is present
                if (nodeFromTable.TS == ((Long) evictedNodes.get(nodeId)).longValue()) {
                    //remove this from the cache copy list
                    metadata.cacheCopyNodes.remove(cachingServer);
                } else {
                    //let it be, a new value was written before we could inform master
                }
            }
        }
    }

    /**
     * Remote server invokes this method on the master site when requesting a cache copy
     *
     * @param nodeId
     * @param cachingSite
     * @return
     */
    public Node getNodeForCaching(String nodeId, int cachingSite) {
        long start = System.currentTimeMillis();
        log.logDebug("***** HashtableServerImpl : getNodeForCaching : Node " + nodeId + " from server " + myId + " requested by " + cachingSite + " *****");
        Node res = null;
        NodeMetadata metadata = metaDataTable.get(nodeId);
        synchronized (metadata) {
            metaDataTable.get(nodeId).cacheCopyNodes.add(cachingSite);
            res = (Node) table.get(nodeId);
        }
        long end = System.currentTimeMillis();
        recordStat(statGetNodeForCachingTime, end - start);
        return res;
    }

    /**
     * Bulk call for getNodeForCaching
     */
    public Vector<Node> getNodesForCaching(Vector<String> nodeIds, int cachingSite) {
        recordStat(statGetNodesForCachingCallSize, nodeIds.size());
        log.logDebug("***** HashtableServerImpl : getNodesForCaching : server " + cachingSite + " is requesting server " + myId + " for " + nodeIds.toString() + " nodes *****");
        Vector<Node> resNodes = new Vector<Node>();
        for (String nodeId : nodeIds) {
            resNodes.add(getNodeForCaching(nodeId, cachingSite));
        }
        return resNodes;
    }

    /**
     * write through to local cache, so that master does need to invalidate the copy on put
     *
     * @param paramString
     * @param value
     */
    public void writeThroughToCache(String paramString, Object value) {
        NodeMetadata metadata = null;
        if (metaDataTable.containsKey(paramString)) {
            metadata = metaDataTable.get(paramString);
            synchronized (metadata) {
                //first check if it exists in the cacheTable
                if (cacheTable.containsKey(paramString)) {
                    cacheTable.put(paramString, value);
                    //reset the flag to stop any caching in progress
                    metadata.cachingInProgressFlag = -1;
                    metaDataTable.put(paramString, metadata);
                }
            }
        }
    }

    /**
     * bulk call for writeThroughToCache
     *
     * @param map
     */
    public void writeThroughToCacheBulk(ConcurrentHashMap<String, Node> map) {
        for (String nodeId : map.keySet()) {
            writeThroughToCache(nodeId, map.get(nodeId));
        }
    }

    /**
     * Storage system invokes this method on the server when requesting nodes and status for those nodes which are not available
     * Only used in caching case
     *
     * @param nodeIds
     * @return
     */
    public Vector<ResultObject> getLocalNodes(Vector<String> nodeIds, long storageSystemThreadId) {
        Vector<ResultObject> resultNodes = new Vector<ResultObject>();
        for (String nodeId : nodeIds) {
            resultNodes.add(getUseCache(nodeId, storageSystemThreadId));
        }

        return resultNodes;
    }


	/*End of cache related functions*/

	/*Helper functions*/

    /**
     * get a node from the table
     *
     * @param nodeId
     * @return
     * @throws RemoteException
     */
    public Node getNodeFromTable(String nodeId) throws RemoteException {
        if (table.containsKey(nodeId)) {
            return (Node) this.table.get(nodeId);
        }
        return null;
    }

    /**
     * Check if the table contains a given key
     *
     * @param key
     * @return
     */
    public boolean contains(String key) {
        return table.containsKey(key);
    }

    /**
     * Get all nodes present in the main table (does not include cache table)
     *
     * @return
     */
    public Vector<Node> getAllNodes() {
        Enumeration nodeObjects = table.keys();
        Vector<Node> nodeList = new Vector<Node>();
        while (nodeObjects.hasMoreElements()) {
            Object key = nodeObjects.nextElement();
            Node node = (Node) table.get(key);
            nodeList.add(node);
        }

        return nodeList;
    }

    /**
     * Get all node ids present in the main table (does not include cache table)
     *
     * @return
     */
    public Vector<String> getAllKeys() {
        Enumeration nodeObjects = table.keys();
        Vector<String> keyList = new Vector<String>();
        while (nodeObjects.hasMoreElements()) {
            Object key = nodeObjects.nextElement();
            keyList.add((String) key);
        }

        return keyList;
    }

    /**
     * Get the stats, called after program run terminates
     *
     * @return
     */
    public String getStats() {
        StringBuffer buffer = new StringBuffer();
        //		buffer.append("No.of change master calls: " + changeMasterCount + "\n");
        buffer.append("\nHashtableServer stats\n");
        buffer.append("**********\n");
        buffer.append("No. of get calls to the server : " + statGetCallTime.getNumSamples() + " \n");
        buffer.append("Avg. time for a get call : " + statGetCallTime.getMean() + " \n");
        buffer.append("No. of put calls to the server : " + statPutCallTime.getNumSamples() + " \n");
        buffer.append("Avg. time for a put call : " + statPutCallTime.getMean() + " \n");
        if (useCaching) {
            buffer.append("\n");
            buffer.append("Avg num of times a cached copy is used : " + statCachedCopyNumTimesUsed.getMean() + " \n");
            buffer.append("Number of times caching was initiated : " + statCountNumCacheStarts + " \n");
            buffer.append("		Number of times caching was successfully completed : " + statCountNumCachePuts + " \n");
            buffer.append("		Percentage of successful cache initiations : " +
                    ((float) statCountNumCachePuts / statCountNumCacheStarts) * 100 + " \n");
            buffer.append("Number of times cache was invalidated : " + statCountNumCacheInvalidations + " \n");
            buffer.append("Avg time taken to invalidate cache copies during put : " + statInvalidateTime.getMean() + " \n");
            buffer.append("		Avg call size for cache invalidation during put call : " + statCountInvalidateCallSize.getMean() + " \n");
            buffer.append("		Avg time taken to invalidate the cache copy of a single node during put : " +
                    statInvalidateTime.getMean() / statCountInvalidateCallSize.getMean() + " \n");
            buffer.append("		Avg Size of the invalidateCacheCopies call : " + invalidateCacheCopiesCallSize.getMean() + " \n");
            buffer.append("No. of times invalidateCacheCopies call is made from the Storage System : " + statInvalidateCacheCopiesTime.getNumSamples() + " \n");
            buffer.append("		Avg time taken for the invalidateCacheCopies call : " + statInvalidateCacheCopiesTime.getMean() + " \n");
            buffer.append("		Avg time taken/Node for the invalidateCacheCopies call : " + statInvalidateCacheCopiesTime.getMean() / invalidateCacheCopiesCallSize.getMean());
            buffer.append("No. of times a node was requested for caching from another server : " +
                    statGetNodeForCachingTime.getNumSamples() + " \n");
            buffer.append("		Avg time taken to service this cache copy : " + statGetNodeForCachingTime.getMean() + " \n");
            buffer.append("		In its parallel call, avg call size : " + statGetNodesForCachingCallSize.getMean() + " \n");
        }
        buffer.append("\n");
        return buffer.toString();
    }

    /**
     * Clear stat containers
     */
    public void clearStatContainers() {
        statCachedCopyNumTimesUsed = new StatContainer();
        statCountInvalidateCallSize = new StatContainer();
        statInvalidateTime = new StatContainer();
        statGetCallTime = new StatContainer();
        statPutCallTime = new StatContainer();
        statGetNodeForCachingTime = new StatContainer();
        statGetNodesForCachingCallSize = new StatContainer();
        statInvalidateCacheCopiesTime = new StatContainer();
        invalidateCacheCopiesCallSize = new StatContainer();
        statCountNumCacheStarts = 0;
        statCountNumCachePuts = 0;
        statCountNumCacheInvalidations = 0;
    }

    private void recordStat(StatContainer stat, float value) {
        if (this.recordStats)
            stat.addOnlyMean(value);
    }

    /**
     * To load initial nodes into the table, used by the file loader
     *
     * @param nodes
     * @throws RemoteException
     */
    public void loadNodes(Vector<Node> nodes) throws RemoteException {
        for (Node n : nodes) {
            table.put(n.nodeId, n);
        }
    }

    public int getServerId() throws RemoteException {
        return this.myId;
    }

    public int getNumServers() throws RemoteException {
        return this.numServers;
    }

    public void setNumServers(int n) throws RemoteException {
        this.numServers = n;
    }

	/* End of helper functions */

	/*Relocation related. Relocation cannot be used along with caching*/

    /**
     * called on the homesite when relocating a node to a remote server
     * homesite makes a note of it in the location table
     *
     * @param nodeId
     * @param serverName
     * @return
     * @deprecated nodeIds are now Strings
     */
    @Deprecated
    public ResultObject relocate(int nodeId, String serverName) {
        String key = Integer.toString(nodeId);
        if (lockedObjects.contains(nodeId)) {
            return new ResultObject(key, Constants.LOCATION_LOCKED);
        }
        if (locationTable.containsKey(key)) {
            // already relocated
            String location = locationTable.get(key);
            Node node = (Node) table.get(key);
            return new ResultObject(Constants.RELOCATED, node, key, location);
        } else {
            locationTable.put(key, serverName);
            Node node = (Node) table.get(key);
            return new ResultObject(node, key);
        }
    }

    /**
     * Inform the homesite that a node has been relocated.
     * homesite makes a note of it in the location table and removes node from main table
     *
     * @param nodeId
     * @param servername
     * @deprecated nodeIds are now Strings
     */
    @Deprecated
    public void locationChange(int nodeId, String servername) {
        String key = Integer.toString(nodeId);
        locationTable.put(key, servername);
        table.remove(key);
        // //System.out.println("Removed from Table"+Integer.toString(nodeId));
        return;
    }


    /**
     * Lock an object to prevent its relocation
     *
     * @param nodeId
     * @return
     * @throws RemoteException
     * @deprecated nodeIds are now Strings
     */
    @Deprecated
    public ResultObject lock(int nodeId) throws RemoteException {
        String key = Integer.toString(nodeId);
        if (locationTable.containsKey(key)) {
            String location = locationTable.get(key);
            //			ResultObject result = new ResultObject(nodeId, location);
            return new ResultObject(key, location);
        } else {
            lockedObjects.add(key);
            return new ResultObject(key, 0);
        }
    }

    /**
     * Remove a key from the location table
     *
     * @param key
     */
    public void removeLocationTableEntry(String key) {

        if (locationTable.containsKey(key))
            locationTable.remove(key);

    }

	/* End of relocation related code*/

    @Override
    public int acquireLocks(SortedSet<LockRequest> lockRequests) throws RemoteException {
        List<LockRequest> acquiredLocks = new ArrayList<>();

        for (LockRequest request : lockRequests) {
            int statusCode = acquireLockForValidation(request.nodeId, request.ts, request.lockMode);
            if (statusCode == Constants.LOCK_ACQUIRED) {
                acquiredLocks.add(request);
            } else {
                for (LockRequest lock : acquiredLocks) {
                    releaseLock(lock);
                }
                return Constants.LOCK_NOT_ACQUIRED;
            }
        }
        return Constants.LOCK_ACQUIRED;
    }

    @Override
    public void releaseLock(SortedSet<LockRequest> lockRequests) throws RemoteException {
        for (LockRequest lockRequest : lockRequests) {
            releaseLock(lockRequest);
        }
    }

    @Override
    public void commitChanges(Set<Node> write, Set<NodeData> writeNd, Map<String, Long> read) throws RemoteException {

        for (Node node : write) {
            table.put(node.getNodeId(), node);
        }

        for (NodeData nodeData : writeNd) {
            Node node = (Node) table.get(nodeData.nodeId);
            HashMap<String, Object> failedUpdates = updateNodePropertiesByReflection(node, nodeData.properties);
            node.TS = nodeData.TS;

            if (failedUpdates == null) {
                table.put(node.nodeId, node);
            } else {
                System.out.println("FAILED UPDATE");
            }
        }

        for (Node node : write) {
            releaseLock(new LockRequest(Constants.VALIDATION_LOCK_WRITE_MODE, node.TS, node.nodeId));
        }

        for (NodeData nodeData : writeNd) {
            releaseLock(new LockRequest(Constants.VALIDATION_LOCK_WRITE_MODE, nodeData.TS, nodeData.nodeId));
        }

        for (String nodeId : read.keySet()) {
            releaseLock(new LockRequest(Constants.VALIDATION_LOCK_READ_MODE, read.get(nodeId), nodeId));
        }
    }

    private void releaseLock(LockRequest lock) {
        ReentrantLock rl = cTableLockList.get(Math.abs(lock.nodeId.hashCode()) % nLocks);
        rl.lock();
        nodeRWLockMap.get(lock.nodeId).releaseLock(lock.lockMode);
        rl.unlock();
    }

    @Override
    public void checkpoint() throws RemoteException {
        try {
            FileOutputStream hashtableServerCPFile = new FileOutputStream(checkpointSaveDir + "/HashtableServer.ser");
            ObjectOutputStream hashtableServerCPStream = new ObjectOutputStream(hashtableServerCPFile);

            hashtableServerCPStream.writeObject(table);
            hashtableServerCPStream.writeObject(locationTable);

            hashtableServerCPStream.close();
            hashtableServerCPFile.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    // TODO DOCUMENTATION
    private int acquireLockForValidation(String nodeId, long TS, int lockMode) throws RemoteException {
        ReentrantLock lock = cTableLockList.get(Math.abs(nodeId.hashCode()) % nLocks);
        lock.lock();

        Node oldNode = (Node) table.get(nodeId);
        if (oldNode.TS != TS) {
            lock.unlock();
            return Constants.LOCK_NOT_ACQUIRED;
        }

        if (!nodeRWLockMap.containsKey(nodeId)) {
            nodeRWLockMap.put(nodeId, new NodeRWValidationLock());
        }

        NodeRWValidationLock rwLock = nodeRWLockMap.get(nodeId);
        int status = rwLock.getLock(lockMode);
        lock.unlock();
        return status;
    }


    //**************************************************************************************************************************************
    // all thrift stuff starts from here

    @Override
    public ResultObject_thrift put_thrift(String nodeId, Node_thrift value) {

        Node n = new Node(value);
        ResultObject res = null;
        try {
            res = put(nodeId, n);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public ResultObject_thrift get_thrift(String key, long storageSystemThreadId)
            throws TException {

        ResultObject res = get(key, storageSystemThreadId);
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public ResultObject_thrift getNodeProperties_thrift(NodeData_thrift n) throws TException {
        ResultObject res = null;
        NodeData rmi_n = new NodeData(n, nodeClassMap);

        try {
            res = getNodeProperties(rmi_n);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public ResultObject_thrift updateNodeProperties_thrift(NodeData_thrift n) throws TException {
        ResultObject res = null;
        NodeData rmi_n = new NodeData(n, nodeClassMap);
        try {
            res = updateNodeProperties(rmi_n);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public List<ResultObject_thrift> getObjectsProperties_thrift(List<NodeData_thrift> keyList) throws TException {

        Vector<NodeData> keyList_rmi = new Vector<NodeData>();
        Vector<ResultObject> res = null;

        for (NodeData_thrift ele : keyList) {
            keyList_rmi.add(new NodeData(ele, nodeClassMap));
        }

        if (keyList_rmi.size() > 0) {
            try {
                res = getObjectsProperties(keyList_rmi);
            } catch (RemoteException e) {
                e.printStackTrace();
            }
        }

        return ConvertRMIToThrift.convertResultObjectList(res);
    }

    @Override
    public List<ResultObject_thrift> updateNodePropertiesParallel_thrift(List<NodeData_thrift> keyList) throws TException {

        Vector<ResultObject> res = null;
        Vector<NodeData> keyList_rmi = new Vector<NodeData>();
        for (NodeData_thrift ele : keyList) {
            keyList_rmi.add(new NodeData(ele, nodeClassMap));
        }
        if (keyList_rmi.size() > 0) {
            try {
                res = updateNodePropertiesParallel(keyList_rmi);
            } catch (RemoteException e) {
                e.printStackTrace();
            }
        }
        return ConvertRMIToThrift.convertResultObjectList(res);
    }

    @Override
    public List<ResultObject_thrift> putObjects_thrift(Map<String, Node_thrift> nodeList_thrift) throws TException {

        ConcurrentHashMap<String, Node> nodeList = new ConcurrentHashMap<String, Node>();
        for (String nodeId : nodeList_thrift.keySet()) {
            nodeList.put(nodeId, new Node(nodeList_thrift.get(nodeId)));
        }

        Vector<ResultObject> res = null;
        try {
            res = putObjects(nodeList);
        } catch (RemoteException e) {
            e.printStackTrace();
        }

        return ConvertRMIToThrift.convertResultObjectList(res);
    }


    @Override
    public List<ResultObject_thrift> getObjects_thrift(List<String> keyList) throws TException {

        Vector<ResultObject> res = getObjects(new Vector<String>(keyList));
        // new vector has to be created, only type-casting to Vector<String> doesn't suffice.
        // this is because we do not know how Thrift implements List internally.

        return ConvertRMIToThrift.convertResultObjectList(res);
    }

    @Override
    public List<String> getAllKeys_thrift() throws TException {
        return getAllKeys();
    }

    @Override
    public ResultObject_thrift relocate_thrift(int nodeId, String serverName) throws TException {
        ResultObject res = relocate(nodeId, serverName);
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public void locationChange_thrift(int nodeId, String serverName) throws TException {
        locationChange(nodeId, serverName);
    }

    @Override
    public ResultObject_thrift lock_thrift(int nodeId) throws TException {
        ResultObject res = null;
        try {
            res = lock(nodeId);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return ConvertRMIToThrift.convertResultObject(res);
    }

    @Override
    public int acquireLocks_thrift(Set<LockRequest_thrift> lockRequests) throws TException {
        SortedSet<LockRequest> jRequests = new TreeSet<>();
        for (LockRequest_thrift lockRequest : lockRequests) {
            jRequests.add(new LockRequest(lockRequest));
        }
        try {
            return acquireLocks(jRequests);
        } catch (RemoteException e) {
            e.printStackTrace();
            return Constants.LOCK_NOT_ACQUIRED;
        }
    }

    @Override
    public void releaseLocks_thrift(Set<LockRequest_thrift> lockRequests) throws TException {
        SortedSet<LockRequest> jRequests = new TreeSet<>();
        for (LockRequest_thrift lockRequest : lockRequests) {
            jRequests.add(new LockRequest(lockRequest));
        }
        try {
            releaseLock(jRequests);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void commitChanges_thrift(Set<NodeData_thrift> writeNd, Map<String, Long> read) throws TException {
        Set<NodeData> jNodeData = new HashSet<>();
        for (NodeData_thrift ndt : writeNd) {
            jNodeData.add(new NodeData(ndt, nodeClassMap));
        }
        try {
            commitChanges(new HashSet<>(), jNodeData, read);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    public static void main(String args[]) throws Exception {

        System.setSecurityManager(new RMISecurityManager());

        // Create new hostlist from the config or similar to application host
        // list construction.
        String myHostname = InetAddress.getLocalHost().getHostName();
        String location = "//" + myHostname + "/HashtableServer";
        HashtableServerImpl HashtableNode = new HashtableServerImpl(location);

        if (args.length == 0) {
            // If no port number is given for the rmiregistry, assume it is on
            // the default port 1099
            Naming.bind("HashtableServer", HashtableNode);
        } else {
            // rmiregistry is on port specified in args[0]. Bind to that
            // registry.
            Registry localRegistry = LocateRegistry.getRegistry(Integer
                    .parseInt(args[0]));
            localRegistry.bind("HashtableServer", HashtableNode);
        }
    }
}

class CachePurgeThread extends Thread {
    private HashtableServerImpl server;

    CachePurgeThread(HashtableServerImpl serverImpl) {
        this.server = serverImpl;
    }

    public void run() {
        while (true) {
            try {
                if (server.cacheTable.size() > server.CACHE_TABLE_SIZE) {
                    //do cache purge
                    //clear off 1/3 of the cache limit size (max size)
                    int numItems = server.cacheTable.size() - (int) (server.CACHE_TABLE_SIZE * server.CACHE_CLEANUP_TARGET);
                    long currentTime = System.currentTimeMillis();
                    String key;
                    NodeMetadata metadata;
                    HashSet<String> secondChance = new HashSet<String>();
                    Vector<Node> evictedNodes = new Vector<Node>();
                    Node evictedNode = null;
                    for (int count = 0; count < numItems; count++) {
                        key = server.cacheUsageQueue.poll();
                        if (key == null) break;//not enough entries
                        metadata = server.metaDataTable.get(key);
                        if (metadata != null) {
                            //check if this qualifies for cache eviction
                            if (currentTime - metadata.lastAccessedTS > server.CACHE_WINDOW || secondChance.contains(key)) {
                                secondChance.remove(key);
                                evictedNode = server.cacheEvict(key);
                                //Add it to list so that master can be informed
                                if (evictedNode != null) {
                                    evictedNodes.add(evictedNode);
                                }
                            } else {
                                //give it a second chance and put it back in the queue
                                secondChance.add(key);
                                //back to the queue
                                server.cacheUsageQueue.add(key);
                                count = count - 1;
                            }
                        } else {
                            //somehow the entry is missing, continue
                            count = count - 1; //in this iteration we could not clear any
                        }
                    }
                    if (evictedNodes.size() > 0) {
                        //bucketise and inform the master nodes
                        Hashtable<Integer, Hashtable<String, Long>> evictionTable = new Hashtable<Integer, Hashtable<String, Long>>();
                        int id;
                        int serverid;
                        Integer ID;
                        for (Node n : evictedNodes) {
                            //							id = n.nodeId;
                            id = Integer.parseInt(n.nodeId);
                            serverid = id % server.numServers;
                            ID = new Integer(serverid);
                            if (evictionTable.get(ID) == null) {
                                evictionTable.put(ID, new Hashtable<String, Long>());
                            }
                            ((Hashtable<String, Long>) evictionTable.get(ID)).put(n.nodeId, new Long(n.TS));
                        }
                        //serial updated of the buckets to the master
                        HashtableServer serverHandle;
                        Node eNode = null;
                        for (Integer i : evictionTable.keySet()) {
                            serverHandle = server.servers.get(i);
                            System.out.println("Telling master that " + ((Hashtable<String, Long>) evictionTable.get(i)).size() + " entries have been evicted from cache");
                            serverHandle.cacheEvictMasterNotify(Integer.toString(server.myId), evictionTable.get(i));
                        }
                        System.out.println("Total number of nodes evicted : " + evictedNodes.size());
                        System.out.println("Size of cache table post replacement : " + server.cacheTable.size());
                    }
                }
                //sleep for 5 seconds
                Thread.sleep(5000);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }


}

class NodeRWValidationLock {
    private boolean writeLock = false;
    private int readCount = 0;

    synchronized int getLock(int lockMode) {
        if (lockMode == Constants.VALIDATION_LOCK_READ_MODE) {
            if (writeLock) {
                return Constants.LOCK_NOT_ACQUIRED;
            } else {
                readCount++;
                return Constants.LOCK_ACQUIRED;
            }
        } else if (lockMode == Constants.VALIDATION_LOCK_WRITE_MODE) {
            if (writeLock || readCount > 0) {    //lock failed
                return Constants.LOCK_NOT_ACQUIRED;
            } else {    //got the lock
                writeLock = true;
                return Constants.LOCK_ACQUIRED;
            }
        } else {
            System.out.println("ERROR :: Unknown Lock mode for Validation getLock() call!");
            return Constants.LOCK_NOT_ACQUIRED;
        }
//        if (lockMode == Constants.VALIDATION_LOCK_READ_MODE) {
//            if (writeLock) {
//                return Constants.VALIDATION_READ_LOCK_FAILED;
//            } else {
//                readCount++;
//                return Constants.VALIDATION_READ_LOCK_SUCCESS;
//            }
//        } else if (lockMode == Constants.VALIDATION_LOCK_WRITE_MODE) {
//
//            if (writeLock == true || readCount > 0) {    //lock failed
//                return Constants.VALIDATION_WRITE_LOCK_FAILED;
//            } else {    //got the lock
//                writeLock = true;
//                return Constants.VALIDATION_WRITE_LOCK_SUCCESS;
//            }
//        } else {
//            System.out.println("ERROR :: Unknown Lock mode for Validation getLock() call!");
//            return Constants.VALIDATION_ABORT;
//        }
    }

    synchronized void releaseLock(int lockMode) {
        if (lockMode == Constants.VALIDATION_LOCK_READ_MODE) {
            if (readCount > 0) {
                readCount--;
            }
        } else if (lockMode == Constants.VALIDATION_LOCK_WRITE_MODE) {
            writeLock = false;
        } else {
            System.out.println("ERROR :: Unknown Lock mode for Validation releaseLock() call!");
        }
    }
}
