/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import beehive.server.HashtableServer;
import beehive.server.ResultObject;

import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

import beehive.thrift.*;

import org.apache.thrift.TException;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.transport.TTransportException;


public final class HashTableThriftHandler implements HashtableServer {

    private LinkedBlockingQueue<HashtableServer_thrift.Client> q = new LinkedBlockingQueue<HashtableServer_thrift.Client>();
    private StorageSystem storagesystem;

    private ConvertRMIToThrift converter = new ConvertRMIToThrift();

    public static String RPC_MECHANISM = "RMI";

    public HashTableThriftHandler(StorageSystem storagesystem, String host, Config configParams) {
        this.storagesystem = storagesystem;
        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

        if (RPC_MECHANISM.equals("THRIFT")) {
            int countThriftHandles = configParams.getIntValue("System.HashtableServer_THRIFT_NUM_HANDLES");
            int portno = configParams.getIntValue("System.HashtableServer_THRIFT_PORT");
            for (int i = 0; i < countThriftHandles; i++) {
                //create htserver client handles. HTserver thrift server handles are created in BeehiveAppLoader.
                try {
                    TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(host, portno);
                    HashtableServer_thrift.Client client = new HashtableServer_thrift.Client(protocol);
                    q.offer(client);
                } catch (TTransportException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    public ResultObject updateNodeProperties(NodeData nodeData) throws RemoteException {

        HashtableServer_thrift.Client client = null;
        try {
            client = q.take();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        NodeData_thrift nd = ConvertRMIToThrift.convertNodeData(nodeData, nodeData.nodeClassId);
        ResultObject_thrift res_thrift;
        try {
            res_thrift = client.updateNodeProperties_thrift(nd);
        } catch (TException e) {
            e.printStackTrace();
            throw new RemoteException();
        }
        ResultObject result = new ResultObject(res_thrift, StorageSystem.nodeClassMap);
        q.offer(client);

        return result;
    }


    public ResultObject getNodeProperties(NodeData nodeData) throws RemoteException {

        HashtableServer_thrift.Client client = null;
        try {
            client = q.take();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        NodeData_thrift nd = ConvertRMIToThrift.convertNodeData(nodeData, nodeData.nodeClassId);

        ResultObject_thrift res_thrift;
        try {
            res_thrift = client.getNodeProperties_thrift(nd);
        } catch (TException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new RemoteException();
        }

        ResultObject result = new ResultObject(res_thrift, StorageSystem.nodeClassMap);
        q.offer(client);
        return result;

    }

    public Vector<ResultObject> updateNodePropertiesParallel(Vector<NodeData> vec) throws RemoteException {

        HashtableServer_thrift.Client client = null;
        try {
            client = q.take();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        Vector<NodeData_thrift> vec_thrift = ConvertRMIToThrift.convertNodeDataList(vec);

        List<ResultObject_thrift> result;
        try {
            result = client.updateNodePropertiesParallel_thrift(vec_thrift);
        } catch (TException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new RemoteException();
        }

        Vector<ResultObject> result_rmi = new Vector<ResultObject>();

        for (ResultObject_thrift ele : result) {
            result_rmi.add(new ResultObject(ele, StorageSystem.nodeClassMap));
        }

        q.offer(client);

        //		System.out.println();
        //		System.out.println("in putNodeDataBatchCall");
        //		System.out.println("recieved "+result_rmi);
        //		System.out.println("\n\n");


        return result_rmi;
    }

    public Vector<ResultObject> getObjectsProperties(Vector<NodeData> vec) throws RemoteException {

        HashtableServer_thrift.Client client = null;
        try {
            client = q.take();
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        Vector<NodeData_thrift> vec_thrift = converter.convertNodeDataList(vec);

        List<ResultObject_thrift> result;
        try {

            result = client.getObjectsProperties_thrift(vec_thrift);
        } catch (TException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new RemoteException();

        }

        Vector<ResultObject> result_rmi = new Vector<ResultObject>();

        for (ResultObject_thrift ele : result) {
            result_rmi.add(new ResultObject(ele, storagesystem.nodeClassMap));
        }

        q.offer(client);

        return result_rmi;

        //		System.out.println("**********************************************************");
        //		System.out.println("in getNodeDataBatchCall");
        //		System.out.println("recieved "+result_rmi);
    }


    @Override
    public Date ping() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public ResultObject put(String key, Object value) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public ResultObject put(String key, Object value, int callingServerId)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public ResultObject get(String key, long storageSystemThreadId)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> getObjects(Vector<String> keyList)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> putObjects(
            ConcurrentHashMap<String, Node> objList) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public ResultObject exec(NodeData nd) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> execBulk(Vector<NodeData> keyList)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> putObjects(
            ConcurrentHashMap<String, Node> objList, int callerServerId)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public boolean contains(String key) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return false;
    }

    @Override
    public Vector<Node> getAllNodes() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<String> getAllKeys() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public String getStats() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void clearStatContainers() throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public Node getNodeFromTable(String nodeId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void loadNodes(Vector<Node> nodes) throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public int getServerId() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return 0;
    }

    @Override
    public int getNumServers() throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return 0;
    }

    @Override
    public void setNumServers(int n) throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public ResultObject relocate(int nodeId, String serverName)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void locationChange(int nodeId, String serverName)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public ResultObject lock(int nodeId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void removeLocationTableEntry(String key) throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public void cacheInvalidate(String nodeID) throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public Node getNodeForCaching(String nodeID, int cachingSite)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public ResultObject cachePut(String nodeID, Node node, int masterCopySite,
                                 long storageSystemThreadId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> cachePut(Map<Integer, Vector<Node>> nodes,
                                         long storageSystemThreadId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<ResultObject> getLocalNodes(Vector<String> nodeIds,
                                              long storageSystemThreadId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<Node> getNodesForCaching(Vector<String> nodeIds,
                                           int callingSite) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void writeThroughToCache(String paramString, Object value)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public void writeThroughToCacheBulk(ConcurrentHashMap<String, Node> map)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public Node getNodeFromCacheTable(String nodeId) throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public Vector<Node> getCachedObjects(Vector<String> nodeIds)
            throws RemoteException {
        System.out.println("ERROR::Not implemented");
        return null;
    }

    @Override
    public void cacheEvictMasterNotify(String cachingServerId,
                                       Hashtable<String, Long> evictedNodes) throws RemoteException {
        System.out.println("ERROR::Not implemented");

    }

    @Override
    public void checkpoint() {

    }

    @Override
    public int acquireLocks(SortedSet<LockRequest> lockRequests) throws RemoteException {

        Set<LockRequest_thrift> tLRs = new HashSet<>();
        for (LockRequest lockRequest : lockRequests) {
            tLRs.add(ConvertRMIToThrift.convertLockRequest(lockRequest));
        }

        try {
            HashtableServer_thrift.Client client = q.take();
            int result = client.acquireLocks_thrift(tLRs);
            q.offer(client);
            return result;
        } catch (InterruptedException | TException e) {
            e.printStackTrace();
            return Constants.LOCK_NOT_ACQUIRED;
        }
    }

    @Override
    public void releaseLock(SortedSet<LockRequest> lockRequests) throws RemoteException {
        Set<LockRequest_thrift> tLRs = new HashSet<>();
        for (LockRequest lockRequest : lockRequests) {
            tLRs.add(ConvertRMIToThrift.convertLockRequest(lockRequest));
        }

        try {
            HashtableServer_thrift.Client client = q.take();
            client.releaseLocks_thrift(tLRs);
            q.offer(client);
        } catch (InterruptedException | TException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void commitChanges(Set<Node> write, Set<NodeData> writeNd, Map<String, Long> read) throws RemoteException {
        if (write != null && !write.isEmpty()) {
            throw new UnsupportedOperationException("Writing nodes cannot be supported by thrift");
        } else {
            Set<NodeData_thrift> tNds = new HashSet<>();
            for (NodeData nodeData : writeNd) {
                tNds.add(ConvertRMIToThrift.convertNodeData(nodeData, nodeData.nodeClassId));
            }

            try {
                HashtableServer_thrift.Client client = q.take();
                client.commitChanges_thrift(tNds, read);
                q.offer(client);
            } catch (InterruptedException | TException e) {
                e.printStackTrace();
            }
        }
    }
}
