/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

import beehive.server.HashtableServer;
import beehive.server.ResultObject;

/**
 * Hostwise Request aggragated in this class. It also has the worker requests.
 *
 * @author sharm443
 */

public class HostDispatchTable {
    final ConcurrentHashMap<Integer, RequestAggregator> dispatchTable;
    ConcurrentHashMap<Long, WorkerRequest> workerRequestTable;
    StorageSystem storageSystem;
    NodeDataCache nodeDatacache;
    int countCacheHit = 0;
    LinkedBlockingQueue<Integer> dispatchRequestQueue;

    int batchDispatchTimeout = 10;
    int batchCountThreshold = 3;
    int nodeDataRequestDispatchPoolSize = 4;
    int nodeDataCacheSize = 10000;
    float nodeDataCachePurgeThreshold = 0.9f;
    int nodeDataCachePurgePeriod = 5000;
    float nodeDataCacheCleanupFraction = 0.1f;
    public static String RPC_MECHANISM = "RMI";


    public HostDispatchTable(StorageSystem strsys) {
        Config configParams = Config.getInstance();
        if (configParams.has("StorageSystem.BatchDispatchTimeout")) {
            this.batchDispatchTimeout = configParams.getIntValue("StorageSystem.BatchDispatchTimeout");
        }
        if (configParams.has("StorageSystem.BatchCountThreshold")) {
            this.batchCountThreshold = configParams.getIntValue("StorageSystem.BatchCountThreshold");
        }
        if (configParams.has("StorageSystem.NodeDataRequestDispatchPoolSize")) {
            this.nodeDataRequestDispatchPoolSize = configParams.getIntValue("StorageSystem.NodeDataRequestDispatchPoolSize");
        }
        if (configParams.has("StorageSystem.NodeDataCacheSize")) {
            this.nodeDataCacheSize = configParams.getIntValue("StorageSystem.NodeDataCacheSize");
        }
        if (configParams.has("StorageSystem.NodeDataCachePurgeThreshold")) {
            this.nodeDataCachePurgeThreshold = configParams.getFloatValue("StorageSystem.NodeDataCachePurgeThreshold");
        }
        if (configParams.has("StorageSystem.NodeDataCachePurgePeriod")) {
            this.nodeDataCachePurgePeriod = configParams.getIntValue("StorageSystem.NodeDataCachePurgePeriod");
        }
        if (configParams.has("StorageSystem.NodeDataCacheCleanupFraction")) {
            this.nodeDataCacheCleanupFraction = configParams.getFloatValue("StorageSystem.NodeDataCacheCleanupFraction");
        }
        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

        this.dispatchTable = new ConcurrentHashMap<Integer, RequestAggregator>();
        this.workerRequestTable = new ConcurrentHashMap<Long, WorkerRequest>();
        this.storageSystem = strsys;
        this.nodeDatacache = new NodeDataCache(nodeDataCacheSize, nodeDataCacheCleanupFraction, nodeDataCachePurgeThreshold, nodeDataCachePurgePeriod);
        dispatchRequestQueue = new LinkedBlockingQueue<Integer>();

        System.out.println("Host " + storageSystem.myId + " ::HostDispatchTable: Creating " + nodeDataRequestDispatchPoolSize + " NodeDataRequestDispatcher threads");
        for (int i = 0; i < nodeDataRequestDispatchPoolSize; i++) {
            NodeDataRequestDispatcher requestDispatcher = new NodeDataRequestDispatcher();
            requestDispatcher.start();
        }

        NodeDataRequestDispatchMonitor requestDispatchMonitor = new NodeDataRequestDispatchMonitor();
        requestDispatchMonitor.start();

    }

    public void doGetNodeData(WorkerRequest wrkReq, ConcurrentHashMap<Integer, Vector<NodeData>> hostNodeDataBucket) throws InterruptedException {
        //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: inside doGetNodeData()");
        long workerId = Thread.currentThread().getId();

        workerRequestTable.put(workerId, wrkReq);

        Iterator it = hostNodeDataBucket.keySet().iterator();
        //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: inside doGetNodeData() : Set => "+hostNodeDataBucket.keySet().toString());
        while (it.hasNext()) {
            Integer hostId = (Integer) it.next();
            Vector<NodeData> vecNodeData = hostNodeDataBucket.get(hostId);

            synchronized (dispatchTable) {
                if (!dispatchTable.containsKey(hostId)) {
                    RequestAggregator reqAgg = new RequestAggregator();
                    dispatchTable.put(hostId, reqAgg);
                }
            }

            RequestAggregator reqAgg = dispatchTable.get(hostId);
            synchronized (reqAgg) {
                //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: doGetNodeData() : Host "+hostId+" ::after getting inside sync block, count : "+reqAgg.count+", batchCountThreshold : "+batchCountThreshold+" flag : "+reqAgg.requestBatchDispatched);
                RequestBatch reqBch = reqAgg.next;
                for (NodeData nd : vecNodeData) {
                    String ndId = nd.nodeId;
                    Vector<Long> vecWorkers;
                    if (reqBch.nodeWorkerWaitingMap.containsKey(ndId)) {
                        vecWorkers = reqBch.nodeWorkerWaitingMap.get(ndId);
                        vecWorkers.add(workerId);
                        reqBch.nodeWorkerWaitingMap.put(ndId, vecWorkers);
                    } else {
                        vecWorkers = new Vector<>();
                        vecWorkers.add(workerId);
                        reqBch.nodeWorkerWaitingMap.put(ndId, vecWorkers);
                        //reqBch.nodeIdSet.add(ndId);
                        reqBch.requestList.add(nd);
                    }
                }
                reqAgg.count++;
                //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: doGetNodeData() : Host "+hostId+" :: count : "+reqAgg.count+", batchCountThreshold : "+batchCountThreshold+" flag : "+reqAgg.requestBatchDispatched);
                if (reqAgg.count == batchCountThreshold) {
                    //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: doGetNodeData() : Host "+hostId+" added to dispatchRequestQueue");
                    dispatchRequestQueue.put(hostId);
                }
            }
        }
    }

    public void dispatchBatch(Integer hostId) throws InterruptedException {
        RequestAggregator reqAgg = dispatchTable.get(hostId);
        if (reqAgg == null) {
            System.out.println("Host " + storageSystem.myId + " ::ERROR::HostDispatchTable::dispatchBatch(): RequestAggregator recieived from dispatchTable found null");
        }

        RequestBatch currentBatch = reqAgg.prepareForDispatch();
        if (currentBatch == null) {
            //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable::dispatchBatch(): Returned null from prepareForDispatch()...");
            //dispatchRequestQueue.put(hostId);
            return;
        }

        ConcurrentHashMap<String, NodeData> nodeDataFromCache = new ConcurrentHashMap<String, NodeData>();

        Vector<NodeData> vecNodeData = currentBatch.requestList;
        for (NodeData nd : vecNodeData) {
            String ndkey = nd.nodeId;
            NodeData ndtemp = nodeDatacache.cacheTable.get(ndkey);

            if (ndtemp != null) {
                countCacheHit++;
                nodeDataFromCache.put(ndkey, ndtemp);
                nd.TS = ndtemp.TS;
            } else {
                nd.TS = -1;
            }
        }

        HashtableServer serverHandle = null;
        if (StorageSystem.RPC_MECHANISM.equals("RMI")) {
            serverHandle = storageSystem.servers.get(hostId);
        } else if (StorageSystem.RPC_MECHANISM.equals("THRIFT")) {
            serverHandle = storageSystem.serversThrift.get(hostId);
        }

        Vector<ResultObject> vecRes = null;

        int retrycounter = 0;
        while (vecRes == null && retrycounter < 100) {
            try {
                vecRes = serverHandle.getObjectsProperties(vecNodeData);
            } catch (Exception exception) {
                retrycounter++;
                vecRes = null;
                exception.printStackTrace();
            }
        }
        if (vecRes == null) {
            System.out.println("FATAL ERROR :: dispatchBatch() : getObjectsProperties() call timed out after multiple retries. Exiting system...");
            System.exit(0);
        }

        int countValidCacheItems = 0;
        for (ResultObject res : vecRes) {
            if (res.status == Constants.SUCCESS) {
                if (res.node == null) {
                    countValidCacheItems++;
                    res.node = nodeDataFromCache.get(res.nodeId);
                } else {    //add node found to cache
                    if (!(storageSystem.myId == hostId))
                        nodeDatacache.addItem((NodeData) res.node);
                }

                Vector<Long> waitingWorkers = currentBatch.nodeWorkerWaitingMap.get(res.nodeId);
                if (waitingWorkers == null) {
                    System.out.println("Host " + storageSystem.myId + " ::ERROR::HostDispatchTable::dispatchBatch(): waitingWorkers is null for nodeId : " + res.nodeId);
                }
                for (Long wrkId : waitingWorkers) {
                    WorkerRequest wrkReq = workerRequestTable.get(wrkId);
                    if (wrkReq != null) {
                        wrkReq.addNodeData(new NodeData((NodeData) res.node));
                    } else {
                        System.out.println("Host " + storageSystem.myId + " ::ERROR::HostDispatchTable::dispatchBatch(): WorkerRequest is null");
                    }
                }
            } else if (res.status == Constants.RELOCATED) {
                //System.out.println("Host "+storageSystem.myId+" ::NodeData relocated!!!");
                String loc = res.location;
                int j = storageSystem.serverMapping.get(loc);

                NodeData ndToAdd = null;
                for (NodeData nd : vecNodeData) {
                    if (nd.nodeId.equals(res.nodeId)) {
                        ndToAdd = nd;
                        break;
                    }
                }

                synchronized (dispatchTable) {
                    if (!dispatchTable.containsKey(j)) {
                        RequestAggregator reqAgg2 = new RequestAggregator();
                        dispatchTable.put(j, reqAgg2);
                    }
                }

                RequestAggregator reqAggOfRequiredHost = dispatchTable.get(j);
                synchronized (reqAggOfRequiredHost) {
                    Vector<Long> vNew = currentBatch.nodeWorkerWaitingMap.get(res.nodeId);
                    ConcurrentHashMap temp = reqAggOfRequiredHost.next.nodeWorkerWaitingMap;
                    Vector<Long> vExisting = (Vector<Long>) temp.get(res.nodeId);
                    if (vExisting == null) {
                        vExisting = new Vector<Long>();
                    }
                    vExisting.addAll(vNew);
                    //System.out.println("Vector contains : "+vExisting.toString());
                    temp.put(new Integer(res.nodeId).toString(), vExisting);
                    reqAggOfRequiredHost.next.requestList.add(ndToAdd);
                    reqAggOfRequiredHost.count++;
                    if (reqAggOfRequiredHost.count == batchCountThreshold) {
                        //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable :: doGetNodeData() : Host "+hostId+" added to dispatchRequestQueue");
                        dispatchRequestQueue.put(hostId);
                    }
                }
                storageSystem.locationTable.put(res.nodeId, loc);

            }
        }
    }

    public void clearWorkerRequest(Long wrkId) {
        //System.out.println("Host "+storageSystem.myId+" ::HostDispatchTable::clearWorkerRequest(): clearing workerRequestTable for worker Id "+wrkId);
        workerRequestTable.remove(wrkId);
    }

    class NodeDataRequestDispatcher extends Thread {
        public void run() {
            while (true) {
                try {
                    Integer hostId = dispatchRequestQueue.take();
                    dispatchBatch(hostId);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    class NodeDataRequestDispatchMonitor extends Thread {
        public void run() {
            while (true) {
                try {
                    Vector<Integer> hostList = new Vector<Integer>(dispatchTable.keySet());
                    for (Integer hostId : hostList) {
                        if (dispatchTable.containsKey(hostId)) {
                            RequestAggregator reqAgg = dispatchTable.get(hostId);
                            long currtime = System.currentTimeMillis();

                            synchronized (reqAgg) {
                                if ((currtime - reqAgg.lastDispatchTime) > batchDispatchTimeout && reqAgg.count > 0) {
                                    //System.out.println("Host "+storageSystem.myId+" ::NodeDataRequestDispatchMonitor::Adding request for host "+hostId+" to dispatchRequestQueue");
                                    dispatchRequestQueue.put(hostId);
                                }
                            }
                        }
                    }
                    Thread.sleep(2 * batchDispatchTimeout);    //setting period as 2 times the batch dispatched timeout
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

}

