/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import java.util.*;
import java.io.*;
import java.util.concurrent.*;

import beehive.thrift.Edge_thrift;
import beehive.thrift.Node_thrift;

public class Node implements Externalizable {
    public String nodeId = "";
    public long TS;
    public ConcurrentHashMap<String, Edge> neighbors;


    public Node() {
        neighbors = new ConcurrentHashMap<String, Edge>();
        TS = 0;
    }

    public Node(Node n) {
        this.nodeId = n.nodeId;
        this.TS = n.TS;
        this.neighbors = new ConcurrentHashMap<String, Edge>();
        //neighbors = new ConcurrentHashMap<>(n.neighbors);
        Iterator<String> it = n.neighbors.keySet().iterator();
        while (it.hasNext()) {
            String nId = it.next();
            this.neighbors.put(nId, n.neighbors.get(nId));
        }
    }

    public Node(Node n, boolean copyNeighbors) {
        this.nodeId = new String(n.nodeId);
        this.TS = new Long(n.TS);
        this.neighbors = new ConcurrentHashMap<String, Edge>();
        //neighbors = new ConcurrentHashMap<>(n.neighbors);
        if (copyNeighbors) {
            Iterator<String> it = n.neighbors.keySet().iterator();
            while (it.hasNext()) {
                String nId = it.next();
                this.neighbors.put(nId, new Edge());
            }
        }
    }

    public Node(String id) {
        this.nodeId = id;
        TS = 0;
        this.neighbors = new ConcurrentHashMap<>();
    }


    public Node(Node_thrift value) {
        this.nodeId = value.nodeId;
        this.TS = value.TS;

        ConcurrentHashMap<String, Edge> m = new ConcurrentHashMap<String, Edge>();
        for (String s : value.neighbors.keySet()) {
            m.put(s, new Edge(new Edge_thrift()));
        }
        neighbors = m;
    }

    public String getNodeId() {
        return nodeId;
    }

    public void setNodeId(String nodeId) {
        this.nodeId = nodeId;
    }

    public ConcurrentHashMap<String, Edge> getNeighbors() {
        return neighbors;
    }

    public void setNeighbors(ConcurrentHashMap<String, Edge> neighbors) {
        this.neighbors = neighbors;
    }

    public Long getTS() {
        return TS;
    }

    public void setTS(Long ts) {
        TS = ts;
    }

    public void addNeighbor(String index, Edge edge) {
        this.neighbors.put(index, edge);
    }

    @Override
    public String toString() {
        return "Node{" +
                "nodeId='" + nodeId + '\'' +
                ", TS=" + TS +
                '}';
    }

    public String printNode() {
        StringBuffer str = new StringBuffer();
        // str.append("Node id: "+ nodeId +"=" + " height="+height+ " excess="+excess);
        str.append("Node id: " + nodeId);
        return str.toString();
    }

    public String printNodeEdges() {
        StringBuffer str = new StringBuffer();
        // str.append("Node id: "+ nodeId +"=" + " height="+height+ " excess="+excess);
        str.append("Node id: " + nodeId);
        Enumeration<String> edges = neighbors.keys();
        while (edges.hasMoreElements()) {
            String key = (String) edges.nextElement();
            Edge e = (Edge) neighbors.get(key);
            str.append(e.toString());
        }
        return str.toString();
    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.nodeId = (String) in.readObject();
        this.TS = in.readLong();

        String[] keys = (String[]) in.readObject();
        Edge[] values = (Edge[]) in.readObject();

        for (int i = 0; i < keys.length; i++) {
            this.neighbors.put(keys[i], values[i]);
        }
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeObject(this.nodeId);
        out.writeLong(this.TS);

        String[] keys = new String[neighbors.size()];
        Edge[] values = new Edge[neighbors.size()];

        int i = 0;
        for (String s : neighbors.keySet()) {
            keys[i] = s;
            values[i] = neighbors.get(s);
            i++;
        }
        out.writeObject(keys);
        out.writeObject(values);
    }

    public NodeData toNodeData(Map map, byte classNameID) {
        NodeData nd = new NodeData(this.nodeId, this.TS, classNameID);
        return nd;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Node node = (Node) o;

        if (TS != node.TS) return false;
        if (nodeId != null ? !nodeId.equals(node.nodeId) : node.nodeId != null) return false;
        return neighbors != null ? neighbors.equals(node.neighbors) : node.neighbors == null;

    }

    @Override
    public int hashCode() {
        int result = nodeId != null ? nodeId.hashCode() : 0;
        result = 31 * result + (int) (TS ^ (TS >>> 32));
        result = 31 * result + (neighbors != null ? neighbors.hashCode() : 0);
        return result;
    }
}