/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.io.*;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

import beehive.thrift.NodeData_thrift;

public class NodeData implements Externalizable {

    public String nodeId;
    public long TS;
    public byte nodeClassId;
    public Map<String, Object> properties;

    public NodeData() {
        TS = 0;
        this.properties = new HashMap<String, Object>();
    }


    public NodeData(String nodeID, byte nodeClassId) {
        this.nodeId = nodeID;
        TS = 0;
        this.nodeClassId = nodeClassId;
        this.properties = new HashMap<String, Object>();
    }

    public NodeData(String nodeID, long ts, byte nodeClassId) {
        this.nodeId = nodeID;
        TS = ts;
        this.nodeClassId = nodeClassId;
        this.properties = new HashMap<String, Object>();
    }

    public NodeData(NodeData n) {
        this.nodeId = n.nodeId;
        this.TS = n.TS;
        this.nodeClassId = n.nodeClassId;
        this.properties = new HashMap<>(n.properties);
    }


    public NodeData(NodeData_thrift n, HashMap<Byte, Class> nodeClassMap) {
        // TODO Auto-generated constructor stub
        this.nodeId = n.nodeId;
        this.TS = n.TS;
        this.nodeClassId = n.nodeClassId;
        this.properties = new HashMap<String, Object>();

        for (Map.Entry<String, String> entry : n.properties.entrySet()) {

            Field field = null;
            try {
                Class current = nodeClassMap.get(n.nodeClassId);

                while (field == null) {
                    try {
                        field = current.getDeclaredField(entry.getKey());
                    } catch (NoSuchFieldException e) {
                        if (current.getSuperclass() != null) {
                            current = current.getSuperclass();
                        } else {
                            System.out.println("ERROR:: Could not find field for " + "Class name : " + current.getName() + ", entry : " + entry);
                            e.printStackTrace();
                            break;
                        }

                        //System.out.println("Class name : "+c.getName()+", entry : "+entry);
                        //System.out.println("All fields of this class : "+c.getDeclaredFields().length);
                        //for(Field f : c.getFields()){
                        //	System.out.print(f.getName()+", ");
                        //}
                        //System.out.println();
                        //e.printStackTrace();
                    }
                }
            } catch (SecurityException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            field.setAccessible(true);
            Class fieldType = field.getType();
            Object obj = stringToValue(fieldType, entry.getValue());
            properties.put(entry.getKey(), obj);
        }
    }

    private Object stringToValue(Class argClass, String argValue) {

        if (argValue == "") {
            return null;
        }

        switch (argClass.getSimpleName()) {

            case "String":
                return argValue;

            case "Boolean":
            case "boolean":
                switch (argValue.toLowerCase()) {
                    case "true":
                        return true;
                    default:
                        return false;
                }

            case "Integer":
            case "int":
                return Integer.parseInt(argValue);

            case "Long":
            case "long":
                return Long.parseLong(argValue);

            case "ConcurrentHashMap":    //considering this as the hashmap of neighbor edges.
                ConcurrentHashMap<String, Edge> neighborMap = new ConcurrentHashMap<String, Edge>();
                String str = argValue.replaceAll("[ }{=]", "");
                if (!str.equals("")) {
                    String mapItems[] = str.split(",");
                    for (String mapItem : mapItems) {
                        neighborMap.put(mapItem, new Edge());
                    }
                }
                return neighborMap;

            case "HashSet":
                HashSet<String> hSet = new HashSet<String>();
                String str2 = argValue.replaceAll("[ \\[\\]]", "");
                if (!str2.equals("")) {
                    String setItems[] = str2.split(",");
                    for (String setItem : setItems) {
                        hSet.add(setItem);
                    }
                }
                return hSet;

            default:
                System.out.println("Received NodeData_thrift (type : " + argClass.getSimpleName() + ") as : " + argValue);
                return null;
        }
    }


    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.nodeId = (String) in.readObject();
        this.TS = in.readLong();
        this.nodeClassId = (byte) in.readByte();

        String[] keys = (String[]) in.readObject();
        Object[] values = (Object[]) in.readObject();

        for (int i = 0; i < keys.length; i++) {
            this.properties.put(keys[i], values[i]);
        }
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeObject(this.nodeId);
        out.writeLong(this.TS);
        out.writeByte(this.nodeClassId);

        String[] keys = new String[properties.size()];
        Object[] values = new Object[properties.size()];

        int i = 0;
        for (String s : properties.keySet()) {
            keys[i] = s;
            values[i] = properties.get(s);
            i++;
        }
        out.writeObject(keys);
        out.writeObject(values);
    }


    public void addProperty(String property, Object value) {
        if (this.properties == null) this.properties = new HashMap<String, Object>();
        this.properties.put(property, value);
    }

    public Object getProperty(String key) {
        return this.properties.get(key);
    }

    public Vector<String> getProperties() {
        //implemented by the derived class
        return null;
    }


    @Override
    public String toString() {
        return "NodeData{" +
                "nodeId='" + nodeId + '\'' +
                ", TS=" + TS +
                ", properties=" + properties +
                '}';
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        NodeData nodeData = (NodeData) o;

        if (TS != nodeData.TS) return false;
        if (nodeClassId != nodeData.nodeClassId) return false;
        if (nodeId != null ? !nodeId.equals(nodeData.nodeId) : nodeData.nodeId != null) return false;
        return properties != null ? properties.equals(nodeData.properties) : nodeData.properties == null;

    }

    @Override
    public int hashCode() {
        int result = nodeId != null ? nodeId.hashCode() : 0;
        result = 31 * result + (int) (TS ^ (TS >>> 32));
        result = 31 * result + (int) nodeClassId;
        result = 31 * result + (properties != null ? properties.hashCode() : 0);
        return result;
    }
}





