/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

/*
 * This is a cache to store nodeData i.e cache only then relevant portions of the node.
 * We use a simple fifo based scheme. 
 * This avoids the invalidation overheard we had before
 * 
 */

public class NodeDataCache {
	public ConcurrentHashMap<String, NodeData> cacheTable; // actual cache
	int limit = 10000;
	float cleanupFraction = 0.1f;
	LinkedBlockingQueue<String> fifoqueue; // for ordering purposes

	public NodeDataCache(int nodeDataCacheSize, float nodeDataCacheCleanupFraction, float nodeDataCachePurgeThreshold, int nodeDataCachePurgePeriod){
		cacheTable = new ConcurrentHashMap<>();
		fifoqueue = new LinkedBlockingQueue<String>();
		limit = nodeDataCacheSize;
		//System.out.println("NodeDataCache : initialized with limit : "+limit);
		this.cleanupFraction = nodeDataCacheCleanupFraction; 
		NodeDataCachePurgeThread purgeThread = new NodeDataCachePurgeThread(cacheTable,fifoqueue,limit,nodeDataCachePurgeThreshold,nodeDataCachePurgePeriod);
		purgeThread.start();
	}


	public void addItem(NodeData nd){
		String id = nd.nodeId.toString();

		if(cacheTable.containsKey(id)){
			// put it at the start of the queue ?
			// synchronize these 2 lines 
			synchronized (fifoqueue) {
				fifoqueue.remove(id);
				fifoqueue.add(id);	
			}
			cacheTable.put(id, nd);
		}
		else{
			
			if(cacheTable.size() < limit){
				fifoqueue.add(id);
				cacheTable.put(id, nd);
			}
			else{
				// remove a fraction of the entries 
				int count = (int) ( (float)limit * cleanupFraction );
				//int count = 1;
				truncateItemByCount(count);
				fifoqueue.add(id);
				cacheTable.put(id, nd);
			}
		}

	}

	public NodeData getNodeData(String nodeId){
		
		if(cacheTable.containsKey(nodeId)){
			return cacheTable.get(nodeId);
		}
		return null;
	}

	public void removeItem(String nodeId){
		if(cacheTable.containsKey(nodeId)){
			cacheTable.remove(nodeId);
			fifoqueue.remove(nodeId);
		}
	}

	public synchronized void clearCache(){
		 // removeAll is not thread safe.  hence synchronize
			fifoqueue.removeAll(fifoqueue);
	}

	public void truncateItemByCount(int count){
		for (int i = 0; i < count; i++) {
			String s = fifoqueue.poll();
			if(s ==  null) continue;
			cacheTable.remove(s);
		}
	}
}


class NodeDataCachePurgeThread extends Thread {
	ConcurrentHashMap<String, NodeData> cacheTable;
	LinkedBlockingQueue<String> fifoqueue;
	int limit;
	int nodeDataCachePurgePeriod=5000;
	float nodeDataCachePurgeThreshold = 0.05f;
	
	public NodeDataCachePurgeThread(ConcurrentHashMap<String, NodeData> cacheTable, LinkedBlockingQueue<String> fifoqueue, int limit, float nodeDataCachePurgeThreshold, int nodeDataCachePurgePeriod) {
		// TODO Auto-generated constructor stub
		this.cacheTable = cacheTable ;
		this.fifoqueue = fifoqueue;
		this.limit = limit;
		this.nodeDataCachePurgeThreshold = nodeDataCachePurgeThreshold;
		this.nodeDataCachePurgePeriod = nodeDataCachePurgePeriod;
	}
	public void run(){
		while(true){
			int currentSize = cacheTable.size();
			//System.out.println("NodeDataCachePurgeThread : NodeDataCache size "+currentSize);
			int targetSize = (int) ( (float)limit * nodeDataCachePurgeThreshold );
			int count =  currentSize - targetSize;
			if(count > 0){
				//System.out.println("NodeDataCachePurgeThread : calling truncateItemByCount, count : "+count);
				truncateItemByCount(count);
			}
			try {
				Thread.sleep(nodeDataCachePurgePeriod);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	public void truncateItemByCount(int count){
		for (int i = 0; i < count; i++) {
			String s = fifoqueue.poll();
			if(s ==  null) continue;
			cacheTable.remove(s);
		}
	}
	
}

