/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import beehive.server.*;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.*;
import java.rmi.registry.*;
import java.util.*;
import java.io.*;
import java.util.concurrent.ConcurrentHashMap;


public class RemoteRequestHandler extends Thread {


    public HashtableServer serverHandle;
    public RequestedOp op;
    public Vector<Object> params;
    public Object result;
    public int status = -1;        // 0 equals failed, 1 equals success
    Barrier barrier;
    public int serverHandleId;
    public StorageSystem storageSys;

    public RemoteRequestHandler(int hostId, RequestedOp op,
                                Vector<Object> params, Barrier barrier, StorageSystem storageSys) {
        this.op = op;
        this.params = params;
        result = null;
        this.barrier = barrier;
        this.status = -1;
        this.storageSys = storageSys;

        if (StorageSystem.RPC_MECHANISM.equals("RMI")) {
            serverHandle = storageSys.servers.get(hostId);
        } else if (StorageSystem.RPC_MECHANISM.equals("THRIFT")) {
            serverHandle = storageSys.serversThrift.get(hostId);
        }
    }

    public void run() {
        int tries = 0;
        while (true) {
            try {
                if (op == RequestedOp.getNode) {
                    //TODO : Remove this? Not used anywhere
                    String param = (String) params.get(0);
                    result = serverHandle.get(param, -1);
                    status = 1;
                } else if (op == RequestedOp.getNodes) {
                    Vector<String> keyList = (Vector<String>) params.get(0);
                    result = serverHandle.getObjects(keyList);
                    status = 1;
                } else if (op == RequestedOp.putNode) {
                    String nodeId = (String) params.get(0);
                    Node node = (Node) params.get(1);
                    result = serverHandle.put(nodeId, node);
                    status = 1;
                } else if (op == RequestedOp.putNodes) {
                    ConcurrentHashMap<String, Node> bucket = (ConcurrentHashMap<String, Node>) params.get(0);
                    result = serverHandle.putObjects(bucket);
                    status = 1;
                }
            /*else if (op == RequestedOp.changeCore) {
                Integer CoreId = (Integer) params.get(0);
                int  coreId = CoreId.intValue();
				Vector<Integer> nodeList = (Vector<Integer>) params.get(1);
				serverHandle.setCoreNodeId(coreId, nodeList);
				status = 1;
            }*/
                else if (op == RequestedOp.updateNodeProperties) {
                    Vector<NodeData> bucket = (Vector<NodeData>) params.get(0);
                    serverHandle.updateNodePropertiesParallel(bucket);
                    status = 1;
                } else if (op == RequestedOp.invalidateCacheCopies) {
                    String nodeId = (String) params.get(0);
                    serverHandle.cacheInvalidate(nodeId);
                    status = 1;
                } else if (op == RequestedOp.getNodesForCaching) {
                    Vector<String> nodeIdsToCache = (Vector<String>) params.get(0);
                    int cachingSite = ((Integer) params.get(1)).intValue();
                    result = serverHandle.getNodesForCaching(nodeIdsToCache, cachingSite);
                    status = 1;
                } else if (op == RequestedOp.getNodeProperties) {
                    Vector<NodeData> keyList = (Vector<NodeData>) params.get(0);
                    result = serverHandle.getObjectsProperties(keyList);
                    status = 1;
                } else if (op == RequestedOp.exec) {
                    Vector<NodeData> keyList = (Vector<NodeData>) params.get(0);
                    result = serverHandle.execBulk(keyList);
                    status = 1;
                }
                break;

            } catch (ConnectException ce) {
                tries++;
                if (tries > 10) {
                    status = 0;
                    result = ce;
                    break;
                }
                // System.out.println("ConnectException:"+ce.getMessage());
                try {
                    Thread.sleep(10);
                } catch (Exception e) {
                }
            } catch (Exception e) {
                e.printStackTrace();
                status = 0;
                result = e;
                break;
            }
        }
        barrier.reportDone();
    }


}
