/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.util;

import java.lang.reflect.Method;
import java.net.InetAddress;
import java.net.URI;
import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.UnexpectedException;
import java.util.*;
import java.util.concurrent.*;

import beehive.checkpoint.CheckpointListener;
import beehive.server.HashtableServerImpl;
import org.apache.thrift.TException;

import beehive.server.HashtableServer;
import beehive.server.ResultObject;

public class StorageSystem implements CheckpointListener {

    private final static String LOG_TAG = StorageSystem.class.getSimpleName();

    public ConcurrentHashMap<Integer, HashtableServer> servers;
    ConcurrentHashMap<String, Integer> serverMapping;
    ConcurrentHashMap<String, String> locationTable;
    public ConcurrentHashMap<Integer, String> serverLocationMapping;
    private Vector<String> relocating;
    public String myURL;
    public int myId;
    public HashtableServer myHashTableServer;
    public int numServers;
    private boolean debug1 = false;
    private boolean useCaching;
    public boolean recordStats;
    private boolean writeThroughCache;
    private boolean canonicalHostNames = true;
    private int remotecall = Constants.SerialRemoteCall;

    /**
     * API call timing stats
     **/
    //single get calls
    private StatContainer statGetNodeTime;
    private StatContainer statGetNodeDataTime;
    //parallel get calls
    private StatContainer statGetNodeParallelTime;
    private StatContainer statGetNodeDataParallelTime;
    //parallel get call size
    private StatContainer statGetNodeParallelCallSize;
    private StatContainer statGetNodeDataParallelCallSize;

    //single put calls
    private StatContainer statPutNodeTime;
    private StatContainer statUpdateNodeDataTime;
    //parallel put calls
    private StatContainer statPutNodeParallelTime;
    private StatContainer statUpdateNodeDataParallelTime;
    //parallel put call size
    private StatContainer statPutNodeParallelCallSize;
    private StatContainer statUpdateNodeDataParallelCallSize;

    //time for getting remote nodes for caching and putting into cache table
    private StatContainer statGetNodeParallelTimeCachePart;

    private StatContainer statGetRemoteNodeAndCache;
    private StatContainer statGetRemoteNode;
    private StatContainer statGetRemoteNodeData;

    private StatContainer statGetLocalNodesTime;
    private StatContainer statGetParallelLocalNodesCount;
    private StatContainer statGetParallelBucketizationTime;

    private Logger log;

    private HostDispatchTable hostDispatchTable;

    private String HASHING_SCHEME = "INTEGER_MODULO";
    //partitioned graph nodes location table
    public int[] partitionedGraphHostNodesTable;
    //it stores the maximum node Id stored at this host

    ConcurrentHashMap<Integer, HashTableThriftHandler> serversThrift;
    public static String RPC_MECHANISM = "RMI";
    private static String VALIDATOR_MECHANISM = "RMI";

    public static HashMap<Byte, String> nodeClassNameMap = new HashMap<Byte, String>();
    public static HashMap<Byte, Class> nodeClassMap = new HashMap<Byte, Class>();
    public static HashMap<String, Byte> nodeClassNametoIDMap = new HashMap<String, Byte>();


    /**
     * Constructor without reference to local hashtable server object, backwards compatible
     *
     * @param hostList
     * @throws RemoteException
     */
    public StorageSystem(Vector<String> hostList, Config configParams) throws RemoteException {
        statGetNodeTime = new StatContainer();
        statGetNodeDataTime = new StatContainer();
        statPutNodeTime = new StatContainer();
        statUpdateNodeDataTime = new StatContainer();
        statGetNodeParallelTime = new StatContainer();
        statGetNodeDataParallelTime = new StatContainer();
        statPutNodeParallelTime = new StatContainer();
        statUpdateNodeDataParallelTime = new StatContainer();
        statGetNodeParallelTimeCachePart = new StatContainer();
        statGetNodeParallelCallSize = new StatContainer();
        statGetNodeDataParallelCallSize = new StatContainer();
        statPutNodeParallelCallSize = new StatContainer();
        statUpdateNodeDataParallelCallSize = new StatContainer();
        statGetRemoteNodeAndCache = new StatContainer();
        statGetRemoteNode = new StatContainer();
        statGetRemoteNodeData = new StatContainer();
        statGetLocalNodesTime = new StatContainer();
        statGetParallelLocalNodesCount = new StatContainer();
        statGetParallelBucketizationTime = new StatContainer();

        this.servers = new ConcurrentHashMap<>();
        this.serverMapping = new ConcurrentHashMap<>();
        this.locationTable = new ConcurrentHashMap<>();//not used in case of caching
        this.serverLocationMapping = new ConcurrentHashMap<Integer, String>();
        this.relocating = new Vector<>();
        System.setSecurityManager(new RMISecurityManager());
        log = new Logger();

        this.numServers = hostList.size();

        if (configParams.has("Server.canonicalHostNames")) {
            this.canonicalHostNames = configParams.getBoolValue("Server.canonicalHostNames");
        }
        if (configParams.has("StorageSystem.remoteCalls")) {
            this.remotecall = configParams.getIntValue("StorageSystem.remoteCalls");
        }
        System.out.println("Server.canonicalHostNames=" + this.canonicalHostNames);
        System.out.println("StorageSystem.remoteCalls=" +
                ((this.remotecall == Constants.ParallelRemoteCall) ? "ParallelRemoteCalls" :
                        ((this.remotecall == Constants.SerialRemoteCall) ? "SerialRemoteCalls" :
                                ((this.remotecall == Constants.ExecutorRemoteServiceCall) ? "ExecotorServiceRemoteCalls" : ""))));

        int i = 0;
        HashMap<Integer, String> serverNameMap = new HashMap<Integer, String>();
        for (String host : hostList) {
            try {
                String url = "//" + host + "/HashtableServer";
                String fullHostName;
                if (this.canonicalHostNames) {
                    fullHostName = InetAddress.getLocalHost().getCanonicalHostName();
                } else {
                    fullHostName = InetAddress.getLocalHost().getHostName();
                }

                HashtableServer server = (HashtableServer) Naming.lookup(url);
                if (fullHostName.equals(host)) {
                    this.myURL = url;
                    this.myId = i;
                    this.myHashTableServer = server;
                }
                this.servers.put(i, server);
                this.serverMapping.put(url, i);
                this.serverLocationMapping.put(i, url);

                serverNameMap.put(i, host);
                i++;
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        hostDispatchTable = new HostDispatchTable(this);

        partitionedGraphHostNodesTable = new int[numServers];
        if (configParams.has("Server.NODE_HASHING_SCHEME")) {
            this.HASHING_SCHEME = configParams.getStringValue("Server.NODE_HASHING_SCHEME");
        }

        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

        if (configParams.has("Validator.RPC_MECHANISM")) {
            VALIDATOR_MECHANISM = configParams.getStringValue("Validator.RPC_MECHANISM");
        }

        this.serversThrift = new ConcurrentHashMap<>();
        if (RPC_MECHANISM.equals("THRIFT")) {
            for (Integer hostid : serverNameMap.keySet()) {
                String hostname = serverNameMap.get(hostid);
                serversThrift.put(hostid, new HashTableThriftHandler(this, hostname, configParams));
            }
        }
    }


    /**
     * Constructor with reference to local hashtable server object
     *
     * @param hostList
     * @param paramHashtableServer
     * @param configParams
     * @param nodeClassNameMap     : Map containing Classnames of all the Node Inherited classes used in the program
     * @param nodeClassMap         : Map containing Classes of all the Node Inherited classes used in the program
     * @throws RemoteException
     */
    public StorageSystem(Vector<String> hostList, HashtableServer paramHashtableServer, Config configParams, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws RemoteException {
        this(hostList, configParams);
        int serverId = this.serverMapping.get(this.myURL);
        this.servers.put(serverId, paramHashtableServer);
        this.myHashTableServer = paramHashtableServer;

        StorageSystem.nodeClassNameMap = nodeClassNameMap;
        StorageSystem.nodeClassMap = nodeClassMap;
        for (Byte key : nodeClassNameMap.keySet()) {
            String value = nodeClassNameMap.get(key);
            nodeClassNametoIDMap.put(value, key);
        }
    }

    /**
     * Only called by BeehiveClientLoader class
     *
     * @param hostList
     * @param configParams
     * @param nodeClassNameMap
     * @param nodeClassMap
     * @throws RemoteException
     */
    public StorageSystem(Vector<String> hostList, Config configParams, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws RemoteException {
        this(hostList, configParams);

        StorageSystem.nodeClassNameMap = nodeClassNameMap;
        StorageSystem.nodeClassMap = nodeClassMap;

        for (Byte key : nodeClassNameMap.keySet()) {
            String value = nodeClassNameMap.get(key);
            nodeClassNametoIDMap.put(value, key);
        }
    }

    /**
     * Constructor for caching case
     *
     * @param hostList
     * @param paramHashtableServer
     * @param writeThrough
     * @throws RemoteException
     */
    public StorageSystem(Vector<String> hostList, HashtableServer paramHashtableServer, boolean writeThrough, Config configParams, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws RemoteException {
        this(hostList, paramHashtableServer, configParams, nodeClassNameMap, nodeClassMap);
        this.useCaching = true;
        this.writeThroughCache = writeThrough;
    }

    // TODO DOCUMENTATION
    public StorageSystem(Vector<String> hostList, boolean writeThrough, Config configParams, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws RemoteException {
        this(hostList, configParams, nodeClassNameMap, nodeClassMap);
        this.useCaching = true;
        this.writeThroughCache = writeThrough;
    }

    public int getNumberOfServers() {
        return numServers;
    }

    /**
     * Turn on stat recording
     */
    public void enableStats() {
        this.recordStats = true;
    }

	/* get/put/update*/

    /**
     * @param nodeId The id of the node that we want to fetch. Can be a local or remote id.
     * @return The node whose nodeId is passed in.
     * @throws RemoteException Thrown if the node could not be fetched.
     */
    public Node getNode(String nodeId) throws RemoteException {
        //Note down the time for the get call
        long start, end;
        Node node = null;
        start = System.currentTimeMillis();
        if (useCaching) {
            node = getNodeUseCache(nodeId);
            end = System.currentTimeMillis();
            recordStat(statGetNodeTime, end - start);
            return node;
        }
        HashtableServer server = getServer(nodeId);
        int serverId = getServerID(nodeId);

        long start_remote = System.currentTimeMillis();
        ResultObject localResultObject = server.get(nodeId, -1);
        long end_remote = System.currentTimeMillis();
        if (serverId != this.myId) {
            recordStat(statGetRemoteNode, end_remote - start_remote);
        }
        if (localResultObject.status == Constants.RELOCATED) {
            String str = localResultObject.location;
            this.locationTable.put(nodeId, str);
            serverId = this.serverMapping.get(str);
            server = this.servers.get(serverId);
            localResultObject = server.get(nodeId, -1);
        }
        if (localResultObject.status == Constants.SUCCESS) {
            node = (Node) localResultObject.node;
        }

        end = System.currentTimeMillis();
        recordStat(statGetNodeTime, end - start);
        return node;
    }

    /**
     * If {@code withCaching} is true and <i>caching</i> is enabled through the configuration, then an attempt to
     * get the node from the local cache will be made. Otherwise, the behavior is identical to {@link #getNode(String)}
     *
     * @param nodeId      The id of the node that we want to fetch. Can be a local or remote id.
     * @param withCaching True if an attempt to fetch from cache is desired, false otherwise.
     * @return The node whose nodeId is passed in.
     * @throws RemoteException Thrown if the node could not be fetched.
     */
    public Node getNode(String nodeId, boolean withCaching) throws RemoteException {
        //Note down the time for the get call
        long start, end;
        start = System.currentTimeMillis();
        if (useCaching && withCaching) {
            Node node = getNodeUseCache(nodeId);
            end = System.currentTimeMillis();
            recordStat(statGetNodeTime, end - start);
            return (node);
        } else {
            return getNode(nodeId);
        }
    }


    /**
     * Gets a node from the local cache. Only called if caching is enabled through the configuration and invoked
     * by {@link #getNode(String, boolean)} where true is passed in.
     *
     * @param nodeId The id of the node we want to fetch from cache. Can be a local or remote id.
     * @return The node whose nodeId is passed in.
     * @throws RemoteException Thrown if the node could not be fetched.
     */
    private Node getNodeUseCache(String nodeId) throws RemoteException {
        log.logDebug("***** StorageSystem : getNodeUseCache : get Node : " + nodeId + " *****");
        //get local hashtable server server
        HashtableServer localServer = this.servers.get(myId);
        ResultObject res;

        res = localServer.get(nodeId, Thread.currentThread().getId());
        if (res.statuscode.equals(Constants.FROM_MASTER)) {
            //this is the homesite server for this node
            if (res.node == null)
                log.logFatal("***** StorageSystem : getNodeUseCache : Got null object from local table ALERT!!! *****");
            return (Node) res.node;
        }
        if (res.statuscode.equals(Constants.FROM_CACHE)) {
            //was present in the server cache, not homesite though
            if (res.node == null)
                log.logFatal("***** StorageSystem : getNodeUseCache : Got null object from local cache ALERT!!! *****");
            return (Node) res.node;
        }

        //Node found neither in table nor in cache, caching needs to be initiated
        //Record the time it takes to get the node from remote and cache it
        long start_remote = System.currentTimeMillis();
        int k = 0;
        // If not present.
        Node nodeFromHomesite = null;
        if (res.statuscode.equals(Constants.INIT_CACHING)) {
            //Get the copy from the homesite
            HashtableServer server = getServer(nodeId);
            k = getServerID(nodeId);

            nodeFromHomesite = server.getNodeForCaching(nodeId, myId);
            if (nodeFromHomesite == null) {
                log.logFatal("***** StorageSystem : getNodeUseCache : Got null object from homesite ALERT!!! *****");
            } else {
                //now put this node in the local cache
                ResultObject resObj = localServer.cachePut(nodeId, nodeFromHomesite, k, Thread.currentThread().getId());
                if (resObj.statuscode.equals(Constants.CONCURRENT_INVALIDATION)) {
                    //Caching was unsuccessful, this can be used later to track num unsuccessful cache puts.
                    log.logWarning("***** StorageSystem : getNodeUseCache : Concurrrent invalidation - did not do a cache put ALERT *****");
                }
            }
        }
        long end_remote = System.currentTimeMillis();
        recordStat(statGetRemoteNodeAndCache, end_remote - start_remote);
        return nodeFromHomesite;
    }

    /**
     * Gets several Nodes whose ids are passed in; does so in a parallel fashion.
     *
     * @param nodeIds The ids of the desired nodes. The ids can be a mixture of local and remote ids.
     * @return All of the nodes associated with the ids passed in.
     * @throws RemoteException Thrown if <i>caching</i> is enabled through the configuration and the nodes could not
     *                         be fetched from the cache.
     */
    public Vector<Node> getNodesParallel(Vector<String> nodeIds) throws RemoteException {
        recordStat(statGetNodeParallelCallSize, nodeIds.size());
        long start = System.currentTimeMillis();
        if (useCaching) {
            Vector<Node> returnVec = getNodesParallelUseCache(nodeIds);
            long end = System.currentTimeMillis();
            recordStat(statGetNodeParallelTime, end - start);
            return returnVec;
        }

        //bucketize nodes according to server numbers
        ConcurrentHashMap<Integer, Vector<String>> serverIdToNodeIds = new ConcurrentHashMap<>();
        for (String nodeId : nodeIds) {
            int serverId = getServerID(nodeId);
            Vector<String> bucketizedNodeIds;
            if (!serverIdToNodeIds.containsKey(serverId)) {
                bucketizedNodeIds = new Vector<>();
            } else {
                bucketizedNodeIds = serverIdToNodeIds.get(serverId);
            }
            bucketizedNodeIds.add(nodeId);
            serverIdToNodeIds.put(serverId, bucketizedNodeIds);
        }

        Vector<Node> nodes = new Vector<>();
        Vector<ResultObject> resultingNodes = doParallelGetDynamic(serverIdToNodeIds);
        serverIdToNodeIds.clear();

        //obtain all required nodes from the corresponding server
        ConcurrentMap<Integer, Vector<String>> serverIdToRelocatedNodeIds = new ConcurrentHashMap<>();
        for (ResultObject res : resultingNodes) {
            int serverId;
            if (res.status == Constants.SUCCESS) {
                nodes.add((Node) res.node);
            } else if (res.status == Constants.RELOCATED) {
                String loc = res.location;
                serverId = this.serverMapping.get(loc);
                this.locationTable.put(res.nodeId, loc);

                Vector<String> bucketizedNodeIds;
                if (!serverIdToRelocatedNodeIds.containsKey(serverId)) {
                    bucketizedNodeIds = new Vector<>();
                } else {
                    bucketizedNodeIds = serverIdToRelocatedNodeIds.get(serverId);
                }
                bucketizedNodeIds.add(res.nodeId);
                serverIdToRelocatedNodeIds.put(serverId, bucketizedNodeIds);        //putting all relocated nodes to fetch from servers
            }
        }

        // Get Relocated nodes
        if (!serverIdToRelocatedNodeIds.isEmpty()) {
            Vector<ResultObject> relocatedResults = doParallelGetDynamic(serverIdToNodeIds);
            for (ResultObject res : relocatedResults) {
                if (res.status == Constants.SUCCESS) {
                    nodes.add((Node) res.node);
                }
            }
        }

        long end = System.currentTimeMillis();
        recordStat(statGetNodeParallelTime, end - start);
        return nodes;
    }

    /**
     * If {@code withCaching} is true and <it>caching</it> is enabled through the configuration, then an attempt to get
     * the nodes whose ids are passed in from the cache will be made. Otherwise, the behavior is identical to
     * {@link #getNodesParallel(Vector)}
     *
     * @param nodeIds     The ids of the nodes we want to fetch. Can be a mixture of local and remote ids.
     * @param withCaching True if an attempt to fetch from cache is desired, false otherwise.
     * @return All of the nodes whose ids are passed in.
     * @throws RemoteException Thrown only if caching is desired and the nodes could not be fetched from cache.
     */
    // TODO POSSIBLY REMOVE
    public Vector<Node> getNodesParallel(Vector<String> nodeIds, boolean withCaching) throws RemoteException {
        recordStat(statGetNodeParallelCallSize, nodeIds.size());
        long start = System.currentTimeMillis();
        if (useCaching && withCaching) {
            Vector<Node> returnVec = getNodesParallelUseCache(nodeIds);
            long end = System.currentTimeMillis();
            recordStat(statGetNodeParallelTime, end - start);
            return returnVec;
        }
        return getNodesParallel(nodeIds);
    }

    /**
     * A sub-routine to fetch the nodes from cache. Called only by {@link #getNodesParallel(Vector)} and
     * {@link #getNodesParallel(Vector, boolean)}
     *
     * @param nodeIds The ids of the nodes that we want to fetch from cache.
     * @return The nodes whose ids are passed in.
     * @throws RemoteException Thrown if the nodes could not be fetched from cache.
     */
    private Vector<Node> getNodesParallelUseCache(Vector<String> nodeIds) throws RemoteException {
        log.logDebug("***** StorageSystem : getNodesParallel : getting these nodes: " + nodeIds);
        Vector<Node> resultNodes = new Vector<Node>();

        // Get the local nodes and set the markers
        long start_getLocalNodes = System.currentTimeMillis();
        Vector<ResultObject> localNodes = myHashTableServer.getLocalNodes(nodeIds, Thread.currentThread().getId());
        long end_getLocalNodes = System.currentTimeMillis();
        recordStat(statGetLocalNodesTime, end_getLocalNodes - start_getLocalNodes);

        HashMap<Integer, Vector<String>> remoteNodeKeyMap = new HashMap<Integer, Vector<String>>();

        // bucketize the remoteNode nodes based on the servers.
        long start_bucket = System.currentTimeMillis();
        for (ResultObject res : localNodes) {
            if (res.statuscode.equals(Constants.FROM_MASTER) || res.statuscode.equals(Constants.FROM_CACHE)) {
                resultNodes.add((Node) res.node);
            } else if (res.statuscode.equals(Constants.INIT_CACHING)) {
                if (!remoteNodeKeyMap.containsKey(getServerID("" + res.nodeId))) {
                    remoteNodeKeyMap.put(getServerID("" + res.nodeId), new Vector<String>());
                }
                remoteNodeKeyMap.get(getServerID("" + res.nodeId)).add(new Integer(res.nodeId).toString());
            }
        }
        long end_bucket = System.currentTimeMillis();
        recordStat(statGetParallelBucketizationTime, end_bucket - start_bucket);

        //note down the number of local and remote nodes
        recordStat(statGetParallelLocalNodesCount, resultNodes.size());

        long start = System.currentTimeMillis();
        HashMap<Integer, Vector<Node>> remoteNodesMap = new HashMap<Integer, Vector<Node>>();

        //Using multi threaded request handlers with barrier
        //Varun : May 17, 2016. Reverting back to parallel calls in place of serial execution
        int keySetSize = remoteNodeKeyMap.keySet().size();
        if (keySetSize > 0) {
            Barrier barrier = new Barrier(keySetSize);

            Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
            for (Integer serverId : remoteNodeKeyMap.keySet()) {
                Vector<String> keys = remoteNodeKeyMap.get(serverId);
                Vector<Object> params = new Vector<Object>();
                params.add(keys);
                params.add(myId);
                RequestedOp requestedOp = RequestedOp.getNodesForCaching;
                RemoteRequestHandler handler = new RemoteRequestHandler(
                        serverId, requestedOp, params, barrier, this);
                handler.serverHandleId = serverId;
                handlerThreads.add(handler);
                handler.start();
            }
            try {
                barrier.barrierWait();
                for (RemoteRequestHandler handler : handlerThreads) {
                    if (handler.status == 1) {
                        Vector<Node> res = (Vector<Node>) handler.result;
                        if (!remoteNodesMap.containsKey(handler.serverHandleId)) {
                            remoteNodesMap.put(handler.serverHandleId, new Vector<Node>());
                        }
                        remoteNodesMap.get(handler.serverHandleId).addAll(res);
                    }
                }

            } catch (Exception exception) {
                exception.printStackTrace();
            }
        }

        // Write to cache and remove marker.
        log.logDebug("***** Storage System : getNodesParallel : Num nodes we need to fetch into cache from other servers are : " + remoteNodesMap.size());
        myHashTableServer.cachePut(remoteNodesMap, Thread.currentThread().getId());
        long end = System.currentTimeMillis();
        recordStat(statGetNodeParallelTimeCachePart, end - start);

        for (int serverId : remoteNodesMap.keySet())
            resultNodes.addAll(remoteNodesMap.get(serverId));

        log.logDebug("***** StorageSystem : getNodesParallel : result nodes being returned: " + resultNodes);
        //do a check, if the num nodes req = num of nodes returned. Also no node should be null
        if (nodeIds.size() != resultNodes.size())
            log.logFatal("***** StorageSystem : getNodesParallel : ALERT!!! Requested " + nodeIds.size() + " nodes and returning " + resultNodes.size() + " nodes *****");
        for (Node retNode : resultNodes) {
            if (retNode == null) {
                log.logFatal("***** StorageSystem : getNodesParallel : ALERT!!! One or more null nodes being returned *****");
                break;
            }
        }

        return resultNodes;
    }

    /**
     * This function is used to fetch the Bucketized remote nodes with options to get the nodes
     * through parallel RemoteRequestHandler calls or serial calls or using ExecutorService pool
     * <p>
     * Created by Rahul R. Sharma on Nov 4 2016
     *
     * @param keysMap
     * @return Vector<ResultObject>
     */
    private Vector<ResultObject> doParallelGetDynamic(ConcurrentHashMap<Integer, Vector<String>> keysMap) {
        if (useCaching)
            System.out.println("StorageSystem : doParallelGet : ALERT!!! This function should not be called for caching case");

        Vector<ResultObject> result = new Vector<ResultObject>();

        if (this.remotecall == Constants.ParallelRemoteCall) {
            Barrier barrier = new Barrier(keysMap.values().size());
            Enumeration<Integer> en = keysMap.keys();

            Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<String> keys = keysMap.get(id);
                Vector<Object> params = new Vector<Object>();
                params.add(keys);
                RequestedOp requestedOp = RequestedOp.getNodes;
                RemoteRequestHandler handler = new RemoteRequestHandler(id, requestedOp, params, barrier, this);
                handlerThreads.add(handler);
                handler.start();
            }
            try {
                barrier.barrierWait();
                for (RemoteRequestHandler handler : handlerThreads) {
                    if (handler.status == 1) {
                        Vector<ResultObject> res = (Vector<ResultObject>) handler.result;
                        result.addAll(res);
                    }
                }

            } catch (Exception exception) {
                exception.printStackTrace();
            }
        } else if (this.remotecall == Constants.SerialRemoteCall) {
            Enumeration<Integer> en = keysMap.keys();

            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<String> keys = keysMap.get(id);
                HashtableServer serverHandle = this.servers.get(id);
                Vector<ResultObject> res;
                try {
                    res = serverHandle.getObjects(keys);
                    result.addAll(res);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        } else if (this.remotecall == Constants.ExecutorRemoteServiceCall) {
            ExecutorService executorService = Executors.newFixedThreadPool(2);
            List<Future<Vector<ResultObject>>> futureList = new ArrayList<Future<Vector<ResultObject>>>();

            Enumeration<Integer> en = keysMap.keys();
            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<String> keys = keysMap.get(id);
                Vector<Object> params = new Vector<Object>();
                params.add(keys);
                RequestedOp requestedOp = RequestedOp.getNodes;
                RemoteRequestExecutorHandler handler = new RemoteRequestExecutorHandler(id, requestedOp, params, this);
                futureList.add(executorService.submit(handler));
            }

            for (Future<Vector<ResultObject>> f : futureList) {
                try {
                    Vector<ResultObject> res = f.get();
                    if (res != null) {
                        result.addAll(res);
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                } catch (ExecutionException e) {
                    e.printStackTrace();
                }
            }
            executorService.shutdown();
        }
        return result;
    }


    /**
     * Gets specific properties of a node
     *
     * @param n A {@link NodeData} object containing:
     *          <ul>
     *          <li> The nodeId of the node whose properties we want to fetch </li>
     *          <li> A mapping from desired properties to required/desired parameters </li>
     *          </ul>
     * @return A NodeData object containing the nodeId of the node and a mapping of the desired
     * properties to their respective values.
     * @throws RemoteException      Thrown if the properties could not be fetched
     * @throws InterruptedException Thrown if the properties could not be fetched
     */
    public NodeData getNodeProperties(NodeData n) throws RemoteException, InterruptedException {
        Vector<NodeData> vec = new Vector<NodeData>();
        vec.add(n);
        Vector<NodeData> resvec = getNodePropertiesParallel(vec);
        return resvec.get(0);
    }

    /**
     * Gets the desired properties of several nodes in parallel.
     *
     * @param nodeDatas Each {@link NodeData} object will contain
     *                  <ul>
     *                  <li> The nodeId of the node whose properties we want to fetch </li>
     *                  <li> A mapping from desired properties to required/desired parameters </li>
     *                  </ul>
     * @return Each NodeData object will contain the nodeId of the node and a mapping of the desired
     * properties to their respective values.
     * @throws RemoteException      Thrown if a node's properties could not be fetched.
     * @throws InterruptedException Thrown if a node's properties could not be fetched.
     */
    public Vector<NodeData> getNodePropertiesParallel(Vector<NodeData> nodeDatas) throws RemoteException, InterruptedException {
        Vector<NodeData> resultVector = new Vector<NodeData>();
        ConcurrentHashMap<Integer, Vector<NodeData>> serverIdToNodeData = new ConcurrentHashMap<>();

        if (nodeDatas.size() == 0) {
            return resultVector;
        }

        //only for remote nodes, for local nodes we get the complete node
        recordStat(statGetNodeDataParallelCallSize, nodeDatas.size());
        long start = System.currentTimeMillis();

        //bucketize nodes according to server numbers
        for (NodeData nd : nodeDatas) {
            int serverId = getServerID(nd.nodeId);
            Vector<NodeData> bucketizedNodeData;
            if (!serverIdToNodeData.containsKey(serverId)) {
                bucketizedNodeData = new Vector<>();
            } else {
                bucketizedNodeData = serverIdToNodeData.get(serverId);
            }
            bucketizedNodeData.add(nd);
            serverIdToNodeData.put(serverId, bucketizedNodeData);
        }

        int reqcount = nodeDatas.size();
        WorkerRequest wrkReq = new WorkerRequest(reqcount);
        hostDispatchTable.doGetNodeData(wrkReq, serverIdToNodeData);
        wrkReq.waitForResult();
        resultVector = wrkReq.result;
        hostDispatchTable.clearWorkerRequest(Thread.currentThread().getId());
        long end = System.currentTimeMillis();
        recordStat(statGetNodeDataParallelTime, end - start);
        return resultVector;
    }

    /**
     * NOT IN USE
     * A sub-routine that fetches the desired properties of each node in a parallel fashion.
     *
     * @param serverIdToNodeData A mapping from serverIds to NodeDatas we want to fetch.
     * @return A vector of ResultObjects, each describing the result of the get and containing the Node.
     * If a node could not be fetched, then its respective {@link ResultObject} will have a null node.
     * @deprecated
     */
    @Deprecated
    private Vector<ResultObject> doParallelGetProperties(ConcurrentHashMap<Integer, Vector<NodeData>> serverIdToNodeData) {

        Barrier barrier = new Barrier(serverIdToNodeData.values().size());
        Enumeration<Integer> en = serverIdToNodeData.keys();

        Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
        while (en.hasMoreElements()) {
            int serverId = en.nextElement();
            Vector<NodeData> nodeData = serverIdToNodeData.get(serverId);
            Vector<Object> params = new Vector<Object>();
            params.add(nodeData);
            RequestedOp requestedOp = RequestedOp.getNodeProperties;
            RemoteRequestHandler handler = new RemoteRequestHandler(serverId, requestedOp, params, barrier, this);
            handlerThreads.add(handler);
            handler.start();
        }
        Vector<ResultObject> result = new Vector<ResultObject>();
        try {
            barrier.barrierWait();
            for (RemoteRequestHandler handler : handlerThreads) {
                if (handler.status == 1) {
                    Vector<ResultObject> res = (Vector<ResultObject>) handler.result;
                    result.addAll(res);
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * This function is used to fetch the Bucketized remote nodes with options to get the nodes
     * through parallel RemoteRequestHandler calls or serial calls or using ExecutorService pool
     * <p>
     * Created by Rahul R. Sharma on Nov 4 2016
     *
     * @param keysMap
     * @return Vector<ResultObject>
     * @deprecated
     */
    @Deprecated
    private Vector<ResultObject> doParallelGetPropertiesDynamic(ConcurrentHashMap<Integer, Vector<NodeData>> keysMap) {
        Vector<ResultObject> result = new Vector<ResultObject>();

        if (this.remotecall == Constants.ParallelRemoteCall) {
            Barrier barrier = new Barrier(keysMap.values().size());
            Enumeration<Integer> en = keysMap.keys();

            Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<NodeData> keys = keysMap.get(id);
                Vector<Object> params = new Vector<Object>();
                params.add(keys);
                RequestedOp requestedOp = RequestedOp.getNodeProperties;
                RemoteRequestHandler handler = new RemoteRequestHandler(id, requestedOp, params, barrier, this);
                handlerThreads.add(handler);
                handler.start();
            }
            try {
                barrier.barrierWait();
                for (RemoteRequestHandler handler : handlerThreads) {
                    if (handler.status == 1) {
                        Vector<ResultObject> res = (Vector<ResultObject>) handler.result;
                        result.addAll(res);
                    }
                }

            } catch (Exception exception) {
                exception.printStackTrace();
            }
        } else if (this.remotecall == Constants.SerialRemoteCall) {
            Enumeration<Integer> en = keysMap.keys();

            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<NodeData> keys = keysMap.get(id);
                HashtableServer serverHandle = (HashtableServer) this.servers.get(id);
                Vector<ResultObject> res;
                try {
                    res = serverHandle.getObjectsProperties(keys);
                    result.addAll(res);
                } catch (Exception exception) {
                    exception.printStackTrace();
                }
            }
        } else if (this.remotecall == Constants.ExecutorRemoteServiceCall) {
            ExecutorService executorService = Executors.newFixedThreadPool(2);
            List<Future<Vector<ResultObject>>> futureList = new ArrayList<Future<Vector<ResultObject>>>();

            Enumeration<Integer> en = keysMap.keys();
            while (en.hasMoreElements()) {
                int id = en.nextElement();
                Vector<NodeData> keys = keysMap.get(id);
                Vector<Object> params = new Vector<Object>();
                params.add(keys);
                RequestedOp requestedOp = RequestedOp.getNodeProperties;
                RemoteRequestExecutorHandler handler = new RemoteRequestExecutorHandler(id, requestedOp, params, this);
                futureList.add(executorService.submit(handler));
            }

            for (Future<Vector<ResultObject>> f : futureList) {
                try {
                    Vector<ResultObject> res = f.get();
                    if (res != null) {
                        result.addAll(res);
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                } catch (ExecutionException e) {
                    e.printStackTrace();
                }
            }
            executorService.shutdown();
        }
        return result;
    }

    /**
     * Remotely executes the methods of a node. If the node is local then the methods are executed locally.
     *
     * @param n Contains the nodeId of the node whose methods are to be executed. The methods that are to be
     *          executed are provided as a mapping of method name to parameters.
     * @return A NodeData object containing the results of the method calls. The NodeData object represents the results
     * as a mapping from method name to return value.
     * @throws RemoteException
     */
    public NodeData exec(NodeData n) throws RemoteException {
        long start, end;
        NodeData nodeData = null;
        start = System.currentTimeMillis();

        HashtableServer server = getServer(n.nodeId);
        int serverId = getServerID(n.nodeId);

        long start_remote = System.currentTimeMillis();
        ResultObject localResultObject = server.exec(n);
        long end_remote = System.currentTimeMillis();
        if (serverId != this.myId) {
            recordStat(statGetRemoteNodeData, end_remote - start_remote);
        }
        if (localResultObject.status == Constants.RELOCATED) {
            String location = localResultObject.location;
            this.locationTable.put(n.nodeId, location);
            int relocatedServerId = this.serverMapping.get(location);
            server = this.servers.get(relocatedServerId);
            localResultObject = server.exec(n);
        }
        if (localResultObject.status == Constants.SUCCESS) {
            nodeData = (NodeData) localResultObject.node;
        }

        end = System.currentTimeMillis();
        recordStat(statGetNodeDataTime, end - start);
        return nodeData;
    }

    public Vector<NodeData> execParallel(Vector<NodeData> nodeDataVector) throws RemoteException {
        //only for remote nodes, for local nodes we get the complete node
        recordStat(statGetNodeDataParallelCallSize, nodeDataVector.size());
        long start = System.currentTimeMillis();

        Vector<NodeData> resultVector = new Vector<>();
        ConcurrentHashMap<Integer, Vector<NodeData>> localHashtable = new ConcurrentHashMap<>();

        // Bucketize nodeData according to their serverId
        for (NodeData nodeData : nodeDataVector) {
            int serverId = getServerID(nodeData.nodeId);
            Vector<NodeData> serverNodeData;
            if (!localHashtable.containsKey(serverId)) {
                serverNodeData = new Vector<>();
            } else {
                serverNodeData = localHashtable.get(serverId);
            }
            serverNodeData.add(nodeData);//adding node data object and not string
            localHashtable.put(serverId, serverNodeData);
        }

        Vector<ResultObject> resultingNodeData = doParallelExec(localHashtable);
        localHashtable.clear();

        // Bucketize relocated nodedata
        ConcurrentHashMap<Integer, Vector<NodeData>> relocatedNodeData = new ConcurrentHashMap<>();
        for (ResultObject resultObject : resultingNodeData) {
            if (resultObject.status == Constants.SUCCESS) {
                NodeData nodeData = (NodeData) resultObject.node;
                resultVector.add(nodeData);
            } else if (resultObject.status == Constants.RELOCATED) {
                String location = resultObject.location;
                int relocatedServerId = this.serverMapping.get(location);
                this.locationTable.put(resultObject.nodeId, location);

                Vector<NodeData> serverNodeData;
                if (!relocatedNodeData.contains(relocatedServerId)) {
                    serverNodeData = new Vector<>();
                } else {
                    serverNodeData = relocatedNodeData.get(relocatedServerId);
                }
                serverNodeData.add((NodeData) resultObject.node);
                relocatedNodeData.put(relocatedServerId, serverNodeData);
            }
        }

        if (!relocatedNodeData.isEmpty()) {
            Vector<ResultObject> relocatedResults = doParallelExec(relocatedNodeData);
            for (ResultObject relocatedResult : relocatedResults) {
                if (relocatedResult.status == Constants.SUCCESS) {
                    NodeData nodeData = (NodeData) relocatedResult.node;
                    resultVector.add(nodeData);
                }
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statGetNodeDataParallelTime, end - start);
        return resultVector;
    }

    /**
     * Executes the methods of each node in parallel.
     *
     * @param keysMap A mapping from serverIds to a vector of NodeData whose nodes are stored on that particular server
     * @return A vector of {@link ResultObject}s describing the results of the execution on each server.
     */
    private Vector<ResultObject> doParallelExec(ConcurrentHashMap<Integer, Vector<NodeData>> keysMap) {

        Barrier barrier = new Barrier(keysMap.values().size());
        Enumeration<Integer> en = keysMap.keys();

        Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
        while (en.hasMoreElements()) {
            int id = en.nextElement();
            Vector<NodeData> keys = keysMap.get(id);
            Vector<Object> params = new Vector<Object>();
            params.add(keys);
            RequestedOp requestedOp = RequestedOp.exec;
            RemoteRequestHandler handler = new RemoteRequestHandler(id, requestedOp, params, barrier, this);
            handlerThreads.add(handler);
            handler.start();
        }
        Vector<ResultObject> result = new Vector<ResultObject>();
        try {
            barrier.barrierWait();
            for (RemoteRequestHandler handler : handlerThreads) {
                if (handler.status == 1) {
                    Vector<ResultObject> res = (Vector<ResultObject>) handler.result;
                    result.addAll(res);
                }
            }

        } catch (Exception exception) {
            exception.printStackTrace();
        }
        return result;
    }

    /**
     * Puts a node into the local server or a remote server based on the {@code nodeId}
     *
     * @param nodeId The identification of the node that is to be put into the server
     * @param node   The node that is to be put into the server
     * @throws RemoteException Thrown if the node could not be put into the server
     */
    public boolean putNode(String nodeId, Node node) throws RemoteException {
        boolean result = true;

        //Note down the time for the put call
        long start, end;
        start = System.currentTimeMillis();
        if (useCaching) {
            putNodeUseCache(nodeId, node);
            end = System.currentTimeMillis();
            recordStat(statPutNodeTime, end - start);
            return result;    //check the status of putNodeCache() method and return the corresponding status
        }
        HashtableServer hashtableServer = getServer(nodeId);

        ResultObject localResultObject = hashtableServer.put(nodeId, node);
        if (localResultObject.status == Constants.RELOCATED) {
            String str = localResultObject.location;
            this.locationTable.put(nodeId, str);
            hashtableServer = getServer(nodeId);
            localResultObject = hashtableServer.put(nodeId, node);
        }

        if (localResultObject.status != Constants.SUCCESS) {
            result = false;
        }
        end = System.currentTimeMillis();
        recordStat(statPutNodeTime, end - start);

        return result;
    }

    /**
     * Puts a node into the local cache table based on the {@code nodeId}. Not accessible from the application and
     * is dependent on the configuration.
     *
     * @param nodeId The identification of the node that is to be put into the cache
     * @param node   The node that is to be put into the cache
     * @throws RemoteException Thrown of the node could not be put into the cache
     */
    private void putNodeUseCache(String nodeId, Node node) throws RemoteException {
        log.logDebug("***** StorageSystem : putNodesUseCache : putting node key : " + nodeId);
        HashtableServer queryServer = getServer(nodeId);
        int serverId = getServerID(nodeId);

        if (this.writeThroughCache && myId != serverId) {
            //first write through to local cache
            myHashTableServer.writeThroughToCache(nodeId, node);
            queryServer.put(nodeId, node, myId);
        } else {
            queryServer.put(nodeId, node);
        }
    }

    /**
     * Bulk call to putNode, called by the application
     * Calls are bucketised to the respective servers
     *
     * @param nodeList The nodes that are to be put into the servers.
     * @throws RemoteException Thrown if a node could not be put into its server
     */
    public void putNodesParallel(Vector<Node> nodeList) throws RemoteException {
        recordStat(statPutNodeParallelCallSize, nodeList.size());
        //Note down the time for the parallel put call
        log.logDebug("***** StorageSystem : putNodesParallel : putting the nodes : " + nodeList.toString());
        long start = System.currentTimeMillis();
        ConcurrentHashMap<Integer, ConcurrentHashMap<String, Node>> serverIdToNodes = new ConcurrentHashMap<Integer, ConcurrentHashMap<String, Node>>();

        for (Node node : nodeList) {
            //group the nodes according to the servers they reside in
            int serverId = getServerID(node.nodeId);

            ConcurrentHashMap<String, Node> nodeIdToNodeMap;
            if (!serverIdToNodes.containsKey(serverId)) {
                nodeIdToNodeMap = new ConcurrentHashMap<>();
            } else {
                nodeIdToNodeMap = serverIdToNodes.get(serverId);
            }
            nodeIdToNodeMap.put(node.nodeId, node);
            serverIdToNodes.put(serverId, nodeIdToNodeMap);
        }

        Vector<ResultObject> resultObjects = doParallelPutDynamic(serverIdToNodes);
        ConcurrentHashMap<Integer, ConcurrentHashMap<String, Node>> relocatedNodes = new ConcurrentHashMap<Integer, ConcurrentHashMap<String, Node>>();

        for (ResultObject resultObject : resultObjects) {
            //Some nodes have relocated, so the put was fail
            if (resultObject.status == Constants.RELOCATED) {
                String location = resultObject.location;
                //whats the new location
                int relocatedServerId = this.serverMapping.get(location);
                this.locationTable.put(resultObject.nodeId, location);

                ConcurrentHashMap<String, Node> nodeIdToNodeMap;
                if (!relocatedNodes.containsKey(relocatedServerId)) {
                    nodeIdToNodeMap = new ConcurrentHashMap<String, Node>();
                } else {
                    nodeIdToNodeMap = relocatedNodes.get(relocatedServerId);
                }

                int n = getServerID(resultObject.nodeId);
                Node node = (serverIdToNodes.get(n)).get(resultObject.nodeId);

                nodeIdToNodeMap.put(resultObject.nodeId, node);
                relocatedNodes.put(relocatedServerId, nodeIdToNodeMap);
            }
        }
        if (!relocatedNodes.isEmpty()) {
            resultObjects = doParallelPutDynamic(relocatedNodes);
            for (ResultObject resultObject : resultObjects) {
                //TODO : What happens when this call also finds relocated nodes? What do you do then?
                // if (resultObject.status == 1)
                // //System.out.println("FATAL ERROR (StorageSystem.getNodesParallel) unexpected error in putting some nodes");
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statPutNodeParallelTime, end - start);
    }

    /**
     * This function is used to put the Bucketized remote nodes with options to put the nodes
     * through parallel RemoteRequestHandler calls or serial calls or using ExecutorService pool
     * <p>
     * Created by Rahul R. Sharma on Nov 4 2016
     *
     * @return Vector<ResultObject>
     */
    private Vector<ResultObject> doParallelPutDynamic(ConcurrentHashMap<Integer, ConcurrentHashMap<String, Node>> nodesMap) {
        Vector<ResultObject> resultList = new Vector<ResultObject>();
        if (useCaching && writeThroughCache) {
            Enumeration<Integer> localEnumeration = nodesMap.keys();

            while (localEnumeration.hasMoreElements()) {
                int id = localEnumeration.nextElement();
                ConcurrentHashMap<String, Node> nodeTable = nodesMap.get(id);

                HashtableServer serverHandle = (HashtableServer) this.servers.get(id);
                try {
                    if (this.writeThroughCache && myId != id) {
                        myHashTableServer.writeThroughToCacheBulk(nodeTable);
                        resultList.addAll(serverHandle.putObjects(nodeTable, myId));
                    } else {
                        resultList.addAll(serverHandle.putObjects(nodeTable));
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            return resultList;
        } else {
            if (this.remotecall == Constants.ParallelRemoteCall) {
                Enumeration<Integer> localEnumeration = nodesMap.keys();
                Barrier barrier = new Barrier(nodesMap.values().size());

                Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
                while (localEnumeration.hasMoreElements()) {
                    int id = localEnumeration.nextElement();
                    ConcurrentHashMap<String, Node> nodeTable = nodesMap.get(id);

                    RequestedOp op = RequestedOp.putNodes;
                    Vector<Object> params = new Vector<Object>();
                    params.add(nodeTable);
                    RemoteRequestHandler remoteRequestHandler = new RemoteRequestHandler(id, op, params, barrier, this);
                    handlerThreads.add(remoteRequestHandler);
                    remoteRequestHandler.start();
                }

                try {
                    barrier.barrierWait();
                    for (RemoteRequestHandler handler : handlerThreads) {
                        if (handler.status == 1) {
                            Vector<ResultObject> results = (Vector<ResultObject>) handler.result;
                            resultList.addAll(results);
                        }
                    }

                } catch (Exception localException) {
                    localException.printStackTrace();
                }
            } else if (this.remotecall == Constants.SerialRemoteCall) {
                Enumeration<Integer> localEnumeration = nodesMap.keys();
                while (localEnumeration.hasMoreElements()) {
                    int id = localEnumeration.nextElement();
                    ConcurrentHashMap<String, Node> nodeTable = nodesMap.get(id);

                    HashtableServer serverHandle = this.servers.get(id);
                    Vector<ResultObject> res;
                    try {
                        res = serverHandle.putObjects(nodeTable);
                        resultList.addAll(res);
                    } catch (Exception exception) {
                        exception.printStackTrace();
                    }

                }
            } else if (this.remotecall == Constants.ExecutorRemoteServiceCall) {
                ExecutorService executorService = Executors.newFixedThreadPool(2);
                List<Future<Vector<ResultObject>>> futureList = new ArrayList<Future<Vector<ResultObject>>>();

                Enumeration<Integer> localEnumeration = nodesMap.keys();
                while (localEnumeration.hasMoreElements()) {
                    int id = localEnumeration.nextElement();
                    ConcurrentHashMap<String, Node> nodeTable = nodesMap.get(id);

                    RequestedOp op = RequestedOp.putNodes;
                    Vector<Object> params = new Vector<Object>();
                    params.add(nodeTable);
                    RemoteRequestExecutorHandler remoteRequestHandler = new RemoteRequestExecutorHandler(id, op, params, this);
                    futureList.add(executorService.submit(remoteRequestHandler));
                }

                for (Future<Vector<ResultObject>> f : futureList) {
                    try {
                        Vector<ResultObject> res = f.get();
                        if (res != null) {
                            resultList.addAll(res);
                        }
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    } catch (ExecutionException e) {
                        e.printStackTrace();
                    }
                }
                executorService.shutdown();
            }
        }
        return resultList;
    }

    /**
     * update Reflection API for the storage system
     *
     * @param n - this is a Node data which tells what the updated properties and TS are
     * @return Returns the same node data but the property map now contains the updates that failed
     * if property map = null, then nothing failed, ALL OK
     * if not null, then it contains exception objects for the failures
     * @throws RemoteException
     * @throws TException
     */
    public boolean updateNodeProperties(NodeData n) throws RemoteException, TException {
        //Note down the time for the put call
        long start, end;
        start = System.currentTimeMillis();

        HashtableServer serverHandle = null;
        if (StorageSystem.RPC_MECHANISM.equals("RMI")) {
            serverHandle = getServer(n.nodeId);
        } else if (StorageSystem.RPC_MECHANISM.equals("THRIFT")) {
            serverHandle = getServerThrift(n.nodeId);
        }

        if (serverHandle == null) {
            throw new UnexpectedException(LOG_TAG + ": HashtableServer handle is null");
        }

        ResultObject localResultObject = serverHandle.updateNodeProperties(n);
        if (localResultObject.status == Constants.RELOCATED) { //Relocated
            String location = localResultObject.location;
            this.locationTable.put(n.nodeId, location);
            int serverId = this.serverMapping.get(location);

            if (StorageSystem.RPC_MECHANISM.equals("RMI")) {
                serverHandle = servers.get(serverId);
            } else if (StorageSystem.RPC_MECHANISM.equals("THRIFT")) {
                serverHandle = serversThrift.get(serverId);
            }
            localResultObject = serverHandle.updateNodeProperties(n);
        }

        end = System.currentTimeMillis();
        recordStat(statUpdateNodeDataTime, end - start);
        return localResultObject.status == Constants.SUCCESS;
    }

    /**
     * updateNodeProperties for the bulk case
     *
     * @param nodeDataVector
     * @throws RemoteException
     */
    public void updateNodePropertiesParallel(Vector<NodeData> nodeDataVector) throws RemoteException {
        recordStat(statUpdateNodeDataParallelCallSize, nodeDataVector.size());
        //Note down the time for the parallel put call
        log.logDebug("***** StorageSystem : putNodesParallel : putting the nodes : " + nodeDataVector.toString());
        long start = System.currentTimeMillis();
        ConcurrentHashMap<Integer, ConcurrentHashMap<String, NodeData>> serverIdToNodeDataMap = new ConcurrentHashMap<Integer, ConcurrentHashMap<String, NodeData>>();

        for (NodeData node : nodeDataVector) {
            //group the nodes according to the servers they reside in
            int serverId = getServerID(node.nodeId);

            ConcurrentHashMap<String, NodeData> serverNodes;
            if (serverIdToNodeDataMap.containsKey(serverId)) {
                serverNodes = serverIdToNodeDataMap.get(serverId);
            } else {
                serverNodes = new ConcurrentHashMap<>();
            }
            serverNodes.put(node.nodeId, node);
            serverIdToNodeDataMap.put(serverId, serverNodes);
        }

        Vector<ResultObject> resultVector = doParallelPutPropertiesDynamic(serverIdToNodeDataMap);
        ConcurrentHashMap<Integer, ConcurrentHashMap<String, NodeData>> relocatedServerNodes = new ConcurrentHashMap<Integer, ConcurrentHashMap<String, NodeData>>();

        for (ResultObject resultObject : resultVector) {
            //Some nodes have relocated, so the put was fail
            if (resultObject.status == Constants.RELOCATED) {
                String location = resultObject.location;
                //whats the new location
                int relocatedServerId = this.serverMapping.get(location);
                this.locationTable.put(resultObject.nodeId, location);

                ConcurrentHashMap<String, NodeData> serverNodes;
                if (!relocatedServerNodes.containsKey(relocatedServerId)) {
                    serverNodes = new ConcurrentHashMap<>();
                } else {
                    serverNodes = relocatedServerNodes.get(relocatedServerId);
                }
                String m = resultObject.nodeId;
                int n = getServerID(m);
                NodeData node = (serverIdToNodeDataMap.get(n)).get(m);
                serverNodes.put(m, node);
                relocatedServerNodes.put(relocatedServerId, serverNodes);
            }
        }
        if (!relocatedServerNodes.isEmpty()) {
            resultVector = doParallelPutPropertiesDynamic(relocatedServerNodes);
            for (ResultObject resultObject : resultVector) {
                //TODO : What happens when this call also finds relocated nodes? What do you do then?
                // if (resultObject.status == 1)
                // //System.out.println("FATAL ERROR (StorageSystem.getNodesParallel) unexpected error in putting some nodes");
            }
        }
        long end = System.currentTimeMillis();
        recordStat(statUpdateNodeDataParallelTime, end - start);
    }

    /**
     * This function is used to put the Bucketized remote nodedata with options to put the nodedata
     * through parallel RemoteRequestHandler calls or serial calls or using ExecutorService pool
     * <p>
     * Created by Rahul R. Sharma on Nov 4 2016
     *
     * @return Vector<ResultObject>
     */
    private Vector<ResultObject> doParallelPutPropertiesDynamic(ConcurrentHashMap<Integer, ConcurrentHashMap<String, NodeData>> serverIdToUpdatedNodeDatas) {
        Vector<ResultObject> resultList = new Vector<ResultObject>();

        if (this.remotecall == Constants.ParallelRemoteCall) {
            Enumeration<Integer> localEnumeration = serverIdToUpdatedNodeDatas.keys();
            Barrier barrier = new Barrier(serverIdToUpdatedNodeDatas.values().size());

            Vector<RemoteRequestHandler> handlerThreads = new Vector<RemoteRequestHandler>();
            Vector<NodeData> argVec = new Vector<NodeData>();
            while (localEnumeration.hasMoreElements()) {
                int id = localEnumeration.nextElement();
                ConcurrentHashMap<String, NodeData> nodeTable = serverIdToUpdatedNodeDatas.get(id);
                for (String key : nodeTable.keySet()) {
                    argVec.add(nodeTable.get(key));
                }
                RequestedOp op = RequestedOp.updateNodeProperties;
                Vector<Object> params = new Vector<Object>();
                params.add(argVec);
                RemoteRequestHandler remoteRequestHandler = new RemoteRequestHandler(id, op, params, barrier, this);
                handlerThreads.add(remoteRequestHandler);
                remoteRequestHandler.start();
            }

            try {
                barrier.barrierWait();
                for (RemoteRequestHandler handler : handlerThreads) {
                    if (handler.status == 1) {
                        Vector<ResultObject> results = (Vector<ResultObject>) handler.result;
                        resultList.addAll(results);
                    }
                }

            } catch (Exception localException) {
                localException.printStackTrace();
            }
        } else if (this.remotecall == Constants.SerialRemoteCall) {
            Enumeration<Integer> localEnumeration = serverIdToUpdatedNodeDatas.keys();
            Vector<NodeData> argVec = null;

            while (localEnumeration.hasMoreElements()) {
                int id = localEnumeration.nextElement();
                ConcurrentHashMap<String, NodeData> nodeTable = serverIdToUpdatedNodeDatas.get(id);
                argVec = new Vector<NodeData>();
                for (String key : nodeTable.keySet()) {
                    argVec.add(nodeTable.get(key));
                }

                HashtableServer serverHandle = null;
                if (RPC_MECHANISM.equals("RMI")) {
                    serverHandle = servers.get(id);
                } else if (RPC_MECHANISM.equals("THRIFT")) {
                    serverHandle = serversThrift.get(id);
                }

                Vector<ResultObject> res = null;
                try {
                    res = serverHandle.updateNodePropertiesParallel(argVec);
                    resultList.addAll(res);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        } else if (this.remotecall == Constants.ExecutorRemoteServiceCall) {
            ExecutorService executorService = Executors.newFixedThreadPool(2);
            List<Future<Vector<ResultObject>>> futureList = new ArrayList<Future<Vector<ResultObject>>>();

            Enumeration<Integer> localEnumeration = serverIdToUpdatedNodeDatas.keys();
            Vector<NodeData> argVec = new Vector<NodeData>();
            while (localEnumeration.hasMoreElements()) {
                int id = localEnumeration.nextElement();
                ConcurrentHashMap<String, NodeData> nodeTable = serverIdToUpdatedNodeDatas.get(id);
                for (String key : nodeTable.keySet()) {
                    argVec.add(nodeTable.get(key));
                }
                RequestedOp op = RequestedOp.updateNodeProperties;
                Vector<Object> params = new Vector<Object>();
                params.add(argVec);
                RemoteRequestExecutorHandler remoteRequestHandler = new RemoteRequestExecutorHandler(id, op, params, this);
                futureList.add(executorService.submit(remoteRequestHandler));
            }
            for (Future<Vector<ResultObject>> f : futureList) {
                try {
                    Vector<ResultObject> res = f.get();
                    if (res != null) {
                        resultList.addAll(res);
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                } catch (ExecutionException e) {
                    e.printStackTrace();
                }
            }
            executorService.shutdown();
        }
        return resultList;
    }


    public Vector<String> getAllKeys(String url) throws RemoteException {
        int serverId = this.serverMapping.get(url);
        HashtableServer hashtableServer = this.servers.get(serverId);
        return hashtableServer.getAllKeys();
    }


    public boolean relocate(String nodeId) throws RemoteException {
        return relocate(nodeId, false);
    }

    public synchronized boolean relocate(String nodeId, boolean lock) throws RemoteException {
        int id = Integer.parseInt(nodeId);
        int mysrvid = this.serverMapping.get(this.myURL);
        HashtableServer myHashtableServer = this.servers.get(mysrvid);
        HashtableServer remoteHashtableServer = null;
        String loc = null;

        int default_srv = getServerID(nodeId);
        ResultObject resultObject = null;

        if (myHashtableServer.contains(nodeId)) {
            return true;
        }

        int attempt = 0;
        int relocatingFrom = -1;
        int k = 0;
        this.relocating.add(nodeId);

        if (this.locationTable.containsKey(nodeId)) {
            loc = this.locationTable.get(nodeId);

            int MAX_ATTEMPTS = 3;
            while (attempt < MAX_ATTEMPTS) {
                relocatingFrom = this.serverMapping.get(loc);
                remoteHashtableServer = this.servers.get(relocatingFrom);
                if (relocatingFrom == default_srv)
                    k = 1;
                if (relocatingFrom != mysrvid) {
                    resultObject = remoteHashtableServer.relocate(id, this.myURL);
                } else {
                    return true;
                }

                if (resultObject.status == Constants.RELOCATED) {
                    loc = resultObject.location;
                } else if (resultObject.status == Constants.SUCCESS) {
                    myHashtableServer.removeLocationTableEntry(nodeId);
                    myHashtableServer.put(nodeId, resultObject.node);
                    remoteHashtableServer.locationChange(id, this.myURL);
                    if (relocatingFrom != default_srv && mysrvid != default_srv) {
                        HashtableServer defServer = servers.get(default_srv);
                        defServer.locationChange(id, this.myURL);
                    }
                    this.locationTable.put(nodeId, this.myURL);
                    this.relocating.remove(nodeId);
                    myHashtableServer.lock(id);
                    return true;
                } else if (resultObject.status == Constants.LOCATION_LOCKED) {
                    // object is locked
                    return false;
                } else {
                    break;
                }
                attempt++;
            }
        }

        if (k == 0) {
            relocatingFrom = default_srv;
            remoteHashtableServer = this.servers.get(relocatingFrom);
            resultObject = remoteHashtableServer.relocate(id, this.myURL);
            if (resultObject == null) {
                System.out.println("result object null returned from relocate ");
                return false;
            }
            if (resultObject.status == Constants.RELOCATED) {
                // //System.out.println("Relocated from default server");
                loc = resultObject.location;
                relocatingFrom = this.serverMapping.get(loc);
                remoteHashtableServer = this.servers.get(relocatingFrom);
                resultObject = remoteHashtableServer.relocate(id, this.myURL);
            }
            if (resultObject.status == Constants.SUCCESS) {

                myHashtableServer.removeLocationTableEntry(nodeId);
                myHashtableServer.put(nodeId, resultObject.node);
                remoteHashtableServer.locationChange(id, this.myURL);
                if (relocatingFrom != default_srv && mysrvid != default_srv) {
                    HashtableServer defServer = servers.get(default_srv);
                    defServer.locationChange(id, this.myURL);
                }
                this.locationTable.put(nodeId, this.myURL);
                this.relocating.remove(nodeId);
                return true;
            } else if (resultObject.status == Constants.LOCATION_LOCKED) {
                return false;
            }
        }
        return false;
    }

    private String getLocationURL(String nodeId) {
        if (locationTable.containsKey(nodeId)) {
            return locationTable.get(nodeId);
        } else {
            int serverId = getServerID(nodeId);
            return serverLocationMapping.get(serverId);
        }
    }

    public String getLocation(String nodeId) {
        String url = getLocationURL(nodeId);
        try {
            URI uri = new URI(url);
            return uri.getHost();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public String getStats() {
        StringBuilder builder = new StringBuilder();
        builder.append("StorageSystem Stats for Server:" + myURL + "\n");
        builder.append("**********\n");
        //No. of single get calls
        builder.append("No. of getNode calls : " + statGetNodeTime.getNumSamples() + "\n");
        builder.append("No. of getNodeProperties calls : " + statGetNodeDataTime.getNumSamples() + "\n");
        //Avg time for single get calls
        builder.append("Avg time of getNode calls : " + statGetNodeTime.getMean() + "\n");
        builder.append("Avg time of getNodeProperties calls : " + statGetNodeDataTime.getMean() + "\n");
        //No. of times single get calls went to remote
        builder.append("No. of getNode call went to remote : " + statGetRemoteNode.getNumSamples() + "\n");
        builder.append("Fraction of getNode calls that went to remote : " + (float) statGetRemoteNode.getNumSamples() / statGetNodeTime.getNumSamples() + "\n");
        builder.append("No. of getNodeProperties call went to remote : " + statGetRemoteNodeData.getNumSamples() + "\n");
        builder.append("Fraction of getNodeProperties calls that went to remote : " + (float) statGetRemoteNodeData.getNumSamples() / statGetNodeDataTime.getNumSamples() + "\n");
        //Avg time to fetch from remote in single get calls
        builder.append("Avg time to fetch from remote in getNode call : " + statGetRemoteNode.getMean() + "\n");
        builder.append("Avg time to fetch from remote in getNodeProperties call : " + statGetRemoteNodeData.getNumSamples() + "\n");

        //Parallel get related calls
        //No. of parallel get calls
        builder.append("No. of getNodesParallel calls : " + statGetNodeParallelTime.getNumSamples() + " \n");
        builder.append("No. of getNodePropertiesParallel calls : " + statGetNodeDataParallelTime.getNumSamples() + " \n");
        //Avg time for parallel get calls
        builder.append("Avg time for getNodesParallel call : " + statGetNodeParallelTime.getMean() + "  \n");
        builder.append("Avg time for getNodePropertiesParallel call : " + statGetNodeDataParallelTime.getMean() + "  \n");
        //Avg size of get parallel calls
        builder.append("	Avg size of getNodesParallel call : " + statGetNodeParallelCallSize.getMean() + " \n");
        builder.append("	Avg time taken for getNodesParallel/Node : " + statGetNodeParallelTime.getMean() / statGetNodeParallelCallSize.getMean() + " \n");
        builder.append("	Avg size of getNodePropertiesParallel call : " + statGetNodeDataParallelCallSize.getMean() + " \n");
        builder.append("	Avg time taken for getNodesPropertiesParallel/Node : " + statGetNodeDataParallelTime.getMean() / statGetNodeDataParallelCallSize.getMean() + " \n");

        //TODO : Varun : instrument remote part of parallel get calls

        //single put call related data
        //no of single put calls
        builder.append("No. of putNode calls : " + statPutNodeTime.getNumSamples() + "\n");
        builder.append("No. of updateNodeProperties calls : " + statUpdateNodeDataTime.getNumSamples() + "\n");
        //avg time for single put calls
        builder.append("Avg time for putNode call : " + statPutNodeTime.getMean() + " \n");
        builder.append("Avg time for updateNodeProperties call : " + statUpdateNodeDataTime.getMean() + " \n");

        //parallel put calls related
        //no of parallel put calls
        builder.append("No. of putNodesParallel calls : " + statPutNodeParallelTime.getNumSamples() + " \n");
        builder.append("No. of updateNodePropertiesParallel calls : " + statUpdateNodeDataParallelTime.getNumSamples() + " \n");
        //Avg time for parallel put calls
        builder.append("No. of putNodesParallel calls : " + statPutNodeParallelTime.getNumSamples() + " \n");
        builder.append("No. of updateNodePropertiesParallel calls : " + statUpdateNodeDataParallelTime.getNumSamples() + " \n");
        //Avg call sized for parallel put calls
        builder.append("	Avg size of putNodesParallel call : " + statPutNodeParallelCallSize.getMean() + " \n");
        builder.append("	Avg time taken for putNodesParallel/Node : " + statPutNodeParallelTime.getMean() / statPutNodeParallelCallSize.getMean() + " \n");
        builder.append("	Avg size of updateNodePropertiesParallel call : " + statUpdateNodeDataParallelCallSize.getMean() + " \n");
        builder.append("	Avg time taken for updateNodePropertiesParallel/Node : " + statUpdateNodeDataParallelTime.getMean() / statUpdateNodeDataParallelCallSize.getMean() + " \n");

        if (useCaching) {
            builder.append("	Avg time for remote get node for caching : " + statGetRemoteNodeAndCache.getMean() + " \n");
            builder.append("		Fraction of get calls that went remote in caching case: " +
                    (float) statGetRemoteNodeAndCache.getNumSamples() / statGetNodeTime.getNumSamples() + " \n");
            builder.append("		Avg time for getting status of nodes (getLocalNodes call of hashtable server) : " + statGetLocalNodesTime.getMean() + " \n");
            builder.append("		Avg time to get status of nodes(getLocalNodes) / Node : " +
                    (statGetLocalNodesTime.getMean() / statGetNodeParallelCallSize.getMean()) + " \n");
            builder.append("		Out of this, Avg time for caching remote nodes : " +
                    statGetNodeParallelTimeCachePart.getMean() + " \n");
            builder.append("		Avg number of local nodes found : " + statGetParallelLocalNodesCount.getMean() + " \n");
            builder.append("		Avg number of remote nodes found : " + (statGetNodeParallelCallSize.getMean()
                    - statGetParallelLocalNodesCount.getMean()) + " \n");
            builder.append("		Avg time taken/Node to fetch remote node and cache it : " +
                    (statGetNodeParallelTimeCachePart.getMean() / (statGetNodeParallelCallSize.getMean()
                            - statGetParallelLocalNodesCount.getMean())) + " \n");
            builder.append("		Fraction of remote nodes fetched per call : " + ((statGetNodeParallelCallSize.getMean()
                    - statGetParallelLocalNodesCount.getMean()) / statGetNodeParallelCallSize.getMean()) + " \n");
            builder.append("		Avg time taken for bucketization during parallel get " + statGetParallelBucketizationTime.getMean() + " \n");
        }

        return builder.toString();
    }

    /**
     * External API to get stats
     *
     * @param includeHashtable
     * @return
     * @throws RemoteException
     */
    public String getStats(boolean includeHashtable) throws RemoteException {
        StringBuilder builder = new StringBuilder();
        builder.append(getStats());//get storage system stats
        if (includeHashtable) {
            //append hashtable stats
            int myid = serverMapping.get(myURL);
            HashtableServer myHashtable = servers.get(myid);
            builder.append(myHashtable.getStats());
        }

        return builder.toString();
    }

    public int getServerID(String nodeId) {
        int serverId = 0;
        switch (HASHING_SCHEME) {
            case "INTEGER_MODULO": {
                //modulo logic
                int nodeID = Integer.parseInt(nodeId);
                serverId = nodeID % numServers;
                if (serverId < 0) {
                    serverId *= -1;
                }
                String knownLoc;
                String targetNode = Integer.toString(nodeID);
                if (this.locationTable.containsKey(targetNode)) {
                    knownLoc = this.locationTable.get(targetNode);
                    serverId = this.serverMapping.get(knownLoc);
                }
                break;
            }
            case "GRAPH_PARTITION": {
                //graph partition logic
                int nodeID = Integer.parseInt(nodeId);
                int start = 0;
                for (int i = 0; i < partitionedGraphHostNodesTable.length; i++) {
                    if (nodeID >= start && nodeID <= partitionedGraphHostNodesTable[i]) {
                        serverId = i;
                        break;
                    }
                    start = partitionedGraphHostNodesTable[i] + 1;
                }
                break;
            }
            case "STRING_MODULO": {
                //modulo logic
                int nodeID = nodeId.hashCode();
                serverId = nodeID % numServers;
                if (serverId < 0) {
                    serverId *= -1;
                }
                String knownLoc;
                String targetNode = Integer.toString(nodeID);
                if (this.locationTable.containsKey(targetNode)) {
                    knownLoc = this.locationTable.get(targetNode);
                    serverId = this.serverMapping.get(knownLoc);
                }
                break;
            }
        }
        return serverId;
    }

    public HashtableServer getServer(String nodeId) {
        int serverId = getServerID(nodeId);
        return this.servers.get(serverId);
    }

    private HashtableServer getServerThrift(String nodeId) {
        int serverId = getServerID(nodeId);
        return this.serversThrift.get(serverId);
    }

    /**
     * Gets all of the local nodes whose ids are passed in. If a remote id is passed in, it is not fetched.
     *
     * @param nodeIds The nodeIds of the local nodes we want to fetch.
     * @return All of the local nodes whose nodeIds are passed in.
     */
    public Vector<Node> getLocalNodes(Vector<String> nodeIds) {
        Vector<Node> nodes = new Vector<>();
        try {
            for (String nodeId : nodeIds) {
                Node node;
                if (useCaching) {
                    node = myHashTableServer.getNodeFromCacheTable(nodeId);
                } else {
                    node = myHashTableServer.getNodeFromTable(nodeId);
                }
                if (node != null) {
                    nodes.add(node);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return nodes;
    }

    /**
     * Gets specific properties of local nodes. The desired properties and nodeIds are passed in as a Vector of
     * NodeData
     *
     * @param nodeData     Each {@link NodeData} contains the nodeId of the node whose properties we want to fetch. The
     *                     desired properties are contained within the same NodeData object.
     * @param remoteVector A pointer to a vector of NodeData. Any remote nodeId from {@code nodeData} will be added
     *                     to this.
     * @return Each NodeData contains the nodeId of a local node from {@code nodeData} and the respective
     * properties and values we wanted to fetch.
     * @deprecated
     */
    @Deprecated
    public Vector<NodeData> getLocalNodesData(Vector<NodeData> nodeData, Vector<NodeData> remoteVector) {

        Vector<NodeData> nodeDataVector = new Vector<>();
        try {
            for (NodeData nd : nodeData) {
                Node retNode = myHashTableServer.getNodeFromTable(nd.nodeId);

                if (retNode != null) {
                    NodeData ndFromNode = readNodePropertiesByReflection(nd.nodeId, retNode, nd.properties, retNode.TS);
                    nodeDataVector.add(ndFromNode);
                } else if (useCaching) {
                    //try to get it from cache table
                    retNode = myHashTableServer.getNodeFromCacheTable(nd.nodeId);
                    if (retNode != null) {
                        NodeData ndFromNode = readNodePropertiesByReflection(nd.nodeId, retNode, nd.properties, retNode.TS);
                        nodeDataVector.add(ndFromNode);
                    }
                } else {
                    remoteVector.add(nd);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return nodeDataVector;
    }

    /**
     * Only called by @see StorageSystem#getLocalNodesData
     * Obtains specific properties of a node using reflection.
     *
     * @param nodeId     The id of the node whose properties we want to fetch.
     * @param n          The node that is being read
     * @param properties The properties of the node we want to fetch. This is represented as a mapping of "getter"
     *                   methods to required/desired parameters.
     * @param TS         The timestamp of this method's invocation
     * @return A NodeData object containing {@code nodeId} and a mapping from the node's desired properties
     * and its respective values.
     */
    private NodeData readNodePropertiesByReflection(String nodeId, Object n, Map<String, Object> properties, long TS) {
        NodeData retNode = new NodeData(nodeId, TS, nodeClassNametoIDMap.get(n.getClass().getName()));
        //create node properties from the fields
        Class c = n.getClass();
        Method m;
        Object argValue;
        Object v = null;
        for (String key : properties.keySet()) {//for now not dealing with function params
            try {
                argValue = properties.get(key);
                StringBuilder builder = new StringBuilder();
                builder.append("get");
                builder.append(key.substring(0, 1).toUpperCase());
                builder.append(key.substring(1));

                if (argValue == null) {
                    m = c.getMethod(builder.toString(), null);
                    v = m.invoke(n);
                } else {
                    m = c.getMethod(builder.toString(), argValue.getClass());
                    v = m.invoke(n, argValue);
                }

                retNode.addProperty(key, v);
            } catch (Exception e) {
                e.printStackTrace();
                retNode.addProperty(key, e);
            }
        }

        return retNode;
    }

    /**
     * Gets all of the local nodes whose ids are passed in. If {@code withCaching} is true and <it>caching</it> is enabled
     * through the configuration, then an attempt to include all of the locally cached nodes will be made. Otherwise,
     * the behavior is identical to {@link #getLocalNodes(Vector)}
     *
     * @param nodeIds     The nodeIds of the local nodes we want to fetch.
     * @param withCaching True if an attempt to include locally cached nodes is desired, false otherwise.
     * @return All of the local nodes whose nodeIds are passed in.
     * @throws RemoteException Thrown if the cached nodes could not be fetched.
     */
    public Vector<Node> getLocalNodes(Vector<String> nodeIds, boolean withCaching) throws RemoteException {

        Vector<Node> resList = new Vector<Node>();
        Vector<String> notFoundList = new Vector<String>();
        int j;
        try {
            for (String nodeId : nodeIds) {
                j = getServerID(nodeId);
                if (j == this.myId) {
                    resList.add(myHashTableServer.getNodeFromTable(nodeId));
                } else {
                    notFoundList.add(nodeId);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        if (useCaching && withCaching) {
            resList.addAll(myHashTableServer.getCachedObjects(notFoundList));
        }

        return resList;
    }

    public Vector<Node> getAllLocalNodes() throws RemoteException {
        int myid = serverMapping.get(myURL);
        HashtableServer myHashtable = servers.get(myid);
        return myHashtable.getAllNodes();
    }

    //clear all the stat containers of the storage system and the hashtable server
    public void clearStatContainers() throws RemoteException {
        statGetNodeTime = new StatContainer();
        statGetNodeDataTime = new StatContainer();
        statPutNodeTime = new StatContainer();
        statUpdateNodeDataTime = new StatContainer();
        statGetNodeParallelTime = new StatContainer();
        statGetNodeDataParallelTime = new StatContainer();
        statPutNodeParallelTime = new StatContainer();
        statUpdateNodeDataParallelTime = new StatContainer();
        statGetNodeParallelTimeCachePart = new StatContainer();
        statGetNodeParallelCallSize = new StatContainer();
        statGetNodeDataParallelCallSize = new StatContainer();
        statPutNodeParallelCallSize = new StatContainer();
        statUpdateNodeDataParallelCallSize = new StatContainer();
        statGetRemoteNodeAndCache = new StatContainer();
        statGetRemoteNode = new StatContainer();
        statGetRemoteNodeData = new StatContainer();
        //clear the counters of the hashtable server as well
        myHashTableServer.clearStatContainers();
    }

    private void recordStat(StatContainer stat, float value) {
        if (this.recordStats)
            stat.addOnlyMean(value);
    }

    /**
     * Attempts to lock each request for validation and commit under a object-locking mechanism
     *
     * @param requests Metadata of nodes that are to be locked
     * @return true if all lock requests were acquired, false otherwise
     */
    public boolean acquireLocks(SortedSet<LockRequest> requests) {
        SortedMap<Integer, SortedSet<LockRequest>> htToReqs = new TreeMap<>();

        for (LockRequest request : requests) {
            String nodeId = request.nodeId;
            int htId = getServerID(nodeId);

            SortedSet<LockRequest> htReqs = htToReqs.getOrDefault(htId, new TreeSet<>());
            htReqs.add(request);
            htToReqs.putIfAbsent(htId, htReqs);
        }

        try {
            List<Integer> acquiredHts = new ArrayList<>();
            for (Integer htId : htToReqs.keySet()) {

                HashtableServer ht;
                if (RPC_MECHANISM.equals("THRIFT")) {
                    ht = serversThrift.get(htId);
                } else {
                    ht = servers.get(htId);
                }
                int status = ht.acquireLocks(htToReqs.get(htId));

                if (status != Constants.LOCK_ACQUIRED) {
                    for (Integer acquiredHt : acquiredHts) {
                        HashtableServer aht;
                        if (RPC_MECHANISM.equals("THRIFT")) {
                            aht = serversThrift.get(acquiredHt);
                        } else {
                            aht = servers.get(acquiredHt);
                        }
                        aht.releaseLock(htToReqs.get(acquiredHt));
                    }
                    return false;
                } else {
                    acquiredHts.add(htId);
                }
            }
            return true;
        } catch (RemoteException e) {
            e.printStackTrace();
            return false;
        }
    }
    
    public void commitChanges(Set<Node> writeSet, Set<NodeData> writeNDSet, Map<String, Long> readSet) {

        // Bucketize sets according to their hosting servers
        Map<Integer, Set<Node>> writeMap = new HashMap<>();
        for (Node node : writeSet) {
            int htId = getServerID(node.getNodeId());
            Set<Node> htNodes = writeMap.getOrDefault(htId, new HashSet<>());
            htNodes.add(node);
            writeMap.putIfAbsent(htId, htNodes);
        }

        Map<Integer, Set<NodeData>> writeNdMap = new HashMap<>();
        for (NodeData nd : writeNDSet) {
            int htId = getServerID(nd.nodeId);
            Set<NodeData> htNodes = writeNdMap.getOrDefault(htId, new HashSet<>());
            htNodes.add(nd);
            writeNdMap.putIfAbsent(htId, htNodes);
        }

        Map<Integer, Map<String, Long>> readMap = new HashMap<>();
        for (String nodeId : readSet.keySet()) {
            int htId = getServerID(nodeId);
            Map<String, Long> htReads = readMap.getOrDefault(htId, new HashMap<>());
            htReads.put(nodeId, readSet.get(nodeId));
            readMap.putIfAbsent(htId, htReads);
        }

        for (Integer htId : servers.keySet()) {
            Set<Node> write = writeMap.getOrDefault(htId, new HashSet<>());
            Set<NodeData> writeNd = writeNdMap.getOrDefault(htId, new HashSet<>());
            Map<String, Long> read = readMap.getOrDefault(htId, new HashMap<>());

            HashtableServer ht = servers.get(htId);
            try {
                ht.commitChanges(write, writeNd, read);
            } catch (RemoteException e) {
                e.printStackTrace();
            }
        }
    }

    public void checkpoint() {
        for (HashtableServer htServers : servers.values()) {
            try {
                htServers.checkpoint();
            } catch (RemoteException e) {
                e.printStackTrace();
            }
        }
    }
}
