/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

import beehive.util.*;

public class ConcurrentValidator {

    // nodeid->ts of the latest writer
    private ConcurrentHashMap<String, Long> conflictTable;
    private ConcurrentHashMap<String, NodeRWLock> cTable;

    private List<ReentrantLock> cTableLockList;
    public Config configParams = null;
    private int nLocks = 256;

    public ConcurrentValidator() {
        configParams = Config.getInstance();
        nLocks = configParams.getIntValue("ConcurrentValidator.nLocks");
        conflictTable = new ConcurrentHashMap<String, Long>();
        cTable = new ConcurrentHashMap<String, NodeRWLock>();
        cTableLockList = new ArrayList<ReentrantLock>();
        for (int i = 0; i < nLocks; i++) {
            cTableLockList.add(new ReentrantLock());
        }
    }

    public long speculativeValidate(long startTS, RWSetInfo rwSet, HashSet<String> conflictSet) {

        long ret = 0; // 0 means no conflict

        for (String nodeId : rwSet.readSet) {
            Long writerTS = conflictTable.get(nodeId);
            // if writerTS > startTS => conflict!

            if (writerTS != null && writerTS > startTS) {
                ret = writerTS;
                conflictSet.add(nodeId);
            }
        }

        for (String nodeId : rwSet.writeSet) {
            Long writerTS = conflictTable.get(nodeId);

            if (writerTS != null && writerTS > startTS) {
                ret = writerTS;
                conflictSet.add(nodeId);
            }
        }

        return ret;

    }

    public long validate(long startTS, RWSetInfo rwSet, HashSet<String> conflictSet) {

        boolean conflict = false;
        long ret = 0;

        boolean lock = acquireLocksOnWriteSet(rwSet);
        if (lock) {
            for (String nodeId : rwSet.readSet) {
                Long writerTS = conflictTable.get(nodeId);
                // if writerTS > startTS => conflict!

                if (writerTS != null && writerTS > startTS) {
                    conflict = true;
                    ret = writerTS;
                    conflictSet.add(nodeId);
                }
            }
            // check write set
            for (String nodeId : rwSet.writeSet) {
                Long writerTS = conflictTable.get(nodeId);
                // if writerTS > startTS => conflict!
                if (writerTS != null && writerTS > startTS) {
                    conflict = true;
                    ret = writerTS;
                    conflictSet.add(nodeId);
                }
            }
        }

        if (conflict) {
            releaseLock(rwSet);
        }

        return ret;
    }

    private boolean acquireLocksOnWriteSet(RWSetInfo rwSet) {

        TreeSet<String> sortedRWSet = new TreeSet<String>();
        sortedRWSet.addAll(rwSet.readSet);
        sortedRWSet.addAll(rwSet.writeSet);
        for (String s : sortedRWSet) {
            if (!(s == null) && !s.isEmpty()) {

                if (cTable.containsKey(s)) {
                    try {
                        LockMode lockMode;
                        if (rwSet.writeSet.contains(s)) {
                            lockMode = LockMode.WRITE;
                        } else {
                            lockMode = LockMode.READ;
                        }
                        cTable.get(s).getLock(lockMode);
                    } catch (Exception e) {
                        e.printStackTrace();
                        return false;
                    }
                } else {
                    ReentrantLock lock = cTableLockList.get(Math.abs(s.hashCode()) % nLocks);
                    lock.lock();
                    if (!cTable.containsKey(s)) {
                        cTable.put(s, new NodeRWLock());
                    }
                    lock.unlock();
                    try {
                        LockMode lockMode;
                        if (rwSet.writeSet.contains(s)) {
                            lockMode = LockMode.WRITE;
                        } else {
                            lockMode = LockMode.READ;
                        }
                        cTable.get(s).getLock(lockMode);
                    } catch (Exception e) {
                        e.printStackTrace();
                        return false;
                    }
                }
            }
        }
        return true;
    }

    private void releaseLock(RWSetInfo rwSet) {
        TreeSet<String> sortedRWSet = new TreeSet<String>();
        sortedRWSet.addAll(rwSet.readSet);
        sortedRWSet.addAll(rwSet.writeSet);
        for (String s : sortedRWSet) {
            if (cTable.containsKey(s)) {
                try {
                    cTable.get(s).releaseLock();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }

    public synchronized void update(long cmtTS, RWSetInfo rwSet) {
        for (String nodeId : rwSet.writeSet) {
            conflictTable.put(nodeId, cmtTS);
        }
    }

    long validateAndUpdate(long startTS, long cmtTS, RWSetInfo rwSet, HashSet<String> refetchSet) {
        long conflictingTS = validate(startTS, rwSet, refetchSet);

        if (conflictingTS == 0) {
            update(cmtTS, rwSet);
            releaseLock(rwSet);   // added here on September 28, 2015
        } else {
            for (String nodeId : rwSet.unusedSet) { // check if any of the item
                // in unused set is modified
                Long writerTS = conflictTable.get(nodeId);
                if (writerTS != null && writerTS > startTS) {
                    refetchSet.add(nodeId);
                }
            }
        }
        return conflictingTS;
    }

    class NodeRWLock {
        boolean writeLock = false;
        int readCount = 0;

        /**
         * wait on this object
         *
         * @throws InterruptedException
         */
        public synchronized void getLock(LockMode lockMode) throws InterruptedException {
            if (lockMode == LockMode.READ) {
                while (writeLock) {
                    try {
                        //waitForLock.waitForLock(node,lockRequestInfo.transactionId);
                        this.wait();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
                readCount++;
//				System.out.println("Acquired READ lock");
            } else {
                while ((writeLock) || (readCount > 0)) {
                    try {
                        //waitForLock.waitForLock(node,lockRequestInfo.transactionId);
                        this.wait();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
                writeLock = true;
//				System.out.println("Acquired WRITE lock");
            }
        }

        /**
         * Release the lock and notify others.
         */
        public synchronized void releaseLock() {
            if (readCount > 0 && writeLock == false) {
                readCount--;
                if (readCount == 0) {
                    notify();
//					System.out.println("Read lock released");
                }
            } else if (readCount == 0 && writeLock == true) {
                writeLock = false;
                notifyAll();
//				System.out.println("Write lock released");
            } else {
//				System.out.println("Conflicting locks granted!!!!!!!!!!!!! readCount="+lockInfo.readCount+" writeLock="+lockInfo.writeLock);

                try {
                    throw new Exception("Conflicting locks granted!");
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }

    enum LockMode {
        READ,
        WRITE
    }
}
