/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

import beehive.util.*;
import beehive.workpool.TimestampManager;

public class ConcurrentValidatorWithTruncation {
	// nodeid->ts of the latest writer
	//	private ConcurrentHashMap<String, Long> conflictTable; //this table has been removed and ts is from cTable
	public ConcurrentHashMap<String, NodeRWLock> cTable; //timestamp of the above table can be recorded here

	//garbage collector thread removes entries from both the above maps
	public ArrayList<Long> sortedActiveTSList;

	public long tts = 0; // truncation time used to trim down the data structures
	public ReentrantLock ttsLock;

	public List<ReentrantLock> cTableLockList;
	public Config configParams=null;
	private int nLocks = 256;

	private TimestampManager timestampMgr;

	long TRUNCATION_WINDOW_SIZE = 1000;

	public ConcurrentValidatorWithTruncation(TimestampManager timestampMgr) {
		configParams = Config.getInstance();
		nLocks = configParams.getIntValue("ConcurrentValidator.nLocks");
		//		conflictTable = new ConcurrentHashMap<String, Long>();
		cTable = new ConcurrentHashMap<String, NodeRWLock>();
		cTableLockList = new ArrayList<ReentrantLock>();
		for(int i=0;i<nLocks;i++){
			cTableLockList.add(new ReentrantLock());
		}
		sortedActiveTSList = new ArrayList<Long>();//maintain this in a sorted manner
		this.timestampMgr = timestampMgr;
		
		ttsLock = new ReentrantLock();
		
		//start the truncation thread
		TruncationThread t = new TruncationThread(this);
		t.start();
	}

	public void advanceTruncationTime() {
		long newTTS = timestampMgr.getSTS() - TRUNCATION_WINDOW_SIZE;
		long minActiveTS;
		if(newTTS > 0) {
			ttsLock.lock();
			tts = newTTS;
			ttsLock.unlock();
			try {
				while(!sortedActiveTSList.isEmpty() && sortedActiveTSList.get(0) <= tts) {
					Thread.sleep(1);
				}
			}catch (Exception e) {
				//This is a safe exception and can be ignored - sortedActiveTSList.isEmpty is false and
				// thread switch happened and then again it became empty and second condition was executed
				//e.printStackTrace();
			}
		}
	}

	// returns true if there is no conflict, else false
	public synchronized long syncValidate(long startTS, RWSetInfo rwSet,
			HashSet<String> conflictSet) {
		return speculativeValidate(startTS, rwSet, conflictSet);
	}

	public long speculativeValidate(long startTS, RWSetInfo rwSet,
			HashSet<String> conflictSet) {

		//		 System.out.println("Speculative checking readset : " + rwSet.readSet);
		//         System.out.println("Speculative checking writeset : " + rwSet.writeSet);

		// check read set
		//truncation timestamp check
		if(startTS < tts) return -1; //returning back and invalid ts to signify abort due to truncation ts check

		boolean conflict = false;
		long ret = 0; // 0 means no conflict
		//Enumeration keys = conflictTable.keys(); BIG NONSENSE, MUST REMOVE from older versions

		for (String nodeId : rwSet.readSet) {
			Long writerTS = null;
			if(cTable.containsKey(nodeId)) {
				writerTS = cTable.get(nodeId).getTimeStamp();
			}
			// if writerTS > startTS => conflict!

			if (writerTS != null && writerTS.longValue() > startTS) {
				conflict = true;
				ret = writerTS.longValue();
				conflictSet.add(nodeId);
			}
		}
		// check write set
		for (String nodeId : rwSet.writeSet) {
			Long writerTS = null;
			if(cTable.containsKey(nodeId)) {
				writerTS = cTable.get(nodeId).getTimeStamp();
			}
			// if writerTS > startTS => conflict!
			if (writerTS != null && writerTS.longValue() > startTS) {
				conflict = true;
				ret = writerTS.longValue();
				conflictSet.add(nodeId);
			}
		}

		return ret;

	}

	public long validate(long startTS, RWSetInfo rwSet,
			HashSet<String> conflictSet) {
		//         System.out.println("Valdation Read set : " + rwSet.readSet);
		//        System.out.println("Validation Write set : " + rwSet.writeSet);
		boolean conflict = false;
		long ret = 0; // 0 means no conflict
		// Acquire locks
		boolean lock = acquireLocksOnWriteSet(rwSet);
		if (lock){
			for (String nodeId : rwSet.readSet) {
				Long writerTS = cTable.get(nodeId).getTimeStamp();
				// if writerTS > startTS => conflict!

				if (writerTS != null && writerTS.longValue() > startTS) {
					conflict = true;
					ret = writerTS.longValue();
					conflictSet.add(nodeId);
				}
			}
			// check write set
			for (String nodeId : rwSet.writeSet) {
				Long writerTS = cTable.get(nodeId).getTimeStamp();
				// if writerTS > startTS => conflict!
				if (writerTS != null && writerTS.longValue() > startTS) {
					conflict = true;
					ret = writerTS.longValue();
					conflictSet.add(nodeId);
				}
			}
		}

		//  September 28, 2015  -  Do not release locks if we have not yet if the transaction will
		// be committing, i.e conflict is false   

		if (conflict) {
			releaseLock(rwSet);
		}

		return ret;
	}

	public boolean acquireLocksOnWriteSet(RWSetInfo rwSet) {
		// System.out.println("Acquiring lock readset:" + rwSet.readSet);
		// System.out.println("Acquring lock writeset: " + rwSet.writeSet);

		TreeSet<String> sortedRWSet = new TreeSet<String>();
		sortedRWSet.addAll(rwSet.readSet);
		sortedRWSet.addAll(rwSet.writeSet);
		//		System.out.println(sortedRWSet);
		for (String s : sortedRWSet) {
			if (!(s==null) && !s.isEmpty()) {
				boolean createEntryFlag = true;
				//				System.out.println("Node is "+s);
				NodeRWLock nodeLock = cTable.get(s);
				if (nodeLock != null) {
					createEntryFlag = false;
					try {
						LockMode lockMode;
						if(rwSet.writeSet.contains(s)){
							lockMode = LockMode.WRITE;
						}
						else{
							lockMode = LockMode.READ;
						}
						//						System.out.println("Trying to acquire lock for "+s+" in "+lockMode);
						if(!nodeLock.getLock(lockMode)) {//entry got deleted by the garbage collector
							createEntryFlag = true;
						}
					} catch (Exception e) {
						e.printStackTrace();
						return false;
					}
				}
				if(createEntryFlag) {
					LockMode lockMode;
					if(rwSet.writeSet.contains(s)){
						lockMode = LockMode.WRITE;
					}
					else{
						lockMode = LockMode.READ;
					}
					ReentrantLock lock = cTableLockList.get(Math.abs(s.hashCode())%nLocks);
					lock.lock();
					if (!cTable.containsKey(s)){
						cTable.put(s, new NodeRWLock(lockMode));
					}
					lock.unlock();
					//					try {
					//						cTable.get(s).getLock(lockMode);
					//					} catch (Exception e) {
					//						e.printStackTrace();
					//						return false;
					//					}
				}
			}
		}
		// System.out.println("Acquired lock readset:" + rwSet.readSet);
		// System.out.println("Acquired lock writeset: " + rwSet.writeSet);

		return true;
	}

	public void releaseLock(RWSetInfo rwSet) {
		//		System.out.println("Release rwSet is called!!!!!!!!");
		TreeSet<String> sortedRWSet = new TreeSet<String>();
		sortedRWSet.addAll(rwSet.readSet);
		sortedRWSet.addAll(rwSet.writeSet);
		for (String s : sortedRWSet) {
			if (cTable.containsKey(s)) {
				try {
					cTable.get(s).releaseLock();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		// System.out.println("Released lock readset:" + rwSet.readSet);
		// System.out.println("Released lock writeset: " + rwSet.writeSet);
	}

	public synchronized void update(long cmtTS, RWSetInfo rwSet) {
		for (String nodeId : rwSet.writeSet) {
			cTable.get(nodeId).setTimeStamp(cmtTS);
		}
	}

	// returns true if no conflict else false
	public long validateAndUpdate(long startTS, long cmtTS,
			RWSetInfo rwSet, HashSet<String> refetchSet) {

		//get ttsLock
		ttsLock.lock();
		if(startTS < tts) {
			ttsLock.unlock();
			return -1; //returning back and invalid ts to signify abort due to truncation ts check
		}
		sortedActiveTSList.add(startTS);
		Collections.sort(sortedActiveTSList);
		ttsLock.unlock();

		long conflictingTS = validate(startTS, rwSet, refetchSet);
		//		System.out.println("Finished validation!");
		// no conflict found, update write items
		//		System.out.println("Conflicting ts:"+conflictingTS);
		if (conflictingTS == 0) {
			update(cmtTS, rwSet);
			releaseLock(rwSet);   // added here on September 28, 2015
		} else {
			for (String nodeId : rwSet.unusedSet) { // check if any of the item
				// in unused set is modified
				Long writerTS = cTable.get(nodeId).getTimeStamp();
				if (writerTS != null && writerTS.longValue() > startTS) {
					refetchSet.add(nodeId);
				}
			}
			// System.out.println("Refetch set: "+refetchSet.toString());
		}

		//remove my entry from the activeTSList
		ttsLock.lock();
		sortedActiveTSList.remove(startTS);
		ttsLock.unlock();

		return conflictingTS;
	}
}

class NodeRWLock {
	boolean writeLock = false;
	int readCount = 0;
	long timeStamp = 0;
	public boolean deleted = false;

	public long getTimeStamp() {
		return timeStamp;
	}

	public void setTimeStamp(long timeStamp) {
		this.timeStamp = timeStamp;
	}

	public NodeRWLock(LockMode lockMode) {
		if (lockMode == LockMode.READ){
			readCount++;
		}
		else{
			writeLock = true;
		}
	}

	/**
	 * wait on this object
	 * 
	 * @throws InterruptedException
	 */
	public synchronized boolean getLock(LockMode lockMode) throws InterruptedException {
		if (lockMode == LockMode.READ){
			while(writeLock){
				try{
					//waitForLock.waitForLock(node,lockRequestInfo.transactionId);
					this.wait();
				}
				catch(Exception e){
					e.printStackTrace();
				}
			}
			readCount++;
			//				System.out.println("Acquired READ lock");
		}
		else{
			while((writeLock) || (readCount>0)){
				try{
					//waitForLock.waitForLock(node,lockRequestInfo.transactionId);
					this.wait();
				}
				catch(Exception e){
					e.printStackTrace();
				}
			}
			writeLock = true;
			//				System.out.println("Acquired WRITE lock");
		}
		
		return !deleted;
	}

	public synchronized boolean tryLock(LockMode lockMode) throws InterruptedException {
		if (lockMode == LockMode.READ){
			if(writeLock){
				return false;
			}
			readCount++;
			return true;
		}
		else{
			if((writeLock) || (readCount>0)){
				return false;	
			}
			writeLock = true;
			return true;
		}
	}

	/**
	 * Release the lock and notify others.
	 */
	public synchronized void releaseLock() {
		if(readCount>0 && writeLock==false){
			readCount--;
			if(readCount == 0){
				notify();
				//					System.out.println("Read lock released");
			}
		}
		else if(readCount==0 && writeLock==true){
			writeLock = false;
			notifyAll();
			//				System.out.println("Write lock released");
		}
		else if(readCount==0 && writeLock==false){
			// Do Nothing - Releasing the non-acquired lock
		}
		else{ // read count > 0  and write lock is true
			//				System.out.println("Conflicting locks granted!!!!!!!!!!!!! readCount="+lockInfo.readCount+" writeLock="+lockInfo.writeLock);
			try{
				throw new Exception("Conflicting locks granted!");
			}
			catch(Exception e){
				e.printStackTrace();
			}
		}
	}
}

enum LockMode{
	READ,
	WRITE
}

class TruncationThread extends Thread {

	ConcurrentValidatorWithTruncation validator;

	int SCAN_LIMIT = 1000;
	int THREAD_SLEEP_TIME = 100;

	public TruncationThread(ConcurrentValidatorWithTruncation validator) {
		this.validator = validator;
	}

	public void run() {
		while(true) {
			Set<String> cTableKeys = validator.cTable.keySet();
			int removeCount = 0;
			int keysIterationCounter = 0;
			for(String s : cTableKeys) {
				NodeRWLock nodeLock = validator.cTable.get(s);
				boolean obtained = false;
				try {
					obtained = nodeLock.tryLock(LockMode.WRITE);
				}catch(Exception e) {
					e.printStackTrace();
				}
				if(obtained) {
					if(nodeLock.getTimeStamp() < validator.tts) {
						nodeLock.deleted = true;
						validator.cTable.remove(s);
						removeCount++;
					}
					nodeLock.releaseLock();
				}
				keysIterationCounter++;
				if(keysIterationCounter >= SCAN_LIMIT) {
					try {
						Thread.sleep(THREAD_SLEEP_TIME);
					}catch(Exception e) {
						e.printStackTrace();
					}
					keysIterationCounter = 0;
					validator.advanceTruncationTime();
					//System.out.println("Truncation time : " + validator.tts + "	Num keys removed : " + removeCount);
				}
			}
		}
	}
}
