/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import beehive.checkpoint.CheckpointPathBuilder;
import beehive.thrift.*;
import beehive.util.*;
import beehive.workpool.LoadInfo;
import beehive.workpool.TimestampManager;
import org.apache.thrift.TException;
import org.apache.thrift.server.TServer;

import java.io.*;
import java.net.InetAddress;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;


public class GlobalValidationService implements ValidationService, ValidationService_thrift.Iface {

    private boolean isConcurrentValidator;
    private boolean isTruncationEnabled;

    private TimestampManager timestampMgr;
    private ConcurrentValidatorWithTruncation concurrentValidatorWithTruncation;
    private ConcurrentValidator concurrentValidator;
    private Validator validator;
    public Config configParams = null;

    public static int DISPATCH_BATCH_SIZE = 20;
    public static int NOAFFINITY_BATCH_SIZE = 5;
    public static long printProgressInterval = 60000;
    public static long idleTimeInterval = 60000;
    public boolean locality = true;
    public boolean logging = false;
    private boolean speculativeChecking = true;
    private StatContainer specCheckTimeStats = new StatContainer();
    private StatContainer completeCheckTimeStats = new StatContainer();
    private QueueStat pendingValQStats = new QueueStat();

    private String hostname;
    private boolean checkpoint = false;
    private String checkpointSaveDir;
    private String checkpointRestoreDir;
    private String checkpointBakDir;

    public GlobalValidationService() throws RemoteException {
        timestampMgr = new TimestampManager();

        configParams = Config.getInstance();
        if (configParams != null) {
            System.out.println("Using ConfigFile");
            DISPATCH_BATCH_SIZE = configParams.getIntValue("GlobalWorkpool.DISPATCH_BATCH_SIZE");
            NOAFFINITY_BATCH_SIZE = configParams.getIntValue("GlobalWorkpool.NOAFFINITY_BATCH_SIZE");
            printProgressInterval = configParams.getLongValue("GlobalWorkpool.printProgressInterval");
            idleTimeInterval = configParams.getLongValue("GlobalWorkpool.idleTimeInterval");
            logging = configParams.getBoolValue("GlobalWorkpool.logging");
            locality = configParams.getBoolValue("GlobalWorkpool.locality");
            speculativeChecking = configParams.getBoolValue("GlobalWorkpool.speculativeChecking");
            isConcurrentValidator = configParams.getBoolValue("GlobalWorkpool.concurrentValidator");
            isTruncationEnabled = configParams.getBoolValue("GlobalWorkpool.truncationEnabled");

            if (configParams.has("System.checkpoint")) {
                this.checkpoint = configParams.getBoolValue("System.checkpoint");
            }
        }

        try {
            hostname = InetAddress.getLocalHost().getHostName();
        } catch (Exception e) {
            e.printStackTrace();
        }


        if (isConcurrentValidator) {
            System.out.println("Using Concurrent Validator!");
            if (!isTruncationEnabled) {
                concurrentValidator = new ConcurrentValidator();
            } else {
                concurrentValidatorWithTruncation = new ConcurrentValidatorWithTruncation(timestampMgr);
            }
        } else {
            validator = new Validator();
        }

        if (checkpoint) {
            initCheckpointingDir();
            if (checkpoint && configParams.getBoolValue("System.recovery")) {
                try {
                    BufferedReader stsIn;
                    if (new File(checkpointRestoreDir).exists()) {
                        stsIn = new BufferedReader(new FileReader(checkpointRestoreDir + "/GlobalWorkpool.ser"));
                    } else {
                        if (new File(checkpointBakDir).exists()) {
                            stsIn = new BufferedReader(new FileReader(checkpointBakDir + "/GlobalWorkpool.ser"));
                        } else {
                            stsIn = new BufferedReader(new FileReader(checkpointSaveDir + "/GlobalWorkpool.ser"));

                        }
                    }
                    String curLine = stsIn.readLine();
                    long archiveSTS = Long.parseLong(curLine);
                    stsIn.close();
                    timestampMgr = new TimestampManager(archiveSTS);
                } catch (IOException i) {
                    i.printStackTrace();
                }
                System.out.println("Deserialized GlobalWorkpool ...");
            }
        }
    }


    private void initCheckpointingDir() {
        checkpointSaveDir = CheckpointPathBuilder.buildCheckpointSavePath(hostname);
        checkpointRestoreDir = CheckpointPathBuilder.buildCheckpointRecoverPath(hostname);
        checkpointBakDir = CheckpointPathBuilder.buildCheckpointBakPath(hostname);

        new File(checkpointSaveDir).mkdirs();
        new File(checkpointRestoreDir).mkdirs();
        new File(checkpointBakDir).mkdirs();
    }

    @Override
    public synchronized long reportCompletion(long commitTS) throws RemoteException {
        timestampMgr.reportDone(commitTS);
        return timestampMgr.getSTS();
    }

    @Override
    public synchronized long reportCompletion(Vector<Long> commitTSList) throws RemoteException {
        for (long commitTS : commitTSList) {
            timestampMgr.reportDone(commitTS);
        }
        return timestampMgr.getSTS();
    }

    @Override
    public ValidateResponse validate(long startTS, RWSetInfo rwSetInfo, LoadInfo myLoadInfo) throws RemoteException {
        pendingValQStats.reportArrival();
        ValidateResponse response = new ValidateResponse();

        HashSet<String> refetchSet = new HashSet<>();
        if (speculativeChecking) {
            long start = System.currentTimeMillis();
            long conflictingTS;
            if (isConcurrentValidator) {
                if (isTruncationEnabled) {
                    conflictingTS = concurrentValidatorWithTruncation.speculativeValidate(startTS, rwSetInfo, refetchSet);
                } else {
                    conflictingTS = concurrentValidator.speculativeValidate(startTS, rwSetInfo, refetchSet);
                }
            } else {
                conflictingTS = validator.validate(startTS, rwSetInfo, refetchSet);    //non-synchronized checking
            }

            boolean commit;
            commit = conflictingTS == 0;
            response.conflictingTS = conflictingTS;
            long end = System.currentTimeMillis();
            specCheckTimeStats.add(end - start);
            if (!commit) {
                response.latestSTS = timestampMgr.getSTS();
                response.speculativeAbort = true;
                pendingValQStats.reportDeparture();
                return response;
            }
        }
        long start = System.currentTimeMillis();
        long commitTS = timestampMgr.getGTS();
        long conflictingTS;
        if (isConcurrentValidator) {
            if (isTruncationEnabled) {
                conflictingTS = concurrentValidatorWithTruncation.validateAndUpdate(startTS, commitTS, rwSetInfo, refetchSet);
            } else {
                conflictingTS = concurrentValidator.validateAndUpdate(startTS, commitTS, rwSetInfo, refetchSet);
            }
        } else {
            conflictingTS = validator.validateAndUpdate(startTS, commitTS, rwSetInfo, refetchSet);    //synchonized validate and update
        }

        boolean commit;
        commit = conflictingTS == 0;
        response.conflictingTS = conflictingTS;
        response.commit = commit;
        response.commitTS = commitTS;
        if (!commit) {    // Validation unsuccesfull
            timestampMgr.reportDone(commitTS);
            response.latestSTS = timestampMgr.getSTS();
            response.refetchSet = refetchSet;
            response.speculativeAbort = false;
        }
        long end = System.currentTimeMillis();
        completeCheckTimeStats.add(end - start);
        pendingValQStats.reportDeparture();

        response.loadVector = null;
        return response;

    }

    @Override
    public Hashtable<String, ValidateResponse> validateBatch(Queue<ValidationRequest> requestBatch, LoadInfo myLoad) throws RemoteException {
        Hashtable<String, ValidateResponse> responseTable = new Hashtable<>();
        for (ValidationRequest request : requestBatch) {
            ValidateResponse response = validate(request.startTS, request.rwSet, myLoad);
            responseTable.put(request.tid, response);
        }
        return responseTable;
    }

    @Override
    public void checkPoint() throws RemoteException {
        try {
            BufferedWriter stsOut = new BufferedWriter(new FileWriter(checkpointSaveDir + "/GlobalWorkpool.ser"));
            stsOut.write(timestampMgr.getSTS() + "\n");
            stsOut.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public synchronized long getSTS() throws RemoteException {
        return timestampMgr.getSTS();
    }

    @Override
    public long getSTS_thrift() {
        try {
            return getSTS();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return 0;
    }

    @Override
    public ValidateResponse_thrift validate_thrift(long startTS, RWSetInfo_thrift rwSet) throws TException {
        ValidateResponse vr = null;
        try {
            vr = validate(startTS, new RWSetInfo(rwSet), null);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return new ValidateResponse_thrift(vr.commit, vr.commitTS, vr.latestSTS, vr.refetchSet, vr.localResponse, vr.speculativeAbort, vr.conflictingTS);
    }

    @Override
    public long reportCompletion_thrift(long commitTS) {
        try {
            return reportCompletion(commitTS);
        } catch (RemoteException e) {
            e.printStackTrace();
            return -1;
        }
    }

    @Override
    public long reportCompletionBatch_thrift(List<Long> commitTSList) {
        try {
            return reportCompletion(new Vector<>(commitTSList));
        } catch (RemoteException e) {
            e.printStackTrace();
            return -1;
        }
    }

    @Override
    public Map<String, ValidateResponse_thrift> batchValidate_thrift(List<ValidationRequest_thrift> requests) {
        return new HashMap<>();
    }

    public static void main(String[] args) throws Exception {

        Registry localRegistry = null;
        System.setSecurityManager(new RMISecurityManager());

        if (args.length == 0) {
            System.out.println("Usage: java beehive.workpool.GlobalWorkpool <configParmaFile> [optional - port-num]");
            System.exit(1);
        }


        Config configParams = new Config(args[0]);
        GlobalValidationService workpool = new GlobalValidationService();

        if (args.length == 1) {
            try {
                localRegistry = LocateRegistry.getRegistry(1099);

                //defining RMI socket factories
                int timeout = 1800000;
                if (configParams.has("System.RMI_SO_TIMEOUT"))
                    timeout = configParams.getIntValue("System.RMI_SO_TIMEOUT");

                RMIClientSocketFactory csf = new CustomClientSocketFactory(timeout);
                RMIServerSocketFactory ssf = new CustomServerSocketFactory();
                ValidationService globalWorkpoolStub = (ValidationService) UnicastRemoteObject.exportObject(workpool, 0, csf, ssf);
                localRegistry.bind("WorkpoolServer", globalWorkpoolStub);

            } catch (RemoteException ee) {
                try {
                    localRegistry = LocateRegistry.createRegistry(1099);
                    localRegistry.bind("WorkpoolServer", workpool);
                } catch (RemoteException ee1) {
                    System.out.println("Couldn't Create Registry");
                    ee1.printStackTrace();
                    System.exit(2);
                }
            }
        } else {
            try {
                localRegistry = LocateRegistry.getRegistry(Integer.parseInt(args[1]));
                localRegistry.bind("WorkpoolServer", workpool);

            } catch (RemoteException ee) {
                try {
                    localRegistry = LocateRegistry.createRegistry(Integer.parseInt(args[1]));
                    localRegistry.bind("WorkpoolServer", workpool);
                } catch (RemoteException ee1) {
                    System.out.println("Couldn't Create Registry");
                    ee1.printStackTrace();
                    System.exit(2);
                }
            }
        }
        System.out.println("RMI Validation Server started..");

        int thriftPort = configParams.getIntValue("Validator.GlobalWorkpoolImpl_THRIFT_PORT");

        ValidationService_thrift.Processor processor = new ValidationService_thrift.Processor(workpool);
        TServer server = BeehiveThriftBuilder.getThriftServer(thriftPort, processor);
        System.out.println("Thrift Validation Server started..");
        server.serve();
    }
}
