/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import beehive.util.*;
import beehive.workpool.LoadInfo;

import java.io.IOException;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;


public class LocalValidator {

    public static int BATCHSIZE = 10;

    // HashSet<String> tentativeRSet;
    private HashSet<String> tentativeWSet;
    private Validator validator;
    private Queue<ValidationRequest> requestBatch;
    private ConcurrentHashMap<String, ValidateResponse> responseMap;

    private ValidationDispatcher dispatcher;
    private ValidationService validationService;
    private boolean useBatching = false;

    private Logger logger;

    // performance statistics
    private StatContainer localReqStats = new StatContainer();
    private StatContainer globalReqStats = new StatContainer();
    private int globalAborts = 0;
    private StatContainer queueSizeStats = new StatContainer();
    private QueueStat dispatcherQStats = new QueueStat();


    private final ReentrantLock lock = new ReentrantLock();
    private Condition responseArrived = lock.newCondition();

    private boolean useLocalValidator = false;


    public LocalValidator(ValidationService validationService) throws IOException {
        tentativeWSet = new HashSet<>();
        validator = new Validator();
        requestBatch = new LinkedList<>();
        responseMap = new ConcurrentHashMap<>();
        this.validationService = validationService;
        if (useBatching) {
            ValidationDispatcher dispatcher = new ValidationDispatcher(validationService);
            this.dispatcher = dispatcher;
            this.dispatcher.setLocalValidator(this);
            dispatcher.start();
        }
        logger = new Logger();
        Config configParams = Config.getInstance();
        if (configParams != null) {
            BATCHSIZE = configParams.getIntValue("LocalValidator.BATCHSIZE");
            useBatching = configParams.getBoolValue("LocalValidator.useBatching");

            if (configParams.has("LocalWorkpool.useLocalValidator")) {
                this.useLocalValidator = configParams.getBoolValue("LocalWorkpool.useWorkPoolValidator");
            }
        }
    }

    private boolean checkTentativeSet(RWSetInfo rwSet) {
        HashSet<String> conflictSet = new HashSet<String>();
        conflictSet.addAll(rwSet.writeSet);
        conflictSet.addAll(rwSet.readSet);
        for (String nodeId : conflictSet) {
            if (tentativeWSet.contains(nodeId)) {
                return false;
            }
        }
        tentativeWSet.addAll(rwSet.writeSet);
        return true;
    }

    public ValidateResponse validate(String tid, long startTS, RWSetInfo rwSet, LoadInfo myLoadInfo) {
        while (true) {
            try {
                long start = System.currentTimeMillis();
                long conflictingTS;
                long end;

                if (useLocalValidator) {
                    HashSet<String> refetchSet = new HashSet<String>();
                    conflictingTS = validator.validate(startTS, rwSet, refetchSet);
                } else {
                    conflictingTS = 0;
                }
                end = System.currentTimeMillis();

                if (conflictingTS != 0) {
                    // local validation failed
                    localReqStats.add(end - start);
                    logger.logDebug("local validation failed..aborting transaction");
                    ValidateResponse response = new ValidateResponse();
                    response.commit = false;    // transaction
                    response.localResponse = true;
                    response.conflictingTS = conflictingTS;
                    return response;
                } else {
                    if (!useBatching) {
                        start = System.currentTimeMillis();
                        ValidateResponse response = validationService.validate(startTS, rwSet, myLoadInfo);
                        end = System.currentTimeMillis();
                        globalReqStats.add(end - start);
                        if (response.commit) {
                            if (useLocalValidator) {
                                validator.update(response.commitTS, rwSet);
                            }
                        } else {
                            globalAborts++;
                        }
                        return response;
                    }
                    // there is no local conflict in the conflict table
                    // Check if there is any conflict with currently pending validation
                    // requests sent to global validator.
                    if (!checkTentativeSet(rwSet)) {
                        // conflict with tentative write set
                        // block the transaction.
                        logger.logDebug("conflict with tentative writeset..blocking transaction");
                        try {
                            responseArrived.await();    // wait on the condition variable responseArrived of the `lock`
                            lock.unlock();
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    } else {
                        logger.logDebug("sending request to global validator");
                        start = System.currentTimeMillis();
                        ValidationRequest request = new ValidationRequest(tid, rwSet, startTS);
                        //put the request in the batch
                        requestBatch.offer(request);
                        dispatcherQStats.reportArrival();
                        if (requestBatch.size() == BATCHSIZE) {
                            dispatcher.flushBatch(myLoadInfo);
                        }
                        while (!responseMap.containsKey(tid)) {
                            try {
                                responseArrived.await();
                            } catch (Exception e) {
                                e.printStackTrace();
                                lock.lock();
                            }
                        }
                        ValidateResponse myResponse = responseMap.remove(tid);
                        end = System.currentTimeMillis();
                        globalReqStats.add(end - start);
                        logger.logDebug("received response from global validator:" + myResponse.commit);
                        myResponse.localResponse = false;
                        if (!myResponse.commit) {
                            globalAborts++;
                        }
                        lock.unlock();
                        return myResponse;
                    }
                }
            } catch (Exception e) {
                if (lock.isHeldByCurrentThread())
                    lock.unlock();
                e.printStackTrace();
                return null;
            } finally {
                if (lock.isHeldByCurrentThread()) {
                    lock.unlock();
                }
            }
        }
    }

    // process global validation response
    void processResponse(String tid, ValidateResponse response, RWSetInfo rwSet) {
        lock.lock();
        if (response.commit) {
            //mark the commit timestamps for updated object
            validator.update(response.commitTS, rwSet);
        }
        for (String nodeId : rwSet.writeSet) {
            // remove this item from tentative writeset.
            tentativeWSet.remove(nodeId);
        }
        responseMap.put(tid, response);    // store the response object
        lock.unlock();
    }

    boolean isBatchReady() {
        return requestBatch.size() >= BATCHSIZE;
    }

    Queue<ValidationRequest> fetchBatch() {
        lock.lock();
        Queue<ValidationRequest> currentBatch = requestBatch;
        requestBatch = new LinkedList<>();
        lock.unlock();
        if (currentBatch.size() > 0) {
            dispatcherQStats.reportDeparture(currentBatch.size());
            queueSizeStats.add(currentBatch.size());
        }
        return currentBatch;
    }

    void notifyWaitingThreads() {
        lock.lock();
        responseArrived.signalAll();
        lock.unlock();
    }

    public String getStats() {
        float localReqs = localReqStats.getNumSamples();
        float globalReqs = globalReqStats.getNumSamples();
        float totalReqs = localReqs + globalReqs;
        float fracLocal = localReqs / totalReqs;
        float fracGlobal = globalAborts / globalReqs;
        StringBuilder buffer = new StringBuilder();
        buffer.append("Local Validator stats:\n");
        buffer.append("Local Aborts: " + localReqs + " Global Aborts:" + globalAborts + " Global Commits:" + (globalReqs - globalAborts) + " \n");
        buffer.append("Fraction of local aborts: " + fracLocal + "\n");        // (localAborts / (localAborts+globalAborts+globalCommit)
        buffer.append("Fraction of global aborts: " + fracGlobal + "\n");    // (globalAborts / (globalAborts+globalCommits)
        buffer.append("Turnaround time for local reqs: avg=" + localReqStats.getMean() + " med=" + localReqStats.getMedian() + "\n");
        buffer.append("Turnaround time for global reqs: avg=" + globalReqStats.getMean() + " med=" + globalReqStats.getMedian() + "\n");
        if (useBatching) {
            buffer.append("Avg dispatcher queue size: " + dispatcherQStats.getAvgQueueLength() + " (with QStat) " + queueSizeStats.getMean() + " (with StatContainer)\n");
            buffer.append("Avg time in dispatcher queue: " + dispatcherQStats.getAvgWaitingTime() + "\n");
        }
        return buffer.toString();
    }
}
