/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import beehive.thrift.BeehiveThriftBuilder;
import beehive.thrift.RWSetInfo_thrift;
import beehive.thrift.ValidateResponse_thrift;
import beehive.thrift.ValidationService_thrift;
import beehive.util.Config;
import beehive.util.RWSetInfo;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

import beehive.workpool.LoadInfo;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.transport.TTransportException;

/**
 * Created by khand055 on 4/4/17.
 */
public class ValidationServiceImpl implements ValidationService {

    private Map<Integer, ValidationService> certifierHandleTable;
    private Config configParams;
    private boolean bUseReplicatedValidator = false;
    private String serverIdentifyPolicy = "RANDOM";
    private ValidationService globalValidator;
    private Random r;
    private Map<Integer, ArrayList<Long>> keyRangeServerMap;
    private long numOfNodes;
    private static final int STS_MANAGER = 0;

    //partitioned graph nodes location table
    public int[] partitionedGraphHostNodesTable;

    public static String RPC_MECHANISM = "RMI";

    private LinkedBlockingQueue<ValidationService_thrift.Client> clientHandles = new LinkedBlockingQueue<ValidationService_thrift.Client>();
    private Map<Integer, LinkedBlockingQueue<ValidationService_thrift.Client>> serviceThreadPools;
    private Map<Integer, String> serverGroup = new HashMap<>();

    public ValidationServiceImpl(String validationHostDetail, int[] partitionedGraphHostNodesTable) throws IOException {
        this.partitionedGraphHostNodesTable = partitionedGraphHostNodesTable;

        configParams = Config.getInstance();
        r = new Random();
        keyRangeServerMap = new HashMap<Integer, ArrayList<Long>>();
        serviceThreadPools = new ConcurrentHashMap<Integer, LinkedBlockingQueue<ValidationService_thrift.Client>>();

        bUseReplicatedValidator = configParams.getBoolValue("Validator.useReplicated");
        serverIdentifyPolicy = configParams.getStringValue("Validator.serverIdentifyPolicy");
        if (configParams.has("Validator.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("Validator.RPC_MECHANISM");
        }


        if (serverIdentifyPolicy == null || serverIdentifyPolicy.length() == 0) {
            serverIdentifyPolicy = "RANDOM";
        }


        if (bUseReplicatedValidator) {

            certifierHandleTable = new HashMap<Integer, ValidationService>();
            prepareCertifierHandle(validationHostDetail);
            numOfNodes = configParams.getIntValue("Validator.nodeSize");

            initKeyRangeServerMap(numOfNodes);

            registerClientHandleReplicatedValidation();
        } else {
            //fetch RMI handle
            String url = "//" + validationHostDetail + "/WorkpoolServer";
            try {
                globalValidator = (ValidationService) Naming.lookup(url);
            } catch (Exception e) {
                e.printStackTrace();
            }

            //fetch thrift handle
            registerClientHandle();
        }


    }

    public void registerClientHandle() {
        String thriftHostname = null;
        int thriftPort;
        thriftHostname = configParams.getStringValue("Validator.GlobalWorkpoolImpl_THRIFT_HOST");
        thriftPort = configParams.getIntValue("Validator.GlobalWorkpoolImpl_THRIFT_PORT");
        int threadcount = configParams.getIntValue("Worker.WORKER_COUNT");

        for (int i = 0; i < threadcount; i++) {
            ValidationService_thrift.Client client = null;
            try {
                TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(thriftHostname, thriftPort);
                client = new ValidationService_thrift.Client(protocol);
                clientHandles.offer(client);
            } catch (Exception e) {
                System.out.println("ValidationServiceImpl :: Error in registerClientHandle() while creating client handle! Please check the hostname and port in configFile for Thrift Validation Server.");
                e.printStackTrace();
            }
        }
    }


    public void registerClientHandleReplicatedValidation() {
        System.out.println("Replicated Validation Client Registration");

        int clientPort = configParams.getIntValue("Validator.thrift.clientPort");
        int thriftConnectionPoolSize = configParams.getIntValue("Validator.clientThriftConnectionPool");

        try {
            for (int i = 0; i < serverGroup.size(); i++) {

                ValidationService_thrift.Client client;

                //Trying to open connection to each peer
                System.out.println("Looking Up Peer : " + serverGroup.get(i) + ":" + (clientPort + i));
                while (true) {
                    try {
                        TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(serverGroup.get(i), clientPort + i);
                        protocol.getTransport().close();
                        System.out.println("Peer Up : " + serverGroup.get(i) + ":" + (clientPort + i));
                        break;
                    } catch (Exception e) {
                        try {
                            Thread.sleep(10);
                        } catch (Exception e1) {
                        }
                    }
                }

                LinkedBlockingQueue<ValidationService_thrift.Client> clientsPool = new LinkedBlockingQueue<ValidationService_thrift.Client>();

                for (int j = 0; j < thriftConnectionPoolSize; j++) {

                    TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(serverGroup.get(i), clientPort + i);
                    client = new ValidationService_thrift.Client(protocol);
                    clientsPool.add(client);
                }
                serviceThreadPools.put(i, clientsPool);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

    }


    @Override
    public ValidateResponse validate(long startTS, RWSetInfo rwSet, LoadInfo myLoad) throws RemoteException {
        int serverId = 0;
        ValidateResponse vr = null;
        ValidationService serverHandle = null;
        ValidationService_thrift.Client thriftServerHandle = null;

        if (bUseReplicatedValidator) {

            if (RPC_MECHANISM.equals("RMI")) {

                switch (serverIdentifyPolicy) {

                    case "RANDOM":
                        serverId = r.nextInt(certifierHandleTable.size() - 2) + 2;
                        serverHandle = certifierHandleTable.get(serverId);
                        break;

                    case "KEY_BASED":
                        serverId = getKeyBasedServer(rwSet);
                        serverHandle = certifierHandleTable.get(serverId);
                        break;
                }

                vr = serverHandle.validate(startTS, rwSet, myLoad);
            } else if (RPC_MECHANISM.equals("THRIFT")) {

                switch (serverIdentifyPolicy) {

                    case "RANDOM":
                        serverId = r.nextInt(certifierHandleTable.size() - 2) + 2;
                        try {
                            thriftServerHandle = serviceThreadPools.get(serverId).take();
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }
                        break;

                    case "KEY_BASED":
                        serverId = getKeyBasedServer(rwSet);
                        try {
                            thriftServerHandle = serviceThreadPools.get(serverId).take();
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }
                        break;
                }

                ValidateResponse_thrift tempRes = null;
                try {
                    System.out.println("Client-Request  ValidationServiceImpl:validate:startTS: " + startTS);
                    System.out.println("Client-Request  ValidationServiceImpl:validate_thrift:rwSet: " + rwSet.readSet + ", " + rwSet.writeSet + ", " + rwSet.unusedSet);
                    tempRes = thriftServerHandle.validate_thrift(startTS, new RWSetInfo_thrift(rwSet.readSet, rwSet.writeSet, rwSet.unusedSet));
                    serviceThreadPools.get(serverId).add(thriftServerHandle);
                } catch (TException e) {
                    e.printStackTrace();
                }
                vr = new ValidateResponse(tempRes);
                //System.out.println("Client-Response  ValidationServiceImpl:validate:startTS: " + startTS + "  Commit: " + vr.commit);
            }


        } else {
            if (RPC_MECHANISM.equals("THRIFT")) {
                ValidationService_thrift.Client client = null;
                try {
                    client = clientHandles.take();
                } catch (InterruptedException e1) {
                    e1.printStackTrace();
                }

                try {
                    vr = new ValidateResponse(client.validate_thrift(startTS, new RWSetInfo_thrift(rwSet.readSet, rwSet.writeSet, rwSet.unusedSet)));
                } catch (TTransportException e) {
                    e.printStackTrace();
                } catch (TException e) {
                    e.printStackTrace();
                }
                clientHandles.offer(client);
            } else {
                serverHandle = globalValidator;
                vr = serverHandle.validate(startTS, rwSet, myLoad);
            }
        }
        return vr;
    }

    private int getKeyBasedServer(RWSetInfo rwSet) {

        int serverId = 0;
        int serverSize = certifierHandleTable.size();
        long[] serverCount = new long[serverSize];
        for (String readItem : rwSet.readSet) {
            serverCount[getServerFromNodeKey(Long.parseLong(readItem))]++;
        }
        for (String writeItem : rwSet.writeSet) {
            serverCount[getServerFromNodeKey(Long.parseLong(writeItem))]++;
        }

        long maxCount = Long.MIN_VALUE;
        for (int i = 0; i < serverSize; i++) {
            if (serverCount[i] > maxCount) {
                maxCount = serverCount[i];
                serverId = i;
            }
        }
        return serverId;
    }

    public int getServerFromNodeKey(long nodeId) {

        List<Long> keyRange;
        for (int serverId : keyRangeServerMap.keySet()) {
            keyRange = keyRangeServerMap.get(serverId);
            if (nodeId >= keyRange.get(0) && nodeId <= keyRange.get(1)) {
                return serverId;
            }
        }
        return -1;
    }

    @Override
    public Hashtable<String, ValidateResponse> validateBatch(Queue<ValidationRequest> requestBatch, LoadInfo myLoad) throws RemoteException {
        return null;
    }

    @Override
    public void checkPoint() throws RemoteException {
        globalValidator.checkPoint();
    }

    @Override
    public long getSTS() throws RemoteException {
        long stsValue = 0;
        ValidationService serverHandle = null;
        ValidationService_thrift.Client thriftServerHandle = null;

        if (bUseReplicatedValidator) {


            if (RPC_MECHANISM.equals("THRIFT")) {

                try {
                    thriftServerHandle = serviceThreadPools.get(STS_MANAGER).take();
                    stsValue = thriftServerHandle.getSTS_thrift();
                    serviceThreadPools.get(STS_MANAGER).add(thriftServerHandle);

                } catch (Exception e) {
                    e.printStackTrace();
                }

            } else if (RPC_MECHANISM.equals("RMI")) {

                serverHandle = certifierHandleTable.get(STS_MANAGER);
                stsValue = serverHandle.getSTS();
            }
        } else {
            if (RPC_MECHANISM.equals("THRIFT")) {
                ValidationService_thrift.Client client = null;
                try {
                    client = clientHandles.take();
                } catch (InterruptedException e1) {
                    e1.printStackTrace();
                }

                try {
                    stsValue = client.getSTS_thrift();
                } catch (TTransportException e) {
                    e.printStackTrace();
                } catch (TException e) {
                    e.printStackTrace();
                }
                clientHandles.offer(client);
            } else {
                serverHandle = globalValidator;
                stsValue = serverHandle.getSTS();
            }
        }
        return stsValue;
    }

    @Override
    public long reportCompletion(long commitTS) throws RemoteException {

        long reportCompletionStatus = 0;
        ValidationService serverHandle = null;
        ValidationService_thrift.Client thriftServerHandle = null;


        if (bUseReplicatedValidator) {

            if (RPC_MECHANISM.equals("THRIFT")) {

                try {

                    thriftServerHandle = serviceThreadPools.get(STS_MANAGER).take();
                    reportCompletionStatus = thriftServerHandle.reportCompletion_thrift(commitTS);
                    serviceThreadPools.get(STS_MANAGER).add(thriftServerHandle);

                } catch (Exception e) {
                    e.printStackTrace();
                }

            } else if (RPC_MECHANISM.equals("RMI")) {

                serverHandle = certifierHandleTable.get(STS_MANAGER);
                reportCompletionStatus = serverHandle.reportCompletion(commitTS);
            }

        } else {
            if (RPC_MECHANISM.equals("THRIFT")) {
                ValidationService_thrift.Client client = null;
                try {
                    client = clientHandles.take();
                } catch (InterruptedException e1) {
                    e1.printStackTrace();
                }

                try {
                    reportCompletionStatus = client.reportCompletion_thrift(commitTS);
                } catch (TTransportException e) {
                    e.printStackTrace();
                } catch (TException e) {
                    e.printStackTrace();
                }
                clientHandles.offer(client);
            } else {
                serverHandle = globalValidator;
                reportCompletionStatus = serverHandle.reportCompletion(commitTS);
            }
        }
        return reportCompletionStatus;
    }

    @Override
    public long reportCompletion(Vector<Long> commitTSList) throws RemoteException {

        long reportCompletionStatus = 0;
        ValidationService serverHandle = null;
        ValidationService_thrift.Client thriftServerHandle = null;


        if (bUseReplicatedValidator) {

            if (RPC_MECHANISM.equals("THRIFT")) {

                try {

                    //System.out.println("Client-Request  ValidationServiceImpl:reportCompletionBatch:commitTSList: " + commitTSList.toString());
                    thriftServerHandle = serviceThreadPools.get(STS_MANAGER).take();
                    reportCompletionStatus = thriftServerHandle.reportCompletionBatch_thrift(commitTSList);
                    serviceThreadPools.get(STS_MANAGER).add(thriftServerHandle);
                    //System.out.println("Client-Response  ValidationServiceImpl:reportCompletionBatch:commitTSList: " + commitTSList.toString());

                } catch (Exception e) {
                    e.printStackTrace();
                }

            } else if (RPC_MECHANISM.equals("RMI")) {

                serverHandle = certifierHandleTable.get(STS_MANAGER);
                reportCompletionStatus = serverHandle.reportCompletion(commitTSList);
            }

        } else {
            if (RPC_MECHANISM.equals("THRIFT")) {
                ValidationService_thrift.Client client = null;
                try {
                    client = clientHandles.take();
                } catch (InterruptedException e1) {
                    e1.printStackTrace();
                }

                try {
                    reportCompletionStatus = client.reportCompletionBatch_thrift(commitTSList);
                } catch (TTransportException e) {
                    e.printStackTrace();
                } catch (TException e) {
                    e.printStackTrace();
                }
                clientHandles.offer(client);
            } else {
                serverHandle = globalValidator;
                reportCompletionStatus = serverHandle.reportCompletion(commitTSList);
            }
        }
        return reportCompletionStatus;
    }


    private void prepareCertifierHandle(String configFile) throws IOException {

        BufferedReader bufferedReader = new BufferedReader(new FileReader(configFile));
        String line;
        String hostname;
        int hostId;
        int hostIdCounter = 0;
        //int i=0;
        while ((line = bufferedReader.readLine()) != null) {
            if (line.charAt(0) == '#') continue;
            String[] splited = line.split("\\s+");
            hostname = splited[0];
            hostId = hostIdCounter++;//Integer.parseInt(splited[1]);
            //serverList.add(hostId);
            String url = "//" + hostname + ":" + Registry.REGISTRY_PORT + "/ValidationService" + hostId;

            ValidationService certifierHandle = lookup(url);
            certifierHandleTable.put(hostId, certifierHandle);
            serverGroup.put(hostId, hostname);

        }
    }

    private ValidationService lookup(String url) {
        System.out.println("Looking up certifier:" + url);
        while (true) {
            try {
                ValidationService certifierHandle = (ValidationService) Naming.lookup(url);
                //System.out.println("certifier up");
                return certifierHandle;
            } catch (Exception e) {
                try {
                    //System.out.print(".");
                    Thread.sleep(10);
                } catch (Exception e1) {

                }
            }
        }
    }

    private void initKeyRangeServerMap(long numOfNodes) {

        String loaderType = configParams.getStringValue("Server.FileLoader").trim();
        String hashingScheme = configParams.getStringValue("Server.NODE_HASHING_SCHEME").trim();

        if (loaderType.equals("Bucketized") && hashingScheme.equals("GRAPH_PARTITION")) {
            setGraphPartitionedServerMap();
        } else {
            setKeyRangeServerMap(numOfNodes);
        }
    }

    private void setGraphPartitionedServerMap() {

        int validationClusterSize = certifierHandleTable.size();
        validationClusterSize = validationClusterSize - 2; //Node 0 will be for timestamp manager STS and 1 for CTS

        int[] clusterPerServiceReplica = new int[validationClusterSize];
        int graphClusterSize = partitionedGraphHostNodesTable.length;
        int iterator = 0;

        while (graphClusterSize > 0) {
            clusterPerServiceReplica[iterator]++;
            iterator = (iterator + 1) % validationClusterSize;
            graphClusterSize--;
        }

        //Initialize range to -1 for all
        for (int i = 2; i < certifierHandleTable.size(); i++) {
            keyRangeServerMap.put(i, new ArrayList<Long>(Arrays.asList(-1L, -1L)));
        }

        int firstIndex = 0;
        int lastIndex = 0;
        int clusterIndex = 0;


        for (int i = 0; i < certifierHandleTable.size() - 2; i++) {

            if (i < clusterPerServiceReplica.length) {
                clusterIndex = clusterIndex + clusterPerServiceReplica[i];
                lastIndex = partitionedGraphHostNodesTable[clusterIndex - 1];
                keyRangeServerMap.put(i + 2, new ArrayList<Long>(Arrays.asList((long) firstIndex, (long) lastIndex)));
                firstIndex = lastIndex + 1;
            }
        }
    }

    private void setKeyRangeServerMap(long numOfNodes) {

        int totalClusterSize = certifierHandleTable.size();

        totalClusterSize = totalClusterSize - 2; //Node 0 will be for timestamp manager STS and 1 for CTS

        long nodePerCluster = numOfNodes / totalClusterSize;
        long residue = numOfNodes % totalClusterSize; // shall be added to last Host

        long start = 0;
        long startIndex = 0;
        long endIndex = 0;

        //Initialize range to -1 fpr all
        for (int i = 2; i < certifierHandleTable.size(); i++) {
            keyRangeServerMap.put(i, new ArrayList<Long>(Arrays.asList(-1L, -1L)));
        }

        if (nodePerCluster > 0) { //atleast one node per cluster
            for (int i = 2; i < certifierHandleTable.size(); i++) {
                startIndex = start;
                endIndex = start + nodePerCluster - 1;
                if (i == certifierHandleTable.size() - 1) {// last item
                    endIndex += residue;
                }
                keyRangeServerMap.put(i, new ArrayList<Long>(Arrays.asList(startIndex, endIndex)));
                start += nodePerCluster;
            }
        } else if (residue > 0) {
            keyRangeServerMap.put(2, new ArrayList<Long>(Arrays.asList(startIndex, endIndex))); //Put residue in first node
        }
    }

}
