/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validation;

import java.util.*;
import java.util.concurrent.*;

import beehive.util.*;

public class Validator {

    private ConcurrentHashMap<String, Long> conflictTable;

    public Validator() {
        conflictTable = new ConcurrentHashMap<>();
    }


    public long validate(long startTS, RWSetInfo rwSet, HashSet<String> conflictSet) {

        // check read set
        long ret = 0; // 0 means no conflict
        for (String nodeId : rwSet.readSet) {
            Long writerTS = conflictTable.get(nodeId);
            if (writerTS != null && writerTS > startTS) {
                ret = writerTS;
                conflictSet.add(nodeId);
            }
        }
        // check write set
        for (String nodeId : rwSet.writeSet) {
            Long writerTS = conflictTable.get(nodeId);
            // if writerTS > startTS => conflict!
            if (writerTS != null && writerTS > startTS) {
                ret = writerTS;
                conflictSet.add(nodeId);
            }
        }

        return ret;

    }

    public void update(long cmtTS, RWSetInfo rwSet) {
        for (String nodeId : rwSet.writeSet) {
            conflictTable.put(nodeId, cmtTS);
        }
    }


    // returns true if no conflict else false
    synchronized long validateAndUpdate(long startTS, long cmtTS, RWSetInfo rwSet, HashSet<String> refetchSet) {
        long conflictingTS = validate(startTS, rwSet, refetchSet);
        // no conflict found, update write items
        if (conflictingTS == 0) {
            update(cmtTS, rwSet);
        } else {
            for (String nodeId : rwSet.unusedSet) {    // check if any of the item in unused set is modified
                Long writerTS = conflictTable.get(nodeId);
                if (writerTS != null && writerTS > startTS) {
                    refetchSet.add(nodeId);
                }
            }
        }
        return conflictingTS;
    }

}
