/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validationService;

import beehive.thrift.*;
import beehive.util.Config;
import beehive.util.Logger;
import beehive.util.RWSetInfo;
import beehive.workpool.LoadInfo;
import beehive.validation.ValidateResponse;
import beehive.validation.ValidationRequest;
import beehive.validation.ValidationService;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TCompactProtocol;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.server.TServer;
import org.apache.thrift.server.TThreadPoolServer;
import org.apache.thrift.transport.TServerSocket;
import org.apache.thrift.transport.TServerTransport;

import java.io.*;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Created by khand055 on 3/23/17.
 */
public class CertifierReplica extends UnicastRemoteObject implements PeerInterface, ValidationService, ValidationService_thrift.Iface, PeerInterface_thrift.Iface {


    static public Map<Integer, LinkedBlockingQueue<PeerInterface_thrift.Client>> peerThreadPools; // Peer thread pool for each host
    static public int peerThriftPoolSize = 20;
    static public int clientPort;
    static public int serverPort;

    long stableTS;
    int myId;
    //Keeping track of all
    ReplicationConfig replicaConfig;
    TimestampManager timestampManager;
    ExecutorService globalExecutors;
    Validator validator;
    Logger logger;
    LinkedBlockingQueue<Long> pendingReportCompletionQueue;
    RMIAdapter rmiAdapter;

    public  PrintStream ps;

    boolean bDebug = false;
    String rpcMechanism;
    boolean bWriteStatsToFile = false;


    Config configParams;

    final AtomicLong clientRequestCount = new AtomicLong();
    final AtomicLong clientCommitCount = new AtomicLong();
    final AtomicLong clientAbortCount = new AtomicLong();

    final AtomicLong peerRequestCount = new AtomicLong();
    final AtomicLong peerCommitCount = new AtomicLong();
    final AtomicLong peerAbortCount = new AtomicLong();

    long transactionId;

    //Map<Long, ArrayList<RWSetInfo>> pendingValidationRequest = new ConcurrentHashMap<Long, ArrayList<RWSetInfo>>();
    //Transaction Id to RWSet
    Map<Long, RWSetInfo> pendingValidationRequest = new ConcurrentHashMap<Long, RWSetInfo>();

    public synchronized long getStableTS() {
        return stableTS;
    }

    public synchronized void setStableTS(long stableTS) {
        if( stableTS > this.stableTS ) {
            this.stableTS = stableTS;
        }
    }

    public synchronized long getTransactionId(){
        long tid = transactionId;
        setTransactionId(transactionId + replicaConfig.serverList.size());
        return tid;
    }

    public synchronized void setTransactionId(long transactionId){
        this.transactionId = transactionId;
    }

    public CertifierReplica(String validationReplicationConfigFile, int serverId, String configFile) throws RemoteException, IOException {

    	//setting THRIFT socket timeout 
        if(configParams.has("System.THRIFT_SO_TIMEOUT")){
        	int timeoutValue=configParams.getIntValue("System.THRIFT_SO_TIMEOUT");
        	BeehiveThriftBuilder.setThriftConnectionTimeout(timeoutValue);
        }
		
    	myId = serverId;
        //Ideally, the below check should be Config constructor itself
        Config configParams = Config.getInstance();
        if( configParams == null){
            configParams = new Config(configFile);
        }

        bDebug = configParams.getBoolValue("Validator.debug");
        bWriteStatsToFile = configParams.getBoolValue("Validator.writeStatsToFile");
        rpcMechanism = configParams.getStringValue("Validator.RPC_MECHANISM");
        globalExecutors = Executors.newCachedThreadPool();
        pendingReportCompletionQueue = new LinkedBlockingQueue<Long>();
        peerThriftPoolSize = configParams.getIntValue("Validator.serverThriftConnectionPool");
        clientPort = configParams.getIntValue("Validator.thrift.clientPort");
        serverPort = configParams.getIntValue("Validator.thrift.serverPort");
        //globalExecutors = Executors.newFixedThreadPool(threadPoolSize);
        validator = new Validator(this);
        logger = new Logger();
        logger.setDebug(bDebug);
        logger.setWarning(false);

        //Initialize timestamp manager
        if(serverId == 0 || serverId == 1){
            timestampManager = new TimestampManager();
        }

        //Initialization done by all server
        replicaConfig = new ReplicationConfig(validationReplicationConfigFile, serverId, this, configFile);

        System.out.println("Starting  + " + replicaConfig.myID);
        //Initialization done by Host 0 server STS Timestamp Manager
        //ps = new PrintStream("Server"+replicaConfig.myID);
        //ps.print("ID:" + replicaConfig.myID);

        //Creating timestamp wrapper for thrift based calls
        rmiAdapter = new RMIAdapter(this, rpcMechanism);

        //Initialization done by Host 1 server CTS Timestamp Manager
        if( serverId == 1){
            //timestampManager = new TimestampManager();
            STSPeriodicFetch stsPeriodicFetch = new STSPeriodicFetch(this);
            stsPeriodicFetch.start();
        }

        //Report Completition dispatcher for other replica
        if( serverId != 0 && serverId != 1) {
            System.out.println("Starting dispatcher + " + replicaConfig.myID);
            ReportCompletionDispatcher reportCompletionDispatcher = new ReportCompletionDispatcher(this);
            reportCompletionDispatcher.start();
        }

        if( bWriteStatsToFile ){
            StatsCollectorThread statsCollectorThread = new StatsCollectorThread(this);
            statsCollectorThread.start();
        }

    }


    public static void main(String[] args) throws Exception {

        int serverId = Integer.parseInt(args[0]);
        String validationReplicationConfigFile = args[1];
        //long numOfNodes = Long.valueOf(args[2]).longValue(); //Number of vertices
        String configFile = args[2];
        //int thriftPortValidationService = Integer.parseInt(args[3]);
        //int thriftPortPeerService = Integer.parseInt(args[4]);

        CertifierReplica certifierReplica = new CertifierReplica(validationReplicationConfigFile, serverId, configFile);
        certifierReplica.setTransactionId(certifierReplica.replicaConfig.myID);

        peerThreadPools = new ConcurrentHashMap<Integer, LinkedBlockingQueue<PeerInterface_thrift.Client>>();
        int serverGroupSize = certifierReplica.replicaConfig.serverGroup.size();
        for( int i=0; i<serverGroupSize; i++){
            peerThreadPools.put(i, new LinkedBlockingQueue<PeerInterface_thrift.Client>());
        }

        // Spawn a thread to start validation service
        InitializeThriftValidationService initValidationService = new InitializeThriftValidationService(certifierReplica, clientPort+serverId);
        initValidationService.start();

        // Spawn a thread to start peer interface
        InitializeThriftPeerInterface initPeerService = new InitializeThriftPeerInterface(certifierReplica, serverPort+serverId);
        initPeerService.start();

        // Spawn a thread to get handle to each of the peer thrift interface
        InitializePeerThriftConnectionPool initConnectionPool = new InitializePeerThriftConnectionPool(certifierReplica);
        initConnectionPool.start();
    }

    ////////////////////////// Peer Interface RMI Begin /////////////////////////////////////////////////////////

    //Will always be called on Replica-1
    @Override
    public long[] getCtsSts() throws RemoteException {

        long[] result = new long[2];
        try{

            result[0]= timestampManager.getGTS();
            result[1] = timestampManager.getSTS();

        }catch (Exception e){
            e.printStackTrace();
        }

        return result;
    }

    @Override
    //Called by Peer Server
    public PeerResponse validate(long startTS, RWSetInfo rwSet, long transactionId) throws RemoteException {

        PeerResponse res = new PeerResponse();
        try {
            peerRequestCount.incrementAndGet();

            logger.logDebug("validate Server REQ: STS : " + startTS + " TID : " + transactionId);//+ " | ReadSet : " + rwSet.readSet.toString() + " | WriteSet : " + rwSet.writeSet.toString());

            pendingValidationRequest.put(transactionId, rwSet);
            long response = validator.validate(startTS, rwSet, new HashSet<>());

            if (response == 0) {
                res.responseType = PeerResponse.ALL_LOCKS_ACQUIRED;
                peerCommitCount.incrementAndGet();
            } else {
                res.responseType = PeerResponse.ABORT;
                peerAbortCount.incrementAndGet();
            }
            //res.responseType = (response == 0) ? PeerResponse.ALL_LOCKS_ACQUIRED : PeerResponse.ABORT;
            res.serverId = replicaConfig.myID;
        }catch (Exception e){
            e.printStackTrace();
        }
        return res;
    }

    //Will always be called on Replica-1
    @Override
    public long getGTS() throws RemoteException{

        long cts=0;
        try{
            cts= timestampManager.getGTS();
        }catch (Exception e){
            e.printStackTrace();
        }
        return cts;
    }

    //Will always be called on Replica-0
    @Override
    public long getSTS() throws RemoteException {

        long sts=0;
        try {
            sts = timestampManager.getSTS();
        }catch (Exception e){
            e.printStackTrace();
        }
        return sts;

        //return timestampManager.getSTS();
    }

    @Override
    public PeerResponse commit(long startTS, long commitTS, long transactionId) throws RemoteException {

        PeerResponse res = new PeerResponse();
        try {
            validator.update(commitTS, pendingValidationRequest.get(transactionId));
            validator.releaseLock(startTS, pendingValidationRequest.get(transactionId));
            pendingValidationRequest.remove(transactionId);
            //logger.logDebug("TID : " + transactionId + " Remove TID after COMMIT ");
            res.responseType = PeerResponse.COMMITED;
        }catch (Exception e){
            e.printStackTrace();
        }
        return res;
    }

    @Override
    public PeerResponse abort(long startTS, long transactionId, boolean bReleaseLock) throws RemoteException {

        PeerResponse res = new PeerResponse();
        try {
            if (bReleaseLock) {
                validator.releaseLock(startTS, pendingValidationRequest.get(transactionId));
            }
            pendingValidationRequest.remove(transactionId);
            res.responseType = PeerResponse.ABORT;
        }catch (Exception e){
            e.printStackTrace();
        }
        return res;

    }

    ////////////////////////// Peer Interface RMI End ///////////////////////////////////////////////////////////





    ////////////////////////// Validation Service RMI Begin /////////////////////////////////////////////////////////

    @Override
    //Called by Client
    public ValidateResponse validate(long startTS, RWSetInfo rwSet, LoadInfo loadInfo) throws RemoteException {

        //System.out.println(startTS);
        ValidateResponse vr = null;



        try {
            clientRequestCount.incrementAndGet();

            logger.logDebug("TRANSACTION_BEGIN");

            long commitTS;
            //logger.logDebug("validate Client REQ: STS : "+startTS );
            //Get the STS and CTS from TimestampManager

            //stableTS = (long) replicaConfig.peerHandleTable.get(0).getSTS();
            //commitTS = (long) replicaConfig.peerHandleTable.get(0).getGTS();
            long[] ctsSts = new long[2];

            ctsSts = rmiAdapter.getCtsSts();


            commitTS = ctsSts[0];
            setStableTS(ctsSts[1]);
               
//        if( replicaConfig.peerHandleTable.size() == 1) { // There is only one replicated server machine
//
//            logger.logDebug("CertifierReplica:validate:: STS : " + startTS + " | ReadSet : " + rwSet.readSet + " | WriteSet : " + rwSet.writeSet);
//
//            vr = new ValidateResponse();
//
//            long response = validator.validate(startTS, rwSet, new HashSet<>());
//
//            if( response == 0) {
//                validator.update(commitTS, rwSet);
//                validator.releaseLock(rwSet);
//                vr.latestSTS = getStableTS();
//                vr.commitTS = commitTS;
//                vr.commit = true;
//            } else {
//                validator.releaseLock(rwSet);
//                reportCompletion(commitTS);
//                vr.latestSTS = getStableTS();
//                vr.commit = false;
//            }
//
//        } else {

            Coordinator coordinator = new Coordinator(startTS, commitTS, rwSet, this, globalExecutors);
            try {
                vr = coordinator.validate();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            //}

            logger.logDebug("TRANSACTION_END");
            if (vr.commit) {
                clientCommitCount.incrementAndGet();
            } else {
                clientAbortCount.incrementAndGet();
            }
        }catch (Exception e){
            e.printStackTrace();
        }

        return vr;
    }


    //Always called on Replica-0
    @Override
    public long reportCompletion(long commitTS) throws RemoteException {

        //logger.logDebug("Received reportCompletion call COMMIT TS : " + commitTS);
        long currentSTS;
        timestampManager.reportDone(commitTS);
        currentSTS = timestampManager.getSTS();
        return currentSTS;
    }

    //Always called on Replica-0
    @Override
    public long reportCompletion(Vector<Long> commitTSList) throws RemoteException {
        for( long commitTS : commitTSList){
            this.reportCompletion(commitTS);
        }

        return timestampManager.getSTS();
    }


    ////////////////////////// Validation Service RMI End /////////////////////////////////////////////////////////

    @Override
    public Hashtable<String, ValidateResponse> validateBatch(Queue<ValidationRequest> requestBatch, LoadInfo myLoad) throws RemoteException {
        return null;
    }

    @Override
    public void checkPoint() throws RemoteException {

    }


    //Will be started by Certifier-1 which has GTS Server
    class STSPeriodicFetch extends Thread {

        CertifierReplica certifierReplica = null;
        //PeerInterface stsTimestampMgr = null;
        long sts;

        public STSPeriodicFetch(CertifierReplica certifierReplica) throws FileNotFoundException {
            this.certifierReplica = certifierReplica;
            //this.stsTimestampMgr = this.certifierReplica.replicaConfig.peerHandleTable.get(0);
        }

        @Override
        public void run() {
            try {
                while (true) {
                    try {
                        Thread.sleep(1); //Call every 1 milliseconds
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }

                    sts = certifierReplica.rmiAdapter.getSTS();
//                    if( sts < this.certifierReplica.timestampManager.getSTS()){
//                        System.out.println("Problem!! ServerID: " + this.certifierReplica.replicaConfig.myID + " STS-Server0: "  + sts + " STS-Server1: " + this.certifierReplica.timestampManager.getSTS());
//                    }
                    this.certifierReplica.timestampManager.setSts( sts );

                }
            }catch (Exception e){
                e.printStackTrace();
            }
        }

    }

    //*********************************************  Thrift Implementation *****************************************************
    @Override
    public long getSTS_thrift() {

        long sts=0;
        try {
            sts = this.getSTS();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return sts;
    }

    @Override
    public ValidateResponse_thrift validate_thrift(long startTS, RWSetInfo_thrift rwSet) throws TException {

        System.out.println("Server  CertifierReplica:validate_thrift:startTS: " + startTS);
        System.out.println("Server  CertifierReplica:validate_thrift:rwSet: " + rwSet.readSet+", "+rwSet.writeSet+", "+rwSet.unusedSet);
        
        ValidateResponse vr = null;
        try {
            vr = validate(startTS, new RWSetInfo(rwSet), null);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        // Vector<LoadInfo_thrift> loadInfoVec = new Vector<LoadInfo_thrift>();
        // for(LoadInfo r : vr.loadVector){
        // 	loadInfoVec.add(new LoadInfo_thrift(r.load, r.tsMatch, r.hostname));
        // }
        System.out.println("Server Return  CertifierReplica:validate_thrift:startTS: " + startTS);
        return new ValidateResponse_thrift(vr.commit, vr.commitTS, vr.latestSTS, vr.refetchSet, vr.localResponse, vr.speculativeAbort, vr.conflictingTS);
    }

    @Override
    public long reportCompletion_thrift(long commitTS) {

        long sts = 0;
        try {
            sts = this.reportCompletion(commitTS);
        } catch (RemoteException e) {
            e.printStackTrace();
        }

        return sts;
    }

    @Override
    public long reportCompletionBatch_thrift(List<Long> commitTSList) {

        //System.out.println("Server  CertifierReplica:reportCompletionBatch_thrift:commitTSList: " + commitTSList.toString());
        Vector<Long> reportCompletitionVector = new Vector<Long>();
        long sts = 0;
        try {
            reportCompletitionVector.addAll(commitTSList);
            sts = this.reportCompletion(reportCompletitionVector);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return sts;

    }

    @Override
    public Map<String, ValidateResponse_thrift> batchValidate_thrift(List<ValidationRequest_thrift> requests) {
        return new HashMap<>();
    }




    ////////////////////////////////////// Peer Interface Thrift Begin /////////////////////////////////////////////////////////

    @Override
    public long reportCompletion_peer_thrift(long commitTS) throws TException {
        long sts = 0;
        try {
            sts = this.reportCompletion(commitTS);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return sts;
    }

    @Override
    public long reportCompletionBatch_peer_thrift(List<Long> commitTS) throws TException {
        Vector<Long> reportCompletitionVector = new Vector<Long>();
        long sts = 0;
        try {
            reportCompletitionVector.addAll(commitTS);
            sts = this.reportCompletion(reportCompletitionVector);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return sts;
    }

    @Override
    public PeerResponse_thrift validate_peer_thrift(long startTS, RWSetInfo_thrift rwSet, long transactionId) throws TException {

        PeerResponse_thrift peerResponse_thrift = new PeerResponse_thrift();
        try {

            PeerResponse pr = this.validate(startTS, new RWSetInfo(rwSet), transactionId);
            peerResponse_thrift.serverId = pr.serverId;
            peerResponse_thrift.responseType = pr.responseType;

        } catch (RemoteException e) {
            e.printStackTrace();
        }

        return peerResponse_thrift;
    }

    @Override
    public long getSTS_peer_thrift() throws TException {

        long sts=0;
        try {
            sts = this.getSTS();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return sts;
    }

    @Override
    public List<Long> getCtsSts_peer_thrift() throws TException {

        List<Long> listCtsSts = new ArrayList<>();
        try{
            long[] longCtsSts = this.getCtsSts();

            listCtsSts.add(longCtsSts[0]);
            listCtsSts.add(longCtsSts[1]);

        }catch (RemoteException e){

        }
        return listCtsSts;
    }

    @Override
    public PeerResponse_thrift commit_peer_thrift(long startTS, long commitTS, long transactionId) throws TException {

        PeerResponse_thrift peerResponse_thrift = new PeerResponse_thrift();
        try {

            PeerResponse pr = this.commit(startTS, commitTS, transactionId);
            peerResponse_thrift.serverId = pr.serverId;
            peerResponse_thrift.responseType = pr.responseType;

        } catch (RemoteException e) {
            e.printStackTrace();
        }

        return peerResponse_thrift;
    }

    @Override
    public PeerResponse_thrift abort_peer_thrift(long startTS, long transactionId, boolean bReleaseLock) throws TException {

        PeerResponse_thrift peerResponse_thrift = new PeerResponse_thrift();
        try {

            PeerResponse pr = this.abort(startTS, transactionId, bReleaseLock);
            peerResponse_thrift.serverId = pr.serverId;
            peerResponse_thrift.responseType = pr.responseType;

        } catch (RemoteException e) {
            e.printStackTrace();
        }

        return peerResponse_thrift;

    }

    ///////////////////////////////////////// Peer Interface Thrift End //////////////////////////////////////////////////////////

    //Initialize Validation Service Server
    static class InitializeThriftValidationService extends Thread {

        CertifierReplica certifierReplica = null;
        int thriftPort;

        public InitializeThriftValidationService(CertifierReplica certifierReplica, int thriftPort) {
            this.certifierReplica = certifierReplica;
            this.thriftPort = thriftPort;
        }

        @Override
        public void run(){

            try {
                ValidationService_thrift.Processor processor;
                processor = new ValidationService_thrift.Processor(certifierReplica);
                TServerTransport serverTransport = new TServerSocket(thriftPort);
                TServer server = new TThreadPoolServer(new TThreadPoolServer.Args(serverTransport).processor(processor).protocolFactory(new TCompactProtocol.Factory()));
                System.out.println("Thrift validation server started at port " + thriftPort + " and host " + certifierReplica.replicaConfig.myHostName);
                server.serve();
            }catch (Exception e){
                e.printStackTrace();
            }

        }
    }

    //Initialize Peer Interface Server
    static class InitializeThriftPeerInterface extends Thread {

        CertifierReplica certifierReplica = null;
        int thriftPort;

        public InitializeThriftPeerInterface(CertifierReplica certifierReplica, int thriftPort) {
            this.certifierReplica = certifierReplica;
            this.thriftPort = thriftPort;
        }

        @Override
        public void run(){

            try {
                PeerInterface_thrift.Processor processor;
                processor = new PeerInterface_thrift.Processor(certifierReplica);
                TServerTransport serverTransport = new TServerSocket(thriftPort);
                TServer server = new TThreadPoolServer(new TThreadPoolServer.Args(serverTransport).processor(processor).protocolFactory(new TCompactProtocol.Factory()));
                System.out.println("Thrift peer server started at port " + thriftPort + " and host " + certifierReplica.replicaConfig.myHostName);
                server.serve();
            }catch (Exception e){
                e.printStackTrace();
            }
        }
    }

    // Prepare thrift thread pool
    static class InitializePeerThriftConnectionPool extends Thread {

        CertifierReplica certifierReplica = null;

        public InitializePeerThriftConnectionPool(CertifierReplica certifierReplica) {
            this.certifierReplica = certifierReplica;
        }

        @Override
        public void run() {
            try {

                Map<Integer, String> serverGroup = certifierReplica.replicaConfig.serverGroup;
                for( int i=0; i<serverGroup.size(); i++){

                    PeerInterface_thrift.Client client;

                    //Trying to open connection to each peer
                    System.out.println("Looking Up Peer : "+serverGroup.get(i) + ":" + (serverPort + i));
                    while(true) {
                        try {
                        	TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(serverGroup.get(i), serverPort+i); 
                            protocol.getTransport().close();
                            System.out.println("Peer Up : "+serverGroup.get(i) + ":" + (serverPort + i));
                            break;
                        } catch (Exception e) {
                            try {
                                Thread.sleep(10);
                            } catch (Exception e1) {
                            }
                        }
                    }

                    LinkedBlockingQueue<PeerInterface_thrift.Client> clientsPool = peerThreadPools.get(i);
                    //LinkedBlockingQueue<PeerInterface_thrift.Client> clientsPool = new LinkedBlockingQueue<PeerInterface_thrift.Client>();
//                    int poolSize = peerThriftPoolSize;
//                    if( i == 1){
//                        poolSize = poolSize*(serverGroup.size()-2);
//                    }
                    for( int j=0; j<peerThriftPoolSize; j++) {

                        TProtocol protocol = BeehiveThriftBuilder.getThriftClientProtocol(serverGroup.get(i), serverPort+i); 
    					client = new PeerInterface_thrift.Client(protocol);
                        clientsPool.add(client);
                    }
                    //peerThreadPools.put(i, clientsPool);
                }

            }catch (Exception e){
                e.printStackTrace();
            }
        }
    }


    //Sending report completion to STS server (Replica-0)
    class ReportCompletionDispatcher extends Thread {

        CertifierReplica certifierReplica = null;
        PeerInterface stsTimestampMgr = null;
        Vector<Long> reportCompletionHolder = null;
        long latestSTS;

        public ReportCompletionDispatcher(CertifierReplica certifierReplica) throws FileNotFoundException {
            this.certifierReplica = certifierReplica;
            this.stsTimestampMgr = this.certifierReplica.replicaConfig.peerHandleTable.get(0);
            this.reportCompletionHolder = new Vector<Long>();
            //System.out.println("Starting Report Completiton Dispatcher for " + certifierReplica.replicaConfig.myID);

        }

        @Override
        public void run() {
            try {
                while (true) {
                    try {
                        Thread.sleep(1); //Call every 1 milliseconds
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                    certifierReplica.pendingReportCompletionQueue.drainTo(reportCompletionHolder);

                    if( reportCompletionHolder.size() > 0) {
                        //latestSTS = certifierReplica.replicaConfig.peerHandleTable.get(0).reportCompletion(reportCompletionHolder);
                        latestSTS = rmiAdapter.reportCompletion(reportCompletionHolder);
                        certifierReplica.setStableTS(latestSTS);
                        reportCompletionHolder.clear();
                    }

                }
            }catch (Exception e){
                e.printStackTrace();
            }
        }

    }

    class StatsCollectorThread extends Thread {

        CertifierReplica certifierReplica = null;
        PrintWriter printWriter = null;
        long clientRequestCount,clientCommitCount,clientAbortCount;
        long peerRequestCount,peerCommitCount,peerAbortCount;
        long abortTruncationCount,abortLockedCount,abortConflictCount;

        public StatsCollectorThread(CertifierReplica certifierReplica) throws FileNotFoundException {
            this.certifierReplica = certifierReplica;
            printWriter = new PrintWriter(new FileOutputStream("CertifierReplica_Stats_" + certifierReplica.replicaConfig.myID ,true),true);
        }

        @Override
        public void run() {

            try {
                while (true) {
                    try {
                        Thread.sleep(10000); //Overwrite file every 10 seconds
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }

                    clientRequestCount = certifierReplica.clientRequestCount.get();
                    clientCommitCount = certifierReplica.clientCommitCount.get();
                    clientAbortCount = certifierReplica.clientAbortCount.get();
                    peerRequestCount = certifierReplica.peerRequestCount.get();
                    peerCommitCount = certifierReplica.peerCommitCount.get();
                    peerAbortCount = certifierReplica.peerAbortCount.get();
                    abortTruncationCount = certifierReplica.validator.abortTruncationCount.get();
                    abortLockedCount = certifierReplica.validator.abortLockedCount.get();
                    abortConflictCount = certifierReplica.validator.abortConflictCount.get();

                    printWriter.println();
                    printWriter.println();

                    printWriter.println("***********COORDINATOR INFO**********");
                    printWriter.println("Total Request: " + clientRequestCount);

                    if (clientRequestCount != 0) {
                        printWriter.println("Total Commit: " + clientCommitCount + "   Perc: " + (clientCommitCount * 100 / clientRequestCount));
                        printWriter.println("Total Abort: " + clientAbortCount + "   Perc: " + (clientAbortCount * 100 / clientRequestCount));
                    }
                    printWriter.println("*************************************");

                    printWriter.println("***********PEER INFO******************");
                    printWriter.println("Total Request: " + peerRequestCount);
                    printWriter.println("Total Commit: " + peerCommitCount);
                    printWriter.println("Total Abort: " + peerAbortCount);
                    if (peerAbortCount != 0) {
                        printWriter.println("Total Abort (Truncation): " + abortTruncationCount + "   Perc: " + (abortTruncationCount * 100 / peerAbortCount));
                        printWriter.println("Total Abort (Locked): " + abortLockedCount + "   Perc: " + (abortLockedCount * 100 / peerAbortCount));
                        printWriter.println("Total Abort (Conflict): " + abortConflictCount + "   Perc: " + (abortConflictCount * 100 / peerAbortCount));
                    }
                    printWriter.println("*************************************");

                    printWriter.flush();
                }

            }catch (Exception e){
                e.printStackTrace();
            }
        }

    }

}
