/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validationService;

import beehive.util.RWSetInfo;
import beehive.validation.ValidateResponse;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.concurrent.ExecutorService;

/**
 * Created by khand055 on 3/23/17.
 */
public class Coordinator {

    ReplicationConfig replicaConfig;
    CertifierReplica certifierReplica;
    ExecutorService globalExecutors;
    long startTS;
    long commitTS;
    RWSetInfo rwSetInfo;
    CoordinationRequestStatus crqStatus;
    HashMap<Integer, RWSetInfo> partitionedRWSetMap;
    long transactionId;


    public Coordinator(long startTS, long commitTS, RWSetInfo rwSetInfo, CertifierReplica certifierReplica, ExecutorService executorService){

        this.certifierReplica = certifierReplica;
        this.replicaConfig = certifierReplica.replicaConfig;
        this.globalExecutors = executorService;
        this.rwSetInfo = rwSetInfo;
        this.startTS = startTS;
        this.commitTS = commitTS;
        crqStatus = new CoordinationRequestStatus();
        crqStatus.status = CoordinationRequestStatus.NOT_INITIATED;
        partitionedRWSetMap = new HashMap<Integer, RWSetInfo>();

        transactionId = certifierReplica.getTransactionId();

        populatePartitionedRWSetMap();
    }

//    public void sendReportCompletionTimeManager() throws RemoteException {
//        long latestSTS = certifierReplica.reportCompletion(commitTS);
//        certifierReplica.setStableTS(latestSTS);
//    }


    public ValidateResponse validate() throws InterruptedException, RemoteException {


        ValidateResponse vr = new ValidateResponse();
//        certifierReplica.logger.logDebug("TID : " + transactionId + " Status : STARTED    ReadSetSize : " +
//                rwSetInfo.readSet.size() + " WriteSetSize : " + rwSetInfo.writeSet.size());

        if( rwSetInfo.readSet.size() == 0 && rwSetInfo.writeSet.size() == 0 ) { //RW set empty send commiy
            vr.latestSTS = certifierReplica.getStableTS();
            vr.commitTS = commitTS;
            vr.commit = true;

            //certifierReplica.logger.logDebug("TID : " + transactionId + " Status : PHASE_1_COMPLETE");
            //certifierReplica.logger.logDebug("TID : " + transactionId + " Status : COMMITED PHASE_2_COMPLETE");
        }

        else {

            //System.out.println("sendValidationRequestToPeer startTS: " + startTS);
            sendValidationRequestToPeer(); //Blocking Call

            switch (crqStatus.validationOutcome) {

                case PeerResponse.ABORT:

                    //System.out.println("sendAbortRequestToPeer startTS: " + startTS);
                    sendAbortRequestToPeer(); //Blocking call
                    //sendReportCompletionTimeManager();
                    //certifierReplica.ps.println(commitTS);
                    certifierReplica.pendingReportCompletionQueue.add(commitTS);
                    vr.latestSTS = certifierReplica.getStableTS();
                    vr.commit = false;
                    vr.commitTS = commitTS;
                    break;

                case PeerResponse.ALL_LOCKS_ACQUIRED:

                    //System.out.println("sendCommitRequestToPeer startTS: " + startTS);
                    sendCommitRequestToPeer(); //Blocking Call
                    vr.latestSTS = certifierReplica.getStableTS();
                    vr.commitTS = commitTS;
                    vr.commit = true;
                    break;
            }
        }

        certifierReplica.logger.logDebug("TID : " + transactionId + " Status : RETURNING TO CLIENT");
        return vr;
    }


    private void sendValidationRequestToPeer() throws InterruptedException{
        crqStatus.status = CoordinationRequestStatus.IN_PHASE_1;
        crqStatus.cmtVoteCnt = 0;

        for (Integer server : partitionedRWSetMap.keySet()) {

            //PeerInterface peerHandle = replicaConfig.peerHandleTable.get(server);

            globalExecutors.submit(new Runnable() {
                public void run() {
                    try {
                        RWSetInfo rwSetInfo = partitionedRWSetMap.get(server);
                        //certifierReplica.logger.logDebug("Coordinator:sendValidationRequestToPeer:: STS : " + startTS + " | ReadSet : " + rwSetInfo.readSet + " | WriteSet : " + rwSetInfo.writeSet);
                        //PeerResponse response = peerHandle.validate(startTS, rwSetInfo, transactionId);
                        PeerResponse response = certifierReplica.rmiAdapter.validate(startTS, rwSetInfo, transactionId, server);
                        processPeerValidationResponse(response);
                    } catch(Exception e){
                        e.printStackTrace();
                    }
                }
            });


        }
        waitForStatus(CoordinationRequestStatus.PHASE_1_COMPLETE);
        certifierReplica.logger.logDebug("TID : " + transactionId + " Status : PHASE_1_COMPLETE");
    }


    private void processPeerValidationResponse(PeerResponse response){

        // proceed to phase 2 either if you get all votes/locks or if you get any abort


        synchronized (crqStatus) {

            //This case can never happen as coordinator wait for all phase 1 response

            if( response.responseType == PeerResponse.ABORT ){
                crqStatus.validationOutcome = PeerResponse.ABORT;
                crqStatus.serverAborted.put(response.serverId, response.serverId);
            } else if( response.responseType == PeerResponse.ALL_LOCKS_ACQUIRED ){
                crqStatus.serverComitted.put(response.serverId, response.serverId);
                if( crqStatus.validationOutcome != PeerResponse.ABORT ){
                    crqStatus.validationOutcome = PeerResponse.ALL_LOCKS_ACQUIRED;
                }
            }
            crqStatus.phase1ResponseCount++;
            if( crqStatus.phase1ResponseCount == partitionedRWSetMap.keySet().size()) {
                crqStatus.status = CoordinationRequestStatus.PHASE_1_COMPLETE;
                crqStatus.notify();
            }
        }
    }


    private void sendCommitRequestToPeer() throws InterruptedException{

        crqStatus.status = CoordinationRequestStatus.IN_PHASE_2;
        crqStatus.completedVoteCnt = 0;

        for (Integer server : partitionedRWSetMap.keySet()) {
            //PeerInterface peerHandle = replicaConfig.peerHandleTable.get(server);
            globalExecutors.submit(new Runnable() {
                public void run() {
                    try {
                        //PeerResponse response = peerHandle.commit(startTS, commitTS, transactionId);
                        PeerResponse response = certifierReplica.rmiAdapter.commit(startTS, commitTS, transactionId, server);
                        //System.out.println(" processPeerCommitResponse startTS: " + startTS + " partiionedKeyMapSize " + partitionedRWSetMap.keySet().size());
                        processPeerCommitResponse(response);
                    } catch(Exception e){
                        e.printStackTrace();
                    }
                }
            });

        }
        waitForStatus(CoordinationRequestStatus.COMPLETED);
        certifierReplica.logger.logDebug("TID : " + transactionId + " Status : COMMITED PHASE_2_COMPLETE");
    }


    private void sendAbortRequestToPeer() throws InterruptedException{

        crqStatus.status = CoordinationRequestStatus.IN_PHASE_2;
        crqStatus.completedVoteCnt = 0;

        for (Integer server : partitionedRWSetMap.keySet()) {
            //PeerInterface peerHandle = replicaConfig.peerHandleTable.get(server);
            globalExecutors.submit(new Runnable() {
                public void run() {
                    try {
                        //PeerResponse response = peerHandle.abort(startTS, transactionId, crqStatus.serverComitted.containsKey(server));
                        PeerResponse response = certifierReplica.rmiAdapter.abort(startTS, transactionId, crqStatus.serverComitted.containsKey(server), server);
                        processPeerAbortResponse(response);
                    } catch(Exception e){
                        e.printStackTrace();
                    }
                }
            });
        }
        waitForStatus(CoordinationRequestStatus.COMPLETED);
        certifierReplica.logger.logDebug("TID : " + transactionId + " Status : ABORTED PHASE_2_COMPLETE");
    }


    private void processPeerCommitResponse(PeerResponse response){

        synchronized (crqStatus) {

            switch (response.responseType){

                case PeerResponse.COMMITED:
                    crqStatus.completedVoteCnt ++;
                    if( crqStatus.completedVoteCnt == partitionedRWSetMap.keySet().size() ){
                        crqStatus.status = CoordinationRequestStatus.COMPLETED;
                        //System.out.println( " send notify startTS" + startTS);
                        crqStatus.notify();
                    }
                    break;
            }

        }
    }



    private void processPeerAbortResponse(PeerResponse response){

        synchronized (crqStatus) {

            switch (response.responseType){

                case PeerResponse.ABORT:
                    crqStatus.completedVoteCnt ++;
                    if( crqStatus.completedVoteCnt == partitionedRWSetMap.keySet().size() ){
                        crqStatus.status = CoordinationRequestStatus.COMPLETED;
                        crqStatus.notify();
                    }
                    break;
            }

        }
    }


    public void waitForStatus(int status) throws InterruptedException {
        synchronized (crqStatus) {
            while (crqStatus.status != status) {
                crqStatus.wait();
                //certifierReplica.logger.logDebug("TID : " + transactionId + " CRQ Phase1RC :" + crqStatus.phase1ResponseCount);
               /* if (crqStatus.status != status) {
                    // System.out.println("T"+tid+" timedout in waiting for status "+status+" crqstatus:"+crqStatus.status);
                    // System.out.println(crqStatus.reqStatusMap);
                    System.out.println("System Exiting !!! - waitForStatus");
                    System.exit(0);
                    break;
                }
*/

            }
        }
    }

    private void populatePartitionedRWSetMap(){

        int serverId=-1;
        for( String nodeId : rwSetInfo.readSet ){
            serverId = replicaConfig.getServerFromNodeKey(Integer.parseInt(nodeId));
            if( serverId >= 2) { //Server O and  1 are STS and CTC
                if (!partitionedRWSetMap.containsKey(serverId)) {
                    partitionedRWSetMap.put(serverId, new RWSetInfo());
                }
                partitionedRWSetMap.get(serverId).readSet.add(nodeId);
            }
        }

        for( String nodeId : rwSetInfo.writeSet ){
            serverId = replicaConfig.getServerFromNodeKey(Integer.parseInt(nodeId));
            if( serverId >= 2) { //Server O and  1 are STS and CTC
                if (!partitionedRWSetMap.containsKey(serverId)) {
                    partitionedRWSetMap.put(serverId, new RWSetInfo());
                }
                partitionedRWSetMap.get(serverId).writeSet.add(nodeId);
            }
        }

    }
}
