/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validationService;

import beehive.thrift.PeerInterface_thrift;
import beehive.thrift.PeerResponse_thrift;
import beehive.thrift.RWSetInfo_thrift;
import beehive.util.RWSetInfo;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/**
 * Created by khand055 on 5/12/17.
 */
public class RMIAdapter {

    String RPC_MECHANISM;
    CertifierReplica certifierReplica;
    public static final int STS_MANAGER = 0;
    public static final int CTS_MANAGER = 1;

    public RMIAdapter(CertifierReplica certifierReplica, String rpc_mechanism){
        this.RPC_MECHANISM = rpc_mechanism;
        this.certifierReplica = certifierReplica;
    }


    public long reportCompletion(long commitTS) throws RemoteException {
        return 0;
    }

    public long reportCompletion(Vector<Long> commitTSList) throws RemoteException {

        long sts = 0;

        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    sts = certifierReplica.replicaConfig.peerHandleTable.get(STS_MANAGER).reportCompletion(commitTSList);
                    break;

                case "THRIFT":
                    List<Long> reportCompletitionList = new ArrayList<>();
                    reportCompletitionList.addAll(commitTSList);
                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(STS_MANAGER).take();
                    sts = client.reportCompletionBatch_peer_thrift(reportCompletitionList);
                    certifierReplica.peerThreadPools.get(STS_MANAGER).add(client);

                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }

        return sts;
    }

    public long getGTS() throws RemoteException {
        return 0;
    }

    public long getSTS() throws RemoteException {

        long sts = 0;
        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    sts = certifierReplica.replicaConfig.peerHandleTable.get(STS_MANAGER).getSTS();
                    break;

                case "THRIFT":

                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(STS_MANAGER).take();
                    sts =  client.getSTS_peer_thrift();//    ctsManagers.getCtsSts_thrift();
                    certifierReplica.peerThreadPools.get(STS_MANAGER).add(client);
                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }


        return sts;
    }

    public long[] getCtsSts() throws RemoteException {

        long[] ctsSts = new long[2];
        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    ctsSts = certifierReplica.replicaConfig.peerHandleTable.get(CTS_MANAGER).getCtsSts();
                    break;

                case "THRIFT":

                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(CTS_MANAGER).take();
                    List<Long> listCtsSts =  client.getCtsSts_peer_thrift();//    ctsManagers.getCtsSts_thrift();
                    certifierReplica.peerThreadPools.get(CTS_MANAGER).add(client);
                    ctsSts[0] = listCtsSts.get(0);
                    ctsSts[1] = listCtsSts.get(1);

                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }


        return ctsSts;
    }

    public PeerResponse commit(long startTS, long commitTS, long transactionId, int serverId) throws RemoteException {

        //Local call
        if( certifierReplica.replicaConfig.myID == serverId) {
            return certifierReplica.commit(startTS, commitTS, transactionId);
        }

        PeerResponse pr = new PeerResponse();

        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    pr = certifierReplica.replicaConfig.peerHandleTable.get(serverId).commit(startTS, commitTS, transactionId);
                    break;

                case "THRIFT":

                    PeerResponse_thrift prt;
                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(serverId).take();
                    prt =  client.commit_peer_thrift(startTS, commitTS, transactionId);//    ctsManagers.getCtsSts_thrift();
                    certifierReplica.peerThreadPools.get(serverId).add(client);

                    pr.serverId = prt.serverId;
                    pr.responseType = prt.responseType;

                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }

        return pr;
    }

    public PeerResponse abort(long startTS, long transactionId, boolean bReleaseLock, int serverId) throws RemoteException {


        //Local call
        if( certifierReplica.replicaConfig.myID == serverId) {
            return certifierReplica.abort(startTS, transactionId, bReleaseLock);
        }

        PeerResponse pr = new PeerResponse();

        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    pr = certifierReplica.replicaConfig.peerHandleTable.get(serverId).abort(startTS, transactionId, bReleaseLock);
                    break;

                case "THRIFT":

                    PeerResponse_thrift prt;
                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(serverId).take();
                    prt =  client.abort_peer_thrift(startTS, transactionId, bReleaseLock);//    ctsManagers.getCtsSts_thrift();
                    certifierReplica.peerThreadPools.get(serverId).add(client);

                    pr.serverId = prt.serverId;
                    pr.responseType = prt.responseType;

                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }

        return pr;
    }

    public PeerResponse validate(long startTS, RWSetInfo rwSet, long transactionId, int serverId) throws RemoteException {

        //Local call
        if( certifierReplica.replicaConfig.myID == serverId) {
            return certifierReplica.validate(startTS, rwSet, transactionId);
        }
        PeerResponse pr = new PeerResponse();

        try {
            switch (RPC_MECHANISM) {
                case "RMI":
                    pr = certifierReplica.replicaConfig.peerHandleTable.get(serverId).validate(startTS, rwSet, transactionId);
                    break;

                case "THRIFT":

                    PeerResponse_thrift prt;
                    PeerInterface_thrift.Client client = certifierReplica.peerThreadPools.get(serverId).take();
                    RWSetInfo_thrift rwSetInfo_thrift = new RWSetInfo_thrift();

                    rwSetInfo_thrift.readSet = rwSet.readSet;
                    rwSetInfo_thrift.writeSet = rwSet.writeSet;
                    rwSetInfo_thrift.unusedSet = rwSet.unusedSet;

                    prt =  client.validate_peer_thrift(startTS, rwSetInfo_thrift, transactionId);//    ctsManagers.getCtsSts_thrift();
                    certifierReplica.peerThreadPools.get(serverId).add(client);

                    pr.serverId = prt.serverId;
                    pr.responseType = prt.responseType;

                    break;
            }
        }catch(Exception e){
            e.printStackTrace();
        }

        return pr;
    }
}
