/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validationService;

import beehive.util.Config;
import beehive.validation.ValidationService;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ExportException;
import java.util.*;

/**
 * Created by khand055 on 3/23/17.
 */
public class ReplicationConfig {

    Map<Integer, PeerInterface> peerHandleTable; //HostId to peer handle mapping
    //HashMap<Integer, String> indexTable; //Server Id to hostname mapping
    String myHostName;
    int myID;
    long numOfNodes;
    List<Integer> serverList = new ArrayList<Integer>();
    Map<Integer, ArrayList<Long>> keyRangeServerMap  = new HashMap<Integer, ArrayList<Long>>();
    public Map<Integer, String> serverGroup = new HashMap<>();

    //partitioned graph nodes location table
    public int[] partitionedGraphHostNodesTable;


    public ReplicationConfig(String validationReplicationConfigFile, int serverId, CertifierReplica localHandle, String configFile)
            throws RemoteException, IOException{

        this.myID = serverId;
        this.peerHandleTable = new HashMap<Integer, PeerInterface>();
        //this.numOfNodes = numOfNode;
        Config configParams = Config.getInstance();
        if( configParams == null){
            configParams = new Config(configFile);
        }
        //configParams = Config.getInstance();
        this.numOfNodes = configParams.getIntValue("Validator.nodeSize");

        selfRegistration(localHandle);
        preparePeerHandle(validationReplicationConfigFile, serverId, localHandle);
        initKeyRangeServerMap(this.numOfNodes);

        for( int key: keyRangeServerMap.keySet()){
            System.out.println(" Server Id: " + key + " Range: " + keyRangeServerMap.get(key).toString());
        }

    }

    public int getServerFromNodeKey(long nodeId){

        List<Long> keyRange;
        for( int serverId : keyRangeServerMap.keySet() ){
            keyRange = keyRangeServerMap.get(serverId);
            if( nodeId >= keyRange.get(0) && nodeId <= keyRange.get(1) ){
                return serverId;
            }
        }
        return -1;
    }

    private void selfRegistration(CertifierReplica localHandle) throws RemoteException{

        //First get the handle to the local registry and bind yourself
        System.setSecurityManager(new RMISecurityManager());
        Registry localRegistry;
        try {
            LocateRegistry.createRegistry(Registry.REGISTRY_PORT);
        } catch (ExportException e){
            //registry already running and do nothing
        }
        localRegistry = LocateRegistry.getRegistry(Registry.REGISTRY_PORT);

        localRegistry.rebind("PeerInterface"+myID, (PeerInterface)localHandle);
        localRegistry.rebind("ValidationService"+myID, (ValidationService)localHandle);


    }

    private void preparePeerHandle(String configFile, int serverId, CertifierReplica localHandle) throws IOException{

        BufferedReader bufferedReader = new BufferedReader(new FileReader(configFile));
        String line;
        String hostname;
        int hostId;
        int hostIdCounter = 0;
        //int i=0;
        while ((line = bufferedReader.readLine()) != null) {
            if( line.charAt(0) == '#') continue;
            String[] splited = line.split("\\s+");
            hostname = splited[0];
            hostId = hostIdCounter++;//Integer.parseInt(splited[1]);
            serverList.add(hostId);
            String url  = "//" + hostname + ":" + Registry.REGISTRY_PORT  + "/PeerInterface" + hostId;
            //indexTable.put(i, host);
            //String fullHostName = InetAddress.getLocalHost().getCanonicalHostName();
            if (myID == hostId) {
                myHostName = hostname;
                peerHandleTable.put(hostId, localHandle);
            } else {
                PeerInterface peerHandle = lookup(url);
                peerHandleTable.put(hostId, peerHandle);
            }

            serverGroup.put(hostId, hostname);

        }

    }

    private PeerInterface lookup(String url) {
        System.out.println("Looking up peer:"+url);
        while(true) {
            try {
                PeerInterface peerHandle = (PeerInterface) Naming.lookup(url);
                System.out.println("Peer up");
                return peerHandle;
            } catch(Exception e){
                try { Thread.sleep(10); } catch(Exception e1){}
            }
        }
    }

    private void initKeyRangeServerMap(long numOfNodes){

        Config configParams = Config.getInstance();
        String loaderType = configParams.getStringValue("Server.FileLoader").trim();
        String hashingScheme = configParams.getStringValue("Server.NODE_HASHING_SCHEME").trim();

        if(loaderType.equals("Bucketized") && hashingScheme.equals("GRAPH_PARTITION")){
            initGraphNodeHostTable();
            setGraphPartitionedServerMap();
        } else {
            setKeyRangeServerMap(numOfNodes);
        }
    }

    private void initGraphNodeHostTable(){

        Config configParams = Config.getInstance();
        List<Integer> hostArrList = new ArrayList<Integer>();

        try{
            String input_filename = configParams.getStringValue("Validator.clusteredGraph") + "/" + "metadata";
            BufferedReader input = new BufferedReader(new FileReader(input_filename));
            String line = "";
            int ctr = 0;

            while( (line = input.readLine()) != null ){
                if( ctr > 0 ) {
                    StringTokenizer lineTok = new StringTokenizer(line, " ");
                    String hostKey = lineTok.nextToken();
                    String fromNode = lineTok.nextToken();
                    String toNode = lineTok.nextToken();
                    int hostid = Integer.parseInt(hostKey);
                    int lastNodeStored = Integer.parseInt(toNode);
                    hostArrList.add(lastNodeStored);
                    //hostArr[hostid]=lastNodeStored;
                }
                ctr++;
            }
            input.close();
        } catch(Exception e){
            e.printStackTrace();
            System.err.println("ERROR :: Failed to read metadata file");
        }
        int hostArr[] = new int[hostArrList.size()];
        for( int i=0; i<hostArrList.size(); i++){
            hostArr[i] = hostArrList.get(i);
        }

        partitionedGraphHostNodesTable = hostArr;
    }

    private void setGraphPartitionedServerMap(){

        int validationClusterSize = peerHandleTable.size();
        validationClusterSize = validationClusterSize - 2; //Node 0 will be for timestamp manager STS and 1 for CTS

        int[] clusterPerServiceReplica = new int[validationClusterSize];
        int graphClusterSize = partitionedGraphHostNodesTable.length;
        int iterator = 0;

        while (graphClusterSize > 0){
            clusterPerServiceReplica[iterator]++;
            iterator = (iterator+1)%validationClusterSize;
            graphClusterSize--;
        }

        //Initialize range to -1 for all
        for( int i=2; i<peerHandleTable.size(); i++ ){
            keyRangeServerMap.put(i, new ArrayList<Long>(Arrays.asList(-1L,-1L)));
        }

        int firstIndex = 0;
        int lastIndex = 0;
        int clusterIndex = 0;


        for( int i=0; i<peerHandleTable.size()-2; i++ ){

            if( i < clusterPerServiceReplica.length) {
                clusterIndex = clusterIndex + clusterPerServiceReplica[i];
                lastIndex = partitionedGraphHostNodesTable[clusterIndex-1];
                keyRangeServerMap.put(i+2, new ArrayList<Long>(Arrays.asList((long) firstIndex, (long) lastIndex)));
                firstIndex = lastIndex + 1;
            }
        }
    }

    private void setKeyRangeServerMap(long numOfNodes){

        int totalClusterSize = peerHandleTable.size();

        totalClusterSize = totalClusterSize - 2; //Node 0 will be for timestamp manager STS and 1 for CTS

        long nodePerCluster = numOfNodes/totalClusterSize;
        long residue = numOfNodes%totalClusterSize; // shall be added to last Host

        long start = 0;
        long startIndex = 0;
        long endIndex = 0;

        //Initialize range to -1 fpr all
        for( int i=2; i<serverList.size(); i++ ){
            keyRangeServerMap.put(serverList.get(i), new ArrayList<Long>(Arrays.asList(-1L,-1L)));
        }

        if( nodePerCluster > 0){ //atleast one node per cluster
            for( int i=2; i<serverList.size(); i++ ){
                startIndex = start;
                endIndex = start + nodePerCluster - 1;
                if( i == serverList.size()-1 ) {// last item
                    endIndex += residue;
                }
                keyRangeServerMap.put(serverList.get(i), new ArrayList<Long>(Arrays.asList(startIndex, endIndex)));
                start += nodePerCluster;
            }
        } else if (residue > 0) {
            keyRangeServerMap.put(serverList.get(2), new ArrayList<Long>(Arrays.asList(startIndex, endIndex))); //Put residue in first node
        }
    }
}
