/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.validationService;

import beehive.util.Config;
import beehive.util.RWSetInfo;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.ReentrantLock;

public class Validator {

    private ConcurrentHashMap<String, NodeRWLock> cTable;
    //private ArrayList<Long> sortedActiveTSList;
    private List<ReentrantLock> cTableLockList;
    private int nLocks = 256;
    private Config configParams;
    public long tts = 0; // truncation time used to trim down the data structures
    public ReentrantLock ttsLock;
    long TRUNCATION_WINDOW_SIZE = 2000;
    int SCAN_LIMIT = 1000;

    final static int ABORT_TRUNCATION = 1;
    final static int ABORT_CONFLICT = 1;
    final static int ABORT_LOCKED = 1;

    final AtomicLong abortTruncationCount = new AtomicLong();
    final AtomicLong abortConflictCount = new AtomicLong();
    final AtomicLong abortLockedCount = new AtomicLong();

    CertifierReplica certifierReplica;
    boolean bTruncationAllowed = false;

    public Validator(CertifierReplica certifierReplica) {

        this.certifierReplica = certifierReplica;
        configParams = Config.getInstance();
        nLocks = configParams.getIntValue("ConcurrentValidator.nLocks");
        bTruncationAllowed = configParams.getBoolValue("Validator.truncation.allowed");
        SCAN_LIMIT = configParams.getIntValue("Validator.truncation.scanLimit");
        TRUNCATION_WINDOW_SIZE = configParams.getIntValue("Validator.truncation.windowSize");
        cTable = new ConcurrentHashMap<String, NodeRWLock>();
        cTableLockList = new ArrayList<ReentrantLock>();
        ttsLock = new ReentrantLock();

        for (int i = 0; i < nLocks; i++) {
            cTableLockList.add(new ReentrantLock());
        }

        System.out.println("bTruncationAllowed" + bTruncationAllowed);
        if (bTruncationAllowed && certifierReplica.myId != 0 && certifierReplica.myId != 1) {
            //start the truncation thread
            //sortedActiveTSList = new ArrayList<Long>();//maintain this in a sorted manner
            //System.out.println("Starting truncation valaditarr");
            TruncationThread t = new TruncationThread(this);
            t.start();
        }
    }

    public long validate(long startTS, RWSetInfo rwSet,
                         HashSet<String> conflictSet) {
        boolean conflict = false;
        long ret = 0; // 0 means no conflict
        long initialTTS = 0;
        long finalTTS = 0;

        if (bTruncationAllowed) {
            //get ttsLock
            ttsLock.lock();
            if (startTS < tts) {
                ttsLock.unlock();
                abortTruncationCount.incrementAndGet();
                return ABORT_TRUNCATION; //returning back and invalid ts to signify abort due to truncation ts check
            }
            //sortedActiveTSList.add(startTS);
            //Collections.sort(sortedActiveTSList);
            initialTTS = tts;
            ttsLock.unlock();
        }

        // Acquire locks
        boolean lock = acquireLocksOnWriteSet(startTS, rwSet);
        if (lock) {
            for (String nodeId : rwSet.readSet) {
                Long writerTS = cTable.get(nodeId).getTimeStamp();
                // if writerTS > startTS => conflict!

                if (writerTS != null && writerTS.longValue() > startTS) {
                    conflict = true;
                    ret = writerTS.longValue();
                    conflictSet.add(nodeId);
                }
            }
            // check write set
            for (String nodeId : rwSet.writeSet) {
                Long writerTS = cTable.get(nodeId).getTimeStamp();
                // if writerTS > startTS => conflict!
                if (writerTS != null && writerTS.longValue() > startTS) {
                    conflict = true;
                    ret = writerTS.longValue();
                    conflictSet.add(nodeId);
                }
            }
        }

        //  Do not release locks if we have not yet if the transaction will
        // be committing, i.e conflict is false
        if (!lock) {
            abortLockedCount.incrementAndGet();
            ret = ABORT_LOCKED;
        } else if (conflict) {
            releaseLock(startTS, rwSet);
            abortConflictCount.incrementAndGet();
            ret = ABORT_CONFLICT;
        } else {
            ttsLock.lock();
            finalTTS = tts;
            if (initialTTS != finalTTS) {
                releaseLock(startTS, rwSet);
                abortTruncationCount.incrementAndGet();
                ret = ABORT_TRUNCATION;
            }
            ttsLock.unlock();
        }

        return ret;
    }

    public boolean acquireLocksOnWriteSet(long startTS, RWSetInfo rwSet) {

        TreeSet<String> sortedRWSet = new TreeSet<String>();
        sortedRWSet.addAll(rwSet.readSet);
        sortedRWSet.addAll(rwSet.writeSet);

        HashSet<String> acquiredLockSet = new HashSet<String>();


        boolean bAcquired;
        for (String s : sortedRWSet) {
            if (!(s == null) && !s.isEmpty()) {
                boolean createEntryFlag = true;
                NodeRWLock nodeLock = cTable.get(s);
                if (nodeLock != null) {
                    createEntryFlag = false;
                    try {
                        LockMode lockMode;
                        if (rwSet.writeSet.contains(s)) {
                            lockMode = LockMode.WRITE;
                        } else {
                            lockMode = LockMode.READ;
                        }
                        bAcquired = nodeLock.tryLock(lockMode);
                        if (!bAcquired) {
                            for (String acquiredLock : acquiredLockSet) { //Releasing the acquired lock
                                cTable.get(acquiredLock).releaseLock();
                            }
                            //Removing entry for truncation to proceed furthur
//                            if( bTruncationAllowed ) {
//                                ttsLock.lock();
//                                sortedActiveTSList.remove(startTS);
//                                ttsLock.unlock();
//                            }
                            return false;
                        } else {
                            acquiredLockSet.add(s);
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        return false;
                    }
                }
                if (createEntryFlag) {

                    LockMode lockMode;
                    if (rwSet.writeSet.contains(s)) {
                        lockMode = LockMode.WRITE;
                    } else {
                        lockMode = LockMode.READ;
                    }

                    ReentrantLock lock = cTableLockList.get(Math.abs(s.hashCode()) % nLocks);
                    lock.lock();
                    if (!cTable.containsKey(s)) {
                        cTable.put(s, new NodeRWLock(lockMode));
                    }
                    lock.unlock();
                    acquiredLockSet.add(s);

                }
            }
        }

        return true;
    }

    public void releaseLock(long startTS, RWSetInfo rwSet) {
        TreeSet<String> sortedRWSet = new TreeSet<String>();
        sortedRWSet.addAll(rwSet.readSet);
        sortedRWSet.addAll(rwSet.writeSet);
        for (String s : sortedRWSet) {
            if (cTable.containsKey(s)) {
                try {
                    cTable.get(s).releaseLock();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }

        //remove my entry from the activeTSList
//        if( bTruncationAllowed) {
//            ttsLock.lock();
//            sortedActiveTSList.remove(startTS);
//            ttsLock.unlock();
//        }
    }

    public synchronized void update(long cmtTS, RWSetInfo rwSet) {
        for (String nodeId : rwSet.writeSet) {
            cTable.get(nodeId).setTimeStamp(cmtTS);
        }
    }


    class NodeRWLock {
        boolean writeLock = false;
        int readCount = 0;
        long timeStamp = 0;
        public boolean deleted = false;

        public NodeRWLock(LockMode lockMode) {
            if (lockMode == LockMode.READ) {
                readCount++;
            } else {
                writeLock = true;
            }
        }

        public long getTimeStamp() {
            return timeStamp;
        }

        public void setTimeStamp(long timeStamp) {
            this.timeStamp = timeStamp;
        }


        public synchronized boolean tryLock(LockMode lockMode) {

            boolean acquiredLock = false;
            switch (lockMode) {
                case READ:
                    acquiredLock = tryReadLock();
                    break;
                case WRITE:
                    acquiredLock = tryWriteLock();
                    break;
            }
            return acquiredLock;
        }

        public synchronized boolean tryReadLock() {
            if (writeLock) {
                return false;
            } else {
                readCount++;
                return !deleted;
            }
        }

        public synchronized boolean tryWriteLock() {
            if ((writeLock) || (readCount > 0)) {
                return false;
            } else {
                writeLock = true;
                return !deleted;
            }
        }

        /**
         * Release the lock and notify others.
         */
        public synchronized void releaseLock() {
            if (readCount > 0 && writeLock == false) {
                readCount--;
            } else if (readCount == 0 && writeLock == true) {
                writeLock = false;
            } else if (readCount == 0 && writeLock == false) {
                // Do Nothing - Releasing the non-acquired lock
            } else { // read count > 0  and write lock is true
                //				System.out.println("Conflicting locks granted!!!!!!!!!!!!! readCount="+lockInfo.readCount+" writeLock="+lockInfo.writeLock);
                try {
                    throw new Exception("Conflicting locks granted!");
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }

    enum LockMode {
        READ,
        WRITE
    }


    public void advanceTruncationTime() {
        long newTTS = certifierReplica.getStableTS() - TRUNCATION_WINDOW_SIZE;
//        long minActiveTS;
        if (newTTS > 0) {
            ttsLock.lock();
            tts = newTTS;
            ttsLock.unlock();
//            try {
//                while(!sortedActiveTSList.isEmpty() && sortedActiveTSList.get(0) <= tts) {
//                    Thread.sleep(1);
//                }
//            }catch (Exception e) {
//                //This is a safe exception and can be ignored - sortedActiveTSList.isEmpty is false and
//                // thread switch happened and then again it became empty and second condition was executed
//                e.printStackTrace();
//            }
        }
    }


    class TruncationThread extends Thread {

        Validator validator;


        int THREAD_SLEEP_TIME = 300;

        public TruncationThread(Validator validator) {
            this.validator = validator;
        }

        public void run() {
            try {
                while (true) {

                    try {
                        Thread.sleep(2 * THREAD_SLEEP_TIME);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                    Set<String> cTableKeys = validator.cTable.keySet();
                    int removeCount = 0;
                    int keysIterationCounter = 0;
                    for (String s : cTableKeys) {
                        NodeRWLock nodeLock = validator.cTable.get(s);
                        boolean obtained = false;
                        try {
                            obtained = nodeLock.tryLock(LockMode.WRITE);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                        if (obtained) {
                            if (nodeLock.getTimeStamp() < validator.tts) {
                                nodeLock.deleted = true;
                                validator.cTable.remove(s);
                                removeCount++;
                            }
                            nodeLock.releaseLock();
                        }
                        keysIterationCounter++;
                        if (keysIterationCounter >= SCAN_LIMIT) {
                            try {
                                Thread.sleep(THREAD_SLEEP_TIME);
                            } catch (Exception e) {
                                e.printStackTrace();
                            }

                            validator.advanceTruncationTime();
                            //System.out.println("Truncation time : " + validator.tts + "	Num keys removed : " + removeCount);
                            removeCount = 0;
                            keysIterationCounter = 0;
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}
