/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.worker;

import beehive.util.*;
import org.apache.thrift.TException;

import java.rmi.RemoteException;
import java.util.*;

/**
 * Validates transactions based on a object-locking mechanism. The strategy will attempt to lock each node
 * in the read/write set at the HashtableServer (HT) in which they are located. A transaction commits/writes
 * its changes if and only if all locks in the read/write set can be acquired. Otherwise, the entire
 * transaciton aborts.
 */
class LockingCommit extends CommitStrategy {

    enum LockAcquisitionPolicy {
        SEQUENTIAL, PARALLEL
    }

    LockingCommit(StorageSystem dataStorage, LockAcquisitionPolicy policy) {
        super(dataStorage);
    }

    LockingCommit(StorageSystem dataStorage) {
        super(dataStorage);
    }


    @Override
    boolean commit(TransactionInfo t, Set<Node> writeSet, Set<NodeData> writeNDSet, Map<String, Long> readSet) {

        // One node case, avoid all bucketiziation costs and go directly to the HT
        if ((writeSet.size() + writeNDSet.size()) == 1 && readSet.isEmpty()) {
            try {
                if (writeSet.size() == 1) {
                    Node node = writeSet.iterator().next();
                    node.TS += 1;
                    return dataStorage.putNode(node.getNodeId(), node);
                } else {
                    NodeData nd = writeNDSet.iterator().next();
                    nd.TS += 1;
                    return dataStorage.updateNodeProperties(nd);
                }
            } catch (RemoteException | TException e) {
                e.printStackTrace();
                return false;
            }
        }

        // Create the requests for each node in the read/write set
        SortedSet<LockRequest> requests = new TreeSet<>();
        for (Node node : writeSet) {
            requests.add(new LockRequest(Constants.VALIDATION_LOCK_WRITE_MODE, node.TS, node.nodeId));
        }
        for (NodeData nodeData : writeNDSet) {
            requests.add(new LockRequest(Constants.VALIDATION_LOCK_WRITE_MODE, nodeData.TS, nodeData.nodeId));
        }
        for (String nodeId : readSet.keySet()) {
            requests.add(new LockRequest(Constants.VALIDATION_LOCK_READ_MODE, readSet.get(nodeId), nodeId));
        }

        boolean locksAcquired = dataStorage.acquireLocks(requests);
        if (locksAcquired) {

            // Writes can safely go through, increment the timestamp and commit the changes
            for (Node node : writeSet) {
                node.TS += 1;
            }
            for (NodeData nodeData : writeNDSet) {
                nodeData.TS += 1;
            }

            dataStorage.commitChanges(writeSet, writeNDSet, readSet);
            return true;
        } else {
            return false;
        }
    }
}