/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.worker;

import beehive.util.*;
import beehive.validation.ValidateResponse;
import beehive.workpool.LocalWorkpoolImpl;
import beehive.workpool.StatusType;
import beehive.workpool.WorkStatus;

import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;

public abstract class Worker extends UnicastRemoteObject implements Runnable {

    public LocalWorkpoolImpl workpool;
    public StorageSystem dataStorage;
    public RWSetInfo rwsetinfo;
    public Hashtable<String, Node> updatedNodes;
    public Hashtable<String, NodeData> updatedNodesData;
    public Map<String, Long> readsetTS; // In the compute method, this map should updated to contain the timestamp of the node being read

    public HashSet<Task> tasksCreated;
    public Config configParams;
    public int workerId;
    public boolean master = false;
    public long startTime = 0;
    public long endTime = 0;

    public WorkerStatus workerStatus;
    public Logger log;

    public static long STS_DELAY = 10;

    //Statistics related
    private StatContainer statGetTaskTime;
    private StatContainer statDoTaskTime;
    private StatContainer statComputeTime;
    private StatContainer statValidateTime;
    private StatContainer statCommitTaskTime;

    public boolean globalValidationEnable = true;

    private static String LOCK_ACQUISITION_STRATEGY = "SEQUENTIAL";

    private CommitStrategy commitStrategy;

    public Worker(String configFile, LocalWorkpoolImpl wpool, int workerNum, StorageSystem dataStore, boolean master) throws RemoteException {
        System.setSecurityManager(new RMISecurityManager());
        workpool = wpool;
        rwsetinfo = new RWSetInfo();
        updatedNodes = new Hashtable<>();
        updatedNodesData = new Hashtable<>();
        tasksCreated = new HashSet<>();
        configParams = Config.getInstance();
        workerStatus = new WorkerStatus(WorkerStatus.IDLE);
        workerId = workerNum;
        this.dataStorage = dataStore;
        this.master = master;
        this.log = new Logger();
        this.readsetTS = new HashMap<>();

        statGetTaskTime = new StatContainer();
        statDoTaskTime = new StatContainer();
        statComputeTime = new StatContainer();
        statValidateTime = new StatContainer();
        statCommitTaskTime = new StatContainer();

        commitStrategy = new ValidatingCommit(dataStorage);
        if (configParams.has("System.globalValidationEnable")) {
            this.globalValidationEnable = configParams.getBoolValue("System.globalValidationEnable");

            if (globalValidationEnable) {
                commitStrategy = new ValidatingCommit(dataStorage);
            } else {
                if (configParams.has("System.lockAcquisitionStrategy")) {
                    LOCK_ACQUISITION_STRATEGY = configParams.getStringValue("System.lockAcquisitionStrategy");
                    LockingCommit.LockAcquisitionPolicy policy = LockingCommit.LockAcquisitionPolicy.SEQUENTIAL;
                    switch (LOCK_ACQUISITION_STRATEGY) {
                        case "SEQUENTIAL":
                            policy = LockingCommit.LockAcquisitionPolicy.SEQUENTIAL;
                            break;
                        case "PARALLEL":
                            policy = LockingCommit.LockAcquisitionPolicy.PARALLEL;
                            break;
                        default:
                            break;
                    }
                    commitStrategy = new LockingCommit(dataStore, policy);
                } else {
                    commitStrategy = new LockingCommit(dataStorage);
                }
            }
        }
    }

    public void cleanUp() {
        System.out.println("Finished cleanup");
    }

    public Vector<Node> getNeighbors(Node u) throws Exception {
        /* get all neighbor nodes from the gloal hashtable */
        Vector<String> nbrKeys = new Vector<>(u.neighbors.keySet());
        Vector<Node> neighborNodes = new Vector<>(nbrKeys.size());
        neighborNodes.addAll(dataStorage.getNodesParallel(nbrKeys));
        return neighborNodes;
    }

    protected Vector<Node> getLocalNeighbors(Node u) throws Exception {
        return dataStorage.getLocalNodes(getNeighborIds(u));
    }

    protected Vector<Node> getRemoteNeighbors(Vector<String> nbrKeys) throws Exception {
        return dataStorage.getNodesParallel(nbrKeys);
    }

    protected Vector<String> getNeighborIds(Node u) {
        return new Vector<>(u.neighbors.keySet());
    }

    public final TransactionInfo beginTransaction() throws RemoteException {
        clearTaskMetadata();
        if (globalValidationEnable) {
            String tid = workpool.getNewTID();
            long sts = workpool.getSTS();
            TransactionInfo txnInfo = new TransactionInfo();
            txnInfo.tid = tid;
            txnInfo.startTS = sts;
            workerStatus.currentTxnInfo = txnInfo;
            return txnInfo;
        } else {
            return null;
        }
    }

    private void clearTaskMetadata() {
        updatedNodes.clear();
        updatedNodesData.clear();
        rwsetinfo.readSet.clear();
        rwsetinfo.writeSet.clear();
        rwsetinfo.unusedSet.clear();
        tasksCreated.clear();
        readsetTS.clear();
    }

    public abstract void doTask(Task tsk) throws Exception;

    public boolean validateTask(Task tsk) throws RemoteException {
        boolean commit;
        if (globalValidationEnable) {
            ValidateResponse res = workpool.validate(tsk, rwsetinfo);
            commit = res.commit;
            if (commit) {
                tsk.txnInfo.commitTS = res.commitTS;
                commitTask(tsk);
            } else {
                tsk.txnInfo = beginTransaction();
            }
        } else {
            commit = commitTask(tsk);
        }

        if (commit) {
            workerStatus.status = WorkerStatus.REPORT_COMPLETION;
            workpool.reportCompletion(tsk, tasksCreated);
        } else {
            if (globalValidationEnable) {
                tsk.txnInfo = beginTransaction();
            }
        }
        clearTaskMetadata();
        return commit;
    }

    public boolean validateTransaction(TransactionInfo txnInfo) throws RemoteException {
        boolean commit;
        if (globalValidationEnable) {
            ValidateResponse res = workpool.validateTransaction(txnInfo, rwsetinfo);
            commit = res.commit;
            if (commit) {
                txnInfo.commitTS = res.commitTS;
                commitTransaction(txnInfo);
            }
        } else {
            commit = commitTransaction(txnInfo);
        }

        if (commit) {
            workerStatus.status = WorkerStatus.REPORT_COMPLETION;


            if (txnInfo == null) {
                workpool.reportTransactionCompletion(tasksCreated);
            } else {
                workpool.reportTransactionCompletion(txnInfo.commitTS, tasksCreated);
            }
        }
        clearTaskMetadata();
        return commit;
    }


    private boolean commitTask(Task t) {
        return commitTransaction(t.txnInfo);
    }

    public boolean commitTransaction(TransactionInfo txnInfo) {

        // Remove any accidental overlaps
        for (String nodeId : updatedNodes.keySet()) {
            updatedNodesData.remove(nodeId);
        }

        for (Node node : updatedNodes.values()) {
            if (readsetTS.containsKey(node.nodeId)) {
                readsetTS.remove(node.nodeId);
            }
        }

        for (NodeData nodeData : updatedNodesData.values()) {
            if (readsetTS.containsKey(nodeData.nodeId)) {
                readsetTS.remove(nodeData.nodeId);
            }
        }

        Vector<Node> updatedNodeList = new Vector<>(updatedNodes.values());
        Vector<NodeData> updatedNodeDataList = new Vector<>(updatedNodesData.values());
        Map<String, Long> readIdTS = new HashMap<>(readsetTS);
        workerStatus.status = WorkerStatus.UPDATE;

        if ((updatedNodeList.size() + updatedNodeDataList.size() + readIdTS.size()) == 0) {
            return true;
        }

        return commitStrategy.commit(txnInfo, new HashSet<>(updatedNodeList), new HashSet<>(updatedNodeDataList), new HashMap<>(readIdTS));
    }

    public void run() {
        Task myTask;
        WorkStatus ws;

        int sleepTime = configParams.getIntValue("Worker.THREADSLEEPTIME");
        while (true) {
            try {
                ws = workpool.getTask();
                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Error: Valid but Null task");
                        continue;
                    }
                    doTask(myTask);
                    workerStatus.status = WorkerStatus.TASK_COMPLETED;
                    tasksCreated.clear();
                } else if (ws.status == StatusType.NO_READY_TASK) {
                    Thread.sleep(sleepTime);
                } else if (ws.status == StatusType.WORK_TERMINATED) {
                    System.out.println("Worker thread proceeding to cleanUp");
                    break;
                }


            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        cleanUp();
        System.exit(0);

    }

    public void recordStat(StatContainer stat, float value) {
        stat.addOnlyMean(value);
    }

    public String getStats() {
        StringBuffer buffer = new StringBuffer();
        buffer.append("Master Worker Stats \n");
        buffer.append("**********\n");

        buffer.append("Avg time taken for getTask : " + statGetTaskTime.getMean() + "\n");
        buffer.append("Avg time taken for doTask : " + statDoTaskTime.getMean() + "\n");
        buffer.append("Avg time taken for compute : " + statComputeTime.getMean() + "\n");
        buffer.append("Avg time taken for validate task : " + statValidateTime.getMean() + "\n");
        buffer.append("Avg time taken for commit task : " + statCommitTaskTime.getMean() + "\n");
        return buffer.toString();
    }
} 

