/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.workpool;

import beehive.util.*;
import beehive.server.*;
import beehive.worker.*;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.rmi.registry.*;
import java.util.*;
import java.net.*;
import java.rmi.server.*;

public class LoadDistributor {

	LocalWorkpoolImpl  wpool;
	int nextPeerRoundRobin=0;
	Logger logger;
	Random rndSource = new Random();

	public LoadDistributor(LocalWorkpoolImpl wpoolHandle){
		this.wpool = wpoolHandle;
	}



	//Called by distribute tasks when you need to select k nodes for distribution
	public Vector<String> selectKNodes(String criteria, int numK){

		Vector<String> targetNodes = new Vector<String>();


		if(criteria.equals("LDAWARE")){
			//	logger.logDebug("Load aware selection of nodes");	
			Vector<LoadInfo> loadInfoVector = new Vector<LoadInfo>(); 
			Enumeration keys = wpool.loadInfoTable.keys();
			while(keys.hasMoreElements()){
				String key = (String)keys.nextElement();
				LoadInfo ldInfo = wpool.loadInfoTable.get(key);
				loadInfoVector.add(ldInfo);
			}		

			Collections.sort(loadInfoVector);	

			//logger.logDebug(loadInfoVector.toString());;
			for(int i=0;i<numK;i++){
				LoadInfo ldInfo = loadInfoVector.get(i);
				String target = ldInfo.hostname;
				targetNodes.add(target);
			}
		}
		else if(criteria.equals("ROUNDROBIN")){
			//logger.logDebug("RoundRobin selection of nodes: Curr peer:"+nextPeerRoundRobin);		
			for(int i=0; i<numK;i++){
				targetNodes.add(wpool.peerList.get(nextPeerRoundRobin));
				nextPeerRoundRobin= (nextPeerRoundRobin+1)%wpool.peerList.size(); 
			}	
		}
		else {	
			//Random selection of nodes
			Vector<String> hosts = new Vector<String>();	
			for(String host: wpool.peerList){	
				hosts.add(host);
			}
			int numPeers = hosts.size();
			for(int i=0;i<numK;i++){
				int ind = rndSource.nextInt(hosts.size());
				String target = hosts.remove(ind);
				targetNodes.add(target);				
			}
		}
		//logger.logDebug("Target nodes: "+targetNodes.toString());
		return targetNodes;
	}


	public void distributeTasks(Vector<Task> tasksCreated) {

		//Table of host to tasks created
		Hashtable<String, Vector<Task>> hostTaskTable;
		hostTaskTable = new Hashtable<String, Vector<Task>>();

		//filter out the strong affinity tasks
		String location;
		Vector<Task> tasksToRemove = new Vector<Task>();
		for(Task t : tasksCreated) {
			if(t.affinityLevel == TaskConstants.STRONG_AFFINITY) {
				location = t.targetLocation;
				if(location.equals(wpool.myHost)) {
					//deposit directly
					wpool.depositTask(t);
				}else {
					if(!hostTaskTable.containsKey(location)) {
						hostTaskTable.put(location, new Vector<Task>());
					}
					((Vector<Task>)hostTaskTable.get(location)).add(t);
				}
				tasksToRemove.add(t);
			}
		}

		tasksCreated.removeAll(tasksToRemove);

		//tasks with strong affinity have been bucketised, we worry only about the other tasks now

		//logger.logDebug("Distributing tasks");
		String distributionPolicy = wpool.configParams.getStringValue("LocalWorkpool.distributionPolicy");
		Vector<String> targetNodes;	
		if(distributionPolicy.equals("KWAYSPLIT") && tasksCreated.size() > 0){
			//logger.logDebug("K-way split strategy");	
			String splitType = wpool.configParams.getStringValue("LocalWorkpool.kwaySplitType");
			int numk = wpool.configParams.getIntValue("LocalWorkpool.ksplit");
			//logger.logDebug("split size:"+numk);
			//Overriding config param taking default 2 way spli
			if(numk<2) {
				numk=2;
				//logger.logDebug("Overriding it to 2");
			}
			//Tara - IMPORTANT ************** 
			//*********quick fix
			if(wpool.peerList.size()<numk)
				numk=1;

			// keep some tasks for ourselves if we are underloaded
			int myLoad = wpool.getMyLoad();
			int numTasks = tasksCreated.size();
			int mynumTasks = (int)Math.ceil((double)numTasks/numk);		
			for (int i=0; i<mynumTasks; i++) { 
				if (tasksCreated.size() > 0 ) {
					Task task = tasksCreated.remove(0);
					wpool.depositTask(task);
				} else {
					wpool.localDepStats.add(i);
					break;
				}
			}
			//quick fix
			//			if(tasksCreated.size()==0)
			//				return;

			targetNodes = selectKNodes(splitType,numk-1);
			//logger.logDebug("Push the remaininig tasks to targetNodes");
			//push the remainings to some tagets
			int numHostWithCeil = numTasks%numk==0?0:(numTasks%numk - 1); 
			int numTasksToPeer;
			//logger.logDebug("NumTasks:"+numTasks+" Numk:"+numk+" numHostWithCiel:"+numHostWithCeil);

			for(String host: targetNodes){

				if(!hostTaskTable.contains(host)) {
					hostTaskTable.put(host, new Vector<Task>());
				}

				if(numHostWithCeil>0) {
					numTasksToPeer = (int)Math.ceil((double)numTasks/numk);
					numHostWithCeil--;
				}
				else
					numTasksToPeer = numTasks/numk;
				if(tasksCreated.size()>0){

					for(int i=0;i<numTasksToPeer;i++) {
						if(tasksCreated.size()>0) {
							Task t = tasksCreated.remove(0);
							((Vector<Task>)hostTaskTable.get(host)).add(t);
						}
					}
				}
				else {
					//logger.logDebug("Pushed all tasks");
					break;
				}	
			}//end of for loop	

		}
		else if(distributionPolicy.equals("LCLDAWARE") && tasksCreated.size() > 0){/*

				Hashtable<String, Vector<Task>> bucketTable = new Hashtable<String, Vector<Task>>();
				//Create buckets for each peer
				for(String host: wpool.peerList) {
					Vector<Task> bucket = new Vector<Task>();
					bucketTable.put(host,bucket);
				}
						//Tasks which could not be bucketized
						Vector<Task> tasksLeft = new Vector<Task>();
						//Tasks destined for myhost
						Vector<Task> myTasks = new Vector<Task>();

						//bucketize based on locality with load awareness
						while(tasksCreated.size()>0)
						{
								Task t = tasksCreated.remove(0);
								String location = t.targetLocation;
								if(location.equals(myHost)) {
									myTasks.add(t);
									continue;
								}
								//decision based on locality and load 
								if(sendToPeer(location)){
									bucketTable.get(location).add(t);
								}
								else 
									tasksLeft.add(t);


						}
						//To check if you can handle tasks
						if(!keepTask())
							tasksLeft.addAll(myTasks); //put it back to TaskLeft for distrbution
						else
							tasksCreated.addAll(myTasks);//add it back to original list		

				//distribute the left tasks to targetNodes
				// targetNodes = selectKNodes("loadaware",numK);

			try {
				if(tasksCreated.size()>0)
				depositTasks(tasksCreated);	
			} catch(Exception e) {e.printStackTrace();}




		 */}

		//deposit the final arrangement of tasks to respective workpools
		//Deposit tasks
		LocalWorkpoolServer peerHandle;
		Vector<Task> tasksForHost;
		for(String host : hostTaskTable.keySet()) {
			tasksForHost = hostTaskTable.get(host);
			if(tasksForHost.size() > 0) {
				peerHandle = wpool.wpoolHandles.get(host);
				try {
					peerHandle.depositTasks(tasksForHost);  
				}catch(Exception e) {
					e.printStackTrace();
				}
			}

		}
	}
}
