/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.workpool;

import beehive.util.*;
import beehive.server.*;
import beehive.worker.*;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.rmi.registry.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.net.*;
import java.rmi.server.*;

public class LoadDistributorWithAggregation {

    //peerlist is workpool is not initialized at the start, so some constructor operating needs to be delayed
    public Boolean distributorInitialized;

    public static int queueDispatchTimeout = 100;
    public static int queueDispatchSize = 10;
    public static float ratioOfLocalTask = 0.5f;

    public static final String POLICY_RANDOM = "RANDOM";
    public static final String POLICY_ROUND_ROBIN = "ROUND_ROBIN";
    public static final String POLICY_MIN_QUEUE = "MIN_QUEUE";
    public static final String POLICY_MAX_QUEUE = "MAX_QUEUE";
    public static final int dispatchThreadSleepTime = 10; //in milliseconds

    public static String buddySelectionPolicy = POLICY_RANDOM;

    long pendingTaskCount; // Total Number of task waiting in the queue to be dispatched

    LocalWorkpoolImpl wpool;

    ConcurrentHashMap<String, LinkedBlockingQueue<Task>> remoteTaskDispatchQueues; //Table of host to tasks created
    ConcurrentHashMap<String, ConcurrentHashMap<String, Task>> remoteTaskDispatchMap; //Table of host to tasks created
    LinkedBlockingQueue<String> pendingHostQueue; //Queue for pending Host task dispatch
    ConcurrentHashMap<String, Long> hostLastDispatchTime; //Last Dispatch time information
    ConcurrentHashMap<String, Boolean> remoteTaskDispatchLock;

    Thread dispatchTaskToHostJob;
    Thread taskDistributorThread;

    Integer nextPeerRoundRobin = 0;
    Logger logger;
    Random rndSource = new Random();

    /**
     * return the count of the pending task in the queue
     *
     * @return
     */
    public long getPendingTaskCount() {
        return pendingTaskCount;
    }

    /**
     * Increase or decrease the pendingTaskCount based on count
     *
     * @param count
     */
    private synchronized void updatePendingTaskCount(long count) {
        pendingTaskCount += count;
    }

    /**
     * Constructor function
     *
     * @param wpoolHandle
     */
    public LoadDistributorWithAggregation(LocalWorkpoolImpl wpoolHandle) {

        distributorInitialized = false;
        wpool = wpoolHandle;
        pendingTaskCount = 0;

        remoteTaskDispatchQueues = new ConcurrentHashMap<String, LinkedBlockingQueue<Task>>();
        remoteTaskDispatchMap = new ConcurrentHashMap<String, ConcurrentHashMap<String, Task>>();
        remoteTaskDispatchLock = new ConcurrentHashMap<String, Boolean>();
        pendingHostQueue = new LinkedBlockingQueue<String>();
        hostLastDispatchTime = new ConcurrentHashMap<String, Long>();

        //Initializing value from the config file
        queueDispatchTimeout = wpool.configParams.getIntValue("LocalWorkpool.LoadDistribution.QueueDispatchTimeout");
        queueDispatchSize = wpool.configParams.getIntValue("LocalWorkpool.LoadDistribution.QueueDispatchSize");
        ratioOfLocalTask = wpool.configParams.getFloatValue("LocalWorkpool.LoadDistribution.LocalTaskDepositRatio");
        buddySelectionPolicy = wpool.configParams.getStringValue("LocalWorkpool.LoadDistribution.Policy");
    }

    /**
     * workpool peerlist must be initialised to execute this function
     */
    private void loadDistributorDelayedInitialization() {

        //Initializing the hostLastDispatchTime and remoteTaskDispatchQueues
        for (String host : wpool.peerList) {
            remoteTaskDispatchQueues.put(host, new LinkedBlockingQueue<Task>());
            remoteTaskDispatchMap.put(host, new ConcurrentHashMap<String, Task>());
            hostLastDispatchTime.put(host, System.currentTimeMillis());
            remoteTaskDispatchLock.put(host, new Boolean(true));
        }

        taskDistributorThread = new Thread(new TaskDistributorThread());
        taskDistributorThread.start();

        for (int i = 0; i < 8; i++) {
            dispatchTaskToHostJob = new Thread(new DispatchTaskToHost());
            dispatchTaskToHostJob.start();
        }
    }

    private String selectBuddyHost() {

        String buddyHost = null;
        int numPeers = wpool.peerList.size();

        switch (buddySelectionPolicy) {

            case POLICY_RANDOM:

                int randomPeerId = rndSource.nextInt(numPeers); // give value between 0 to numPeers - 1
                buddyHost = wpool.peerList.get(randomPeerId);
                break;

            case POLICY_ROUND_ROBIN:

                synchronized (nextPeerRoundRobin) {
                    buddyHost = wpool.peerList.get(nextPeerRoundRobin);
                    nextPeerRoundRobin = (nextPeerRoundRobin + 1) % wpool.peerList.size();
                }
                break;

            case POLICY_MIN_QUEUE:

                int minSize = Integer.MAX_VALUE;

                for (String host : remoteTaskDispatchQueues.keySet()) {

                    LinkedBlockingQueue<Task> hostQueueHandle = remoteTaskDispatchQueues.get(host);
                    if (hostQueueHandle.size() < minSize) {
                        minSize = hostQueueHandle.size();
                        buddyHost = host;
                    }
                }
                break;

            case POLICY_MAX_QUEUE:

                int maxSize = Integer.MIN_VALUE;

                for (String host : remoteTaskDispatchQueues.keySet()) {

                    LinkedBlockingQueue<Task> hostQueueHandle = remoteTaskDispatchQueues.get(host);
                    if (hostQueueHandle.size() > maxSize) {
                        maxSize = hostQueueHandle.size();
                        buddyHost = host;
                    }
                }
                break;

            default:
                buddyHost = wpool.peerList.get(0);
        }
        return buddyHost;
    }

    public int addTaskToHostQueue(String location, Task t) throws InterruptedException {
        int countTaskAdded = 0;
        Boolean hostLock = remoteTaskDispatchLock.get(location);
        synchronized (hostLock) {
            LinkedBlockingQueue<Task> remoteHostQueue = remoteTaskDispatchQueues.get(location);
            ConcurrentHashMap<String, Task> remoteHostMap = remoteTaskDispatchMap.get(location);

            int comparatorStatus = -1;

            if (wpool.enableTaskAggregation == false) {
                remoteHostQueue.put(t);
            } else {
                Task oldTask = remoteHostMap.get(t.nodeId);
                if (wpool.taskComparator != null) {
                    comparatorStatus = wpool.taskComparator.combine(oldTask, t);

                    if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_DONT_UPDATE_TASKMAP) {
                        remoteHostQueue.put(t);
                    } else if (comparatorStatus == TaskComparator.IGNORE_NEW_TASK) {
                        //do nothing
                    } else if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_UPDATE_TASKMAP) {
                        remoteHostQueue.put(t);
                        remoteHostMap.put(t.nodeId, t);
                    }
                } else {
                    remoteHostQueue.put(t);
                    //System.out.println("ERROR :: Task Aggregation is enabled. Please instantiate TaskComparator!");
                }
            }

            if (comparatorStatus == TaskComparator.IGNORE_NEW_TASK) {
                //do not increase load count
            } else {
                countTaskAdded++;
            }
            return countTaskAdded;
        }
    }

    /**
     * Distribute the task in the queue based on affinity,
     * then divide among self and chosen buddy
     *
     * @param tasksCreated
     * @throws RemoteException
     */
    public void distributeTasks(Vector<Task> tasksCreated) throws RemoteException {

        //Manu Khandelwal - Revamped on 7th March 2017 - Involving least computing
        //Straight away adding task to the string affinity task to respective queue
        if (!distributorInitialized) {
            synchronized (this) {
                if (!distributorInitialized) {
                    loadDistributorDelayedInitialization();
                    distributorInitialized = true;
                }

            }
        }

        if (wpool.peerList.size() == 0) { //There is only one host for the whole application

            //Put everything in the current workpool
            wpool.depositTasks(tasksCreated);

        } else {

            String location;
            long remoteTaskCount = 0;
            Vector<Task> tasksToRemove = new Vector<Task>();
            for (Task t : tasksCreated) {
                if (t.affinityLevel == TaskConstants.STRONG_AFFINITY) {
                    location = t.targetLocation;
                    if (location.equals(wpool.myHost)) {
                        wpool.depositTask(t); //deposit directly
                    } else {
                        try {
                            int countTaskAdded = addTaskToHostQueue(location, t);
                            remoteTaskCount = remoteTaskCount + countTaskAdded;

                        } catch (InterruptedException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }
                    }
                    tasksToRemove.add(t);
                }
            }

            updatePendingTaskCount(remoteTaskCount);
            tasksCreated.removeAll(tasksToRemove);
            //tasks with strong affinity have been put into dispatch-queue, we worry only about the other tasks now

            //First add config specified fraction of the task to your local workpool
            int remainingTaskSize = tasksCreated.size();

            int taskForLocalPool = (int) Math.ceil(remainingTaskSize * ratioOfLocalTask);

            //Depositing directly to the local workpool as no RMI involved
            for (int i = 0; i < taskForLocalPool; ++i) {
                Task task = tasksCreated.get(i);
                wpool.depositTask(task);
            }

            remoteTaskCount = 0;
            //Second add config specified fraction to the node based on the policy

            String hostBuddy = selectBuddyHost();
            int totalTaskRemaining = tasksCreated.size();

            //Depositing to the dispatch queue of Buddy
            for (int i = taskForLocalPool; i < totalTaskRemaining; ++i) {
                Task task = tasksCreated.get(i);

                try {
                    int countTaskAdded = addTaskToHostQueue(hostBuddy, task);
                    remoteTaskCount = remoteTaskCount + countTaskAdded;
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            updatePendingTaskCount(remoteTaskCount);
            tasksCreated.clear();
        }
    }

    //For debugging purpose
    private void printQueueStatus() {

        for (String host : remoteTaskDispatchQueues.keySet()) {
            System.out.println(wpool.myHost + " Load Distributor " + host + " \n last Dispatch: + " + hostLastDispatchTime.get(host)
                    + " Queue" + remoteTaskDispatchQueues.get(host).size() + " Pending Q Size " + pendingHostQueue.size());
        }
    }

    class DispatchTaskToHost implements Runnable {
        @Override
        // This will continuously loop through pending Host Queue and dispatch across
        public void run() {

            while (true) {
                try {
                    String hostLocation = pendingHostQueue.take();
                    Vector<Task> tasksToBeDispatched = new Vector<Task>();

                    Boolean hostLock = remoteTaskDispatchLock.get(hostLocation);
                    synchronized (hostLock) {
                        LinkedBlockingQueue<Task> hostPendingTaskQueue = remoteTaskDispatchQueues.get(hostLocation);
                        ConcurrentHashMap<String, Task> hostPendingTaskMap = remoteTaskDispatchMap.get(hostLocation);

                        hostPendingTaskQueue.drainTo(tasksToBeDispatched);
                        hostPendingTaskMap.clear();
                    }

                    LocalWorkpoolServer peerHandle = wpool.wpoolHandles.get(hostLocation);
                    boolean success = false;
                    while (!success) {
                        try {
                            peerHandle.depositTasks(tasksToBeDispatched);
                            hostLastDispatchTime.put(hostLocation, System.currentTimeMillis());
                            updatePendingTaskCount(-1 * tasksToBeDispatched.size());
                            success = true;
                        } catch (Exception e) {
                            System.out.println(wpool.myHost + " :: WARNING :: LoadDistributorV2 :: depositTasks() to remote host " + hostLocation + " failed!!! Retrying operation...");
                            e.printStackTrace();
                            Thread.sleep(10);
                        }
                    }
                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
    }

    class TaskDistributorThread implements Runnable {
        @Override
        // This will continuously loop through the hostLastDispatchTime and put to pending queue if its longer than the last dispatched time
        public void run() {

            while (true) {
                try {
                    Thread.sleep(dispatchThreadSleepTime);

                    for (String host : wpool.peerList) {
                        int hostPendingQueueSize = remoteTaskDispatchQueues.get(host).size();
                        Long lastTimeDispatched = hostLastDispatchTime.get(host);
                        Long currentTime = System.currentTimeMillis();

                        //Either timeout happens and size greater than zero or size greater than dispatchQueueSze
                        if (hostPendingQueueSize >= queueDispatchSize ||
                                (hostPendingQueueSize > 0 && (currentTime - lastTimeDispatched) >= queueDispatchTimeout)) {
                            pendingHostQueue.put(host);
                        }
                    }

                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

        }
    }
}
