/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.workpool;

import java.io.*;
import java.net.InetAddress;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;

import beehive.checkpoint.CheckpointListener;
import beehive.checkpoint.CheckpointPathBuilder;
import beehive.util.*;
import beehive.validation.LocalValidator;
import beehive.validation.ValidateResponse;
import beehive.validation.ValidationService;
import beehive.validation.ValidationServiceImpl;
import org.apache.log4j.*;
import org.apache.log4j.Logger;

public class LocalWorkpoolImpl implements LocalWorkpoolServer, CheckpointListener {

    // Id fields
    private String myURL = "";
    public ValidationService validationService;
    private LocalValidator localValidator;
    public String myHost = "";
    private String masterHost = "";
    boolean master = false;


    // Distributor fields
    ConcurrentHashMap<String, LocalWorkpoolServer> wpoolHandles;
    private final LoadInfo myLoadInfo;
    private LoadDistributor loadDistributor;
    private LoadDistributorWithAggregation loadDistributorWithAggregation;
    ConcurrentHashMap<String, LoadInfo> loadInfoTable;
    Vector<String> peerList;


    // Statistics
    StatContainer localDepStats = new StatContainer();
    QueueStat readyPoolStats = new QueueStat();
    private long abort_count = 0;
    private long prevPrintTime;
    private long prevMonitorTime;
    private int totalTasks = 0;
    private int totalAborts = 0;
    private Integer abortOnStartCounter = 0;
    public PrintProgressStatusThread progressStatusThread = new PrintProgressStatusThread();


    // Taskpool fields
    private ConcurrentHashMap<String, Task> activeTaskPool;
    private Queue<Task> archiveTaskPool = new LinkedBlockingQueue<>();
    private LinkedBlockingQueue<Task> readyTaskPool;
    private ConcurrentHashMap<String, Task> readyTaskPoolTaskMap;
    private ConcurrentHashMap<String, Task> bufferTaskPoolTaskMap;
    private LinkedBlockingQueue<Task> bufferTaskPool;
    private ConcurrentHashMap<String, Integer> barrierStatus;
    private int barrierCount = 0;


    // Configurable fields
    public Config configParams;
    public static long SLEEP_TIME = 1;
    public static long GAP_TIME = 0;
    public static long STS_DELAY = 10;
    private boolean pullTask = true;
    public String modelType = "TRANSACTION";
    private boolean logging = false;
    private int progressMonitorInterval = 30000; // this is the interval for monitoring abort rate and applying adaptive contrlo on parallelism
    private int printStatusInterval = 30000; // Default value of 30 sec to print periodic progress
    private boolean canonicalHostNames = true;
    private static String LOAD_DISTRIBUTOR_VERSION = "V1";
    public static String RPC_MECHANISM = "RMI";
    public boolean globalValidationEnable = true;
    boolean enableTaskAggregation = false;
    boolean checkpoint = false;


    // Checkpointing fields
    private Map<String, Integer> quiescentStatus = new HashMap<>();
    private String checkpointSaveDir;
    private String checkpointRestoreDir;
    private String checkpointBakDir;


    // Workpool statuses
    public int status = INIT;
    private static final int INIT_PENDING = 0;
    private static final int INIT = 1;
    static final int JOB_STARTED = 2;
    private static final int JOB_FINISHED = 3;
    private static final Integer QUIESCENT_PENDING = 4;
    private static final Integer QUIESCENT = 5;
    private static final Integer BARRIER_PENDING = 0;
    private static final Integer BARRIER_FINISHED = 1;


    // STS fields
    private Long latestSTS = new Long(0);
    private Long tid = new Long(0);
    private HashSet<Long> committedTSList = new HashSet<Long>();
    private final Integer STSLOCK = -1;
    private long lastSTSUpdateTime = 0;
    private long lastLoadVectorUpdateTime = 0;
    private final Object COMMITED_LOCK = new Object();
    private int termQueryTS = 1; //termination detection counter
    private int termQueryCounter = 1;    // only to be used by master


    public TaskComparator taskComparator = null;
    private StringBuffer logBuffer;
    public boolean recovery = false;
    private boolean isProgressPrinting = true;

    public LocalWorkpoolImpl(String validationHostDetail, int workerCount, int[] partitionedGraphHostNodesTable) throws IOException {
        BasicConfigurator.configure();

        configParams = Config.getInstance();
        if (configParams.has("Logger.logLevels")) {
            if (configParams.getStringValue("Logger.logLevels").equals("debug")) {
                Logger.getRootLogger().setLevel(Level.DEBUG);
            }
        }

        if (configParams != null) {
            System.out.println("Using config File Params");
            SLEEP_TIME = configParams.getIntValue("LocalWorkpool.SLEEP_TIME");
            GAP_TIME = configParams.getIntValue("LocalWorkpool.GAP_TIME");
            STS_DELAY = configParams.getIntValue("LocalWorkpool.STS_DELAY");
            pullTask = configParams.getBoolValue("LocalWorkpool.pullTask");
            modelType = configParams.getStringValue("LocalWorkpool.modelType");
            logging = configParams.getBoolValue("LocalWorkpool.logging");
            progressMonitorInterval = configParams.getIntValue("LocalWorkpool.progressMonitorInterval");
            printStatusInterval = configParams.getIntValue("LocalWorkpool.printStatusInterval");
            this.canonicalHostNames = configParams.getBoolValue("Server.canonicalHostNames");
            LOAD_DISTRIBUTOR_VERSION = configParams.getStringValue("LocalWorkpool.LoadDistribution.Version");
            if (LOAD_DISTRIBUTOR_VERSION == null || LOAD_DISTRIBUTOR_VERSION.length() == 0) {
                LOAD_DISTRIBUTOR_VERSION = "V1";
            }

            //thrift check
            if (configParams.has("System.RPC_MECHANISM")) {
                RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
                //System.out.println("in ValidationServiceImpl() : using "+RPC_MECHANISM+" for RPC");
            }
            if (configParams.has("System.globalValidationEnable")) {
                this.globalValidationEnable = configParams.getBoolValue("System.globalValidationEnable");
            }

            if (configParams.has("LocalWorkpool.enableTaskAggregation")) {
                this.enableTaskAggregation = configParams.getBoolValue("LocalWorkpool.enableTaskAggregation");
            }

            if (configParams.has("System.checkpoint")) {
                this.checkpoint = configParams.getBoolValue("System.checkpoint");
            }
        }

        if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
            loadDistributorWithAggregation = new LoadDistributorWithAggregation(this);
        } else {
            loadDistributor = new LoadDistributor(this);
        }

        try {
            if (this.canonicalHostNames) {
                myHost = InetAddress.getLocalHost().getCanonicalHostName();
            } else {
                myHost = InetAddress.getLocalHost().getHostName();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        myURL = "//" + myHost + "/WorkpoolServer";

        if (globalValidationEnable) {
            validationService = new ValidationServiceImpl(validationHostDetail, partitionedGraphHostNodesTable);
            localValidator = new LocalValidator(validationService);
        }

        activeTaskPool = new ConcurrentHashMap<String, Task>();
        readyTaskPool = new LinkedBlockingQueue<Task>();
        readyTaskPoolTaskMap = new ConcurrentHashMap<String, Task>();
        bufferTaskPoolTaskMap = new ConcurrentHashMap<String, Task>();
        bufferTaskPool = new LinkedBlockingQueue<Task>();
        barrierStatus = new ConcurrentHashMap<String, Integer>();

        logBuffer = new StringBuffer();
        status = INIT_PENDING;
        prevMonitorTime = System.currentTimeMillis();
        prevPrintTime = System.currentTimeMillis();
        wpoolHandles = new ConcurrentHashMap<String, LocalWorkpoolServer>();
        loadInfoTable = new ConcurrentHashMap<String, LoadInfo>();
        myLoadInfo = new LoadInfo(0, myHost);
        myLoadInfo.load = 0;

        System.out.println("Local workpool constructed");

        progressStatusThread.start();

        BufferedTaskTransferToMapThread bufferedTaskTransferToMapThread = new BufferedTaskTransferToMapThread();
        bufferedTaskTransferToMapThread.setPriority(Thread.MAX_PRIORITY);
        bufferedTaskTransferToMapThread.start();

        if (checkpoint) {
            initCheckpointingDir();
        }
    }

    public LocalWorkpoolImpl(String validationHostDetail, int workerCount, int[] partitionedGraphHostNodesTable, boolean recover) throws IOException {
        this(validationHostDetail, 0, partitionedGraphHostNodesTable);

        this.recovery = recover;
        if (recover) {
            try {
                FileInputStream localWorkpoolCPFile;
                if (new File(checkpointRestoreDir).exists()) {
                    localWorkpoolCPFile = new FileInputStream(checkpointRestoreDir + "/LocalWorkpool.ser");
                    System.out.println("***** Restarting From Restore File *****");

                } else {
                    if (new File(checkpointBakDir).exists()) {
                        localWorkpoolCPFile = new FileInputStream(checkpointBakDir + "/LocalWorkpool.ser");
                        System.out.println("***** Restarting From Restore Bak File *****");
                    } else {
                        localWorkpoolCPFile = new FileInputStream(checkpointSaveDir + "/LocalWorkpool.ser");
                        System.out.println("***** Restarting From Save File *****");

                    }
                }
                ObjectInputStream localWorkpoolCPStream = new ObjectInputStream(localWorkpoolCPFile);

                // load the checkpoint for tasks
                archiveTaskPool = (Queue<Task>) localWorkpoolCPStream.readObject();

                localWorkpoolCPStream.close();
                localWorkpoolCPFile.close();

            } catch (IOException i) {
                i.printStackTrace();
            } catch (ClassNotFoundException c) {
                System.out.println("class not found, recovery failure");
                c.printStackTrace();
            }

            System.out.println("Deserialized LocalWorkpool ...");
        }

    }

    private void initCheckpointingDir() {
        checkpointSaveDir = CheckpointPathBuilder.buildCheckpointSavePath(myHost);
        checkpointRestoreDir = CheckpointPathBuilder.buildCheckpointRecoverPath(myHost);
        checkpointBakDir = CheckpointPathBuilder.buildCheckpointBakPath(myHost);

        new File(checkpointSaveDir).mkdirs();
        new File(checkpointRestoreDir).mkdirs();
        new File(checkpointBakDir).mkdirs();
    }

    /***************************************************************************************/
    /*     RMI Interface metthods                                                          */

    /***************************************************************************************/

    @Override
    public LoadInfo depositTasks(Vector<Task> taskList) throws RemoteException {
        if (status == QUIESCENT || status == QUIESCENT_PENDING) {
            archiveTaskPool.addAll(taskList);
        } else {
            bufferTaskPool.addAll(taskList);

            synchronized (this) {
                if (status == INIT_PENDING || status == INIT || status == JOB_FINISHED) {
                    status = JOB_STARTED;
                    if (!isProgressPrinting) {
                        isProgressPrinting = true;
                        PrintProgressStatusThread ppst = new PrintProgressStatusThread();
                        ppst.start();
                    }
                }
                termQueryTS = 0;
                //notify();
            }
        }
        return null;
    }

    // This call is used by global workpool to dispatch task
    void depositTask(Task t) {
        int comparatorStatus = -1;
        if (status == QUIESCENT || status == QUIESCENT_PENDING) {
            archiveTaskPool.add(t);
        } else {
            synchronized (this) {
                if (status == INIT_PENDING || status == INIT || status == JOB_FINISHED) {
                    status = JOB_STARTED;
                    if (!isProgressPrinting) {
                        isProgressPrinting = true;
                        PrintProgressStatusThread ppst = new PrintProgressStatusThread();
                        ppst.start();
                    }
                }
                try {
                    if (enableTaskAggregation == false) {
                        readyTaskPool.put(t);
                    } else {
                        Task oldTask = readyTaskPoolTaskMap.get(t.nodeId);
                        if (taskComparator != null) {
                            comparatorStatus = taskComparator.combine(oldTask, t);

                            if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_DONT_UPDATE_TASKMAP) {
                                readyTaskPool.put(t);
                            } else if (comparatorStatus == TaskComparator.IGNORE_NEW_TASK) {
                                //do nothing
                            } else if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_UPDATE_TASKMAP) {
                                readyTaskPool.put(t);
                                readyTaskPoolTaskMap.put(t.nodeId, t);
                            }
                        } else {
                            readyTaskPool.put(t);
                        }
                    }

                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

                termQueryTS = 0;
                notify();
            }
            synchronized (myLoadInfo) {
                if (comparatorStatus != TaskComparator.IGNORE_NEW_TASK) {
                    myLoadInfo.load++;
                }
            }
        }
    }

    private void depositTasksToBufferTaskMap(int moveCount) {
        int comparatorStatus = -1;
        if (bufferTaskPool.size() != 0) {
            Vector<Task> vecTaskToAddToReadyTaskPool = new Vector<Task>();

            while (bufferTaskPool.size() != 0) {
                if (moveCount == 0) {
                    break;
                }
                Task bufferedTask = bufferTaskPool.poll();

                if (bufferedTask != null) {
                    if (enableTaskAggregation == false) {
                        vecTaskToAddToReadyTaskPool.add(bufferedTask);
                    } else {
                        try {
                            Task oldTask = bufferTaskPoolTaskMap.get(bufferedTask.nodeId);
                            if (taskComparator != null) {
                                comparatorStatus = taskComparator.combine(oldTask, bufferedTask);

                                if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_DONT_UPDATE_TASKMAP) {
                                    vecTaskToAddToReadyTaskPool.add(bufferedTask);
                                } else if (comparatorStatus == TaskComparator.IGNORE_NEW_TASK) {
                                    //do nothing
                                } else if (comparatorStatus == TaskComparator.ADD_TO_TASKPOOL_AND_UPDATE_TASKMAP) {
                                    bufferTaskPoolTaskMap.put(bufferedTask.nodeId, bufferedTask);
                                }
                            } else {
                                vecTaskToAddToReadyTaskPool.add(bufferedTask);
                            }
                            //}
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                } else {
                    break;
                }
                moveCount--;
            }

            synchronized (this) {
                try {
                    for (Task task : vecTaskToAddToReadyTaskPool) {
                        readyTaskPool.put(task);
                    }
                    notifyAll();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
            synchronized (myLoadInfo) {
                myLoadInfo.load = myLoadInfo.load + vecTaskToAddToReadyTaskPool.size();
            }
            vecTaskToAddToReadyTaskPool.clear();
        }
    }

    private void transferTaskFromBufferTaskPoolTaskMapToReadyTaskPool() {
        if (bufferTaskPoolTaskMap.size() != 0) {
            for (String nodeId : bufferTaskPoolTaskMap.keySet()) {
                Task newTask = bufferTaskPoolTaskMap.get(nodeId);
                if (newTask != null) {
                    depositTask(newTask);
                    bufferTaskPoolTaskMap.remove(newTask.nodeId);
                }
            }
            bufferTaskPoolTaskMap.clear();
        }
    }

    // called by the global wpool to report that has job finished.'
    @Override
    public synchronized void reportJobCompletion() throws RemoteException {
        status = JOB_FINISHED;
        notifyAll();
    }

    public void reportAbortOnStart() {
        synchronized (abortOnStartCounter) {
            abortOnStartCounter++;
        }
    }


    public LoadInfo getLoadInfo() throws RemoteException {
        return myLoadInfo;
    }

    // sent by master to check termination status
    public synchronized LoadInfo getTerminationStatus(int oldts, int newts) throws RemoteException {
        LoadInfo loadInfo = new LoadInfo(getMyLoad(), myHost);
        if (loadInfo.load > 0) {
            loadInfo.tsMatch = false;
            termQueryTS = 0;
        } else {    // our load is 0
            loadInfo.tsMatch = termQueryTS == oldts;
            termQueryTS = newts;
        }
        return loadInfo;
    }

    @Override
    public synchronized void barrierJoin(String remoteHost) {
        barrierStatus.put(remoteHost, BARRIER_FINISHED);
        barrierCount--;
        if (barrierCount == 0) {
            for (String key : wpoolHandles.keySet()) {
                Integer status = barrierStatus.get(key);
                if (status.equals(BARRIER_PENDING)) {
                    barrierCount++;
                }
            }
        }

        if ((barrierCount == 0) && master) {
            try {
                for (LocalWorkpoolServer peerHandle : wpoolHandles.values()) {
                    peerHandle.reportJobCompletion();
                }
                reportJobCompletion();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    public void pause() throws RemoteException {

        for (String peerId : wpoolHandles.keySet()) {
            quiescentStatus.put(peerId, QUIESCENT_PENDING);
        }
        quiescentStatus.put(myHost, QUIESCENT_PENDING);

        // put system in quiescent state
        try {

            for (LocalWorkpoolServer peerHandle : wpoolHandles.values()) {
                //System.out.println("LocalWorkpoolImpl: in pause, calling remote beginQuiescent");
                peerHandle.beginQuiescent();
            }

//            System.out.println("LocalWorkpoolImpl: in pause, calling beginQuiescent on ourself");
            beginQuiescent();

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void start() throws RemoteException {
        for (LocalWorkpoolServer peer : wpoolHandles.values()) {
            peer.endQuiescent();
        }
        endQuiescent();
    }

    @Override
    public void saveState() throws RemoteException {
        try {
            FileOutputStream localWorkpoolCPFile = new FileOutputStream(checkpointSaveDir + "/LocalWorkpool.ser");
            ObjectOutputStream localWorkpoolCPStream = new ObjectOutputStream(localWorkpoolCPFile);

            // make the checkpoint for tasks
            localWorkpoolCPStream.writeObject(archiveTaskPool);

            localWorkpoolCPStream.close();
            localWorkpoolCPFile.close();
        } catch (IOException i) {
            i.printStackTrace();
        }
    }

    @Override
    public void checkpoint() {
        for (LocalWorkpoolServer peer : wpoolHandles.values()) {
            try {
                peer.saveState();
            } catch (RemoteException e) {
                e.printStackTrace();
            }
        }

        try {
            saveState();
            if (globalValidationEnable) {
                validationService.checkPoint();
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }

    }

    public synchronized void beginQuiescent() throws RemoteException {
        status = QUIESCENT_PENDING;
        Iterator<Task> rtp = readyTaskPool.iterator();
        while (rtp.hasNext()) {
            archiveTaskPool.add(rtp.next());
            rtp.remove();
        }
    }

    // called when workpool is quiescent to report to master
    public synchronized void reportQuiescent(String remoteHost) {

        quiescentStatus.put(remoteHost, QUIESCENT);

        boolean allQuiescent = true;
        for (String peerId : quiescentStatus.keySet()) {
            int peerStatus = quiescentStatus.get(peerId);
            if (peerStatus == QUIESCENT_PENDING) {
                allQuiescent = false;
                break;
            }
        }

        if (allQuiescent && master) {
            this.status = QUIESCENT;
            notifyAll();
        }
    }

    public synchronized void endQuiescent() {
        status = JOB_STARTED;
        Iterator<Task> atp = archiveTaskPool.iterator();

        while (atp.hasNext()) {
            bufferTaskPool.add(atp.next());
            atp.remove();
            readyPoolStats.reportArrival();
            myLoadInfo.load++;
            termQueryTS = 0;
        }
        if (master) {
            notifyAll();
        }
        // notifyAll();
    }

    public void checkQuiescent() {
        if (status == QUIESCENT_PENDING && activeTaskPool.isEmpty()) {
            if (!masterHost.equals(myHost)) {

                LocalWorkpoolServer masterHandle = wpoolHandles.get(masterHost);
                try {
                    masterHandle.reportQuiescent(myHost);
                    status = QUIESCENT;
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else {
                reportQuiescent(myHost);
            }
        }

    }


    /***************************************************************************************/
    /*  End of RMI Interface methods                                                       */

    /***************************************************************************************/

    public void lookupPeers(Vector<String> hostList) {
        masterHost = hostList.get(0);
        if (myHost.equals(masterHost)) {
            master = true;
            System.out.println("I am master..");
        }
        peerList = new Vector<String>();
        peerList.addAll(hostList);
        peerList.remove(myHost);
        for (String host : peerList) {
            String url = "//" + host + "/WorkpoolServer";
            System.out.print("\nLooking for peer:" + host);
            while (true) {
                try {
                    Thread.sleep(50);
                    LocalWorkpoolServer peerHandle = (LocalWorkpoolServer) Naming.lookup(url);
                    wpoolHandles.put(host, peerHandle);
                    LoadInfo peerLoadInfo = new LoadInfo(0, host);
                    loadInfoTable.put(host, peerLoadInfo);
                    break;
                } catch (Exception e) {
                    System.out.print(".");
                }
            }
        }
//        if (master) {
//            MonitorThread monThread = new MonitorThread(this);
//            monThread.start();
//        }
        MonitorThread monThread = new MonitorThread(this);
        monThread.start();
    }

    /*Called by master machine on all slaves to initialize*/
    public void reportInitialization() throws RemoteException {
        System.out.println("report initialization from master setting status to INIT");
        setStatusInit();
    }

    public void reportStartup() throws RemoteException {
        //set your status to INIT
        status = INIT;
        Enumeration<String> keys = wpoolHandles.keys();
        try {
            while (keys.hasMoreElements()) {
                String key = keys.nextElement();
                LocalWorkpoolServer rpool = wpoolHandles.get(key);
                System.out.println("Calling report intiializtion on host: " + key);
                rpool.reportInitialization();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    //only called by proceedToClustering
    public synchronized int getStatus() {
        return status;
    }

    public WorkStatus getWorkStatus() {
        StatusType type;
        if (status == QUIESCENT) {
            type = StatusType.QUIESCENT;
        } else if (status == QUIESCENT_PENDING) {
            type = StatusType.QUIESCENT_PENDING;
        } else if (status == JOB_FINISHED) {
            type = StatusType.WORK_TERMINATED;
        } else {
            type = null;
        }
        return new WorkStatus(type);
    }

    public synchronized void setStatusInit() {
        status = INIT;
    }

    public void broadcast(Task t) throws RemoteException {

        if ((t.phase.equals("STARTUP")) && (modelType.equals("BARRIER-JOIN")) && master) {
            barrierCount = 0;

            for (String key : wpoolHandles.keySet()) {
                barrierCount++;
                barrierStatus.put(key, BARRIER_PENDING);
            }
            barrierStatus.put(myHost, BARRIER_PENDING);
            barrierCount++;
        }

        try {
            for (String key : wpoolHandles.keySet()) {
                LocalWorkpoolServer rpool = wpoolHandles.get(key);
                Vector<Task> taskVector = new Vector<Task>();
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;
                t.targetLocation = key;
                taskVector.add(t);
                rpool.depositTasks(taskVector);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        t.targetLocation = myHost;
        t.affinityLevel = TaskConstants.STRONG_AFFINITY;
        addTask(t);
    }

    public void addTask(Task t) throws RemoteException {
        Vector<Task> vec = new Vector<Task>();
        vec.add(t);
        addTask(vec);
    }

    public void addTask(Vector<Task> taskVector) throws RemoteException {
        synchronized (this) {
            status = JOB_STARTED;
            if (!isProgressPrinting) {
                isProgressPrinting = true;
                PrintProgressStatusThread ppst = new PrintProgressStatusThread();
                ppst.start();
            }
        }

        if (taskVector.size() > 0) {
            if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
                loadDistributorWithAggregation.distributeTasks(taskVector);
            } else {
                loadDistributor.distributeTasks(taskVector);
            }
        }
    }

    public WorkStatus getTask() throws RemoteException {
        Task t;

        if (status == QUIESCENT) {
            StatusType type = StatusType.QUIESCENT;
            WorkStatus ret = new WorkStatus(type);
            return ret;
        } else if (status == QUIESCENT_PENDING) {
            StatusType type = StatusType.QUIESCENT_PENDING;
            WorkStatus ret = new WorkStatus(type);
            return ret;
        }


        synchronized (this) {
            while (status == INIT_PENDING || status == INIT) {
                try {
                    wait(24);
                } catch (InterruptedException ignored) {
                }
            }
            if (status == JOB_FINISHED) {
                StatusType type = StatusType.WORK_TERMINATED;
                return new WorkStatus(type);
            } else {
                try {
                    while (readyTaskPool.peek() == null) {
                        wait();

                        if (status == JOB_FINISHED) {
                            StatusType type = StatusType.WORK_TERMINATED;
                            return new WorkStatus(type);
                        }
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            t = readyTaskPool.poll();
            readyTaskPoolTaskMap.remove(t.nodeId);
        }
        activeTaskPool.put(t.nodeId, t);
        return new WorkStatus(t);
    }

    public synchronized String getNewTID() {
        tid++;
        return myHost + tid;
    }

    public ValidateResponse validate(Task t, RWSetInfo rwSetInfo) throws RemoteException {

        LoadInfo currLoadInfo = new LoadInfo(getMyLoad(), myHost);
        ValidateResponse response = localValidator.validate(t.txnInfo.tid, t.txnInfo.startTS, rwSetInfo, currLoadInfo);

        if (response.commit == false) {
            abort_count++;
        }
        //update STS only if it went to the validator service
        if (response.latestSTS > 0) {
            updateSTS(response.latestSTS);
        }
        return response;
    }


    /**
     * Provides the ability to validate a transaction/computation that does not rely on a Task structure
     *
     * @param txnInfo   The transaction's information
     * @param rwSetInfo A set of nodeIds that may have been read or written to during the computation
     * @return Information about the validation. Most important is the whether or not the task can be
     * committed.
     * @throws RemoteException
     */
    public ValidateResponse validateTransaction(TransactionInfo txnInfo, RWSetInfo rwSetInfo) throws RemoteException {
        LoadInfo currLoadInfo = new LoadInfo(getMyLoad(), myHost);

        ValidateResponse response = localValidator.validate(txnInfo.tid, txnInfo.startTS, rwSetInfo, currLoadInfo);
        if (response.commit == false)
            abort_count++;
        //update STS only if it went to the validator service
        if (response.latestSTS > 0) {
            updateSTS(response.latestSTS);
        }
        return response;
    }

    /**
     * Synchronizes the local stable timestamp with that of the global timestamp
     *
     * @return The current Stable timestamp of the system
     * @throws RemoteException Thrown if the {@link ValidationService} could not be contacted
     */
    public long getSTS() throws RemoteException {
        synchronized (STSLOCK) {
            if (System.currentTimeMillis() - lastSTSUpdateTime > STS_DELAY) {
                latestSTS = validationService.getSTS();
                lastSTSUpdateTime = System.currentTimeMillis();
            }
            return latestSTS;
        }
    }

    /**
     * Updates the local stable timestamp of the system. If {@code sts} is earlier than the current local stable
     * timestamp, then nothing happens.
     *
     * @param sts The proposed timestamp to update to.
     */
    private void updateSTS(long sts) {
        synchronized (STSLOCK) {
            if (sts > latestSTS) {
                latestSTS = sts;
                lastSTSUpdateTime = System.currentTimeMillis();
            }
        }
    }

    /**
     * Each workpool caches the {@link LoadInfo} of its peers. This is a means for this workpool to refresh its
     * cache.
     *
     * @param loadVector The new loads of each workpool.
     */
    public void updateLoadInfo(Vector<LoadInfo> loadVector) {
        notify();
        for (LoadInfo ld : loadVector) {
            if (!ld.hostname.equals(myHost))
                loadInfoTable.put(ld.hostname, ld);
        }
        lastLoadVectorUpdateTime = System.currentTimeMillis();
    }

    /**
     * Removes a task from the active workpool. If the workpools are configured to use <it>BARRIER-JOIN</it>, then if
     * the load becomes 0, then the caller is blocked until all of its peers have joined the barrier.
     *
     * @param t The task that is to be removed.
     */
    public void removeTask(Task t) {
        t.status = TaskConstants.TASK_FINISHED;
        activeTaskPool.remove(t.getNodeId());
        printPool();
        synchronized (myLoadInfo) {
            recordProgress();
            myLoadInfo.load--;
        }

        if (modelType.equals("BARRIER-JOIN") && (myLoadInfo.load == 0)) {
            if (!masterHost.equals(myHost)) {
                LocalWorkpoolServer masterHandle = wpoolHandles.get(masterHost);
                try {
                    masterHandle.barrierJoin(myHost);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else {
                barrierJoin(myHost);
            }
        }
    }

    /**
     * Reports all of the committed timestamps to the {@link ValidationService}
     *
     * @throws RemoteException Thrown if the validator could not be contacted
     * @see ValidationService#reportCompletion(Vector)
     */
    private void reportCommittedTS() throws RemoteException {
        Vector<Long> reportedTS = null;
        synchronized (COMMITED_LOCK) {
            reportedTS = new Vector<Long>(committedTSList);
        }
        if (!reportedTS.isEmpty()) {
            validationService.reportCompletion(reportedTS);
        }
        synchronized (COMMITED_LOCK) {
            for (Long ts : reportedTS) {
                committedTSList.remove(ts);
            }
        }
    }

    /**
     * Allows a task to report the completion of a transaction phase. A task may involve multiple phases. This method
     * does not remove the task from the active workpool. {@link #reportCompletion(Task, HashSet)} also calls this method,
     * but it also removes the task from the workpool
     *
     * @param commitTS The timestamp at which this transaction was reported
     */
    private void reportTransactionCompletion(long commitTS) {
        synchronized (COMMITED_LOCK) {
            committedTSList.add(commitTS);
        }
        while (true) {
            try {
                reportCommittedTS();
                break;
            } catch (RemoteException re) {
                re.printStackTrace();
            }
        }
    }

    /**
     * A means to report the completion of the execution of a task. If the task created additional tasks upon
     * completion, then they will be distributed to all of the workpools in the cluster.
     *
     * @param t            The task that is to reported for completion
     * @param tasksCreated Any tasks created by the completed task
     * @throws RemoteException Thrown if the created tasks could not be distributed
     */
    public void reportCompletion(Task t, HashSet<Task> tasksCreated) throws RemoteException {
        if (modelType.equals("TRANSACTION") && globalValidationEnable) {
            reportTransactionCompletion(t.txnInfo.commitTS);
        }
        t.status = TaskConstants.TASK_FINISHED;
        // distribute task to peers
        Vector<Task> newTasks = new Vector<Task>(tasksCreated);
        if (newTasks.size() > 0) {
            if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
                loadDistributorWithAggregation.distributeTasks(newTasks);
            } else {
                loadDistributor.distributeTasks(newTasks);
            }
        }

        activeTaskPool.remove(t.getNodeId());
        synchronized (myLoadInfo) {
            recordProgress();
            myLoadInfo.load--;
        }
    }

    private void recordProgress() {
        totalTasks++;
        totalAborts += abort_count;
        abort_count = 0;
    }

    /**
     * Provides the ability to report that a transaction/computation which does not rely on a Task structure has
     * completed
     *
     * @param commitTS     The time at which the transaction was completed
     * @param tasksCreated Any tasks created by the completed task
     * @throws RemoteException Thrown if the created tasks could not be distributed
     */
    public void reportTransactionCompletion(long commitTS, HashSet<Task> tasksCreated) throws RemoteException {
        if (modelType.equals("TRANSACTION") && globalValidationEnable) {
            synchronized (COMMITED_LOCK) {
                committedTSList.add(commitTS);
            }
            while (true) {
                try {
                    if (modelType.equals("TRANSACTION") && globalValidationEnable) {
                        reportCommittedTS();
                    }
                    break;
                } catch (RemoteException re) {
                    re.printStackTrace();
                }
            }
        }

        // distribute task to peers
        Vector<Task> newTasks = new Vector<Task>(tasksCreated);
        if (newTasks.size() > 0) {
            if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
                loadDistributorWithAggregation.distributeTasks(newTasks);
            } else {
                loadDistributor.distributeTasks(newTasks);
            }
        }
    }

    public void reportTransactionCompletion(HashSet<Task> tasksCreated) throws RemoteException {
        // distribute task to peers
        Vector<Task> newTasks = new Vector<Task>(tasksCreated);
        if (newTasks.size() > 0) {
            if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
                loadDistributorWithAggregation.distributeTasks(newTasks);
            } else {
                loadDistributor.distributeTasks(newTasks);
            }
        }
    }

    private boolean printProgressString(long prevPrintTime) {
        long curr_time = System.currentTimeMillis();
        if (curr_time - prevPrintTime > printStatusInterval) {

            long heapFreeSize = Runtime.getRuntime().freeMemory() / ((long) (1024 * 1024));
            System.out.println("PROGRESS => " + myHost
                    + " - Tasks completed: " + totalTasks
                    + " Tasks in Taskpool:" + readyTaskPool.size()
                    + " Tasks in BufferPool:" + bufferTaskPool.size()
                    + " ActivePool size:" + activeTaskPool.size()
                    + " ReadyMap size:" + readyTaskPoolTaskMap.size()
                    + " BufferMap size:" + bufferTaskPoolTaskMap.size()
                    + "\n            "
                    + " Load=" + getMyLoad()
                    + " Free Heap=" + heapFreeSize + " MB"
                    + " Total Aborted: " + totalAborts
                    + " Aborts on Start : " + abortOnStartCounter
                    + " STS: " + latestSTS
                    + " STATUS: " + status
            );
            return true;
        }
        return false;
    }


    private void printPool() {
        //		logger.logDebug("LocalWorkpool: "+"Ready Pool:" + readyTaskPool.values());
        //		logger.logDebug("LocalWorkpool: "+"Active Pool:" + activeTaskPool.values());
        //logger.logDebug("LocalWorkpool: "+"My load:" +getMyLoad());
        //logger.logDebug("LocalWorkpool: "+"Avg tasks deposited to local pool:"+localDepStats.getMean());
    }


    public String getStats() {
        StringBuffer buffer = new StringBuffer();
        buffer.append(localValidator.getStats());
        buffer.append("\n");
        buffer.append("Aborts on Start : " + abortOnStartCounter);
        buffer.append("\n");
        buffer.append("Local Workpool Stats:\n");
        //		buffer.append("Ready pool stats: avg qlen = "+readyPoolStats.getAvgQueueLength()+" avg time in pool = "+readyPoolStats.getAvgWaitingTime()
        //						+" utilization = "+readyPoolStats.getUtilization()+"\n");
        //		buffer.append("Active pool stats: avg qlen = "+activePoolStats.getAvgQueueLength()+" avg time in pool = "+activePoolStats.getAvgWaitingTime()
        //						+" utilization = "+activePoolStats.getUtilization()+"\n");
        return buffer.toString();
    }

    int getMyLoad() {

        int load;
        if (LOAD_DISTRIBUTOR_VERSION.equals("V2")) {
            load = getMyAggregatedLoad() + bufferTaskPool.size() + bufferTaskPoolTaskMap.size() + archiveTaskPool.size();
        } else {
            load = readyTaskPool.size() + activeTaskPool.size() + bufferTaskPool.size() + bufferTaskPoolTaskMap.size() + archiveTaskPool.size();
        }

        return load;
    }

    //New API to consider tasks sitting in Load Distributor as well
    private int getMyAggregatedLoad() {
        //		System.out.println(this.myHost + " Distributor load "+ (int)loadDistributorV2.getPendingTaskCount() +
        //					"Active pool load "+ activeTaskPool.size() +
        //						"Ready Pool load "+ readyTaskPool.size());
        return readyTaskPool.size() + activeTaskPool.size() + (int) loadDistributorWithAggregation.getPendingTaskCount();

    }

    /**
     * A means to check if there are no tasks left in either workpool, if so, the system is in the termination phase.
     *
     * @return True if the system is in termination, false otherwise.
     * @throws RemoteException
     */
    boolean checkTermination() throws RemoteException {
        if (status != JOB_STARTED) {
            return false;
        }
        int currentLoad = getMyLoad();
        if (currentLoad > 0) {
            return false;
        }
        int oldts = termQueryCounter;
        termQueryCounter++;
        int numRounds = 2;
        boolean termDetected = true;
        for (int i = 0; i < numRounds; i++) {
            if (getMyLoad() > 0) {
                return false;
            }
            termDetected = true;
            for (String host : peerList) {
                LocalWorkpoolServer peerHandle = wpoolHandles.get(host);
                LoadInfo peerStatus = peerHandle.getTerminationStatus(oldts, termQueryCounter);
                if (peerStatus.load > 0) {
                    return false;
                }
                if (peerStatus.tsMatch == false) {
                    termDetected = false;
                }
            }
            if (getMyLoad() > 0) {
                return false;
            }

            if (termDetected) {
                break;
            }
        }
        if (termDetected) {
//            System.out.println("=============================================");
//            System.out.println("*******   Termination Detected ********");
//            System.out.println("=============================================");
            // everyone's load is 0 and ts matched
            for (LocalWorkpoolServer peerHandle : wpoolHandles.values()) {
                peerHandle.reportJobCompletion();
            }
            reportJobCompletion();
            return true;
        } else {
            // everyone's load is 0 but ts did not match
            return false;
        }
    }

    private class BufferedTaskTransferToMapThread extends Thread {
        public void run() {
            while (true) {
                try {
                    Thread.sleep(10);
                    int bufferSize = bufferTaskPool.size();
                    if (bufferSize > 0) {
                        depositTasksToBufferTaskMap(bufferSize);
                    }

                    if (readyTaskPool.size() == 0) {
                        transferTaskFromBufferTaskPoolTaskMapToReadyTaskPool();
                    }

                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }

    public class PrintProgressStatusThread extends Thread {
        public boolean stopPrinting = false;

        public void run() {
            long prevPrintTime = 0L;
            while (true) {
                if (status == JOB_FINISHED) {
                    System.out.println(myHost + " ending progress thread...");
                    break;
                }
                try {
                    Thread.sleep(1000);
                    boolean printed = printProgressString(prevPrintTime);
                    if (printed) {
                        prevPrintTime = System.currentTimeMillis();
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }
}

class MonitorThread extends Thread {
    private LocalWorkpoolImpl localWorkpool;

    MonitorThread(LocalWorkpoolImpl localWorkpool) {
        this.localWorkpool = localWorkpool;
    }

    public void run() {
        long lastPrintTime = System.currentTimeMillis();
        while (true) {
            try {
                Thread.sleep(100);
                if (localWorkpool.master) {
                    int currentLoad = localWorkpool.getMyLoad();

                    if (currentLoad == 0 && localWorkpool.status == LocalWorkpoolImpl.JOB_STARTED) {
                        boolean terminated = localWorkpool.checkTermination();
                    }
                }
                // check (without acquiring lock) if load is 0 and should we start checking for termination
                if (localWorkpool.checkpoint) {
                    localWorkpool.checkQuiescent();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}


