/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.APSP;

import beehive.util.Node;
import beehive.util.StorageSystem;
import beehive.util.Task;
import beehive.worker.Worker;
import beehive.workpool.LocalWorkpoolImpl;
import beehive.workpool.StatusType;
import beehive.workpool.WorkStatus;

import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.nio.channels.FileLock;
import java.rmi.RemoteException;
import java.util.Vector;


public class APSPWorker extends Worker implements Runnable {

    boolean debug = false;
    private long ioStartTime = 0;
    private int numNodes;
    private String myStorageURL;
    private int iteration = 0;

    // Global current iteration APSP Node
    private static APSPNode currentIterationNode;

    public APSPWorker(String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, int workerId, int numNodes, String myStorageURL) throws RemoteException {
        super(configFile, wpool, workerId, datastorage, master);
        this.numNodes = numNodes;
        this.myStorageURL = myStorageURL;
        this.ioStartTime = System.currentTimeMillis();
    }

    private void calculateDistances(APSPNode i, int round) throws Exception {
        APSPNode k = currentIterationNode;
        // for every node: j
        for (int j = 0; j < numNodes; j++) {
            long altDist = i.dist[round] + k.dist[j];
            if (altDist < i.dist[j]) {
                // update distance for i to j in local node
                i.dist[j] = altDist;
            }
        }
    }

    public void doTask(Task tsk) throws Exception {

        Node origNode = dataStorage.getNode(tsk.getNodeId());
        //Vector<Node> backNeighbors = getBackNeighbors(origNode);
        compute(tsk, origNode);
        Vector<Node> updatedNodeList = new Vector<Node>();
        updatedNodeList.add(origNode);
        dataStorage.putNodesParallel(updatedNodeList);
        workpool.removeTask(tsk);
    }


    public void compute(Task t, Node node) {
        try {
            APSPNode u = (APSPNode) node;
            APSPTask prt = (APSPTask) t;
            int itrCount = prt.itrCount;

            calculateDistances(u, itrCount);

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public void createTasksAllVertices(Task tsk) throws Exception {

        APSPTask myTask = (APSPTask) tsk;
        int itrCount = myTask.itrCount;

        // The following two lines are new code
        Node n = dataStorage.getNode(Integer.toString(itrCount));
        currentIterationNode = (APSPNode) n;

        Vector<String> myLocalKeys = dataStorage.getAllKeys(myStorageURL);
        for (String key : myLocalKeys) {
            APSPTask t = new APSPTask(itrCount, key);
            t.phase = "GENERIC";
            workpool.addTask(t);
        }
        workpool.removeTask(tsk);
    }

    public void cleanUp() {
        // Final Computation of the flow out of source node.
        try {
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            long ioEnd = System.currentTimeMillis();
            long duration = ioEnd - ioStartTime;


            Vector<String> nodeKeys = new Vector<String>();
            for (int i = 0; i < numNodes; i++) {
                String key = Integer.toString(i);
                nodeKeys.add(key);
            }

            //System.out.println("Printing ranks...\n");
            printAllNodes();

            System.out.println("Time taken: " + duration + " msec");
            out.println("Time taken: " + duration + " msec");
            out.flush();
            lock.release();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("Finished cleanup");
    }

    private void printAllNodes() throws RemoteException {
        for (int i = 0; i < numNodes; i++) {
            Node n = dataStorage.getNode(Integer.toString(i));
            System.out.println(n);
        }
    }


    public void printStats() {
        try {
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            out.println("Server:" + dataStorage.myURL);
            out.println(dataStorage.getStats());
            out.println(workpool.getStats());
            out.flush();
            lock.release();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    public void run() {

        Task myTask;
        WorkStatus ws;

        if (master) {
            try {
                //initTasks();
                APSPTask t = new APSPTask(iteration);
                t.phase = "STARTUP";
                // t.nodeId = "STARTUP_TASK";
                t.nodeId = Integer.toString(numNodes + iteration);

                System.out.println("Broadcasting startup task: " + iteration);
                workpool.broadcast(t);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        while (true) {
            try {
                String myURL = dataStorage.myURL;
                ws = workpool.getTask();

                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Error: Valid but Null task");
                        continue;
                    }

                    switch (myTask.phase) {
                        case "STARTUP":
                            createTasksAllVertices(myTask);
                            break;
                        case "GENERIC":
                            //    System.out.println("Performing task for :" +myTask.getNodeId());
                            doTask(myTask);
                            //tasksCreated.clear();
                            break;
                        case "TERMINATION":
                            //Stop task
                            System.out.println("Termination task");
                            // printStats();
                            System.exit(0);
                        default:

                            System.out.println("Unknown phase");
                            break;
                    }
                } else if (ws.status == StatusType.NO_READY_TASK) {
                    // System.out.println("Going to sleep 1 sec");
                    // Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
                } else if (ws.status == StatusType.WORK_TERMINATED) {
                    if (master) {
                        //jt more phases??
                        if (iteration == (numNodes - 1)) {
                            //     System.out.println("Master crating TERMINTATE TASK - Worker thread proceeding to cleanUp");
                            cleanUp(); //do clean up before broadcasting termination task
                            APSPTask t = new APSPTask(iteration);
                            t.nodeId = "TERMINATE_TASK";
                            t.phase = "TERMINATION";
                            System.exit(0);
                        } else {
                            iteration++;
                            APSPTask t = new APSPTask(iteration);
                            t.nodeId = Integer.toString(numNodes + iteration);
                            t.phase = "STARTUP";
                            workpool.broadcast(t);
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}//end of class

