/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.ConnectedComponent;

import beehive.util.*;
import beehive.server.*;
import beehive.workpool.*;
import beehive.worker.*;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.rmi.Remote;
import java.rmi.registry.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

import java.io.*;
import java.nio.channels.*;
import java.lang.*;

public class CCWorkerReflection extends Worker implements Runnable {

    private String myStorageURL;
    private int numNodes = 0;

    private long phaseStartTime[] = new long[6];
    private long phaseEndTime[] = new long[6];

    private final HashSet<String> unlabelled;
    private final HashSet<String> currentlyLabelled = new HashSet<String>();
    private ConcurrentHashMap<String, HashSet<String>> labeledNodes;

    private int labelingInProgressId = -(dataStorage.myId + dataStorage.numServers);
    private int labelClashId = -(dataStorage.numServers * 2);

    private int seedTaskCount = 1;

    public CCWorkerReflection(String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, int workerId, String myStorageURL, int numOfNodes, ConcurrentHashMap<String, HashSet<String>> labeledNodes, HashSet<String> unlabeledNodes) throws java.rmi.RemoteException {
        super(configFile, wpool, workerId, datastorage, master);
        this.myStorageURL = myStorageURL;
        this.numNodes = numOfNodes;
        //this.labelled = labeledNodes;
        this.unlabelled = unlabeledNodes;
        this.labeledNodes = labeledNodes;
    }

    private synchronized void addToMap(String label, String nodeId) {
        HashSet<String> vec = labeledNodes.get(label);
        if (vec == null)
            vec = new HashSet<>();
        vec.add(nodeId);
        labeledNodes.put(label, vec);
    }

    private synchronized void removeFromMap(String oldlabel, String nodeId) {
        HashSet<String> vec = labeledNodes.get(oldlabel);
        if (vec != null && nodeId != null) {
            vec.remove(nodeId);
            labeledNodes.put(oldlabel, vec);
        }
    }

    private String computeCC(Task t, Node node, Vector<Node> neighborNodes, Vector<NodeData> remoteNeighborNodesData) {
        String resultlbl = "";
        try {
            CCTask tsk = (CCTask) t;
            CCNode ccNode = new CCNode((CCNode) node);
            Vector<CCTask> initialTasks = new Vector<CCTask>();

            if (tsk.label == -1) {
                ccNode.label = Integer.parseInt(ccNode.nodeId);
                resultlbl = "" + ccNode.label;
                for (String nbrId : ccNode.neighbors.keySet()) {
                    if (!tsk.srcNodeId.equals(nbrId)) {
                        String location = dataStorage.getLocation(nbrId);
                        CCTask t1 = new CCTask(nbrId, "" + ccNode.nodeId);
                        t1.label = Integer.parseInt(ccNode.nodeId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        initialTasks.add(t1);
                    }
                }
            } else {
                if (ccNode.label == -1) {
                    ccNode.label = tsk.label;
                    resultlbl = "" + ccNode.label;
                    //System.out.println("New label : "+ccNode.label+", Nbrs : "+ccNode.neighbors.keySet().toString());
                    for (String nbrId : ccNode.neighbors.keySet()) {
                        if (!tsk.srcNodeId.equals(nbrId)) {
                            String location = dataStorage.getLocation(nbrId);
                            CCTask t1 = new CCTask(nbrId, "" + ccNode.nodeId);
                            t1.label = tsk.label;
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            initialTasks.add(t1);
                        }
                    }
                } else {
                    //the node already has a label. So, there is a label clash. Store this clash info for equivalence class calculation.
                    int lbl = tsk.label;
                    String clashres = "";
                    if (ccNode.label > lbl) {
                        clashres = lbl + "," + ccNode.label;
                    } else {
                        clashres = ccNode.label + "," + lbl;
                    }
                    resultlbl = "" + ccNode.label;
                    NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                    nd.properties.put("addToCountMap", clashres);
                    nd.TS = nd.TS + 1;
                    dataStorage.exec(nd);

                    Vector<String> vec = ccNode.labelClashWithNbr.get(tsk.srcNodeId);
                    if (vec == null) {
                        vec = new Vector<String>();
                    }
                    vec.add(clashres);
                    ccNode.labelClashWithNbr.put(tsk.srcNodeId, vec);
                }
            }

            updatedNodes.put(ccNode.nodeId, ccNode);
            rwsetinfo.writeSet.add(ccNode.nodeId);
            tasksCreated.addAll(initialTasks);

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return resultlbl;
    }

    private void relabelCC(Task t, Node node, Vector<Node> neighborNodes, Vector<NodeData> remoteNeighborNodesData) {
        try {
            RelabelTask tsk = (RelabelTask) t;
            CCNode ccNode = new CCNode((CCNode) node);
            ccNode.label = Integer.parseInt(tsk.label);

            updatedNodes.put(ccNode.nodeId, ccNode);
            rwsetinfo.writeSet.add(ccNode.nodeId);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    @Override
    public void doTask(Task tsk) throws Exception {
        boolean commit = false;
        tasksCreated.clear();
        updatedNodes.clear();
        //TransactionInfo txnInfo = null;
        //txnInfo = beginTransaction();
        //tsk.txnInfo = txnInfo;
        String lbl = "";
        HashMap<String, String> tempNodesLabeledtoLabel = new HashMap<String, String>();

        Class c = tsk.getClass();
        String strclassnamearr[] = c.toString().split(" ")[1].split("\\.");
        String taskClassName = strclassnamearr[strclassnamearr.length - 1];
        //System.out.println("Task class name : "+taskClassName);

        Node origNode = dataStorage.getNode(tsk.getNodeId());
        String oldLabel = "" + ((CCNode) origNode).label;

        while (!commit) {
            tsk.txnInfo = beginTransaction();
            try {
                switch (taskClassName) {
                    case "UpdateGraphTask":
                        //System.out.println("Inside UpdateGraphTask");
                        //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" picked update graph task");
                        tempNodesLabeledtoLabel = updateGraph(tsk, origNode, null);
                        break;
                    case "RecomputeTask":
                        tempNodesLabeledtoLabel = recompute(tsk, origNode, null);
                        break;
                    case "RelabelTask":
                        relabelCC(tsk, origNode, null, null);
                        break;
                    default:
                        lbl = computeCC(tsk, origNode, null, null);
                        break;
                }

                commit = validateTask(tsk);
                if (!commit) {
                    origNode = dataStorage.getNode(tsk.getNodeId());
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } //Transaction committed

        if (!lbl.equals("") && !(taskClassName.equals("RecomputeTask")) && !(taskClassName.equals("UpdateGraphTask"))) {    //for normal scenario
            synchronized (unlabelled) {
                unlabelled.remove(tsk.nodeId);
            }
            addToMap(lbl, tsk.nodeId);    //add to labelled Nodes hashmap
        } else if (taskClassName.equals("RecomputeTask")) {    //for recompute scenario
            synchronized (currentlyLabelled) {
                for (String nd : tempNodesLabeledtoLabel.keySet())
                    currentlyLabelled.add(nd);
            }

            //System.out.println("in doTask RecomputeTask if case : tempNodesLabeledtoLabel : "+tempNodesLabeledtoLabel.keySet());
            //delete from previous labeled set and add to new labeled set.
            for (String nd : tempNodesLabeledtoLabel.keySet())
                removeFromMap(oldLabel, nd);

            //add to labelled Nodes hashmap
            for (String nd : tempNodesLabeledtoLabel.keySet()) {
                String templbl = tempNodesLabeledtoLabel.get(nd);
                addToMap(templbl, nd);
            }
        } else if (taskClassName.equals("UpdateGraphTask")) {    //for update graph scenario fpr delete edge label change to current node's label
            synchronized (currentlyLabelled) {
                for (String nd : tempNodesLabeledtoLabel.keySet())
                    currentlyLabelled.add(nd);
            }

            //System.out.println("in doTask UpdateGraphTask if case : tempNodesLabeledtoLabel : "+tempNodesLabeledtoLabel.keySet());
            //delete from previous labeled set and add to new labeled set.
            for (String nd : tempNodesLabeledtoLabel.keySet())
                removeFromMap(oldLabel, nd);

            //add to labelled Nodes hashmap
            for (String nd : tempNodesLabeledtoLabel.keySet()) {
                String templbl = tempNodesLabeledtoLabel.get(nd);
                addToMap(templbl, nd);
            }
        }

        //System.out.println("Commiting task for node : "+origNode.nodeId+", "+"New tasks created : "+tasksCreated.toString()+", "+"W/R set : "+rwsetinfo.writeSet.toString()+"  "+rwsetinfo.readSet.toString());
        //commitTask(tsk);
    }

    private HashMap<String, String> updateGraph(Task T, Node node, Vector<Node> neighborNodes) {
        HashMap<String, String> templabeledNodes = new HashMap<String, String>();

        CCNode tgtNode = null;
        if (node != null) {
            tgtNode = new CCNode((CCNode) node);    //you need a copy of orig node
        }

        try {
            UpdateGraphTask t = (UpdateGraphTask) T;

            switch (t.opname) {
                case CCTaskConstants.addEdge: {
                    String nbrid = (String) t.params.get(0);
                    //System.out.println("inside addEdge:: node : "+node.nodeId+", nbrid : "+nbrid);
                    CCNode nbrNode = new CCNode((CCNode) dataStorage.getNode(nbrid));

                    nbrNode.addNeighbor(tgtNode.nodeId, new CCEdge());
                    tgtNode.addNeighbor(nbrNode.nodeId, new CCEdge());

                    if (!(nbrNode.label == tgtNode.label)) {
                        String clashres = "";
                        if (tgtNode.label > nbrNode.label)
                            clashres = nbrNode.label + "," + tgtNode.label;
                        else
                            clashres = tgtNode.label + "," + nbrNode.label;
                        NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                        nd.properties.put("addToCountMap", clashres);
//                        nd.TS = nd.TS + 1;
                        dataStorage.exec(nd);
                        //System.out.println("Clash added...");
                    }

                    rwsetinfo.writeSet.add(tgtNode.nodeId);
                    rwsetinfo.writeSet.add(nbrNode.nodeId);
                    updatedNodes.put(tgtNode.nodeId, tgtNode);
                    updatedNodes.put(nbrNode.nodeId, nbrNode);

                }
                break;
                case CCTaskConstants.deleteEdge: {
                    String nbrid = (String) t.params.get(0);
                    //System.out.println("inside deleteEdge::"+" nbrid : "+nbrid);
                    CCNode nbrNode = new CCNode((CCNode) dataStorage.getNode(nbrid));

                    nbrNode.neighbors.remove(tgtNode.nodeId);
                    tgtNode.neighbors.remove(nbrNode.nodeId);

                    tgtNode.removedNeighbors.add(nbrNode.nodeId);
                    nbrNode.removedNeighbors.add(tgtNode.nodeId);

                    //if deleted node has a clash earlier, remove this clash from global clash metadata node, and then remove this entry from both the node's labelClashWithNbr hashmap
                    //for tgtNode
                    {
                        Vector<String> vec = tgtNode.labelClashWithNbr.get(nbrNode.nodeId);
                        if (vec != null) {
                            for (String clashentry : vec) {
                                NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                                nd.properties.put("decrementFromCountMap", clashentry);
//                                nd.TS = nd.TS + 1;
                                dataStorage.exec(nd);
                            }
                        }
                        tgtNode.labelClashWithNbr.remove(nbrNode.nodeId);
                    }
                    //for nbrNode
                    {
                        Vector<String> vec = nbrNode.labelClashWithNbr.get(tgtNode.nodeId);
                        if (vec != null) {
                            for (String clashentry : vec) {
                                NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                                nd.properties.put("decrementFromCountMap", clashentry);
//                                nd.TS = nd.TS + 1;
                                dataStorage.exec(nd);
                            }
                        }
                        nbrNode.labelClashWithNbr.remove(tgtNode.nodeId);
                    }

                    rwsetinfo.writeSet.add(tgtNode.nodeId);
                    rwsetinfo.writeSet.add(nbrNode.nodeId);
                    updatedNodes.put(tgtNode.nodeId, tgtNode);
                    updatedNodes.put(nbrNode.nodeId, nbrNode);
                    //System.out.println("Worker "+workerId+" : Inside deleteEdge :: deleting edge between target : "+tgtNode.nodeId+" and nbr : "+nbrNode.nodeId);

                    if (Integer.parseInt(tgtNode.nodeId) != tgtNode.label) {
                        int oldlabel = tgtNode.label;
                        tgtNode.label = Integer.parseInt(tgtNode.nodeId);
                        tgtNode.history.add("" + tgtNode.label);
                        tgtNode.history.add("" + oldlabel);
                        templabeledNodes.put(tgtNode.nodeId, "" + tgtNode.label);

                        //create task for all neighbors
                        for (String nbrId : tgtNode.neighbors.keySet()) {
                            RecomputeTask tnewTgt = new RecomputeTask(nbrId, tgtNode.nodeId, nbrNode.nodeId, "" + tgtNode.label, "" + oldlabel, tgtNode.history, new HashSet<String>(), "" + tgtNode.nodeId);
                            tnewTgt.phase = "GENERIC";
                            tnewTgt.targetLocation = dataStorage.getLocation(nbrId);
                            tnewTgt.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(tnewTgt);
                        }
                    }
                    if (Integer.parseInt(nbrNode.nodeId) != nbrNode.label) {
                        int oldlabel = nbrNode.label;
                        nbrNode.label = Integer.parseInt(nbrNode.nodeId);
                        nbrNode.history.add("" + nbrNode.label);
                        nbrNode.history.add("" + oldlabel);

                        //creating a task only for the purpose of updating the info of labeled nodes at the hosts.
                        //This task will only update the labeled tasks and will do nothing else and exit with nothing in the r/w sets.
                        RecomputeTask temptnewTgt = new RecomputeTask(nbrNode.nodeId, nbrNode.nodeId, tgtNode.nodeId, "" + nbrNode.label, "" + oldlabel, nbrNode.history, new HashSet<String>(), "" + nbrNode.nodeId);
                        temptnewTgt.phase = "GENERIC";
                        temptnewTgt.targetLocation = dataStorage.getLocation(nbrNode.nodeId);
                        temptnewTgt.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        tasksCreated.add(temptnewTgt);

                        //create task for all neighbors
                        for (String nbrId : nbrNode.neighbors.keySet()) {
                            RecomputeTask tnewTgt = new RecomputeTask(nbrId, nbrNode.nodeId, tgtNode.nodeId, "" + nbrNode.label, "" + oldlabel, nbrNode.history, new HashSet<String>(), "" + nbrNode.nodeId);
                            tnewTgt.phase = "GENERIC";
                            tnewTgt.targetLocation = dataStorage.getLocation(nbrId);
                            tnewTgt.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(tnewTgt);
                        }
                    }
                }
                break;
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return templabeledNodes;
    }

    private HashMap<String, String> recompute(Task t, Node node, Vector<Node> neighborNodes) {
        HashMap<String, String> templabeledNodes = new HashMap<String, String>();
        //HashSet<String> remotelabeledNodes = new HashSet<String>();
        try {
            RecomputeTask tsk = (RecomputeTask) t;
            //templabeledNodes.addAll(tsk.remotelabeledNodes);
            CCNode ccNode = new CCNode((CCNode) node);
            Vector<RecomputeTask> initialTasks = new Vector<RecomputeTask>();

            if (ccNode.removedNeighbors.contains(tsk.senderId)) {    //task from deleted edge
                //skip this task. do nothing
            } else {
                boolean nodeLabelPresentInTaskHistory = tsk.history.contains("" + ccNode.label);
                if (ccNode.label == Integer.parseInt(tsk.label)) {

                    //done to maintain local info of labeled nodes
                    synchronized (currentlyLabelled) {
                        currentlyLabelled.add(ccNode.nodeId);
                    }
                    //delete from previous labeled set and add to new labeled set.
                    removeFromMap(tsk.oldLabel, ccNode.nodeId);
                    addToMap("" + ccNode.label, ccNode.nodeId);

                    //label is same as current label. So, do nothing
                } else if (ccNode.label != Integer.parseInt(tsk.label) && !nodeLabelPresentInTaskHistory) {
                    //there is a label clash. Some other concurrent delete operation changed its label.
                    //No change in label for this node

                    //System.out.println("Label clash for : "+tsk.nodeId);

                    int lbl = Integer.parseInt(tsk.label);
                    String clashres = "";
                    if (ccNode.label <= lbl) {
                        clashres = ccNode.label + "," + lbl;
                    } else {
                        clashres = lbl + "," + ccNode.label;
                    }
                    NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                    nd.properties.put("addToCountMap", clashres);
//                    nd.TS = nd.TS + 1;
                    dataStorage.exec(nd);

                    Vector<String> vec = ccNode.labelClashWithNbr.get(tsk.senderId);
                    if (vec == null) {
                        vec = new Vector<>();
                    }
                    vec.add(clashres);
                    ccNode.labelClashWithNbr.put(tsk.senderId, vec);

                    updatedNodes.put(ccNode.nodeId, ccNode);
                    rwsetinfo.writeSet.add(ccNode.nodeId);
                } else {
                    boolean tskLabelPresentInNodeHistory = ccNode.history.contains(tsk.label);
                    if (!tskLabelPresentInNodeHistory) {
                        ccNode.label = Integer.parseInt(tsk.label);
                        templabeledNodes.put("" + ccNode.nodeId, "" + ccNode.label);

                        //merge task and node's history
                        ccNode.history.add("" + ccNode.label);
                        ccNode.history.addAll(tsk.history);

                        //create task for all neighbors
                        for (String nbrId : ccNode.neighbors.keySet()) {
                            if (!tsk.deletedEdgeSeedNodeId.equals(nbrId) && !tsk.deletedEdgeOtherNodeId.equals(nbrId)) {
                                RecomputeTask tnewTgt = new RecomputeTask(nbrId, tsk.deletedEdgeSeedNodeId, tsk.deletedEdgeOtherNodeId, tsk.label, tsk.oldLabel, ccNode.history, new HashSet<String>(), "" + ccNode.nodeId);
                                tnewTgt.phase = "GENERIC";
                                String location = dataStorage.getLocation(nbrId);
                                tnewTgt.targetLocation = location;
                                tnewTgt.affinityLevel = TaskConstants.STRONG_AFFINITY;
                                initialTasks.add(tnewTgt);
                            }
                        }

                        updatedNodes.put(ccNode.nodeId, ccNode);
                        rwsetinfo.writeSet.add(ccNode.nodeId);
                    }
                }

                tasksCreated.addAll(initialTasks);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return templabeledNodes;
    }

    private Vector<NodeData> getRemoteNeighborsData(Vector<String> nbrKeys, String className) throws Exception {
        Vector<NodeData> reqList = new Vector<NodeData>();
        NodeData nd;
        for (String key : nbrKeys) {
            nd = new NodeData(key, StorageSystem.nodeClassNametoIDMap.get(className));
            nd.addProperty("label", null);
            nd.addProperty("history", null);
            reqList.add(nd);
        }
        return dataStorage.getNodePropertiesParallel(reqList);
    }

    @Override
    public void run() {
        startTime = System.currentTimeMillis();

        Task myTask = null;
        WorkStatus ws;
        int phase = 0;
        String myURL = dataStorage.myURL;

        //add labelingInProgress metadatanode to HTServer
        MetaDataNode metaLabelingInProgress = new MetaDataNode("" + labelingInProgressId);
        metaLabelingInProgress.TS = metaLabelingInProgress.TS + 1;
        Vector<Node> vec_node_batch = new Vector<Node>();
        vec_node_batch.add(metaLabelingInProgress);
        HashtableServer metaHTServer = dataStorage.getServer("" + labelingInProgressId);
        try {
            metaHTServer.loadNodes(vec_node_batch);
        } catch (RemoteException e1) {
            e1.printStackTrace();
        }

        synchronized (unlabelled) {
            try {
                unlabelled.addAll(dataStorage.getAllKeys(myStorageURL));
                unlabelled.remove("" + labelingInProgressId);
                unlabelled.remove("" + labelClashId);
            } catch (RemoteException e) {
                e.printStackTrace();
            }

        }

        if (master) {
            try {
                MetaDataNode labelClash = new MetaDataNode("" + labelClashId);
                labelClash.TS = labelClash.TS + 1;
                Vector<Node> node_batch = new Vector<Node>();
                node_batch.add(labelClash);
                HashtableServer htServer = dataStorage.getServer("" + labelClashId);
                htServer.loadNodes(node_batch);

                Task t = new Task();
                t.phase = "STARTUP";
                t.nodeId = "STARTUP_TASK";
                workpool.broadcast(t);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        int maxphase = 2;
        if (configParams.has("CCWorker.MAX_PHASE"))
            maxphase = configParams.getIntValue("CCWorker.MAX_PHASE") + 1;

        if (configParams.has("CCWorker.SEED_TASK_COUNT"))
            seedTaskCount = configParams.getIntValue("CCWorker.SEED_TASK_COUNT");

        while (true) {
            try {
                ws = workpool.getTask();
                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Worker " + workerId + ":  Error: Valid but Null task");
                    } else if (myTask.phase.equals("SHUTDOWN")) {
                        phase = maxphase;
                        workpool.progressStatusThread.stopPrinting = true;
                        //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" received SHUTDOWN. Closing all operations after reaching phase "+phase);
                        break;
                    } else if (myTask.phase.equals("STARTUP")) {
                        if (master)
                            System.out.println("Worker " + workerId + ":  Startup task picked up by master itself");
                        else
                            System.out.println("Worker " + workerId + ":  Startup task picked up by a non master.");

                        //at start, the metadata node labelingInProgress will have one item in set.
                        //When labeling is over, we will remove this item from the set so that the set is empty
                        MetaDataNode labelingInProgress = (MetaDataNode) dataStorage.getNode("" + labelingInProgressId);
                        labelingInProgress.addToSet("1");
                        labelingInProgress.TS = labelingInProgress.TS + 1;
                        dataStorage.putNode("" + labelingInProgress.nodeId, labelingInProgress);

                        boolean result = createSeedTask(myTask);
                        if (result == false) {
                            System.out.println("No node present in file...");
                            labelingInProgress = (MetaDataNode) dataStorage.getNode("" + labelingInProgressId);
                            labelingInProgress.clearSet();
                            labelingInProgress.TS = labelingInProgress.TS + 1;
                            dataStorage.putNode("" + labelingInProgress.nodeId, labelingInProgress);
                        }
                        System.out.println("Worker " + workerId + ":  Created initial tasks");
                    } else if (myTask.phase.equals("START_RELABELING_TASK")) {
                        CCStartRelabelingTask tsk = (CCStartRelabelingTask) myTask;
                        Vector<Task> newTasks = new Vector<Task>();

                        for (String lbl : labeledNodes.keySet()) {
                            String newlbl = tsk.labelsUpdated.get(lbl);
                            if (newlbl != null) {
                                HashSet<String> nodesToRelabel = labeledNodes.get(lbl);

                                HashSet<String> oldset = labeledNodes.get(newlbl);
                                if (oldset == null) {
                                    oldset = new HashSet<String>();
                                }
                                oldset.addAll(nodesToRelabel);
                                labeledNodes.put(newlbl, oldset);
                                labeledNodes.remove(lbl);

                                for (String nodeId : nodesToRelabel) {
                                    RelabelTask t = new RelabelTask(nodeId, newlbl);
                                    t.phase = "GENERIC";
                                    t.targetLocation = dataStorage.getLocation(nodeId);
                                    t.affinityLevel = TaskConstants.STRONG_AFFINITY;
                                    newTasks.add(t);
                                }
                            }
                        }

                        //System.out.println("New relabel tasks created : "+newTasks.toString());
                        //labeledNodes.clear();
                        workpool.removeTask(tsk);
                        workpool.addTask(newTasks);
                    } else if (myTask.phase.equals("SELECT_NEXT_SEED_TASK")) {
                        boolean result = createSeedTask(myTask);
                        if (result == false) {
                            MetaDataNode labelingInProgress = (MetaDataNode) dataStorage.getNode("" + labelingInProgressId);
                            labelingInProgress.clearSet();
                            labelingInProgress.TS = labelingInProgress.TS + 1;
                            dataStorage.putNode("" + labelingInProgress.nodeId, labelingInProgress);
                            //System.out.println("Host "+dataStorage.myId+" :: clearing labelingInProgress node, new TS is "+labelingInProgress.TS++);
                        }
                    } else if (myTask.phase.equals("GENERIC")) {
                        doTask(myTask);
                        //tasksCreated.clear();
                    } else if (myTask.phase.equals("TERMINATION")) {
                        System.out.println("Worker " + workerId + ":  Termination task");
                        //System.exit(0);
                    } else {
                        System.out.println("Worker " + workerId + ":  Unknown phase");
                    }

                } else if (ws.status == StatusType.NO_READY_TASK) {
                    // Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
                } else if (ws.status == StatusType.WORK_TERMINATED) {
                    if (master && (phase == maxphase)) {
                        //phaseEndTime[phase] = System.currentTimeMillis();
                        //printResult(phase);

                        //deleting the nodes that were to be deleted
                        //System.out.println("Master removing the nodes to be deleted.");
                        //deleteNodesToBeRemoved();

                        System.out.println("Master Proceeding to cleanUp after reaching phase : " + phase);
                        Task t = new Task();
                        t.phase = "SHUTDOWN";
                        t.nodeId = "SHUTDOWN_TASK";
                        for (int i = 0; i < configParams.getIntValue("Worker.WORKER_COUNT"); i++) {
                            workpool.broadcast(t);
                        }
                        cleanUp();
                        workpool.progressStatusThread.stopPrinting = true;
                        if (configParams.getBoolValue("CCWorker.WRITE_RESULT_TO_FILE") == true)
                            writeResult(phase - 1);
                        break;
                    } else if (master && (phase < maxphase)) {
                        MetaDataNode labelClash = (MetaDataNode) dataStorage.getNode("" + labelClashId);
                        //System.out.println("labelClash.set.size() :: "+labelClash.set.size());
                        if (labelClash.countmap.size() > 0) {    //check for label clashes
                            System.out.println("Worker " + workerId + ":  Master Merging Equivalence classes...");
                            Set<String> tempset = new HashSet<String>();
                            for (String clashKey : labelClash.countmap.keySet()) {
                                int count = labelClash.countmap.get(clashKey);
                                if (count > 0) {
                                    tempset.add(clashKey);
                                }
                            }

                            Vector<String> labelpair = new Vector<>(tempset);
                            HashMap<String, String> resultmap = calculateEquivalentSets(labelpair);
                            //System.out.println("Initial size : "+resultmap.size());
                            //deleting the elements from the hashmap whose label will not be updated.
                            HashSet<String> labelsNotModified = new HashSet<String>();
                            for (String lbl : resultmap.keySet()) {
                                String newlbl = resultmap.get(lbl);
                                if (lbl.equals(newlbl)) {
                                    labelsNotModified.add(lbl);
                                }
                            }
                            for (String s : labelsNotModified) {
                                resultmap.remove(s);
                            }
                            //System.out.println("New size : "+resultmap.size());
                            //now, we only have those labels that are modified.
                            //Create a Task and broadcast it to all hosts to create relabel tasks for the nodes whose label will be updated

                            CCStartRelabelingTask t = new CCStartRelabelingTask(resultmap);
                            t.phase = "START_RELABELING_TASK";
                            t.nodeId = "START_RELABELING_TASK";

                            NodeData nd = new NodeData("" + labelClashId, StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                            nd.properties.put("clearCountMap", null);
                            nd.TS = nd.TS + 1;
                            dataStorage.exec(nd);

                            workpool.broadcast(t);
                        } else {    //check if all nodes are labeled. If not, then broadcast a task to pick another seed task. Otherwise, proceed to next phase.
                            //System.out.println("Inside labelClash else part...");
                            boolean allnodeslabelled = true;
                            for (int i = 0; i < dataStorage.numServers; i++) {
                                NodeData nd2 = new NodeData("" + (-(dataStorage.numServers + i)), StorageSystem.nodeClassNametoIDMap.get(MetaDataNode.class.getName()));
                                nd2.properties.put("isSetEmpty", null);
                                nd2.TS = nd2.TS + 1;
                                boolean res = (Boolean) dataStorage.exec(nd2).properties.get("isSetEmpty");
                                if (res == false) {
                                    //System.out.println("Set is not empty for : "+-(dataStorage.numServers+i));
                                    allnodeslabelled = false;
                                    break;
                                }
                            }
                            if (allnodeslabelled) {    //all nodes in all hosts are labeled. So, goto next phase
                                if (master) {
                                    phaseEndTime[phase] = System.currentTimeMillis();
                                }
                                phase = phase + 1;
                                System.out.println("Host " + dataStorage.myId + ", Worker " + workerId + " completed phase " + phase);
                                if (phase < maxphase && master) {
                                    if (configParams.getBoolValue("CCWorker.WRITE_RESULT_TO_FILE") == true)
                                        writeResult(phase - 1);
                                    //createGraphUpdateTasks( dataStorage.myId, phase );
                                    Vector<UpdateGraphTask> vtasks = createGraphUpdateTasks(dataStorage.myId, phase);
                                    Vector<Task> finalVTasks = new Vector<Task>();
                                    for (UpdateGraphTask vtask : vtasks) {
                                        if (vtask != null) {
                                            finalVTasks.add(vtask);
                                        }
                                    }

                                    System.out.println("Master creating " + finalVTasks.size() + " new tasks...");
                                    phaseStartTime[phase] = System.currentTimeMillis();
                                    try {
                                        workpool.addTask(finalVTasks);
                                    } catch (RemoteException e) {
                                        e.printStackTrace();
                                    }
                                }
                                /*if(master){
									phaseStartTime[phase] = System.currentTimeMillis();
								}*/
                            } else {    //all nodes are not labeled. So, create task to pick next seed node on every host.
                                Task t = new Task();
                                t.phase = "SELECT_NEXT_SEED_TASK";
                                t.nodeId = "SELECT_NEXT_SEED_TASK";

                                for (String str : currentlyLabelled) {
                                    CCNode origNode = (CCNode) dataStorage.getNode(str);
                                    origNode.history = new HashSet<String>();
                                }
                                currentlyLabelled.clear();

                                workpool.broadcast(t);
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    private Vector<UpdateGraphTask> createGraphUpdateTasks(int hostId, int phaseno) {
        if (master) {
            //printResult(phaseno-1);
            System.out.println("\nAdding new changes in the graph...\n");
        }

        //loadGenerator
        Vector<UpdateGraphTask> vtasks = new Vector<UpdateGraphTask>();
        int countAddEdgeOperations = 0, countDeleteEdgeOperations = 0;
        //int countAddNodeOperations=0, countDeleteNodeOperations=0;
        boolean deleteEdgeUsingFile = false;
        boolean showOutput = false;
        //int countMaxNeighbors=0;
        String deleteEdgeUsingFileName = "";
        int numServers = dataStorage.getNumberOfServers();
        showOutput = configParams.getBoolValue("CCWorker.SHOW_OUTPUT");
        countAddEdgeOperations = configParams.getIntValue("CCWorker.ADD_EDGE")/*/numServers*/;
        countDeleteEdgeOperations = configParams.getIntValue("CCWorker.DELETE_EDGE")/*/numServers*/;
        //countAddNodeOperations = configParams.getIntValue("CCWorker.ADD_NODE")/*/numServers*/;
        //countDeleteNodeOperations = configParams.getIntValue("CCWorker.DELETE_NODE")/*/numServers*/;
        deleteEdgeUsingFile = configParams.getBoolValue("CCWorker.DELETE_EDGE_USING_FILE");
        //countMaxNeighbors = configParams.getIntValue("CCWorker.MAX_NEIGHBORS");
        if (deleteEdgeUsingFile) {
            deleteEdgeUsingFileName = configParams.getStringValue("CCWorker.DELETE_EDGE_USING_FILENAME");
        }

        Random r = new Random();
        StringBuffer strbuf = new StringBuffer();
        UpdateGraphTask t = null;

        switch (phaseno) {
            case 1: //add Edge
                for (int i = 0; i < countAddEdgeOperations; i++) {
                    int src = r.nextInt(this.numNodes);
                    int dest = r.nextInt(this.numNodes);
                    while (dest == src) {
                        dest = r.nextInt(this.numNodes);
                    }
                    t = addEdge("" + src, "" + dest);
                    if (t == null) {
                        i--;
                    } else {
                        vtasks.add(t);
                        strbuf.append("Host " + dataStorage.myId + " :: Adding edge between Node " + src + " and " + dest + "\n");
                    }
                }

//            // Testing Purposes only, comment out the for-loop above
//                t = addEdge("10", "11");
//                vtasks.add(t);
//                t = addEdge("49", "50");
//                vtasks.add(t);
//                t = addEdge("11", "12");
//                vtasks.add(t);
//                t = addEdge("15", "16");
//                vtasks.add(t);
//                t = addEdge("18", "19");
//                vtasks.add(t);
//                t = addEdge("22", "23");
//                vtasks.add(t);
//                t = addEdge("26", "27");
//                vtasks.add(t);
//                t = addEdge("7", "9");
//                vtasks.add(t);
//                t = addEdge("7", "10");
//                vtasks.add(t);
//                t = addEdge("7", "11");
//                vtasks.add(t);
//                t = addEdge("10", "14");
//                vtasks.add(t);
//                t = addEdge("12", "8");
//                vtasks.add(t);
//                t = addEdge("6", "13");
//                vtasks.add(t);
//                t = addEdge("1", "5");
//                vtasks.add(t);

                break;
            case 2: //delete Edge
                if (deleteEdgeUsingFile && (countDeleteEdgeOperations != 0)) {
                    //delete Edges read from file upto the count value
                    int i = 0;
                    String line = null;
                    int src = 0, dest = 0;
                    try {
                        BufferedReader br = new BufferedReader(new FileReader(deleteEdgeUsingFileName));
                        while ((line = br.readLine()) != null) {
                            String str[] = line.split(",");
                            src = Integer.parseInt(str[0]);
                            dest = Integer.parseInt(str[1]);
                            t = deleteEdge("" + src, "" + dest);
                            if (t == null) {
                                System.out.println("Task is null for delete edge case! Please check task list file...");
                                //i is not incremented
                            } else {
                                vtasks.add(t);
                                i++;
                                strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                            }
                            if (i >= countDeleteEdgeOperations) {
                                break;
                            }
                        }
                        br.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    //randomly delete edges
                    for (int i = 0; i < countDeleteEdgeOperations; i++) {
                        int src = r.nextInt(this.numNodes);
                        int dest = r.nextInt(this.numNodes);
                        while (dest == src) {
                            dest = r.nextInt(this.numNodes);
                        }
                        t = deleteEdge("" + src, "" + dest);
                        if (t == null) {
                            i--;
                        } else {
                            vtasks.add(t);
                            strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                        }
                    }

//            // Testing Purposes only, comment out the for-loop above
//                    vtasks.add(deleteEdge("20", "21"));
//                    vtasks.add(deleteEdge("40", "41"));
//                    vtasks.add(deleteEdge("60", "61"));
//                    vtasks.add(deleteEdge("80", "81"));
//			        t = deleteEdge("29", "30");
//                    vtasks.add(t);
//                    t = deleteEdge("69", "70");
//                    vtasks.add(t);
//                    t = deleteEdge("10", "11");
//                    vtasks.add(t);
//                    t = deleteEdge("40", "60");
//                    vtasks.add(t);
//                    t = deleteEdge("40", "41");
//                    vtasks.add(t);
//                    t = deleteEdge("49", "50");
//                    vtasks.add(t);
//                    t = deleteEdge("59", "60");
//                    vtasks.add(t);
//                    t = deleteEdge("4", "5");
//                    vtasks.add(t);
//                    t = deleteEdge("5", "6");
//                    vtasks.add(t);
//                    t = deleteEdge("6", "7");
//                    vtasks.add(t);
//                    t = deleteEdge("7", "8");
//                    vtasks.add(t);
//                    t = deleteEdge("8", "9");
//                    vtasks.add(t);
//                    t = deleteEdge("9", "10");
//                    vtasks.add(t);
//                    t = deleteEdge("10", "11");
//                    vtasks.add(t);
//                    t = deleteEdge("11", "12");
//                    vtasks.add(t);
//                    t = deleteEdge("12", "13");
//                    vtasks.add(t);
//                    t = deleteEdge("13", "14");
//                    vtasks.add(t);
//                    t = deleteEdge("14", "0");
//                    vtasks.add(t);
                }


                break;
        }

        if (showOutput) {
            System.out.println(strbuf);
        }

        return vtasks;
    }

    public UpdateGraphTask addEdge(String sourceNode, String destNode) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        UpdateGraphTask t = null;

        CCNode srcNode;
        try {
            srcNode = (CCNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = null;
            } else {
                t = new UpdateGraphTask(sourceNode, CCTaskConstants.addEdge, params);
                t.targetLocation = dataStorage.getLocation(t.nodeId);
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;
                t.phase = "GENERIC";
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    public UpdateGraphTask deleteEdge(String sourceNode, String destNode) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        UpdateGraphTask t = null;

        CCNode srcNode;
        try {
            srcNode = (CCNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = new UpdateGraphTask(sourceNode, CCTaskConstants.deleteEdge, params);
                t.targetLocation = dataStorage.getLocation(t.nodeId);
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;
                t.phase = "GENERIC";
            } else {
                t = null;
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    private HashMap<String, String> calculateEquivalentSets(Vector<String> baseInput) {
        HashMap<String, String> baseMap = new HashMap<String, String>();
        HashMap<String, Vector<String>> convergedBaseMap = new HashMap<String, Vector<String>>();

        for (String pair : baseInput) {

            //Split the string into two
            String[] splitedString = pair.split(",");

            String splitedPairFirst = splitedString[0];
            String splitedPairSecond = splitedString[1];

            if (baseMap.get(splitedPairFirst) == null) {
                baseMap.put(splitedPairFirst, splitedPairFirst);
            }

            if (baseMap.get(splitedPairSecond) == null) {
                baseMap.put(splitedPairSecond, baseMap.get(splitedPairFirst));
            } else {
                String initialParent = baseMap.get(splitedPairSecond);
                String currentParent = baseMap.get(splitedPairFirst);

                if (currentParent.compareTo(initialParent) < 0) {//Current parent is less

                    //Get all the initial parent value and map to current parent
                    Vector<String> childsIinitialParent = convergedBaseMap.get(initialParent);
                    for (String child : childsIinitialParent) {
                        baseMap.put(child, currentParent);
                    }

                    Vector<String> childsCurrentParent = convergedBaseMap.get(currentParent);
                    if (childsCurrentParent == null) {
                        Vector<String> tempList = new Vector<String>();
                        tempList.add(currentParent);
                        convergedBaseMap.put(currentParent, tempList);
                        childsCurrentParent = convergedBaseMap.get(currentParent);
                    }
                    for (String child : childsIinitialParent) {
                        childsCurrentParent.add(child);
                    }

                    //Remove entry of initial parent
                    convergedBaseMap.remove(initialParent);
                    continue; //convergedBaseMap processing is done here itself

                } else if (currentParent.compareTo(initialParent) > 0) { //Current parent is greater than initial parent

                    //Get all the current parent value and map to initial parent
                    Vector<String> childsCurrentParent = convergedBaseMap.get(currentParent);
                    if (childsCurrentParent == null) {
                        Vector<String> tempList = new Vector<String>();
                        tempList.add(currentParent);
                        convergedBaseMap.put(currentParent, tempList);
                        childsCurrentParent = convergedBaseMap.get(currentParent);
                    }
                    for (String child : childsCurrentParent) {
                        baseMap.put(child, initialParent);
                    }

                    Vector<String> childsInitialParent = convergedBaseMap.get(initialParent);
                    for (String child : childsCurrentParent) {
                        childsInitialParent.add(child);
                    }

                    //Remove entry of initial parent
                    convergedBaseMap.remove(currentParent);
                    continue; //convergedBaseMap processing is done here itself
                }
            }

            //ConvergeBaseAddition
            if (convergedBaseMap.get(baseMap.get(splitedPairSecond)) == null) {
                Vector<String> tempList = new Vector<String>();
                tempList.add(baseMap.get(splitedPairSecond));

                tempList.add(splitedPairSecond);
                convergedBaseMap.put(baseMap.get(splitedPairSecond), tempList);
            } else {
                Vector<String> tempList = convergedBaseMap.get(baseMap.get(splitedPairSecond));
                tempList.add(splitedPairSecond);
                convergedBaseMap.put(baseMap.get(splitedPairSecond), tempList);
            }

        }

		/*StringBuffer sb = new StringBuffer();
		for( String s1 : baseMap.keySet() ){
			sb.append(s1+","+baseMap.get(s1)+"\n");
		}
		System.out.println(sb.toString());

		System.out.println("------------------------------------------------------------------");

		for( String first : convergedBaseMap.keySet()){
			System.out.print("**"+first + "**"+"\n");
			for(String second : convergedBaseMap.get(first)) {
				System.out.println(second);
			}
			System.out.println("");
		}*/

        return baseMap;
    }

    private void writeResult(int phaseno) {
        try {
            ConcurrentHashMap<Integer, HashtableServer> servers = dataStorage.servers;
            Collection<HashtableServer> serverCollection = servers.values();
            Vector<Node> l = new Vector<Node>();
            for (HashtableServer server : serverCollection) {
                l.addAll(server.getAllNodes());
            }
            Collections.sort(l, new NodeComparator());

            String filename = "cc_result_" + phaseno + ".txt";
            System.out.println("Writing results of connected components computation to file " + filename + " ...");

            FileOutputStream fout = new FileOutputStream(filename, false);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            //out.println("==================================");

            StringBuffer sb = new StringBuffer();
            int i = 0;
            for (Node n : l) {
                if (Integer.parseInt(n.nodeId) < 0)
                    continue;
                CCNode cc = (CCNode) n;
                sb.append(cc.nodeId + " " + cc.label);
                out.println(sb.toString());
                //System.out.println(sb.toString());
                sb = new StringBuffer();
                i++;
                if (i % 1000 == 0) {
                    out.flush();
                }
            }

            //out.println("==================================");
            out.flush();
            lock.release();
            out.close();
            fout.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("Worker " + workerId + ":  Finished writing to file");
        System.out.println("*****************************");

    }

    public void cleanUp() {
        endTime = System.currentTimeMillis();
        long duration = endTime - startTime;

        long durationInitial = phaseEndTime[0] - startTime;
        long durationP1 = phaseEndTime[1] - phaseStartTime[1];
        long durationP2 = phaseEndTime[2] - phaseStartTime[2];
        long durationP3 = phaseEndTime[3] - phaseStartTime[3];
        long durationP4 = phaseEndTime[4] - phaseStartTime[4];


        System.out.println("*******************************");
        System.out.println("Total Execution Time taken : " + duration + " msec");
        System.out.println("Initial Computation Time taken : " + durationInitial + " msec");
        System.out.println("Phase 1 Time taken : " + durationP1 + " msec");
        System.out.println("Phase 2 Time taken : " + durationP2 + " msec");
        System.out.println("Phase 3 Time taken : " + durationP3 + " msec");
        System.out.println("Phase 4 Time taken : " + durationP4 + " msec");
        //String systemStats = "";
        try {
            //systemStats = dataStorage.getStats(true);
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            out.println("==================================");
            out.println("Total Execution Time taken : " + duration + " msec");
            out.println("Initial Computation Time taken : " + durationInitial + " msec");
            out.println("Phase 1 Time taken : " + durationP1 + " msec");
            out.println("Phase 2 Time taken : " + durationP2 + " msec");
            out.println("Phase 3 Time taken : " + durationP3 + " msec");
            out.println("Phase 4 Time taken : " + durationP4 + " msec");
            out.println("==================================");
            //out.println(systemStats);
            out.println("==================================");
            out.flush();
            lock.release();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("Worker " + workerId + ":  Finished cleanup");

        System.out.println("*****************************");

        //System.out.println(systemStats);
    }

    private boolean createSeedTask(Task tsk) throws Exception {
        int count = seedTaskCount;
        synchronized (unlabelled) {
            //System.out.println("Host "+dataStorage.myId+" :: unlabelled.size() :: "+unlabelled.size()+" set has : "+unlabelled.toString());
            if (unlabelled.size() == 0) {
                workpool.removeTask(tsk);
                return false;
            }
            if (unlabelled.size() < count) {    //reset count value if there are less than count number of unlabeled nodes
                count = unlabelled.size();
            }
            //System.out.println("Host "+dataStorage.myId+" :: after return unlabelled.size() :: "+unlabelled.size()+" set has : "+unlabelled.toString());
            //System.out.println("unlabelled : "+unlabelled.toString());

            Vector<Task> vecTasks = new Vector<Task>();
            String seedNodeId = null;

            List<String> unlabelledList = new ArrayList<String>(unlabelled);
            Collections.shuffle(unlabelledList);

            for (int i = 0; i < count; i++) {
                seedNodeId = unlabelledList.get(i);
                //System.out.println("Host "+dataStorage.myId+" :: seedNodeId :: "+seedNodeId);
                if (seedNodeId != null) {
                    Task t = new CCTask(seedNodeId, "-1"); //source node is null. So, -1
                    t.phase = "GENERIC";
                    t.targetLocation = dataStorage.getLocation(seedNodeId);
                    t.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    vecTasks.add(t);
                }
            }

            workpool.addTask(vecTasks);
            workpool.removeTask(tsk);
            return true;
        }
    }

    public class NodeComparator implements Comparator<Node> {
        public int compare(Node n1, Node n2) {
            if (Integer.parseInt(n1.nodeId) > Integer.parseInt(n2.nodeId)) return 1;
            else return -1;
        }
    }

}