/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.KNearestNeighborsV2;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.io.*;

import beehive.util.*;


public class KNNSortedSet implements Externalizable {

    public int capacity;
    public String nodeId;
    public SortedSet<KNNData> sortedNeighbors = new TreeSet<KNNData>(new KNNDataComparator());

    public KNNSortedSet(int capacity, String nodeId, ConcurrentHashMap<String, Edge> edges) {
        this.capacity = capacity;
        this.nodeId = nodeId;
        for (Map.Entry<String, Edge> entry : edges.entrySet()) {
            if (entry.getKey().equals(nodeId)) continue;
            this.add(new KNNData(entry.getKey(), ((KNNEdge) entry.getValue()).weight, entry.getKey()));
        }
//		System.out.println("Sorted Set "+sortedNeighbors+" created for "+Integer.toString(nodeId)+" with capacity "+capacity);
    }

    public KNNSortedSet(int capacity, String nodeId) {
        this.capacity = capacity;
        this.nodeId = nodeId;
    }

    public KNNSortedSet(KNNSortedSet k) {
        this.nodeId = k.nodeId;
        this.capacity = k.capacity;
        for (KNNData d : k.sortedNeighbors) {
            if (d.nodeId.equals(nodeId))
                continue;
            this.add(new KNNData(d));
        }
    }

    public KNNSortedSet() {
        nodeId = "";
    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.capacity = in.readInt();
        this.nodeId = (String) in.readObject();
        this.sortedNeighbors = (SortedSet<KNNData>) in.readObject();
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeInt(this.capacity);
        out.writeObject(this.nodeId);
        out.writeObject(this.sortedNeighbors);
    }

    public boolean add(KNNData data) {
        if (data.nodeId.equals(nodeId))
            return false;

        if (sortedNeighbors.size() == capacity && Integer.compare(data.distance, sortedNeighbors.last().distance) < 0) {
            KNNData old = null;
            for (KNNData d : sortedNeighbors) {
                if (d.nodeId.equals(data.nodeId)) {
                    old = d;
                    break;
                }
            }

            if (old != null) {
                if (old.distance <= data.distance) {
                    return false;
                } else {
                    Iterator<KNNData> it = sortedNeighbors.iterator();
                    while (it.hasNext()) {
                        KNNData d = it.next();
                        if (d.nodeId.equals(old.nodeId)) {
                            it.remove();
                        }
                    }
                    sortedNeighbors.add(data);
                    return true;
                }
            } else {
                sortedNeighbors.remove(sortedNeighbors.last());
                sortedNeighbors.add(data);
                return true;
            }
        } else if (sortedNeighbors.size() < capacity) {
            KNNData old = null;
            for (KNNData d : sortedNeighbors) {
                if (d.nodeId.equals(data.nodeId)) {
                    old = d;
                    break;
                }
            }
            if (old != null) {
                if (old.distance <= data.distance) {
                    return false;
                } else {
                    Iterator<KNNData> it = sortedNeighbors.iterator();
                    while (it.hasNext()) {
                        KNNData d = it.next();
                        if (d.nodeId.equals(old.nodeId)) {
                            it.remove();
                        }
                    }
                    sortedNeighbors.add(data);
                    return true;
                }
            } else {
                sortedNeighbors.add(data);
                return true;
            }
        }
        return false;
    }

    public boolean addAll(SortedSet<KNNData> nodes, String pathThrough, int pathThroughDistance) {
//		System.out.println("Add "+nodes+" with nodeId "+nodeId);
        boolean ret = false;
        for (KNNData node : nodes) {
            KNNData data = new KNNData(node);
            if (pathThrough.equals("-1")) {
                data.pathThrough = data.nodeId;
            } else {
                data.pathThrough = pathThrough;
                data.distance = data.distance + pathThroughDistance;
            }

            boolean localret = this.add(data);
            ret = ret || localret;
        }
        return ret;
    }

    public int size() {
        return sortedNeighbors.size();
    }

    public SortedSet<KNNData> getNeighbors() {
        return sortedNeighbors;
    }

    public class KNNDataComparator implements Comparator<KNNData>, Serializable {

        @Override
        public int compare(KNNData data1, KNNData data2) {
            if (data1.nodeId.equals(data2.nodeId)) return 0;
            else if (data1.distance < data2.distance) {
                return -1;
            }
            return 1;
        }

    }
}
