/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.KNearestNeighborsV2;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import beehive.util.Task;
import beehive.workpool.TaskComparator;

public class KNNTaskComparator extends TaskComparator{
	public boolean taskAggrReplaceAndEnsureKItems = false;
	
	public KNNTaskComparator(boolean taskAggrReplaceAndEnsureKItems) {
		this.taskAggrReplaceAndEnsureKItems = taskAggrReplaceAndEnsureKItems;
		if(taskAggrReplaceAndEnsureKItems){
			System.out.println("Task Aggregator initialized with replace duplicate nodeId and ensure only K items logic...");
		}
		else{
			System.out.println("Task Aggregator initialized with replace duplicate nodeId logic...");
		}
	}
	
	@Override
	public int combine(Task oldTask, Task newTask) {
		Class c = newTask.getClass();
		String strclassnamearr[] = c.toString().split(" ")[1].split("\\.");
		String newTaskClassName = strclassnamearr[strclassnamearr.length-1];
		
		if(newTaskClassName.equals("KNNTask")){
			if(oldTask == null){
				return ADD_TO_TASKPOOL_AND_UPDATE_TASKMAP;
			}
			else{
				KNNTask oldSPTask = (KNNTask) oldTask;
				KNNTask newSPTask = (KNNTask) newTask;
				if(taskAggrReplaceAndEnsureKItems){
					replaceAndEnsureKItems(oldSPTask, newSPTask);
				}
				else{
					replace(oldSPTask, newSPTask);
				}
				
				return IGNORE_NEW_TASK;
			}
		}
		else{
			return ADD_TO_TASKPOOL_AND_DONT_UPDATE_TASKMAP;
		}
	}
	
	public void replaceAndEnsureKItems(KNNTask oldSPTask, KNNTask newSPTask){
		HashMap<String, KNNItem> toNodeIdMap = new HashMap<String, KNNItem>();
		HashMap<String, Integer> toNodeIdTotalDistanceMap = new HashMap<String, Integer>();
		
		Iterator<String> itNew = newSPTask.sortedNeighborsMap.keySet().iterator();
		String newSPTaskSender = itNew.next(); 
		KNNSortedSet newSPTaskKSet = newSPTask.sortedNeighborsMap.get(newSPTaskSender);
		int k = newSPTaskKSet.capacity;
		
		//the new SPTask has only one sender and the KSet of only that sender with unique nodeIds in the KSet, and KSet size is atmost k
		//so putting these items directly into toNodeIdMap and toNodeIdTotalDistanceMap
		int newSPTaskEdgeLength = newSPTask.edgeLengthMap.get(newSPTaskSender);
		Iterator<KNNData> itnew = newSPTaskKSet.sortedNeighbors.iterator();
		while(itnew.hasNext()){
			KNNData d = itnew.next();
			KNNItem item = new KNNItem(newSPTaskSender, newSPTaskEdgeLength, d.distance, d.nodeId, d.pathThrough);
			toNodeIdMap.put(d.nodeId, item);
			toNodeIdTotalDistanceMap.put(d.nodeId, item.totalDist);
		}
		
		//performing operation for old task
		Iterator<String> iterOld = oldSPTask.sortedNeighborsMap.keySet().iterator();
		while(iterOld.hasNext()){
			String sender = iterOld.next();
			KNNSortedSet KSet = oldSPTask.sortedNeighborsMap.get(sender);
			Iterator<KNNData> iterKset = KSet.sortedNeighbors.iterator();
			while(iterKset.hasNext()){
				KNNData d = iterKset.next();
				KNNItem item = new KNNItem(sender, oldSPTask.edgeLengthMap.get(sender), d.distance, d.nodeId, d.pathThrough);
				
				//removing duplicate items
				KNNItem olditem = toNodeIdMap.get(d.nodeId);
				if(olditem != null){
					int oldTotalDist = toNodeIdTotalDistanceMap.get(d.nodeId);
					if(oldTotalDist > item.totalDist){
						toNodeIdMap.put(d.nodeId, item);
						toNodeIdTotalDistanceMap.put(d.nodeId, item.totalDist);
					}
				}
				else{
					toNodeIdMap.put(d.nodeId, item);
					toNodeIdTotalDistanceMap.put(d.nodeId, item.totalDist);
				}
				
				//ensuring that there are atmost only k items
				if(toNodeIdMap.size() > k){	
					//here, size has become k+1. So, deleting the largest item
					int largestTotalDist = Integer.MIN_VALUE;
					String largestNodeId = null;
					Iterator<String> iterKNNItem = toNodeIdTotalDistanceMap.keySet().iterator();
					while(iterKNNItem.hasNext()){
						String nId = iterKNNItem.next();
						int totDist = toNodeIdTotalDistanceMap.get(nId);
						if(totDist > largestTotalDist){
							largestTotalDist = totDist;
							largestNodeId = nId;
						}
					}
					toNodeIdMap.remove(largestNodeId);
					toNodeIdTotalDistanceMap.remove(largestNodeId);
				}
			}
		}
		
		//rebuilding the KSortedSet
		HashMap<String, KNNSortedSet> sortedNeighborsMap = new HashMap<String, KNNSortedSet>();
		HashMap<String, Integer> edgeLengthMap = new HashMap<String, Integer>();
		
		Iterator<String> iterKNNItem = toNodeIdMap.keySet().iterator();
		while(iterKNNItem.hasNext()){
			String nodeId = iterKNNItem.next();
			KNNItem item = toNodeIdMap.get(nodeId);
			
			KNNSortedSet tempKSet = sortedNeighborsMap.get(nodeId);
			if(tempKSet==null){
				KNNSortedSet newKSet = new KNNSortedSet(k, nodeId);
				KNNData d = new KNNData(item.toNodeId, item.distance, item.pathThrough);
				newKSet.sortedNeighbors.add(d);
				
				sortedNeighborsMap.put(nodeId, newKSet);
				edgeLengthMap.put(nodeId, item.edgelength);
			}
			else{
				KNNData d = new KNNData(item.toNodeId, item.distance, item.pathThrough);
				tempKSet.sortedNeighbors.add(d);
			}
		}
		
		//resetting items in oldSPTask to the current values
		oldSPTask.sortedNeighborsMap = sortedNeighborsMap;
		oldSPTask.edgeLengthMap = edgeLengthMap;
	}
	
	public class KNNItem{
		String sender;
		int edgelength;
		int distance;
		String toNodeId;
		String pathThrough;
		int totalDist;
		
		public KNNItem(String sender, int edgelength, int distance, String toNodeId, String pathThrough){
			this.sender = sender;
			this.edgelength = edgelength;
			this.distance = distance;
			this.toNodeId = toNodeId;
			this.pathThrough = pathThrough;
			this.totalDist = edgelength+distance;
		}
	}
	
	//older KNNTask replace logic
	//this method only removes all nodes which are present in both old and new SPTask 
	//such that every nodeId in KNNData is present in only one of the two tasks.
	public void replace(KNNTask oldSPTask, KNNTask newSPTask){
		Iterator<String> itNew = newSPTask.sortedNeighborsMap.keySet().iterator();
		String newSPTaskSender = itNew.next(); 
		KNNSortedSet newSPTaskKSet = newSPTask.sortedNeighborsMap.get(newSPTaskSender);
		int newSPTaskEdgeLength = newSPTask.edgeLengthMap.get(newSPTaskSender);
		HashMap<String, KNNData> newSPTaskKNNDataMap = new HashMap<String, KNNData>();
		Iterator<KNNData> itnew = newSPTaskKSet.sortedNeighbors.iterator();
		while(itnew.hasNext()){
			KNNData d = itnew.next();
			newSPTaskKNNDataMap.put(d.nodeId, d);
		}
		
		Vector<String> oldSPTaskSenderTaskToRemove = new Vector<String>();
		Iterator<String> itOld = oldSPTask.sortedNeighborsMap.keySet().iterator();
		while(itOld.hasNext()){
			String oldSPTaskSender = itOld.next();
			KNNSortedSet oldSPTaskKSet = oldSPTask.sortedNeighborsMap.get(oldSPTaskSender);
			int oldSPTaskEdgeLength = oldSPTask.edgeLengthMap.get(oldSPTaskSender);
			
			Iterator<KNNData> it = oldSPTaskKSet.sortedNeighbors.iterator();
			while(it.hasNext()){
				KNNData d = it.next();
				KNNData newTaskKNNData = newSPTaskKNNDataMap.get(d.nodeId);
				if(newTaskKNNData != null){
					if( (d.distance + oldSPTaskEdgeLength) >= (newTaskKNNData.distance + newSPTaskEdgeLength) ){
						it.remove();	//remove oldTask's KNNData
					}
					else{ //remove newTask's KNNData
						Iterator<KNNData> itnewSP = newSPTaskKSet.sortedNeighbors.iterator();
						while(itnewSP.hasNext()){
							KNNData dNew = itnewSP.next();
							if(dNew.nodeId.equals(d.nodeId)){
								itnewSP.remove();
							}
						}
						newSPTaskKNNDataMap.remove(d.nodeId);
					}
				}
			}
			if(oldSPTaskKSet.sortedNeighbors.size()==0){
				oldSPTaskSenderTaskToRemove.add(oldSPTaskSender);
			}
		}
		
		for(String oldSPTaskSender : oldSPTaskSenderTaskToRemove){
			oldSPTask.sortedNeighborsMap.remove(oldSPTaskSender);
			oldSPTask.edgeLengthMap.remove(oldSPTaskSender);
		}
		
		oldSPTask.sortedNeighborsMap.put(newSPTaskSender, new KNNSortedSet(newSPTaskKSet));
		oldSPTask.edgeLengthMap.put(newSPTaskSender, newSPTaskEdgeLength);
	}
}
