/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.KNearestNeighborsV2;

/***
 * Author: Charandeep Parisineti
 */

import TestProgs.ShortestPathIncremental.UpdateGraphTask;
import beehive.server.HashtableServer;
import beehive.util.Node;
import beehive.util.StorageSystem;
import beehive.util.Task;
import beehive.util.TaskConstants;
import beehive.worker.Worker;
import beehive.worker.WorkerStatus;
import beehive.workpool.LocalWorkpoolImpl;
import beehive.workpool.StatusType;
import beehive.workpool.WorkStatus;

import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintWriter;
import java.nio.channels.FileLock;
import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

public class KNNWorker extends Worker implements Runnable {

    private boolean master = false;
    private int workerId = 0;
    private String myStorageURL;
    private int numNodes = 0;
    int k;

    private long phaseStartTime[] = new long[6];
    private long phaseEndTime[] = new long[6];

    public KNNWorker(String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, int workerId, String myStorageURL, int k, int numOfNodes) throws RemoteException {
        super(configFile, wpool, workerId, datastorage, master);
        this.master = master;
        this.dataStorage = datastorage;
        this.workerId = workerId;
        this.myStorageURL = myStorageURL;
        this.numNodes = numOfNodes;
        this.k = k;
    }

    public void createTasksAllVertices(Task tsk) throws Exception {

        Vector<String> myLocalKeys = dataStorage.getAllKeys(myStorageURL);
        Vector<Task> taskVector = new Vector<Task>();
        int batch_size = 1000;
        for (String key : myLocalKeys) {
            if (key.equals(Integer.toString(numNodes))) {
                continue;
            }
            KNNNode node = (KNNNode) dataStorage.getNode(key);
            String location = dataStorage.getLocation(key);
            KNNTask t = new KNNTask(k, node.nodeId, node.neighbors, node.nodeId, 0);
            t.targetLocation = location;
            t.affinityLevel = TaskConstants.STRONG_AFFINITY;
            t.phase = "GENERIC";
            taskVector.add(t);
            if (taskVector.size() == batch_size) {
                workpool.addTask(taskVector);
                taskVector.clear();
            }
        }
        if (taskVector.size() > 0) {
            workpool.addTask(taskVector);
        }
        workpool.removeTask(tsk);
    }

    public Node compute(Task T, Node node) {
        boolean ret = false;
        KNNTask t = (KNNTask) T;
        KNNNode u = new KNNNode((KNNNode) node);

        try {
            for (String nbrId : t.sortedNeighborsMap.keySet()) {
                KNNSortedSet kSet = t.sortedNeighborsMap.get(nbrId);

                String senderId = kSet.nodeId;
                if (kSet.nodeId.equals(node.nodeId)) {
                    boolean localret = u.sortedNeighbors.addAll(kSet.sortedNeighbors, "-1", 0);
                    ret = ret || localret;
                } else {
                    if (u.neighbors.get(kSet.nodeId) != null) {    //checking if the edge is not deleted parallely
                        int dist = ((KNNEdge) u.neighbors.get(kSet.nodeId)).weight;
                        for (KNNData data : kSet.sortedNeighbors) {
                            boolean localret = u.sortedNeighbors.add(new KNNData(data.nodeId, data.distance + dist, senderId));
                            ret = ret || localret;
                        }
                    }
                }
            }

            if (ret) {
                updatedNodes.put(u.nodeId, u);
                rwsetinfo.writeSet.add(u.nodeId);

                for (String nbrId : u.neighbors.keySet()) {
                    if (nbrId.equals(u.nodeId))
                        continue;

                    //System.out.println("Task at node "+ u.nodeId +  " creating a task for " + nbrId );
                    KNNTask t1 = new KNNTask(nbrId, u.sortedNeighbors, u.nodeId, ((KNNEdge) u.neighbors.get(nbrId)).weight);
                    t1.targetLocation = dataStorage.getLocation(nbrId);
                    t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    tasksCreated.add(t1);
                }
                return u;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    private Node recompute(Task T, Node node) {
        Vector<Task> initialTasks = new Vector<Task>();
        try {
            RecomputeTask t = (RecomputeTask) T;
            KNNNode tgtNode = new KNNNode((KNNNode) node);
            //System.out.println("Target  Node: " + u.nodeId + "Task node-id: " + t.nodeId +  " sortedNeighbors nodeId: "+ t.sortedNeighbors.nodeId);

            //building deleted set
            HashSet<String> deletedSet = new HashSet<String>();
            Set<KNNData> toBeDeleted = new HashSet<KNNData>();
            for (KNNData d : tgtNode.sortedNeighbors.sortedNeighbors) {
                if (t.deletedSet.contains(d.nodeId) && d.pathThrough.equals(t.senderId)) {
                    deletedSet.add(d.nodeId);
                    toBeDeleted.add(d);
                }
            }

            if (deletedSet.size() > 0) { //KNN set is modified
                updatedNodes.put(tgtNode.nodeId, tgtNode);
                rwsetinfo.writeSet.add(tgtNode.getNodeId());

                //deleting all nodes obtained above in deleted set from KNN set
                Iterator<KNNData> iter = tgtNode.sortedNeighbors.sortedNeighbors.iterator();
                while (iter.hasNext()) {
                    KNNData d = iter.next();
                    if (toBeDeleted.contains(d)) {
                        iter.remove();
                    }
                }

                //adding all neighbor edges to SS
                Iterator it = tgtNode.neighbors.keySet().iterator();
                while (it.hasNext()) {
                    String nbr = (String) it.next();
                    KNNEdge nbrEdge = (KNNEdge) tgtNode.neighbors.get(nbr);
                    tgtNode.sortedNeighbors.add(new KNNData(nbr, nbrEdge.weight, nbr));
                }

                //adding KNN set set from neighbor who has sent this task
                {
                    KNNSortedSet ss = new KNNSortedSet(t.sortedNeighbors);

                    Set<KNNData> toBeDeleted2 = new HashSet<KNNData>();
                    for (KNNData d : ss.sortedNeighbors) {
                        if (d.pathThrough == tgtNode.nodeId) {
                            toBeDeleted2.add(d);
                        }
                    }
                    Iterator<KNNData> iter2 = ss.sortedNeighbors.iterator();
                    while (iter2.hasNext()) {
                        KNNData d = iter2.next();
                        if (toBeDeleted2.contains(d)) {
                            iter2.remove();
                        }
                    }

                    if (tgtNode.neighbors.get(t.senderId) != null) {
                        int dist = ((KNNEdge) tgtNode.neighbors.get(t.senderId)).weight;
                        tgtNode.sortedNeighbors.addAll(ss.sortedNeighbors, t.senderId, dist);
                    }
                }

                //creating Recompute task for neighbors
                Enumeration<String> nbrEnum = tgtNode.neighbors.keys();
                while (nbrEnum.hasMoreElements()) {
                    String nbrId = nbrEnum.nextElement();
                    if (nbrId.equals(tgtNode.nodeId))
                        continue;

                    RecomputeTask t1 = new RecomputeTask(nbrId, tgtNode.sortedNeighbors, tgtNode.nodeId, deletedSet);
                    String location = dataStorage.getLocation(nbrId);
                    t1.targetLocation = location;
                    t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    tasksCreated.add(t1);
                }

                return tgtNode;
            } else {    //deletedSet == 0. So, no recompute task will be created. There may be a optimizedCompute task if the KNN set changes
                //adding KNN set set from neighbor who has sent this task
                boolean ret = false;
                KNNSortedSet ss = new KNNSortedSet(t.sortedNeighbors);

                Set<KNNData> toBeDeleted2 = new HashSet<KNNData>();
                for (KNNData d : ss.sortedNeighbors) {
                    if (d.pathThrough == tgtNode.nodeId) {
                        toBeDeleted2.add(d);
                    }
                }
                Iterator<KNNData> iter2 = ss.sortedNeighbors.iterator();
                while (iter2.hasNext()) {
                    KNNData d = iter2.next();
                    if (toBeDeleted2.contains(d)) {
                        iter2.remove();
                    }
                }

                if (tgtNode.neighbors.get(t.senderId) != null) {
                    int dist = ((KNNEdge) tgtNode.neighbors.get(t.senderId)).weight;
                    boolean localret = tgtNode.sortedNeighbors.addAll(ss.sortedNeighbors, t.senderId, dist);
                    ret = ret || localret;

                    //send task to sender, to update the sender's KSet using the tgtNode's KSet
                    KNNTask tSndr = new KNNTask(t.senderId, tgtNode.sortedNeighbors, tgtNode.nodeId, ((KNNEdge) tgtNode.neighbors.get(t.senderId)).weight);
                    String locationSndr = dataStorage.getLocation(t.senderId);
                    tSndr.targetLocation = locationSndr;
                    tSndr.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    tasksCreated.add(tSndr);

                    if (ret) {
                        updatedNodes.put(tgtNode.nodeId, tgtNode);
                        rwsetinfo.writeSet.add(tgtNode.nodeId);
                        Enumeration<String> nbrEnum = tgtNode.neighbors.keys();
                        while (nbrEnum.hasMoreElements()) {
                            String nbrId = nbrEnum.nextElement();
                            if (nbrId.equals(tgtNode.nodeId) || nbrId.equals(t.senderId))
                                continue;

                            //System.out.println("Task at node "+ u.nodeId +  " creating a task for " + nbrId );
                            KNNTask t1 = new KNNTask(nbrId, tgtNode.sortedNeighbors, tgtNode.nodeId, ((KNNEdge) tgtNode.neighbors.get(nbrId)).weight);
                            String location = dataStorage.getLocation(nbrId);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                        return tgtNode;
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            //System.exit(1);
        }
        return null;
    }

    public void doTask(Task tsk) throws Exception {
        tasksCreated.clear();
        updatedNodes.clear();
        boolean commit = false;
        Node resNode = null;

        Class c = tsk.getClass();
        String strclassnamearr[] = c.toString().split(" ")[1].split("\\.");
        String taskClassName = strclassnamearr[strclassnamearr.length - 1];
        //System.out.println("Task class name : "+taskClassName);

        while (!commit) {
            Node tgtNode = dataStorage.getNode(tsk.getNodeId());
            tsk.txnInfo = beginTransaction();
            try {
                switch (taskClassName) {
                    case "UpdateGraphTask":
                        resNode = updateGraph(tsk, tgtNode);
                        break;
                    case "RecomputeTask":
                        resNode = recompute(tsk, tgtNode);
                        break;
                    default:
                        resNode = compute(tsk, tgtNode);
                        break;
                }

                if (resNode != null) {
                    commit = validateTask(tsk);
                } else {
                    commit = true;
                    workerStatus.status = WorkerStatus.REPORT_COMPLETION;
                    workpool.reportCompletion(tsk, tasksCreated);
                    break;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }

        }
    }

    private Vector<UpdateGraphTask> createGraphUpdateTasks(int hostId, int phaseno) {
        //printResult(phaseno-1);
        //only master will print this line
        if (master) {
            System.out.println("\nAdding new changes in the graph...\n");
        }
        //loadGenerator
        Vector<UpdateGraphTask> vtasks = new Vector<UpdateGraphTask>();
        int countAddEdgeOperations = 0, countDeleteEdgeOperations = 0, countUpdateEdgeOperations = 0;
        //int countAddNodeOperations=0, countDeleteNodeOperations=0;
        boolean deleteEdgeUsingFile = false;
        boolean showOutput = false;
        //int countMaxNeighbors=0;
        String deleteEdgeUsingFileName = "";
        int maxEdgeLength = 100;
        int numServers = dataStorage.getNumberOfServers();
        showOutput = configParams.getBoolValue("KNNWorker.SHOW_OUTPUT");
        countAddEdgeOperations = configParams.getIntValue("KNNWorker.ADD_EDGE")/*/numServers*/;
        maxEdgeLength = configParams.getIntValue("KNNWorker.EDGE_MAX_LENGTH");
        countDeleteEdgeOperations = configParams.getIntValue("KNNWorker.DELETE_EDGE")/*/numServers*/;
        countUpdateEdgeOperations = configParams.getIntValue("KNNWorker.UPDATE_EDGE")/*/numServers*/;
        deleteEdgeUsingFile = configParams.getBoolValue("KNNWorker.DELETE_EDGE_USING_FILE");
        if (deleteEdgeUsingFile) {
            deleteEdgeUsingFileName = configParams.getStringValue("KNNWorker.DELETE_EDGE_USING_FILENAME");
        }

        Random r = new Random();
        StringBuffer strbuf = new StringBuffer();
        UpdateGraphTask t = null;

        switch (phaseno) {
            case 1: //add Edge
                for (int i = 0; i < countAddEdgeOperations; i++) {
                    int src = r.nextInt(this.numNodes);
                    int dest = r.nextInt(this.numNodes);
                    while (dest == src) {
                        dest = r.nextInt(this.numNodes);
                    }
                    int len = r.nextInt(maxEdgeLength) + 1;
                    t = addEdge("" + src, "" + dest, len);
                    if (t == null) {
                        i--;
                    } else {
                        vtasks.add(t);
                        strbuf.append("Host " + dataStorage.myId + " :: Adding edge between Node " + src + " and " + dest + "\n");
                    }
                }
                break;
            case 2: //delete Edge
                if (deleteEdgeUsingFile && (countDeleteEdgeOperations != 0)) {
                    //delete Edges read from file upto the count value
                    int i = 0;
                    String line = null;
                    int src = 0, dest = 0;
                    try {
                        BufferedReader br = new BufferedReader(new FileReader(deleteEdgeUsingFileName));
                        while ((line = br.readLine()) != null) {
                            String str[] = line.split(",");
                            src = Integer.parseInt(str[0]);
                            dest = Integer.parseInt(str[1]);
                            t = deleteEdge("" + src, "" + dest);
                            if (t == null) {
                                System.out.println("Task is null for delete edge scenario! Please check edge list file...");
                            } else {
                                vtasks.add(t);
                                i++;
                                strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                            }
                            if (i >= countDeleteEdgeOperations) {
                                break;
                            }
                        }
                        br.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    //randomly delete edges
                    for (int i = 0; i < countDeleteEdgeOperations; i++) {
                        int src = r.nextInt(this.numNodes);
                        int dest = r.nextInt(this.numNodes);
                        while (dest == src) {
                            dest = r.nextInt(this.numNodes);
                        }
                        t = deleteEdge("" + src, "" + dest);
                        if (t == null) {
                            i--;
                        } else {
                            vtasks.add(t);
                            strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                        }
                    }
                }
                break;
            case 3: //update Edge
                long[] lengthIncreasedCount = new long[1];    //create arrays to pass so that it is a pass by reference call
                long[] lengthDecreasedCount = new long[1];

                if (deleteEdgeUsingFile && (countUpdateEdgeOperations != 0)) {
                    //update Edges read from file upto the count value
                    int i = 0;
                    int skipedges = 0;
                    String line = null;
                    int src = 0, dest = 0;
                    try {
                        BufferedReader br = new BufferedReader(new FileReader(deleteEdgeUsingFileName));
                        while ((line = br.readLine()) != null) {
                            if (skipedges <= countDeleteEdgeOperations) {
                                skipedges++;
                                continue;
                            }
                            String str[] = line.split(",");
                            src = Integer.parseInt(str[0]);
                        /*if((src%numServers)!=dataStorage.myId)
                            continue;*/
                            dest = Integer.parseInt(str[1]);
                            int len = r.nextInt(maxEdgeLength) + 1;
                            t = updateEdge("" + src, "" + dest, len, lengthIncreasedCount, lengthDecreasedCount);
                            if (t == null) {
                                System.out.println("Task is null for update edge scenario! Please check edge list file...");
                            } else {
                                vtasks.add(t);
                                i++;
                                strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                            }
                            if (i >= countDeleteEdgeOperations) {
                                break;
                            }
                        }
                        br.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    for (int i = 0; i < countUpdateEdgeOperations; i++) {
                        int src = r.nextInt(this.numNodes);
                        int dest = r.nextInt(this.numNodes);
                        while (dest == src) {
                            dest = r.nextInt(this.numNodes);
                        }
                        int len = r.nextInt(maxEdgeLength) + 1;
                        t = updateEdge("" + src, "" + dest, len, lengthIncreasedCount, lengthDecreasedCount);
                        if (t == null) {
                            i--;
                        } else {
                            vtasks.add(t);
                            strbuf.append("Host " + dataStorage.myId + " :: Updating edge between Node " + src + " and " + dest + "\n");
                        }
                    }

                }

                //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" :: In Update Edge operations, edge length : increased = "+lengthIncreasedCount[0]+", decreased = "+lengthDecreasedCount[0]);
                break;
        }

        if (showOutput) {
            System.out.println(strbuf);
        }

        return vtasks;
    }

    public UpdateGraphTask addEdge(String sourceNode, String destNode, int edgeLength) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        params.add(edgeLength);
        UpdateGraphTask t = null;

        KNNNode srcNode;
        try {
            srcNode = (KNNNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                //t = new UpdateGraphTask(sourceNode, KNNTaskConstants.updateEdge, params);
                t = null;
            } else {
                t = new UpdateGraphTask(sourceNode, KNNTaskConstants.addEdge, params);
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    public UpdateGraphTask deleteEdge(String sourceNode, String destNode) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        UpdateGraphTask t = null;

        KNNNode srcNode;
        try {
            srcNode = (KNNNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = new UpdateGraphTask(sourceNode, KNNTaskConstants.deleteEdge, params);
            } else {
                t = null;
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    private UpdateGraphTask updateEdge(String sourceNode, String destNode, int edgeLength, long[] lengthIncreasedCount, long[] lengthDecreasedCount) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        params.add(edgeLength);
        UpdateGraphTask t = null;

        KNNNode srcNode;
        try {
            srcNode = (KNNNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = new UpdateGraphTask(sourceNode, KNNTaskConstants.updateEdge, params);
                KNNEdge e = (KNNEdge) srcNode.neighbors.get(destNode);
                if (e.weight > edgeLength) {
                    lengthDecreasedCount[0]++;
                } else if (e.weight < edgeLength) {
                    lengthIncreasedCount[0]++;
                }
            } else {
                //t = new UpdateGraphTask(sourceNode, KNNTaskConstants.addEdge, params);
                t = null;
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }


    private Node updateGraph(Task T, Node node) {
        KNNNode tgtNode = null;
        if (node != null)
            tgtNode = new KNNNode((KNNNode) node);    //you need a copy of orig node

        try {
            UpdateGraphTask t = (UpdateGraphTask) T;

            switch (t.opname) {
                case KNNTaskConstants.addEdge:
                    addEdge(tgtNode, t);
                    break;
                case KNNTaskConstants.deleteEdge:
                    deleteEdge(tgtNode, t);
                    break;
                case KNNTaskConstants.updateEdge:
                    updateEdge(tgtNode, t);
                    break;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return tgtNode;
    }

    private void addEdge(KNNNode tgtNode, UpdateGraphTask t) throws RemoteException {
        String nbrid = (String) t.params.get(0);
        Integer edge = (Integer) t.params.get(1);
        int edgelength = edge;
        //System.out.println("inside addEdge:: node : "+node.nodeId+", nbrid : "+nbrid);
        KNNNode nbrNode = new KNNNode((KNNNode) dataStorage.getNode(nbrid));

        KNNEdge edgeToNbr = new KNNEdge(edgelength);
        nbrNode.addNeighbor(tgtNode.nodeId, edgeToNbr);
        KNNEdge edgeToTgt = new KNNEdge(edgelength);
        tgtNode.addNeighbor(nbrNode.nodeId, edgeToTgt);

        updatedNodes.put(tgtNode.nodeId, tgtNode);
        updatedNodes.put(nbrNode.nodeId, nbrNode);

        rwsetinfo.writeSet.add(tgtNode.getNodeId());
        rwsetinfo.writeSet.add(nbrNode.getNodeId());

        createNbrKNNTasks(tgtNode, edgelength, nbrNode);
        createNbrKNNTasks(nbrNode, edgelength, tgtNode);
    }


    private void deleteEdge(KNNNode tgtNode, UpdateGraphTask t) throws RemoteException {
        String nbrid = (String) t.params.get(0);
        //System.out.println("inside deleteEdge::"+" nbrid : "+nbrid);
        KNNNode nbrNode = new KNNNode((KNNNode) dataStorage.getNode(nbrid));

        nbrNode.neighbors.remove(tgtNode.nodeId);
        tgtNode.neighbors.remove(nbrNode.nodeId);


        updatedNodes.put(tgtNode.nodeId, tgtNode);
        updatedNodes.put(nbrNode.nodeId, nbrNode);
        rwsetinfo.writeSet.add(tgtNode.getNodeId());
        rwsetinfo.writeSet.add(nbrNode.getNodeId());
        //System.out.println("Worker "+workerId+" : Inside deleteEdge :: deleting edge between target : "+tgtNode.nodeId+" and nbr : "+nbrNode.nodeId);

        createNbrRecomputeTasks(tgtNode, nbrNode);
        createNbrRecomputeTasks(nbrNode, tgtNode);
    }

    /**
     * Attempts to add nbr and its neighbors as neighbors of target in a sorted fashion. If successful, then
     * KNNTasks are created for each neighbor of the target
     *
     * @param target     The node to add neighbors to
     * @param edgelength The edgeLength between target and nbr
     * @param nbr        The node that is being added as a neighbor to the target
     */
    private void createNbrKNNTasks(KNNNode target, int edgelength, KNNNode nbr) {
        boolean didAddSucceed = target.sortedNeighbors.add(new KNNData("" + nbr.nodeId, edgelength, nbr.nodeId));

        if (didAddSucceed) {
            didAddSucceed = target.sortedNeighbors.addAll(nbr.sortedNeighbors.sortedNeighbors, nbr.nodeId, edgelength);

            if (didAddSucceed) {
                for (String nbrId : target.neighbors.keySet()) {
                    if (nbrId.equals(target.nodeId))
                        continue;

                    KNNTask nbrTask = new KNNTask(nbrId, target.sortedNeighbors, target.nodeId, ((KNNEdge) target.neighbors.get(nbrId)).weight);
                    nbrTask.targetLocation = dataStorage.getLocation(nbrId);
                    nbrTask.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    tasksCreated.add(nbrTask);
                }
            }
        }
    }

    /**
     * Attempts to remove nbr as a neighbor of target. If successful, then RecomputeKNN
     * tasks are created for each neighbor of the target who have nbr as their neighbor
     *
     * @param target The node from which to remove nbr as a neighbor
     * @param nbr    The neighbor to be removed
     */
    private void createNbrRecomputeTasks(KNNNode target, KNNNode nbr) {
        boolean containsDeletedEdge = false;
        HashSet<String> deletedSet = new HashSet<>();
        for (KNNData nbrId : target.sortedNeighbors.sortedNeighbors) {
            if (nbrId.nodeId.equals(nbr.nodeId)) {
                containsDeletedEdge = true;
                break;
            }
        }

        if (containsDeletedEdge) {
            // Find all neighbors of target that are also neighbors of nbr
            Set<KNNData> toBeDeleted = new HashSet<>();
            for (KNNData nbrId : target.sortedNeighbors.sortedNeighbors) {
                if (nbrId.pathThrough.equals(nbr.nodeId)) {
                    deletedSet.add(nbrId.nodeId);
                    toBeDeleted.add(nbrId);
                }
            }

            //deleting all nodes obtained above in deleted set from KNN set
            Iterator<KNNData> iter = target.sortedNeighbors.sortedNeighbors.iterator();
            while (iter.hasNext()) {
                KNNData d = iter.next();
                if (toBeDeleted.contains(d)) {
                    iter.remove();
                }
            }

            //adding all neighbor edges to SS
            for (String nbrId : target.neighbors.keySet()) {
                KNNEdge nbrEdge = (KNNEdge) target.neighbors.get(nbrId);
                target.sortedNeighbors.add(new KNNData(nbrId, nbrEdge.weight, nbrId));
            }

            //creating Recompute task for neighbors
            for (String nbrId : target.neighbors.keySet()) {
                if (nbrId.equals(target.nodeId))
                    continue;

                RecomputeTask t1 = new RecomputeTask(nbrId, target.sortedNeighbors, target.nodeId, deletedSet);
                t1.targetLocation = dataStorage.getLocation(nbrId);
                t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                tasksCreated.add(t1);
            }
        }
    }

    private void updateEdge(KNNNode tgtNode, UpdateGraphTask t) throws RemoteException {
        String nbrid = (String) t.params.get(0);
        Integer edge = (Integer) t.params.get(1);
        int edgelength = edge;

        //System.out.println("inside updateEdge:: node : "+node.nodeId+", nbrid : "+nbrid);
        KNNNode nbrNode = new KNNNode((KNNNode) dataStorage.getNode(nbrid));

        int edgeOldWeight = ((KNNEdge) tgtNode.neighbors.get(nbrid)).weight;

        //adding the same edge with new weight with replace the older value in hashtable
        KNNEdge edgeToNbr = new KNNEdge(edgelength);
        nbrNode.addNeighbor(tgtNode.nodeId, edgeToNbr);
        KNNEdge edgeToTgt = new KNNEdge(edgelength);
        tgtNode.addNeighbor(nbrNode.nodeId, edgeToTgt);


        updatedNodes.put(tgtNode.nodeId, tgtNode);
        updatedNodes.put(nbrNode.nodeId, nbrNode);
        rwsetinfo.writeSet.add(tgtNode.getNodeId());
        rwsetinfo.writeSet.add(nbrNode.getNodeId());

        //System.out.println("tgt : "+tgtNode.nodeId+", nbr : "+nbrNode.nodeId+", edge : "+edgelength);
        if (edgeOldWeight > edgelength) {    //edge is updated to smaller value. So, doing an operation just like addEdge
            //System.out.println("Updated to smaller value");
            //doing for tgtNode
            {
                boolean ret = false;
                boolean localret = tgtNode.sortedNeighbors.add(new KNNData("" + nbrNode.nodeId, edgelength, nbrNode.nodeId));
                ret = ret || localret;
                //System.out.println("ret : "+ret+", "+tgtNode.sortedNeighbors.getNeighbors());
                if (ret == false) {
                    //adding to the KNN set did not update the KNN set. So, nothing to do.
                } else {    //ret is true
                    boolean localret2 = tgtNode.sortedNeighbors.addAll(nbrNode.sortedNeighbors.sortedNeighbors, nbrNode.nodeId, edgelength);
                    ret = ret || localret2;

                    if (ret) {
                        Enumeration<String> nbrEnum = tgtNode.neighbors.keys();
                        while (nbrEnum.hasMoreElements()) {
                            String nbrId = nbrEnum.nextElement();
                            if (nbrId.equals(tgtNode.nodeId))
                                continue;

                            KNNTask t1 = new KNNTask(nbrId, tgtNode.sortedNeighbors, tgtNode.nodeId, ((KNNEdge) tgtNode.neighbors.get(nbrId)).weight);
                            String location = dataStorage.getLocation(nbrId);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                    }
                }
            }

            //doing for nbrNode
            {
                boolean ret = false;
                boolean localret = nbrNode.sortedNeighbors.add(new KNNData("" + tgtNode.nodeId, edgelength, tgtNode.nodeId));
                ret = ret || localret;
                //System.out.println("ret : "+ret+", "+nbrNode.sortedNeighbors.getNeighbors());
                if (ret == false) {
                    //adding to the KNN set did not update the KNN set. So, nothing to do.
                } else {    //ret is true
                    boolean localret2 = nbrNode.sortedNeighbors.addAll(tgtNode.sortedNeighbors.sortedNeighbors, tgtNode.nodeId, edgelength);
                    ret = ret || localret2;

                    if (ret) {
                        Enumeration<String> nbrEnum = nbrNode.neighbors.keys();
                        while (nbrEnum.hasMoreElements()) {
                            String nbrId = nbrEnum.nextElement();
                            if (nbrId.equals(nbrNode.nodeId))
                                continue;

                            KNNTask t1 = new KNNTask(nbrId, nbrNode.sortedNeighbors, nbrNode.nodeId, ((KNNEdge) nbrNode.neighbors.get(nbrId)).weight);
                            String location = dataStorage.getLocation(nbrId);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                    }
                }
            }
        } else if (edgeOldWeight < edgelength) {    //edge is updated to larger value. So, doing an operation just like deleteEdge
            //System.out.println("Updated to larger value");
            //doing for tgtNode
            {
                boolean containsUpdatedEdge = false;
                HashSet<String> updatedSet = new HashSet<String>();
                for (KNNData d : tgtNode.sortedNeighbors.sortedNeighbors) {
                    if (d.nodeId.equals(nbrNode.nodeId)) {
                        containsUpdatedEdge = true;
                        break;
                    }
                }

                if (containsUpdatedEdge == false) {
                    //deleted neighbor is not present in this node's KNN set. So, nothing to do.
                } else { //containsUpdatedEdge is true
                    //building deleted set
                    Set<KNNData> toBeDeleted = new HashSet<KNNData>();
                    for (KNNData d : tgtNode.sortedNeighbors.sortedNeighbors) {
                        if (d.pathThrough == nbrNode.nodeId) {
                            updatedSet.add(d.nodeId);
                            toBeDeleted.add(d);
                        }
                    }

                    //deleting all nodes obtained above in deleted set from KNN set
                    Iterator<KNNData> iter = tgtNode.sortedNeighbors.sortedNeighbors.iterator();
                    while (iter.hasNext()) {
                        KNNData d = iter.next();
                        if (toBeDeleted.contains(d)) {
                            iter.remove();
                        }
                    }

                    //adding all neighbor edges to SS
                    Iterator it = tgtNode.neighbors.keySet().iterator();
                    while (it.hasNext()) {
                        String nbr = (String) it.next();
                        KNNEdge nbrEdge = (KNNEdge) tgtNode.neighbors.get(nbr);
                        tgtNode.sortedNeighbors.add(new KNNData(nbr, nbrEdge.weight, nbr));
                    }

                    //creating Recompute task for neighbors
                    Enumeration<String> nbrEnum = tgtNode.neighbors.keys();
                    while (nbrEnum.hasMoreElements()) {
                        String nbrId = nbrEnum.nextElement();
                        if (nbrId.equals(tgtNode.nodeId))
                            continue;

                        RecomputeTask t1 = new RecomputeTask(nbrId, tgtNode.sortedNeighbors, tgtNode.nodeId, updatedSet);
                        String location = dataStorage.getLocation(nbrId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        tasksCreated.add(t1);
                    }
                }
            }

            //doing for nbrNode
            {
                boolean containsUpdatedEdge = false;
                HashSet<String> updatedSet = new HashSet<String>();
                for (KNNData d : nbrNode.sortedNeighbors.sortedNeighbors) {
                    if (d.nodeId.equals(tgtNode.nodeId)) {
                        containsUpdatedEdge = true;
                        break;
                    }
                }

                if (containsUpdatedEdge == false) {
                    //deleted neighbor is not present in this node's KNN set. So, nothing to do.
                } else { //containsUpdatedEdge is true
                    //building deleted set
                    Set<KNNData> toBeDeleted = new HashSet<KNNData>();
                    for (KNNData d : nbrNode.sortedNeighbors.sortedNeighbors) {
                        if (d.pathThrough == tgtNode.nodeId) {
                            updatedSet.add(d.nodeId);
                            toBeDeleted.add(d);
                        }
                    }

                    //deleting all nodes obtained above in deleted set from KNN set
                    Iterator<KNNData> iter = nbrNode.sortedNeighbors.sortedNeighbors.iterator();
                    while (iter.hasNext()) {
                        KNNData d = iter.next();
                        if (toBeDeleted.contains(d)) {
                            iter.remove();
                        }
                    }

                    //adding all neighbor edges to SS
                    Iterator it = nbrNode.neighbors.keySet().iterator();
                    while (it.hasNext()) {
                        String nbr = (String) it.next();
                        KNNEdge nbrEdge = (KNNEdge) nbrNode.neighbors.get(nbr);
                        nbrNode.sortedNeighbors.add(new KNNData(nbr, nbrEdge.weight, nbr));
                    }

                    //creating Recompute task for neighbors
                    Enumeration<String> nbrEnum = nbrNode.neighbors.keys();
                    while (nbrEnum.hasMoreElements()) {
                        String nbrId = nbrEnum.nextElement();
                        if (nbrId.equals(nbrNode.nodeId))
                            continue;

                        RecomputeTask t1 = new RecomputeTask(nbrId, nbrNode.sortedNeighbors, nbrNode.nodeId, updatedSet);
                        String location = dataStorage.getLocation(nbrId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        tasksCreated.add(t1);
                    }
                }
            }
        }
    }

    public void cleanUp() {
        endTime = System.currentTimeMillis();
        long duration = endTime - startTime;

        long durationInitial = phaseEndTime[0] - startTime;
        long durationP1 = phaseEndTime[1] - phaseStartTime[1];
        long durationP2 = phaseEndTime[2] - phaseStartTime[2];
        long durationP3 = phaseEndTime[3] - phaseStartTime[3];
        long durationP4 = phaseEndTime[4] - phaseStartTime[4];

        System.out.println("*******************************");
        System.out.println("Total Execution Time taken : " + duration + " msec");
        System.out.println("Initial Computation Time taken : " + durationInitial + " msec");
        System.out.println("Phase 1 Time taken : " + durationP1 + " msec");
        System.out.println("Phase 2 Time taken : " + durationP2 + " msec");
        System.out.println("Phase 3 Time taken : " + durationP3 + " msec");
        System.out.println("Phase 4 Time taken : " + durationP4 + " msec");
        //String systemStats = "";
        try {
            //systemStats = dataStorage.getStats(true);
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            out.println("==================================");
            out.println("Total Execution Time taken : " + duration + " msec");
            out.println("Initial Computation Time taken : " + durationInitial + " msec");
            out.println("Phase 1 Time taken : " + durationP1 + " msec");
            out.println("Phase 2 Time taken : " + durationP2 + " msec");
            out.println("Phase 3 Time taken : " + durationP3 + " msec");
            out.println("Phase 4 Time taken : " + durationP4 + " msec");
            out.println("==================================");
            //out.println(systemStats);
            out.println("==================================");
            out.flush();
            lock.release();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("Worker " + workerId + ":  Finished cleanup");
        System.out.println("*****************************");
        //System.out.println(systemStats);
    }

    private void writeResult(int phaseno) {
        try {
            ConcurrentHashMap<Integer, HashtableServer> servers = dataStorage.servers;
            Collection<HashtableServer> serverCollection = servers.values();
            Vector<Node> l = new Vector<Node>();
            for (HashtableServer server : serverCollection) {
                l.addAll(server.getAllNodes());
            }
            Collections.sort(l, new NodeComparator());

            String filename = "knn_result_" + phaseno + ".txt";
            System.out.println("Writing results of KNN computation to file " + filename + " ...");

            FileOutputStream fout = new FileOutputStream(filename, false);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            //out.println("==================================");

            StringBuffer sb = new StringBuffer();
            int i = 0;
            for (Node n : l) {
                if (Integer.parseInt(n.nodeId) < 0)
                    continue;
                sb.append(n.nodeId + " :: " + k + "-NN are " + ((KNNNode) n).sortedNeighbors.getNeighbors());
                out.println(sb.toString());
                //System.out.println(sb.toString());
                sb = new StringBuffer();
                i++;
                if (i % 1000 == 0) {
                    out.flush();
                }
            }

            //out.println("==================================");
            out.flush();
            lock.release();
            out.close();
            fout.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("Worker " + workerId + ":  Finished writing to file");
        System.out.println("*****************************");

    }

    public void run() {

        startTime = System.currentTimeMillis();

        Task myTask;
        WorkStatus ws;
        int phase = 0;

        if (master) {
            try {
                Task t = new Task();
                t.phase = "STARTUP";
                t.nodeId = "STARTUP_TASK";
                workpool.broadcast(t);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        int maxphase = 1;
        if (configParams.has("KNNWorker.MAX_PHASE"))
            maxphase = configParams.getIntValue("KNNWorker.MAX_PHASE");

        boolean writeresult = false;
        if (configParams.has("KNNWorker.writeToFile"))
            writeresult = configParams.getBoolValue("KNNWorker.writeToFile");


        while (true) {
            try {
                String myURL = dataStorage.myURL;
                ws = workpool.getTask();
                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Worker " + workerId + ":  Error: Valid but Null task");
                    } else if (myTask.phase.equals("SHUTDOWN")) {
                        phase = maxphase;
                        workpool.progressStatusThread.stopPrinting = true;
                        //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" received SHUTDOWN. Closing all operations after reaching phase "+phase);
                        break;
                    } else if (myTask.phase.equals("STARTUP")) {
                        if (master)
                            System.out.println("Worker " + workerId + ":  Startup task picked up by master itself");
                        else
                            System.out.println("Worker " + workerId + ":  Startup task picked up by a non master.");
                        createTasksAllVertices(myTask);
                        System.out.println("Worker " + workerId + ":  Created initial tasks");
                    } else if (myTask.phase.equals("GENERIC")) {
                        doTask(myTask);
                        tasksCreated.clear();
                    } else if (myTask.phase.equals("TERMINATION")) {
                        //Stop task

                        System.out.println("Worker " + workerId + ":  Termination task");
                        //System.out.println("Worker "+workerId+":  Calling verify()");
                        //verify();
                        //System.exit(0);
                    } else {
                        System.out.println("Worker " + workerId + ":  Unknown phase");
                    }


                } else if (ws.status == StatusType.NO_READY_TASK) {
                    // Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
                } else if (ws.status == StatusType.WORK_TERMINATED) {

                    if (master /*&& !addtasktemp*/ && (phase == maxphase)) {
                        phaseEndTime[phase] = System.currentTimeMillis();

                        System.out.println("Master Proceeding to cleanUp after reaching phase : " + phase);
                        Task t = new Task();
                        t.phase = "SHUTDOWN";
                        t.nodeId = "STARTUP_TASK";
                        for (int i = 0; i < configParams.getIntValue("Worker.WORKER_COUNT"); i++) {
                            workpool.broadcast(t);
                        }
                        workpool.progressStatusThread.stopPrinting = true;
                        break;
                    }
                    if ((workerId == 0) && (phase < maxphase)  /*&& addtasktemp*/) {

                        if (master) {
                            phaseEndTime[phase] = System.currentTimeMillis();
                            System.out.println("Host " + dataStorage.myId + ", Worker " + workerId + " completed phase " + phase);
                            if (writeresult)
                                writeResult(phase);
                        }
                        //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" completed phase "+phase);

                        phase = phase + 1;
                        if (master) {
                            Vector<UpdateGraphTask> vtasks = createGraphUpdateTasks(dataStorage.myId, phase);
                            Vector<Task> finalVTasks = new Vector<Task>();
                            for (UpdateGraphTask vtask : vtasks) {
                                Task t = vtask;
                                if (t != null) {
                                    finalVTasks.add(t);
                                }
                            }

                            System.out.println("Master creating " + finalVTasks.size() + " new tasks...");
                            phaseStartTime[phase] = System.currentTimeMillis();
                            try {
                                workpool.addTask(finalVTasks);
                            } catch (RemoteException e) {
                                e.printStackTrace();
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (master) {
            cleanUp();
        }
    }

    public class NodeComparator implements Comparator<Node> {
        public int compare(Node n1, Node n2) {
            if (Integer.parseInt(n1.nodeId) > Integer.parseInt(n2.nodeId)) return 1;
            else return -1;
        }
    }
}
