/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.MST.MSTAsync;

import beehive.util.*;
import beehive.workpool.*;
import beehive.worker.*;

import java.util.*;
import java.io.*;
import java.nio.channels.*;
import java.lang.*;

public class MSTWorker extends Worker implements Runnable {

    boolean master = false;
    boolean debug = false;
    long ioStartTime = 0;
    int calls = 0;
    long ioTime = 0;
    float seedTaskCount = 0.5f;    //fraction of total nodes at storage server
    String myStorageURL;
    int numNodes = 0;
    public static Object mstobjLock = new Object();
    public static Vector<MSTEdge> mstEdges = new Vector<MSTEdge>();
    public Vector<String> mergedVertices;

    public MSTWorker(String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, String myStorageURL, int numOfNodes) throws java.rmi.RemoteException {
        super(configFile, wpool, 0, datastorage, master);
        this.master = master;
        this.dataStorage = datastorage;
//	this.workerId = workerId;
        this.myStorageURL = myStorageURL;
        this.numNodes = numOfNodes;
    }

    public MSTNode getCoreNode(MSTNode v) throws java.rmi.RemoteException {

        if (v == null) {
            System.out.println("Worker " + workerId + ": Null node passed in getCoreNode");
            System.exit(1);
        }
        Vector<Node> nodeList = new Vector<Node>();
        while (v.coreId != v.nodeId) {
            nodeList.add(new MSTNode((MSTNode) v));
            // System.out.println("Worker "+workerId+":  following cluster pointers from "+v.nodeId+" to "+v.coreId);
            v = (MSTNode) dataStorage.getNode(v.coreId);

        }
        for (Node n : nodeList) {
            MSTNode mn = (MSTNode) n;
            mn.coreId = v.coreId;
        }

        AsyncUpdateThread asyncUpdateThread = new AsyncUpdateThread(nodeList);
        asyncUpdateThread.start();

        return (new MSTNode(v)); //return a copy of the node

    }

    public String getCoreId(String id) throws java.rmi.RemoteException {
        MSTNode v = (MSTNode) dataStorage.getNode(id);
        if (v == null) {
            System.out.println("Worker " + workerId + ": Null node from storager in getCoreId for id:" + id);
            System.exit(1);
        }

        MSTNode coreNode = getCoreNode(v);
        return coreNode.coreId;
    }

    public void addMSTEdge(MSTEdge edge) {
        synchronized (mstobjLock) {
            mstEdges.add(edge);
        }

    }

    public void mergeCluster(MSTNode u, MSTNode v) throws java.rmi.RemoteException {

        v = getCoreNode(v);
//	System.out.println("Worker "+workerId+":  Core node of closest node"+v.printNode());
        //System.out.println("Worker "+workerId+":  U - Node before merging: "+u.printNodeEdges());
        //System.out.println("Worker "+workerId+":  V - Node before merging: "+v.printNodeEdges());

	/* Vinit (Jun 24, 2013): 
     add all the edges of v to my edge list except those pointing to nodes in my cluster, dont worry about the duplicates
	 This will avoid following the cluster pointers 
	 first get rid of all the edges in v's edge list that are pointing to me   */

        for (String key : u.clusterKeys) {
            if (v.neighbors.containsKey(key)) {
                v.neighbors.remove(key);
            }
        }

//        Iterator<String> vNbrKeys = v.neighbors.keySet().iterator();
//        while (vNbrKeys.hasNext()) {
//            String nbrKey = vNbrKeys.next();
//            if (u.clusterKeys.contains(nbrKey)) {
//                v.neighbors.remove(nbrKey);
////                vNbrKeys.remove();
//            }
//        }


	/*
     Now go through all the edges or v adding them to my edges.
		if I have an edge for the same neighbour take the one with smaller weight */
        Enumeration otherEdges = v.neighbors.keys();
        //System.out.println("Worker "+workerId+": Iterating over V edges");
        for (String key : v.neighbors.keySet()) {
            MSTEdge otherEdge = (MSTEdge) v.neighbors.get(key);
            MSTEdge myEdge = (MSTEdge) u.neighbors.get(key);
            if (myEdge == null || myEdge.weight > otherEdge.weight) {
                u.neighbors.put(key, otherEdge);
            }
        }


	/* 	
    Enumeration otherEdges = v.neighbors.keys();
    //System.out.println("Worker "+workerId+": Iterating over V edges");
    while(otherEdges.hasMoreElements()){
        String key = (String)otherEdges.nextElement();
        MSTEdge otherEdge = (MSTEdge) v.neighbors.get(key);
        int coreId = getCoreId(Integer.parseInt(key));
        MSTEdge myEdge = (MSTEdge)u.neighbors.get(Integer.toString(coreId));
        if(coreId != u.coreId && coreId != v.coreId) {
            if(myEdge==null||myEdge.weight>otherEdge.weight){
                otherEdge.neighborId = coreId; //overwriting the edge neighbor with coreid
                //System.out.println("Worker "+workerId+": OverWriting edge: key:"+coreId+" Edge: "+otherEdge.toString());
                u.neighbors.put(Integer.toString(coreId),otherEdge); //add the edge if u dont have it or you dont have the best weight
            }
        }
    }
	*/

        //System.out.println("Worker "+workerId+": Iterating over U edges");

	/* Removeing the internal edges */
        for (String key : v.clusterKeys) {
            if (u.neighbors.containsKey(key)) {
                u.neighbors.remove(key);
            }
        }


//        Iterator<String> uNbrKeys = u.neighbors.keySet().iterator();
//        while (uNbrKeys.hasNext()) {
//            String nbrKey = uNbrKeys.next();
//            if (v.clusterKeys.contains(nbrKey)) {
//                u.neighbors.remove(nbrKey);
////                vNbrKeys.remove();
//            }
//        }

    /*
    //	Vinit (Jun 24 2013): )This code is commented because it requires following the cluster pointers
	Enumeration myEdges = u.neighbors.keys();
	while(myEdges.hasMoreElements()){
		String key = (String) myEdges.nextElement();
		MSTEdge myEdge = (MSTEdge)u.neighbors.get(key);
	 	if(getCoreId(Integer.parseInt(key))==v.coreId) {
			//System.out.println("Worker "+workerId+": Removing edge: "+myEdge.toString());
			u.neighbors.remove(key); // removing all those edges which go to the 2nd cluster v.
		}
	}		
	*/
        //finally update core id of v to u
        this.mergedVertices = v.clusterKeys;
        u.clusterKeys.addAll(v.clusterKeys);

        //System.out.println("Worker "+workerId+":  Merging if transaction commits: Updating cored id of "+v.coreId+" to "+u.coreId);
        //System.out.println("Worker "+workerId+": Updated node: "+u.printNodeEdges());
        v.coreId = u.coreId;
        //System.out.print("Worker "+workerId+": adding u.MSTCost+"+u.MSTCost+ " to v.MSTCost="+v.MSTCost);
        u.MSTCost += v.MSTCost;
        //System.out.println("Worker "+workerId+": total cost so far:"+u.MSTCost);
        rwsetinfo.writeSet.add(u.nodeId);
        rwsetinfo.writeSet.add(v.nodeId);
        updatedNodes.put(v.nodeId, v);
        updatedNodes.put(u.nodeId, u);

    }

    public void compute(Task t, Node node) {
        try {
            MSTNode u = new MSTNode((MSTNode) node);    //VERIFY
            MSTEdge pickededge = null;
            if (!u.coreId.equals(u.nodeId)) {
                //System.out.println("Worker "+workerId+":  This task node id: "+u.nodeId+" core id: "+u.coreId+" has nothing to do...Returning");
                return;
            } else {
                int closestDist = Integer.MAX_VALUE;
                String closestId = "-1";
                String sourceId = "-1";
                Enumeration keys = u.neighbors.keys();
                while (keys.hasMoreElements()) {
                    String key = (String) keys.nextElement();
                    MSTEdge edge = (MSTEdge) u.neighbors.get(key);
                    if (edge.weight < closestDist) {
                        pickededge = edge;
                        closestDist = edge.weight;
                        closestId = edge.nbrId;
                        sourceId = edge.sourceId; // Although we dont need this at this point
                    }

                }
                // System.out.println("Worker "+workerId+":  Current node: "+u.printNode());
                //	 System.out.println( "Closest Vrtex " + closestId + " dist:" + closestDist );
                if (!closestId.equals("-1")) {
                    //		System.out.println(" Edge picked: "+pickededge.toString());
                    u.MSTCost += closestDist;
                    Node cNode = dataStorage.getNode(closestId); // CHECK
                    if (cNode == null) {
                        System.out.println("Null node from storage for node id:" + closestId);
                        System.exit(1);
                    }
                    MSTNode closestNode = new MSTNode((MSTNode) cNode); //make a copy of the closest node

                    //System.out.println("Worker "+workerId+":  Merging u.node: "+u.nodeId+" u.coreId: "+u.coreId+" and closestId: "+closestId);
                    mergeCluster(u, closestNode);
                    // Vinit (Jun 21, 2013): the worker will continue executing the task to merge further rather than depositing a new task for the same.
                    // tasksCreated.add(t);

                } else {
                    // A connected component is formed
                    System.out.println("A connected component is formed");
                    SuperNode sn = (SuperNode) dataStorage.getNode(Integer.toString(numNodes));
                    sn.clusterIds.add(u.coreId);
                    sn.TS += 1;
                    dataStorage.putNode(Integer.toString(numNodes), sn);
                    System.out.println("Worker " + workerId + ":  ******************\n*****Cluster formed core Id: " + u.printNode() + " Cost: " + u.MSTCost + "**********\n*********************************");
                    return;
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }


    }

    public void doTask(Task tsk) throws Exception {
        workerStatus.status = WorkerStatus.TASK_PICKED;
        workerStatus.taskId = tsk.getNodeId();
        boolean commit = false;
        if (tsk == null)
            System.out.println("Worker " + workerId + ":  doTask(): null task");

//        Node origNode = dataStorage.getNode(tsk.getNodeId());
        while (true) {
            TransactionInfo txnInfo = beginTransaction();
            tsk.txnInfo = txnInfo;
            Node origNode = dataStorage.getNode(tsk.getNodeId());
            try {
                updatedNodes.clear();
                rwsetinfo.readSet.clear();
                rwsetinfo.writeSet.clear();
                rwsetinfo.unusedSet.clear();
                tasksCreated.clear();
                workerStatus.status = WorkerStatus.COMPUTE;

                compute(tsk, origNode);

                if (rwsetinfo.writeSet.isEmpty() && rwsetinfo.readSet.isEmpty()) {
//                    System.out.println("Worker " + workerId + ":  transaction has no read/writes sets,  nothing to do ...so terminated");
                    commit = true;
                    workpool.removeTask(tsk);
                    return; // nothing to do
                }

                commit = validateTask(tsk);

                if (commit) {
                    Task t = new Task(tsk.nodeId);
                    t.phase = "GENERIC";
                    workpool.addTask(t);
                    break;
                }


////                System.out.println("Worker " + workerId + ":  validating");
//                workerStatus.status = WorkerStatus.TASK_VALIDATION;
//                ValidateResponse res = workpool.validate(tsk, rwsetinfo);
//                commit = res.commit;
//
//                if (commit) {
////                    System.out.println("Worker " + workerId + ":  commit: true");
//                    tsk.txnInfo.commitTS = res.commitTS;
//                    commitTransaction(tsk.txnInfo);
//                    workpool.reportTransactionCompletion(tsk.txnInfo.commitTS, tasksCreated);
//                    // Delay untill STS is advanced
//                    long curSTS = workpool.getSTS();
//                    while (curSTS < res.commitTS) {
//                        try {
//                            Thread.sleep(1);
//                            curSTS = workpool.getSTS();
//                        } catch (InterruptedException ie) {
//                        }
//                    }
//
////                    System.out.println("Starting new transaction on node: " + tsk.nodeId);
//                    txnInfo = beginTransaction();
//                    tsk.txnInfo = txnInfo;
//                    commit = false;
////                    origNode = updatedNodes.get(origNode.nodeId);
//                } else {
////                    System.out.println("Transaction aborted on node: " + tsk.nodeId);
//                    try {
//                        Thread.sleep(5);
//                    } catch (Exception e) {
//                    }
//                    txnInfo = beginTransaction();
//                    tsk.txnInfo = txnInfo;
//                    origNode = dataStorage.getNode(tsk.getNodeId());
//                    commit = false;
//                }
                //update STS is with latest STS
            } catch (Exception e) {
                e.printStackTrace();
            }
        } //Transaction committed
    }

    public void updateCoreMergedVertices(String coreId) {

        AsyncUpdateMergedVerticesThread mergerThread = new AsyncUpdateMergedVerticesThread(this.mergedVertices, coreId);
        mergerThread.start();

    }

    public void cleanUp() {
        // Final Computation of the flow out of source node.
        try {
            SuperNode n = (SuperNode) dataStorage.getNode(Integer.toString(numNodes));
            Vector<String> clusterIds = n.getClusterIds();
            int finalCost = 0;
            System.out.println("*****************************");
            System.out.println("***********RESULT*************");
            System.out.println("Cluster Ids: " + clusterIds);
            long currentTime = System.currentTimeMillis();
            long timeTaken = currentTime - ioStartTime;
            System.out.println("Time taken: " + timeTaken + " msec");
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            for (String key : clusterIds) {
                MSTNode node = (MSTNode) dataStorage.getNode(key);
                out.println("Component Cost: " + node.MSTCost);
                System.out.println("Worker " + workerId + ":  Component Cost:" + node.MSTCost);
                finalCost += node.MSTCost;
            }
            out.println("Total MSTPhase cost: " + finalCost);
            out.println("Time taken: " + timeTaken + " msec");

            out.println("========================================================================");
            out.flush();
            lock.release();
            out.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        System.out.println("Worker " + workerId + ":  Finished cleanup");

        System.out.println("*****************************");
    }

    public void createSeedTasks(Task tsk) throws Exception {


        Vector<String> myLocalKeys = dataStorage.getAllKeys(myStorageURL);
        int taskLimit = (int) (myLocalKeys.size() * seedTaskCount);

        System.out.println("Worker " + workerId + ":  Startup Task: creating " + taskLimit + " seed tasks");
        Vector<Task> taskVector = new Vector<Task>();
        for (int i = 0; i < taskLimit; i++) {
            String key = myLocalKeys.get(i);
            if (key.equals(Integer.toString(numNodes)))
                continue;
            Task t = new Task(key);
            t.phase = "GENERIC";
            taskVector.add(t);

        }
        workpool.addTask(taskVector);
        workpool.removeTask(tsk);
    }


    public void run() {
        ioStartTime = System.currentTimeMillis();

        Task myTask;
        WorkStatus ws;


        if (master) {
            try {
                SuperNode sn = new SuperNode(Integer.toString(numNodes));
                dataStorage.putNode(sn.nodeId, sn);
                sn = (SuperNode) dataStorage.getNode(Integer.toString(numNodes));
                if (sn == null) {
                    System.out.println("Null super node");
                    System.exit(1);
                }


                Task t = new Task();
                t.phase = "STARTUP";
                t.nodeId = "STARTUP_TASK";
                workpool.broadcast(t);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }


        while (true) {
            workerStatus.status = WorkerStatus.IDLE;
            try {
                //	ws = workpool.getTask();
                String myURL = dataStorage.myURL;
                ws = workpool.getTask();
                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Worker " + workerId + ":  Error: Valid but Null task");
                        continue;
                    }
                    if (myTask.phase.equals("STARTUP")) {
                        if (master)
                            System.out.println("Worker " + workerId + ":  Startup task picked up by master itself");
                        else
                            System.out.println("Worker " + workerId + ":  Startup task picked up by a non master.");
                        createSeedTasks(myTask);
                        System.out.println("Worker " + workerId + ":  Created initial tasks");
                    } else if (myTask.phase.equals("GENERIC")) {
                        // System.out.println("Worker "+workerId+":  Performing task for :" +myTask.getNodeId());
                        doTask(myTask);
                        tasksCreated.clear();
                    } else if (myTask.phase.equals("TERMINATION")) {
                        //Stop task
                        System.out.println("Worker " + workerId + ":  Termination task");
                        System.exit(0);
                    } else {

                        System.out.println("Worker " + workerId + ":  Unknown phase");
                    }


                } else if (ws.status == StatusType.NO_READY_TASK) {
                    // System.out.println("Worker "+workerId+":  Going to sleep 1 sec");
                    // Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
                } else if (ws.status == StatusType.WORK_TERMINATED) {
                    if (master) {
                        System.out.println("Worker " + workerId + ":  Master crating TERMINTATE TASK - Worker thread proceeding to ncleanUp");


                        cleanUp(); //do clean up before broadcasting termination task
                        /**FIX***/
                        Task t = new Task();
                        t.nodeId = "TERMINATE_TASK";
                        t.phase = "TERMINATION";
                        workpool.broadcast(t);
                        System.exit(0);

                    }
                    //System.out.println("Worker "+workerId+":  Worker thread proceeding to cleanUp");
                    //break;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    class AsyncUpdateThread extends Thread {
        Vector<Node> nodeList;

        public AsyncUpdateThread(Vector<Node> nodeList) {
            this.nodeList = nodeList;
        }

        public void run() {
            try {
                dataStorage.putNodesParallel(nodeList);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

    }


    class AsyncUpdateMergedVerticesThread extends Thread {
        public Vector<String> mergedVertices;

        public static final int BATCHSIZE = 100;
        public String coreId;

        public AsyncUpdateMergedVerticesThread(Vector<String> mergedVertices, String newCoreId) {
            this.mergedVertices = mergedVertices;
            this.coreId = newCoreId;
        }

        public void run() {

            Vector<Node> currBatchNodes = new Vector<Node>();
            Vector<String> currBatchVertices = new Vector<String>();

            try {

                while (mergedVertices.size() > 0) {
                    for (int i = 0; i < BATCHSIZE; i++) {
                        if (mergedVertices.size() > 0) {
                            String vertex = mergedVertices.remove(0);
                            currBatchVertices.add(vertex);
                        } else
                            break;
                    }

                    Vector<NodeData> execBatch = new Vector<>(currBatchVertices.size());
                    for (String execKey : currBatchVertices) {
                        NodeData toExec = new NodeData(execKey, (byte) 1);
                        toExec.addProperty("setCoreId", this.coreId);
                        execBatch.add(toExec);
                    }

                    dataStorage.execParallel(execBatch);

//
//                    currBatchNodes = dataStorage.getNodesParallel(currBatchVertices);
//                    for (Node n : currBatchNodes) {
//                        MSTNode mstnode = (MSTNode) n;
//                        mstnode.coreId = this.coreId;
//                    }
//                    dataStorage.putNodesParallel(currBatchNodes);
//                    currBatchNodes.clear();
                    currBatchVertices.clear();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

    }

}//end of class
