/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.MST.MSTOptimized;

import java.util.*;
import java.io.*;

import beehive.util.*;

public class InputFileReader {

    StorageSystem storage;
    String fileName;
    public Vector<String> totalNodes;
    public Vector<String> nodesCreated;
    Hashtable<String, Node> table;
    boolean debug = false;

    public InputFileReader(String filename, int numNodes, StorageSystem storage) {
        this.fileName = filename;
        this.storage = storage;
        table = new Hashtable<String, Node>();
        for (int i = 0; i < numNodes; i++) {
            MSTNode n = new MSTNode(Integer.toString(i));
            String nodeKey = Integer.toString(i);
            table.put(nodeKey, n);
        }
    }

    public void readAndLoadData() throws IOException {
        totalNodes = new Vector<String>();
        nodesCreated = new Vector<String>();

        BufferedReader br = new BufferedReader(new FileReader(fileName));
        String line = null;


        while ((line = br.readLine()) != null) {
            StringTokenizer tok = new StringTokenizer(line, " ");
            int currentNodeId = Integer.parseInt(tok.nextToken());
            String currentNodeKey = Integer.toString(currentNodeId);
            nodesCreated.add(currentNodeKey);
            if (!totalNodes.contains(currentNodeKey)) {
                totalNodes.add(currentNodeKey);
            }

            MSTNode currentNode = new MSTNode(currentNodeKey);
            if (table.containsKey(currentNodeKey)) {
                currentNode = (MSTNode) table.get(currentNodeKey);

            }
            //update the neighbor list for this node;
            String nbrList = "";
            if (tok.hasMoreTokens()) {
                nbrList = tok.nextToken();
            } else
                continue;

            StringTokenizer tok2 = new StringTokenizer(nbrList, ",");
            // loop over the list of neighbours
            while (tok2.hasMoreTokens()) {
                String neighborInfo = tok2.nextToken();
                StringTokenizer tok3 = new StringTokenizer(neighborInfo, ":");
                int nbrId = Integer.parseInt(tok3.nextToken());  /* Neighbor Id */

                String nbrKey = Integer.toString(nbrId);
                if (!totalNodes.contains(currentNodeKey)) {
                    totalNodes.add(currentNodeKey);
                }


                int weight = Integer.parseInt(tok3.nextToken());
                MSTEdge toNbr = new MSTEdge(Integer.toString(currentNodeId), Integer.toString(nbrId));
                toNbr.weight = weight;
                currentNode.addNeighbor(toNbr.nbrId, toNbr);

				/* If the neighbor alread has an edge for the current node */
                MSTNode neighborNode = (MSTNode) table.get(nbrKey);

                MSTEdge fromNbr = (MSTEdge) neighborNode.neighbors.get(currentNodeKey);
                if (fromNbr == null) {
                    if (debug)
                        System.out.println("No edge was found from" + neighborNode.nodeId + " to " + currentNodeKey);
                    fromNbr = new MSTEdge(neighborNode.nodeId, Integer.toString(currentNodeId));
                    fromNbr.weight = weight;
                    neighborNode.addNeighbor(fromNbr.nbrId, fromNbr);
                    table.put(nbrKey, neighborNode);
                }

            }
            table.put(currentNodeKey, currentNode);
            if (debug)
                System.out.println("Added node: " + currentNodeKey);
        }

        System.out.println("Putting nodes into the storage system...");
        Enumeration<String> enumKeys = table.keys();
        while (enumKeys.hasMoreElements()) {
            String key = enumKeys.nextElement();
            MSTNode node = (MSTNode) table.get(key);
            storage.putNode(key, node);
        }
    }


    public void printTrace() {
        for (String nodeId : nodesCreated) {
            try {
                MSTNode node = (MSTNode) storage.getNode(nodeId);
//                System.out.println("Node " + node.printNodeEdges());
            } catch (Exception e) {
                System.out.println("Exception encountered while fetching node " + nodeId);
                e.printStackTrace();
            }
        }
    }

}

