package TestProgs.PageRankV2;

import beehive.util.*;
import beehive.server.*;
import beehive.workpool.*;
import beehive.worker.*;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.rmi.registry.*;
import java.util.*;
import java.io.*;
import java.nio.channels.*;


public class PageRankWorker extends Worker implements Runnable {

	boolean debug = false;
	long ioStartTime = 0;
	int calls = 0;	 
	long ioTime = 0;
	int numNodes ;
	String myStorageURL;
	int iteration = 0;
	public PageRankWorker (String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, int workerId, int numNodes, String myStorageURL )  throws java.rmi.RemoteException {

		super(configFile, wpool, workerId, datastorage, master);
		this.numNodes = numNodes;
		this.myStorageURL = myStorageURL;
		this.ioStartTime = System.currentTimeMillis();
	}

	public Vector<Node> getBackNeighbors(Node u) throws Exception {

		/* get all neighbor nodes from the gloal hashtable */
		PRNode pr = (PRNode)u;
		Vector<Node> neighborNodes = new Vector<Node>();
		Enumeration edges = pr.backNeighbors.keys();
		Vector<String>  nbrKeys  = new Vector();;
		while ( edges.hasMoreElements() ){
			String key = (String) edges.nextElement();
			nbrKeys.add( key );
		}

		neighborNodes.addAll(dataStorage.getNodesParallel(nbrKeys));
		return neighborNodes;
	}

	public void calculateRank(PRNode u,int phase, Vector<Node> backNeighborNodeList) throws Exception {

		// sanity check
		double total=0;

		double tolerance;
		double tau = u.tau;
		double rank = 0;

		for(Node n: backNeighborNodeList) {
			PRNode prNode = (PRNode) n;
			int degree = prNode.neighbors.size();
			double neighborRank;
			if(phase==0)
				neighborRank = prNode.rank[1];
			else
				neighborRank = prNode.rank[0];
			//System.out.println(" Neighbor Node id: "+prNode.nodeId+" Rank 0: "+prNode.rank[0]+" Rank 1: "+prNode.rank[1]);		
			//System.out.println("Neighbor rank"+neighborRank);
			double frac = (neighborRank/(double)degree);  
			rank += frac;
		}


		rank  = (0.85*rank) + (0.15*tau);

		//updating our rank	
		u.rank[phase]=rank;

		//System.out.println("Computed Curr rank: "+u.rank[phase]+" Prev rank"+u.rank[(phase+1)%2]);

	}

	@Override
	public void doTask(Task tsk) throws Exception{

		Node origNode = dataStorage.getNode(tsk.getNodeId());
		Vector<Node> backNeighbors = getBackNeighbors(origNode);
		compute(tsk, origNode,	backNeighbors);
		Vector<Node> updatedNodeList = new Vector<Node>();
		updatedNodeList.add(origNode);
		dataStorage.putNodesParallel(updatedNodeList);
		workpool.removeTask(tsk);
	}


	public void  compute(Task t, Node node, Vector<Node> backNeighborNodeList) { 
		try{
			//dont need a copy 
			PRNode u = (PRNode)node; 	
			PRTask prt = (PRTask)t;
			int itrCount = prt.itrCount;
			//System.out.println("Computing rank for node: "+ u.printNode()+" Iteration :"+itrCount+ " Phase: "+(itrCount%2));

			calculateRank(u, itrCount%2, backNeighborNodeList); 

		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}


	}

	public void createTasksAllVertices(Task tsk) throws Exception{

		PRTask myTask = (PRTask)tsk;
		int itrCount = myTask.itrCount;
		Vector<String> myLocalKeys = dataStorage.getAllKeys(myStorageURL);	
		//	System.out.println("Startup Task: creating tasks for the following keys "+myLocalKeys.toString());
		for (String key: myLocalKeys) {
			PRTask t = new PRTask(itrCount,key);
			t.phase = "GENERIC";
			workpool.addTask(t);
		}
		workpool.removeTask(tsk);
	}

	public void cleanUp()  {
		// Final Computation of the flow out of source node.
		try {
			FileOutputStream fout = new FileOutputStream("result.txt",true);
			FileLock lock = fout.getChannel().lock();
			PrintWriter out = new PrintWriter(fout);
			long ioEnd = System.currentTimeMillis();
			long duration = ioEnd - ioStartTime;
			System.out.println("Time taken: "+duration+" msec");
			out.println("Time taken: "+duration+" msec");

			System.out.println("Printing ranks... SUPPRESSED \n");

			out.flush();
			lock.release();
			out.close();
		} catch(Exception e) {
			e.printStackTrace();
		}

		System.out.println("Finished cleanup");
	}


	public void printStats() {
		try{
			FileOutputStream fout = new FileOutputStream("result.txt",true);
			FileLock lock = fout.getChannel().lock();
			PrintWriter out = new PrintWriter(fout);
			out.println("Server:"+dataStorage.myURL);
			out.println(dataStorage.getStats());
			out.println(workpool.getStats());
			out.flush();
			lock.release();
			out.close();
		}catch (Exception e) {
			e.printStackTrace();
		}
	}


	public void run ()  {

		Task  myTask;
		WorkStatus ws;

		if(master) {
			try{
				PRTask t = new PRTask(iteration);
				t.phase = "STARTUP";
				t.nodeId = "STARTUP_TASK";
				workpool.broadcast(t);
			} catch(Exception e) {
				e.printStackTrace();	
			}	
		}


		while (true)  {
			try {
				String myURL = dataStorage.myURL;
				ws = workpool.getTask();

				if (ws.status ==  StatusType.VALID_TASK) {
					myTask = ws.task;

					if (myTask == null) {
						System.out.println("Error: Valid but Null task");
						continue;
					}

					if(myTask.phase.equals("STARTUP")){
						createTasksAllVertices(myTask);
					}
					else if(myTask.phase.equals("GENERIC")){
						// System.out.println("Performing task for :" +myTask.getNodeId());
						doTask( myTask );
						//tasksCreated.clear();	
					}
					else if (myTask.phase.equals("TERMINATION")) {
						//Stop task
						System.out.println("Termination task");
						// printStats();
						System.exit(0);
					}
					else{

						System.out.println("Unknown phase");
					}


				}
				else if (ws.status == StatusType.NO_READY_TASK) {
					// System.out.println("Going to sleep 1 sec");
					// Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
				}
				else if (ws.status == StatusType.WORK_TERMINATED) {

					if(master){
						if(iteration==50) {
							System.out.println("Master crating TERMINTATE TASK - Worker thread proceeding to cleanUp");	
							cleanUp(); //do clean up before broadcasting termination task
							PRTask t = new PRTask(iteration);
							t.nodeId = "TERMINATE_TASK";
							t.phase = "TERMINATION";
							//workpool.broadcast(t);
							// (Vinit Jun 28 commented out due to a bug while sending broadcast task
							System.exit(0);
						}
						else {
							iteration++;
							PRTask t = new PRTask(iteration);
							System.out.println("Master creating STARTUP task - iteration count:"+iteration);
							t.nodeId="STARTUP_TASK";
							t.phase = "STARTUP";
							workpool.broadcast(t);
							//create startup task for this iteratiion
						}
					}
					else {

					}			


				}				
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}


}