/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.ShortestPathIncremental;

import beehive.util.*;
import beehive.server.*;
import beehive.workpool.*;
import beehive.worker.*;
import org.slf4j.*;

import java.rmi.RemoteException;
import java.util.*;
import java.io.*;
import java.nio.channels.*;
import java.lang.*;
import java.util.concurrent.ConcurrentHashMap;

public class SPWorker extends Worker implements Runnable {

    private long jobStartTime = 0;
    private long phaseStartTime[] = new long[6];
    private long phaseEndTime[] = new long[6];

    private String myStorageURL;
    private int numNodes = 0;
    private String srcNode = null;
    private ConcurrentHashMap<String, NeighborInfo> neighborDistances;

    public SPWorker(String configFile, LocalWorkpoolImpl wpool, boolean master, StorageSystem datastorage, int workerId, String myStorageURL, String srcNode, int numOfNodes, ConcurrentHashMap<String, NeighborInfo> neighborDistances) throws RemoteException {
        super(configFile, wpool, workerId, datastorage, master);
        this.myStorageURL = myStorageURL;
        this.numNodes = numOfNodes;
        this.srcNode = srcNode;
        this.neighborDistances = neighborDistances;
    }

    public void run() {

        Task myTask;
        WorkStatus ws;
        int phase = 0;
        jobStartTime = System.currentTimeMillis();
        if (master) {
            try {
                //System.out.println("*********** Starting ********** PreFlowPushWorker: run: calling getNode for nodeId "+ srcId );
                SPNode u = (SPNode) dataStorage.getNode(srcNode);
                u.distFromSrc = 0;
                u.predecessorId = "0";
                dataStorage.putNode(u.nodeId, u);

                initialPush(u);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        int maxphase = configParams.getIntValue("SPWorker.MAX_PHASE");
        while (true) {
            try {
                ws = workpool.getTask();
                if (ws.status == StatusType.VALID_TASK) {
                    myTask = ws.task;

                    if (myTask == null) {
                        System.out.println("Error: Valid but Null task");
                    } else if (myTask.phase.equals("SHUTDOWN")) {
                        phase = maxphase;
                        workpool.progressStatusThread.stopPrinting = true;
                        //System.out.println("Host "+dataStorage.myId+", Worker "+workerId+" received SHUTDOWN. Closing all operations after reaching phase "+phase);
                        break;
                    } else {
                        doTask(myTask);
                    }
                } else if (ws.status == StatusType.NO_READY_TASK) {
                    // Thread.sleep(configParams.getIntValue("Worker.THREADSLEEPTIME"));
                } else if (ws.status == StatusType.WORK_TERMINATED) {
                    if (master && (phase == maxphase)) {
                        phaseEndTime[phase] = System.currentTimeMillis();

                        System.out.println("Master Proceeding to cleanUp after reaching phase : " + phase);
                        Task t = new Task();
                        t.phase = "SHUTDOWN";
                        t.nodeId = "STARTUP_TASK";
                        for (int i = 0; i < configParams.getIntValue("Worker.WORKER_COUNT"); i++) {
                            workpool.broadcast(t);
                        }
                        workpool.progressStatusThread.stopPrinting = true;
                        break;
                    }
                    if ((workerId == 0) && (phase < maxphase)) {
                        if (master) {
                            phaseEndTime[phase] = System.currentTimeMillis();
                        }
                        phase = phase + 1;
                        if (master) {
                            Vector<UpdateGraphTask> vtasks = createGraphUpdateTasks(dataStorage.myId, phase);
                            Vector<Task> finalVTasks = new Vector<>();
                            for (UpdateGraphTask vtask : vtasks) {
                                if (vtask != null) {
                                    finalVTasks.add(vtask);
                                }
                            }

                            System.out.println("Master creating " + finalVTasks.size() + " new tasks...");
                            phaseStartTime[phase] = System.currentTimeMillis();
                            try {
                                workpool.addTask(finalVTasks);
                            } catch (RemoteException e) {
                                e.printStackTrace();
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (master) {
            cleanUp(phase);
        }
    }

    private void initialPush(SPNode srcnode) throws RemoteException {
        System.out.println("Master creating initial tasks ...");
        createSPTasksForNbrs(srcnode);
        workpool.addTask(new Vector<>(tasksCreated));
    }

    public Node compute(Task t, Node node) {
        try {
            SPTask spTask = (SPTask) t;
            SPNode target = new SPNode((SPNode) node);    //you need a copy of orig node

            // Cache the neighbor's distance
            NeighborInfo oldInfo = neighborDistances.get(spTask.senderId);
            if ((oldInfo == null) || (oldInfo.nbrTS < spTask.TS)) {
                neighborDistances.put(spTask.senderId, new NeighborInfo(spTask.distFromSrc, spTask.TS));
            }

            if ((spTask.distFromSrc != Integer.MAX_VALUE) && (target.distFromSrc > spTask.distFromSrc + spTask.edgeWeight)) {
                // Distance to the target node can be improved
                target.distFromSrc = spTask.distFromSrc + spTask.edgeWeight;
                target.predecessorId = spTask.senderId;
                updatedNodes.put(target.nodeId, target);

                createSPTasksForNbrs(target);
                return target;
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return null;
    }

    private Node recompute(Task T, Node node) {
        SPNode tgtNode = null;
        try {
            RecomputeSPTask t = (RecomputeSPTask) T;
            tgtNode = new SPNode((SPNode) node);

            SPEdge nbrEdge = (SPEdge) tgtNode.neighbors.get(t.senderId);
            if (nbrEdge == null) {    //this task is from a non-existing neighbor edge (deleted neighbor). So, do nothing.
                return null;
            }

            //from the task, record the neighbor's distance. Might be used later
            NeighborInfo oldInfo = neighborDistances.get(t.senderId);
            if ((oldInfo == null) || (oldInfo.nbrTS < t.TS)) {
                neighborDistances.put(t.senderId, new NeighborInfo(t.distFromSrc, t.TS));
            }

            if ((!tgtNode.predecessorId.equals(t.senderId)) && (tgtNode.distFromSrc != Integer.MAX_VALUE) && ((tgtNode.distFromSrc + nbrEdge.weight) < t.distFromSrc)) {
                SPTask t1 = new SPTask(t.senderId, tgtNode.distFromSrc, nbrEdge.weight, tgtNode.nodeId, tgtNode.TS);
                t1.targetLocation = dataStorage.getLocation(tgtNode.nodeId);
                t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                //System.out.println("addEdge :: Adding task to tasksCreated for : "+t1.nodeId);
                tasksCreated.add(t1);
            } else if (tgtNode.predecessorId.equals(t.senderId)) {
                if (t.distFromSrc != Integer.MAX_VALUE) {
                    tgtNode.distFromSrc = t.distFromSrc + t.edgeWeight;
                    tgtNode.predecessorId = t.senderId;
                } else {
                    tgtNode.distFromSrc = Integer.MAX_VALUE;
                    tgtNode.predecessorId = "-1";
                }

                updatedNodes.put(tgtNode.nodeId, tgtNode);

                for (String nbr : tgtNode.neighbors.keySet()) {
                    if (nbr.equals(t.senderId))
                        continue;
                    RecomputeSPTask t1 = new RecomputeSPTask(nbr, tgtNode.distFromSrc, ((SPEdge) tgtNode.neighbors.get(nbr)).weight, tgtNode.nodeId, tgtNode.TS);
                    String location = dataStorage.getLocation(nbr);
                    t1.targetLocation = location;
                    t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                    tasksCreated.add(t1);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return tgtNode;
    }

    private Node updateGraph(Task T, Node node) {
        SPNode tgtNode = null;
        if (node != null)
            tgtNode = new SPNode((SPNode) node);    //you need a copy of orig node

        try {
            UpdateGraphTask t = (UpdateGraphTask) T;

            switch (t.opname) {
                case SPTaskConstants.addEdge: {
                    String nbrid = (String) t.params.get(0);
                    //System.out.println("inside addEdge::"+" nbrid : "+nbrid);
                    Integer edge = (Integer) t.params.get(1);
                    int edgelength = edge.intValue();
                    SPNode nbrNode = new SPNode((SPNode) dataStorage.getNode(nbrid));

                    SPEdge edgeToNbr = new SPEdge(edgelength);
                    nbrNode.addNeighbor(tgtNode.nodeId, edgeToNbr);
                    SPEdge edgeToTgt = new SPEdge(edgelength);
                    tgtNode.addNeighbor(nbrNode.nodeId, edgeToTgt);

                    updatedNodes.put(tgtNode.nodeId, tgtNode);
                    updatedNodes.put(nbrNode.nodeId, nbrNode);

                    if ((nbrNode.distFromSrc != Integer.MAX_VALUE) && (nbrNode.distFromSrc + edgelength < tgtNode.distFromSrc)) {
                        //create a new SPTask for tgtNode
                        SPTask t1 = new SPTask(tgtNode.nodeId, nbrNode.distFromSrc, edgelength, nbrNode.nodeId, nbrNode.TS);
                        String location = dataStorage.getLocation(tgtNode.nodeId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        //System.out.println("addEdge :: Adding task to tasksCreated for : "+t1.nodeId);
                        tasksCreated.add(t1);
                    } else if ((tgtNode.distFromSrc != Integer.MAX_VALUE) && (tgtNode.distFromSrc + edgelength < nbrNode.distFromSrc)) {
                        //create a new SPTask for nbrNode
                        SPTask t1 = new SPTask((nbrNode.nodeId), tgtNode.distFromSrc, edgelength, tgtNode.nodeId, nbrNode.TS);
                        String location = dataStorage.getLocation(nbrNode.nodeId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        //System.out.println("addEdge :: Adding task to tasksCreated for : "+t1.nodeId);
                        tasksCreated.add(t1);
                    }
                }
                break;
                case SPTaskConstants.deleteEdge: {
                    String nbrid = (String) t.params.get(0);
                    //System.out.println("inside deleteEdge::"+" nbrid : "+nbrid+", tgtNode : "+tgtNode.nodeId);
                    SPNode nbrNode = new SPNode((SPNode) dataStorage.getNode(nbrid));

                    nbrNode.neighbors.remove(tgtNode.nodeId);
                    tgtNode.neighbors.remove(nbrNode.nodeId);


                    updatedNodes.put(tgtNode.nodeId, tgtNode);
                    updatedNodes.put(nbrNode.nodeId, nbrNode);

                    //System.out.println(nbrNode+"(pred : "+nbrNode.predecessorId+"), "+tgtNode+"(pred : "+tgtNode.predecessorId+")");

                    if (nbrNode.predecessorId.equals(tgtNode.nodeId)) {
                        //int oldDistFromSrc = nbrNode.distFromSrc;
                        nbrNode.predecessorId = "-1";
                        nbrNode.distFromSrc = Integer.MAX_VALUE;
                        //System.out.println(" inside nbrNode.predecessorId == tgtNode.nodeId");

                        for (String nbr : nbrNode.neighbors.keySet()) {
                            if (nbr.equals(tgtNode.nodeId))
                                continue;
                            RecomputeSPTask t1 = new RecomputeSPTask(nbr, nbrNode.distFromSrc, ((SPEdge) nbrNode.neighbors.get(nbr)).weight, nbrNode.nodeId, nbrNode.TS);
                            String location = dataStorage.getLocation(nbr);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                    } else if (tgtNode.predecessorId.equals(nbrNode.nodeId)) {
                        //int oldDistFromSrc = tgtNode.distFromSrc;
                        tgtNode.predecessorId = "-1";
                        tgtNode.distFromSrc = Integer.MAX_VALUE;
                        //System.out.println(" inside tgtNode.predecessorId == nbrNode.nodeId");

                        for (String nbr : tgtNode.neighbors.keySet()) {
                            if (nbr.equals(nbrNode.nodeId))
                                continue;
                            RecomputeSPTask t1 = new RecomputeSPTask(nbr, tgtNode.distFromSrc, ((SPEdge) tgtNode.neighbors.get(nbr)).weight, tgtNode.nodeId, tgtNode.TS);
                            String location = dataStorage.getLocation(nbr);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);

                            //System.out.println(tgtNode.nodeId +" :: Created Recompute task for nbr : "+nbr);
                        }
                    }
                }
                break;
                case SPTaskConstants.updateEdge: {
                    String nbrid = (String) t.params.get(0);
                    //System.out.println("inside updateEdge::"+" nbrid : "+nbrid);
                    Integer edge = (Integer) t.params.get(1);
                    int edgelength = edge.intValue();
                    SPNode nbrNode = new SPNode((SPNode) dataStorage.getNode(nbrid));

                    SPEdge edgeToNbr = new SPEdge(edgelength);
                    nbrNode.addNeighbor(tgtNode.nodeId, edgeToNbr);
                    SPEdge edgeToTgt = new SPEdge(edgelength);
                    tgtNode.addNeighbor(nbrNode.nodeId, edgeToTgt);

                    updatedNodes.put(tgtNode.nodeId, tgtNode);
                    updatedNodes.put(nbrNode.nodeId, nbrNode);

					/* logic for tgtNode */
                    if ((nbrNode.distFromSrc != Integer.MAX_VALUE) && (nbrNode.distFromSrc + edgelength < tgtNode.distFromSrc)) {
                        //if the distance updated such that the shortest path decreased further, create a new SPTask for tgtNode
                        SPTask t1 = new SPTask((tgtNode.nodeId), nbrNode.distFromSrc, edgelength, nbrNode.nodeId, nbrNode.TS);
                        String location = dataStorage.getLocation(tgtNode.nodeId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        //System.out.println("updateEdge :: Adding task to tasksCreated for : "+t1.nodeId);
                        tasksCreated.add(t1);
                    } else if ((tgtNode.predecessorId.equals(nbrNode.nodeId)) && (nbrNode.distFromSrc != Integer.MAX_VALUE) && (nbrNode.distFromSrc + edgelength > tgtNode.distFromSrc)) {
                        //if the distance updated is for the predecessor edge and the distance increased
                        tgtNode.predecessorId = nbrNode.nodeId;
                        tgtNode.distFromSrc = nbrNode.distFromSrc + edgelength;

                        for (String nbr : tgtNode.neighbors.keySet()) {
                            if (nbr.equals(nbrNode.nodeId))
                                continue;
                            RecomputeSPTask t1 = new RecomputeSPTask(nbr, tgtNode.distFromSrc, ((SPEdge) tgtNode.neighbors.get(nbr)).weight, tgtNode.nodeId, tgtNode.TS);
                            String location = dataStorage.getLocation(nbr);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                    }

					/* logic for nbrNode */
                    else if ((tgtNode.distFromSrc != Integer.MAX_VALUE) && (tgtNode.distFromSrc + edgelength < nbrNode.distFromSrc)) {
                        //if the distance updated such that the shortest path decreased further

                        //create a new SPTask for nbrNode
                        SPTask t1 = new SPTask((nbrNode.nodeId), tgtNode.distFromSrc, edgelength, tgtNode.nodeId, tgtNode.TS);
                        String location = dataStorage.getLocation(nbrNode.nodeId);
                        t1.targetLocation = location;
                        t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                        //System.out.println("updateEdge :: Adding task to tasksCreated for : "+t1.nodeId);
                        tasksCreated.add(t1);
                    } else if ((nbrNode.predecessorId.equals(tgtNode.nodeId)) && (tgtNode.distFromSrc != Integer.MAX_VALUE) && (tgtNode.distFromSrc + edgelength > nbrNode.distFromSrc)) {
                        //if the distance updated is for the predecessor edge and the distance increased

                        nbrNode.predecessorId = tgtNode.nodeId;
                        nbrNode.distFromSrc = tgtNode.distFromSrc + edgelength;

                        Vector<RecomputeSPTask> initialTasks = new Vector<RecomputeSPTask>();

                        for (String nbr : nbrNode.neighbors.keySet()) {
                            if (nbr.equals(tgtNode.nodeId))
                                continue;
                            RecomputeSPTask t1 = new RecomputeSPTask(nbr, nbrNode.distFromSrc, ((SPEdge) nbrNode.neighbors.get(nbr)).weight, nbrNode.nodeId, nbrNode.TS);
                            String location = dataStorage.getLocation(nbr);
                            t1.targetLocation = location;
                            t1.affinityLevel = TaskConstants.STRONG_AFFINITY;
                            tasksCreated.add(t1);
                        }
                    }
                }
                break;
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return tgtNode;
    }

    /**
     * Given a node, creates an SP task for each neighbor whose path distance to the final node will
     * improve/decrease based on the current my (the node's) distance from the source; excludes self-loops
     *
     * @param node The node for which to create SPTasks for
     */
    private void createSPTasksForNbrs(Node node) {
        if (node == null) {
            return;
        }

        SPNode target = new SPNode((SPNode) node);
        for (String nbrId : target.neighbors.keySet()) {
            if (nbrId.equals(node.nodeId)) {
                continue;
            }
            SPEdge e = (SPEdge) target.neighbors.get(nbrId);
            if (e != null) {
                int weight = e.weight;

                //Do not send task to neighbor, if its already at a lower distance
                if (neighborDistances.containsKey(nbrId) && (neighborDistances.get(nbrId).nbrDistFromSrc <= (target.distFromSrc + weight))) {
                    continue;
                }

                // Let my neighbor's neighbors know about my change in distance
                SPTask nbrTask = new SPTask(nbrId, target.distFromSrc, weight, target.nodeId, target.TS);
                nbrTask.targetLocation = dataStorage.getLocation(nbrId);
                nbrTask.affinityLevel = TaskConstants.STRONG_AFFINITY;
                tasksCreated.add(nbrTask);
            }
        }
    }

    public void doTask(Task tsk) throws Exception {
        tasksCreated.clear();
        updatedNodes.clear();
        boolean commit = false;
        Node resNode = null;

        Class c = tsk.getClass();
        String strclassnamearr[] = c.toString().split(" ")[1].split("\\.");
        String taskClassName = strclassnamearr[strclassnamearr.length - 1];
        //System.out.println("Task class name : "+taskClassName);

        while (!commit) {
            Node tgtNode = dataStorage.getNode(tsk.getNodeId());
            tsk.txnInfo = beginTransaction();
            try {
                switch (taskClassName) {
                    case "SPTask":
                        resNode = compute(tsk, tgtNode);
                        break;
                    case "UpdateGraphTask":
                        resNode = updateGraph(tsk, tgtNode);
                        break;
                    case "RecomputeSPTask":
                        resNode = recompute(tsk, tgtNode);
                        break;
                }


                if (resNode != null) {
                    commit = validateTask(tsk);
                } else {
                    // The distance from the source did not improve. There is nothing to commit, but the
                    // transaction needs to completed
                    commit = true;
                    workerStatus.status = WorkerStatus.REPORT_COMPLETION;
                    workpool.reportCompletion(tsk, tasksCreated);
                    break;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }


    /****************
     * Incremental aux methods
     ****************/

    public UpdateGraphTask addEdge(String sourceNode, String destNode, int edgeLength) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        params.add(edgeLength);
        UpdateGraphTask t = null;

        SPNode srcNode;
        try {
            srcNode = (SPNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = null;
            } else {
                t = new UpdateGraphTask(sourceNode, SPTaskConstants.addEdge, params);
                t.targetLocation = dataStorage.getLocation(sourceNode);
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;

            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    public UpdateGraphTask deleteEdge(String sourceNode, String destNode) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        UpdateGraphTask t = null;

        SPNode srcNode;
        try {
            srcNode = (SPNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = new UpdateGraphTask(sourceNode, SPTaskConstants.deleteEdge, params);
                String location = dataStorage.getLocation(sourceNode);
                t.targetLocation = location;
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;
            } else {
                //no edge to delete between source and destination. So, nothing to do.
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    private UpdateGraphTask updateEdge(String sourceNode, String destNode, int edgeLength) {
        Vector<Object> params = new Vector<Object>();
        params.add(destNode);
        params.add(edgeLength);
        UpdateGraphTask t = null;

        SPNode srcNode;
        try {
            srcNode = (SPNode) dataStorage.getNode(sourceNode);
            if (srcNode.neighbors.containsKey(destNode)) {
                t = new UpdateGraphTask(sourceNode, SPTaskConstants.updateEdge, params);
                String location = dataStorage.getLocation(sourceNode);
                t.targetLocation = location;
                t.affinityLevel = TaskConstants.STRONG_AFFINITY;
            } else {
                t = null;
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return t;
    }

    private Vector<UpdateGraphTask> createGraphUpdateTasks(int hostId, int phaseno) {
        if (master) {
            printResult(phaseno - 1);
            System.out.println("\nAdding new changes in the graph...\n");
        }

        //loadGenerator
        Vector<UpdateGraphTask> vtasks = new Vector<UpdateGraphTask>();
        int countAddEdgeOperations = 0, countDeleteEdgeOperations = 0, countUpdateEdgeOperations = 0;
        int maxEdgeLength = 100;
        boolean deleteEdgeUsingFile = false;
        boolean showOutput = false;
        String deleteEdgeUsingFileName = "";
        int numServers = dataStorage.getNumberOfServers();
        showOutput = configParams.getBoolValue("SPWorker.SHOW_OUTPUT");
        countAddEdgeOperations = configParams.getIntValue("SPWorker.ADD_EDGE")/*/numServers*/;
        maxEdgeLength = configParams.getIntValue("SPWorker.EDGE_MAX_LENGTH");
        countDeleteEdgeOperations = configParams.getIntValue("SPWorker.DELETE_EDGE")/*/numServers*/;
        deleteEdgeUsingFile = configParams.getBoolValue("SPWorker.DELETE_EDGE_USING_FILE");
        if (deleteEdgeUsingFile) {
            deleteEdgeUsingFileName = configParams.getStringValue("SPWorker.DELETE_EDGE_USING_FILENAME");
        }
        countUpdateEdgeOperations = configParams.getIntValue("SPWorker.UPDATE_EDGE")/*/numServers*/;

        Random r = new Random();
        StringBuffer strbuf = new StringBuffer();
        UpdateGraphTask t = null;

        switch (phaseno) {
            case 1: //add Edge

                for (int i = 0; i < countAddEdgeOperations; i++) {
                    int src = r.nextInt(this.numNodes);
                    int dest = r.nextInt(this.numNodes);
                    while (dest == src) {
                        dest = r.nextInt(this.numNodes);
                    }
                    int len = r.nextInt(maxEdgeLength) + 1;
                    strbuf.append("Host " + dataStorage.myId + " :: Adding edge between Node " + src + " and " + dest + " of length " + len + "\n");
                    t = addEdge("" + src, "" + dest, len);

                    if (t == null) {
                        i--;
                    } else {
                        vtasks.add(t);
                    }
                }
                break;
            case 2: //delete Edge

                if (deleteEdgeUsingFile && (countDeleteEdgeOperations != 0)) {
                    //delete Edges read from file upto the count value
                    int i = 0;
                    String line = null;
                    int src = 0, dest = 0;
                    try {
                        BufferedReader br = new BufferedReader(new FileReader(deleteEdgeUsingFileName));
                        while ((line = br.readLine()) != null) {
                            String str[] = line.split(",");
                            src = Integer.parseInt(str[0]);

                            dest = Integer.parseInt(str[1]);
                            strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                            t = deleteEdge("" + src, "" + dest);
                            if (t == null) {
                                System.out.println("Task is null for Delete Edge scenario. Please check the edge list file...");
                            } else {
                                vtasks.add(t);
                                i++;
                            }

                            if (i >= countDeleteEdgeOperations) {
                                break;
                            }
                        }
                        br.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    //randomly delete edges
                    for (int i = 0; i < countDeleteEdgeOperations; i++) {
                        int src = r.nextInt(this.numNodes);
                        int dest = r.nextInt(this.numNodes);
                        while (dest == src) {
                            dest = r.nextInt(this.numNodes);
                        }
                        t = deleteEdge("" + src, "" + dest);
                        if (t == null) {
                            i--;
                        } else {
                            strbuf.append("Host " + dataStorage.myId + " :: Deleting edge between Node " + src + " and " + dest + "\n");
                            vtasks.add(t);
                        }
                    }
                    System.out.println("NUMBER OF DELETE EDGE TASKS: " + vtasks.size());
                }
                break;
            case 3: //update Edge

                if (deleteEdgeUsingFile && (countUpdateEdgeOperations != 0)) {
                    //delete Edges read from file upto the count value
                    int i = 0;
                    int skipedges = 0;
                    String line = null;
                    int src = 0, dest = 0;
                    try {
                        BufferedReader br = new BufferedReader(new FileReader(deleteEdgeUsingFileName));
                        while ((line = br.readLine()) != null) {
                            //skipping the lines used by delete edge transactions
                            if (skipedges <= countDeleteEdgeOperations) {
                                skipedges++;
                                continue;
                            }
                            String str[] = line.split(",");
                            src = Integer.parseInt(str[0]);

                            dest = Integer.parseInt(str[1]);
                            int len = r.nextInt(maxEdgeLength) + 1;
                            strbuf.append("Host " + dataStorage.myId + " :: Updating edge between Node " + src + " and " + dest + " to length " + len + "\n");
                            t = updateEdge("" + src, "" + dest, len);
                            if (t == null) {
                                System.out.println("Task is null for Update Edge scenario. Please check the edge list file...");
                            } else {
                                vtasks.add(t);
                                i++;
                            }
                            if (i >= countUpdateEdgeOperations) {
                                break;
                            }
                        }
                        br.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    for (int i = 0; i < countUpdateEdgeOperations; i++) {
                        int src = r.nextInt(this.numNodes);
                        int dest = r.nextInt(this.numNodes);
                        while (dest == src) {
                            dest = r.nextInt(this.numNodes);
                        }
                        int len = r.nextInt(maxEdgeLength) + 1;
                        strbuf.append("Host " + dataStorage.myId + " :: Updating edge between Node " + src + " and " + dest + " to length " + len + "\n");
                        t = updateEdge("" + src, "" + dest, len);
                        if (t == null) {
                            i--;
                        } else {
                            vtasks.add(t);
                        }
                    }
                }
                break;
        }

        if (showOutput) {
            System.out.println(strbuf);
        }

        return vtasks;
    }


    /****************
     * Aux methods
     ****************/

    public void cleanUp(int phase) {
        long jobFinishTime = System.currentTimeMillis();
        long duration = jobFinishTime - jobStartTime;
        long durationInitial = phaseEndTime[0] - jobStartTime;
        long durationP1 = phaseEndTime[1] - phaseStartTime[1];
        long durationP2 = phaseEndTime[2] - phaseStartTime[2];
        long durationP3 = phaseEndTime[3] - phaseStartTime[3];
        long durationP4 = phaseEndTime[4] - phaseStartTime[4];
        long durationP5 = phaseEndTime[5] - phaseStartTime[5];

        if (configParams.getBoolValue("SPWorker.SHOW_OUTPUT")) {
            try {
                ConcurrentHashMap<Integer, HashtableServer> servers = dataStorage.servers;
                Collection<HashtableServer> serverCollection = servers.values();
                Vector<Node> l = new Vector<Node>();
                for (HashtableServer server : serverCollection) {
                    l.addAll(server.getAllNodes());
                }
                Collections.sort(l, new NodeComparator());
                System.out.println("*******************************");
                System.out.println("Result of phase : " + phase);
                for (Node n : l) {
                    System.out.println("Distance from " + srcNode + " for node : " + n.nodeId + " = " + ((SPNode) n).distFromSrc + " Predecessor = " +
                            ((SPNode) n).predecessorId);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        System.out.println("*******************************");
        System.out.println("Total Execution Time taken : " + duration + " msec");
        System.out.println("Initial Computation Time taken : " + durationInitial + " msec");
        System.out.println("Phase 1 Time taken : " + durationP1 + " msec");
        System.out.println("Phase 2 Time taken : " + durationP2 + " msec");
        System.out.println("Phase 3 Time taken : " + durationP3 + " msec");
        System.out.println("Phase 4 Time taken : " + durationP4 + " msec");
        System.out.println("Phase 5 Time taken : " + durationP5 + " msec");

        try {
            FileOutputStream fout = new FileOutputStream("result.txt", true);
            FileLock lock = fout.getChannel().lock();
            PrintWriter out = new PrintWriter(fout);
            out.println("==================================");
            out.println("Total Execution Time taken : " + duration + " msec");
            out.println("Initial Computation Time taken : " + durationInitial + " msec");
            out.println("Phase 1 Time taken : " + durationP1 + " msec");
            out.println("Phase 2 Time taken : " + durationP2 + " msec");
            out.println("Phase 3 Time taken : " + durationP3 + " msec");
            out.println("Phase 4 Time taken : " + durationP4 + " msec");
            out.println("Phase 5 Time taken : " + durationP5 + " msec");

            out.println("==================================");
            out.flush();
            lock.release();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }


        System.out.println("Worker " + workerId + ":  Finished cleanup");

        System.out.println("*****************************");
        System.exit(0);
    }

    private void printResult(int phase) {
        if (configParams.getBoolValue("SPWorker.SHOW_OUTPUT")) {
            try {
                ConcurrentHashMap<Integer, HashtableServer> servers = dataStorage.servers;
                Collection<HashtableServer> serverCollection = servers.values();
                Vector<Node> l = new Vector<Node>();
                for (HashtableServer server : serverCollection) {
                    l.addAll(server.getAllNodes());
                }
                Collections.sort(l, new NodeComparator());
                System.out.println("*******************************");
                System.out.println("Result of phase : " + phase);
                for (Node n : l) {
                    System.out.println("Distance from " + srcNode + " for node : " + n.nodeId + " = " + ((SPNode) n).distFromSrc + " Predecessor = " +
                            ((SPNode) n).predecessorId);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }


    /****************
     * Aux classes
     ****************/

    private static class NodeComparator implements Comparator<Node> {
        public int compare(Node n1, Node n2) {
            if (Integer.parseInt(n1.nodeId) > Integer.parseInt(n2.nodeId)) {
                return 1;
            }
            else {
                return -1;
            }
        }
    }

    static class NeighborInfo {
        final int nbrDistFromSrc;
        final long nbrTS;

        public NeighborInfo(int nbrDistFromSrc, long nbrTS) {
            this.nbrDistFromSrc = nbrDistFromSrc;
            this.nbrTS = nbrTS;
        }
    }
}
