/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.util;

import beehive.checkpoint.CheckpointListener;
import beehive.checkpoint.CheckpointPathBuilder;
import beehive.checkpoint.FailureDetector;
import beehive.server.*;
import beehive.thrift.BeehiveThriftBuilder;
import beehive.thrift.HashtableServer_thrift;
import beehive.util.*;
import beehive.workpool.*;

import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;
import java.rmi.server.RMISocketFactory;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.apache.thrift.protocol.TCompactProtocol;
import org.apache.thrift.server.TServer;
import org.apache.thrift.server.TThreadPoolServer;
import org.apache.thrift.transport.TServerSocket;
import org.apache.thrift.transport.TServerTransport;
import org.apache.thrift.transport.TTransportException;

import java.io.*;
import java.net.*;

abstract public class BeehiveAppLoader {
    public String hostlistFile;
    public LocalWorkpoolImpl workpool = null;
    public StorageSystem dataStorage = null;
    public String location;
    public int numNodes;
    public boolean useNodeData = false;
    public boolean master;
    public int numThreads = 0;
    public int myId;
    public boolean useCaching = false;
    public boolean recordStats = false;
    public boolean writeThroughCache = false;
    public boolean canonicalHostNames = true;
    public boolean oldFileLoader = true;

    public String coordinator = null;
    private boolean recovery = false;
    private boolean checkpoint = false;
    private boolean failureDetectionEnable = false;

    public void init(String args[], String classname, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws Exception {

		/*Initializing Config Parameters*/
        String configParamsFile = args[1];
        Config configParams = new Config(configParamsFile);
        readConfigParams(configParams);

        System.out.println("Number of workers:" + numThreads);

        //defining RMI socket factories
        int timeout = 1800000;
        if (configParams.has("System.RMI_SO_TIMEOUT"))
            timeout = configParams.getIntValue("System.RMI_SO_TIMEOUT");

        RMIClientSocketFactory csf = new CustomClientSocketFactory(timeout);
        RMIServerSocketFactory ssf = new CustomServerSocketFactory();

        //setting THRIFT socket timeout 
        if (configParams.has("System.THRIFT_SO_TIMEOUT")) {
            int timeoutValue = configParams.getIntValue("System.THRIFT_SO_TIMEOUT");
            BeehiveThriftBuilder.setThriftConnectionTimeout(timeoutValue);
        }

        
        /*Initialize host data*/
        this.hostlistFile = args[0];
        this.numNodes = Integer.parseInt(args[4]);

        String myHostname;
        if (this.canonicalHostNames) {
            myHostname = InetAddress.getLocalHost().getCanonicalHostName();
        } else {
            myHostname = InetAddress.getLocalHost().getHostName();

        }

        //System.out.println("My hostname is: " + myHostname );

        Vector<String> hostList = createHostList(myHostname);
        if (master) {
            printToResultFile(args, numThreads, classname);
        }

        this.location = "//" + myHostname + "/HashtableServer";
        //System.out.println("My URL is: " + location + " my hostlist size is " + hostList.size()  );

        HashtableServerImpl HashtableNode = new HashtableServerImpl(location, myId, configParams, useCaching, nodeClassNameMap, nodeClassMap, recovery);

        if (recordStats) HashtableNode.enableStats();

        HashtableNode.setNumServers(hostList.size());


        //synchronized lookup and graph loading logic
        boolean isCoordinator = false;
        if (myHostname.equals(coordinator))
            isCoordinator = true;

        DistributedSynchronizer distributedSync = null;

        int coordinatorPort = 33333;

        if (isCoordinator) {
            //Initialize RMI registry
            Registry localRegistryCoordinator = null;
            try {
                localRegistryCoordinator = LocateRegistry.createRegistry(coordinatorPort);
            } catch (RemoteException e) {

                System.out.println("Couldn't Create Registry");
                e.printStackTrace();
            }
            System.setSecurityManager(new RMISecurityManager());

            distributedSync = new DistributedSynchronizerImpl(hostList.size(), isCoordinator);
            //Creating a stub from DistributedSynchronizerImpl and exporting it with custom client and server socket configurations
            DistributedSynchronizer stub = (DistributedSynchronizer) UnicastRemoteObject.exportObject(distributedSync, 0, csf, ssf);
            localRegistryCoordinator.bind("DistributedSynchronizer", stub);
        }


        if (oldFileLoader == false) {
            if (isCoordinator) {
                distributedSync.synchronize(myHostname);
            } else {
                Thread.sleep(1000);
                String url = "//" + coordinator + ":" + coordinatorPort + "/DistributedSynchronizer";
                distributedSync = (DistributedSynchronizer) Naming.lookup(url);

                distributedSync.synchronize(myHostname);
            }

            if (isCoordinator) {
                distributedSync.reset();
            }
        }

        if (oldFileLoader == false) {
            if (!recovery) {
                System.out.println(myHostname + " started loading graph...");
                long startGraphLoad = System.currentTimeMillis();
                loadGraphData(args[3], HashtableNode, numNodes, null);        //Implementation specific method to be specified by caller
                long endGraphLoad = System.currentTimeMillis();
                System.out.println(myHostname + " completed graph load in " + (endGraphLoad - startGraphLoad) + " ms");
            }
        }

        //synchronized lookup and graph loading logic
        if (oldFileLoader == false) {
            if (isCoordinator) {
                distributedSync.synchronize(myHostname);
            } else {
                Thread.sleep(1000);
                distributedSync.synchronize(myHostname);
            }

            if (isCoordinator) {
                distributedSync.reset();
            }
        }


        //global socket factory configuration. Effective for all RMI connections
        RMISocketFactory.setSocketFactory(new RMISocketFactory() {
            public Socket createSocket(String host, int port)
                    throws IOException {
                int timeout = 30000;
                Socket socket = new Socket();
                socket.setSoTimeout(timeout);
                socket.setSoLinger(false, 0);
                socket.connect(new InetSocketAddress(host, port), timeout);
                return socket;
            }

            public ServerSocket createServerSocket(int port)
                    throws IOException {
                return new ServerSocket(port);
            }
        });


        //Initialize RMI registry
        Registry localRegistry = null;
        try {
            localRegistry = LocateRegistry.createRegistry(1099);
        } catch (RemoteException e) {

            System.out.println("Couldn't Create Registry");
            e.printStackTrace();
        }
        System.setSecurityManager(new RMISecurityManager());

        //binding HashTableServer to RMI registry
        HashtableServer htStub = (HashtableServer) UnicastRemoteObject.exportObject(HashtableNode, 0, csf, ssf);
        localRegistry.bind("HashtableServer", htStub);

        System.out.println("Hashtable server started");
        Thread.sleep(10000);

        //configure this server with handles of other servers while others are also doing the same
        HashtableNode.configureServer(hostList);
        System.out.println("Server configure complete");


        // start thrift server
        String RPC_MECHANISM = "RMI";
        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

        if (RPC_MECHANISM.equals("THRIFT")) {
            StartServer thriftServer = new StartServer(HashtableNode);
            thriftServer.start();
            Thread.sleep(10000);
        }


        //storage system does not need to know about the useCaching flag since we have separate APIs
        StorageSystem dataStorage;
        if (useCaching) {
            dataStorage = new StorageSystem(hostList, HashtableNode, writeThroughCache, configParams, nodeClassNameMap, nodeClassMap);
        } else {
            dataStorage = new StorageSystem(hostList, HashtableNode, configParams, nodeClassNameMap, nodeClassMap);
        }
        if (recordStats) dataStorage.enableStats();


        this.numNodes = Integer.parseInt(args[4]);

        //if graph partition hashing scheme is used, creating the partition table
        if (configParams.has("Server.FileLoader") && configParams.has("Server.NODE_HASHING_SCHEME")) {
            String loaderType = configParams.getStringValue("Server.FileLoader").trim();
            String hashingScheme = configParams.getStringValue("Server.NODE_HASHING_SCHEME").trim();
            if (loaderType.equals("Bucketized") && hashingScheme.equals("GRAPH_PARTITION")) {
                List<Integer> hostArrList = new ArrayList<Integer>();

                try {
                    String input_filename = args[3] + "/" + "metadata";
                    BufferedReader input = new BufferedReader(new FileReader(input_filename));
                    String line = "";
                    int ctr = 0;

                    while ((line = input.readLine()) != null) {
                        if (line.charAt(0) == '#' || line.trim().equals("") || line.length() < 5)
                            continue;

                        StringTokenizer lineTok = new StringTokenizer(line, " ");
                        String hostKey = lineTok.nextToken();
                        String fromNode = lineTok.nextToken();
                        String toNode = lineTok.nextToken();
                        int hostid = Integer.parseInt(hostKey);
                        int lastNodeStored = Integer.parseInt(toNode);
                        hostArrList.add(hostid, lastNodeStored);
                    }
                    input.close();
                } catch (Exception e) {
                    e.printStackTrace();
                    System.err.println("ERROR :: Failed to read metadata file");
                }

                int hostArr[] = new int[hostArrList.size()];
                for (int i = 0; i < hostArrList.size(); i++) {
                    hostArr[i] = hostArrList.get(i);
                }
                dataStorage.partitionedGraphHostNodesTable = hostArr;
            }
        }


        //binding Local Workpool to the RMI registry
        LocalWorkpoolImpl workpool = new LocalWorkpoolImpl(args[2], numThreads, dataStorage.partitionedGraphHostNodesTable, recovery);
        LocalWorkpoolServer localWorkpoolStub = (LocalWorkpoolServer) UnicastRemoteObject.exportObject(workpool, 0, csf, ssf);
        localRegistry.bind("WorkpoolServer", localWorkpoolStub);

        Thread.sleep(10000);
        workpool.lookupPeers(hostList);

        this.workpool = workpool;
        this.dataStorage = dataStorage;


        if (oldFileLoader == true) {
            if (master) {
                if (!recovery) {
                    System.out.println(myHostname + " master started loading graph...");
                    long startGraphLoad = System.currentTimeMillis();
                    loadGraphData(args[3], null, numNodes, dataStorage);    //Implementation specific method to be declared by caller
                    long endGraphLoad = System.currentTimeMillis();
                    System.out.println(myHostname + " completed graph load in " + (endGraphLoad - startGraphLoad) + " ms");
                }
            }
        }


        if (master && checkpoint) {
            new Checkpointer(workpool, dataStorage).start();
        }

        if (failureDetectionEnable) {
            String failureServerHostname = configParams.getStringValue("FailureDetection.hostname");
            String failure_url = "//" + failureServerHostname + "/FailureDetectionServer";
            try {
                System.out.println(myHostname + " looking up failure detector server...");
                FailureDetector failureDetectionServer = (FailureDetector) Naming.lookup(failure_url);
                System.out.println(myHostname + " connected to failure detector server");

                Heartbeat heartbeat;
                if (configParams.has("FailureDetection.heartbeatFrequency")) {
                    heartbeat = new Heartbeat(myHostname, failureDetectionServer, configParams.getIntValue("FailureDetection.heartbeatFrequency"));
                } else {
                    heartbeat = new Heartbeat(myHostname, failureDetectionServer);
                }
                heartbeat.start();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    public void init(String args[], String classname, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap, boolean recover) throws Exception {

		/*Initializing Config Parameters*/
        String configParamsFile = args[1];
        Config configParams = new Config(configParamsFile);
        readConfigParams(configParams);
        recovery = recover;

        System.out.println("Number of workers:" + numThreads);

        //defining RMI socket factories
        int timeout = 1800000;
        if (configParams.has("System.RMI_SO_TIMEOUT"))
            timeout = configParams.getIntValue("System.RMI_SO_TIMEOUT");

        RMIClientSocketFactory csf = new CustomClientSocketFactory(timeout);
        RMIServerSocketFactory ssf = new CustomServerSocketFactory();

        //setting THRIFT socket timeout
        if (configParams.has("System.THRIFT_SO_TIMEOUT")) {
            int timeoutValue = configParams.getIntValue("System.THRIFT_SO_TIMEOUT");
            BeehiveThriftBuilder.setThriftConnectionTimeout(timeoutValue);
        }


        /*Initialize host data*/
        this.hostlistFile = args[0];

        String myHostname;
        if (this.canonicalHostNames) {
            myHostname = InetAddress.getLocalHost().getCanonicalHostName();
        } else {
            myHostname = InetAddress.getLocalHost().getHostName();

        }

        //System.out.println("My hostname is: " + myHostname );

        Vector<String> hostList = createHostList(myHostname);
        if (master) {
            printToResultFile(args, numThreads, classname);
        }

        this.location = "//" + myHostname + "/HashtableServer";
        //System.out.println("My URL is: " + location + " my hostlist size is " + hostList.size()  );

        HashtableServerImpl HashtableNode = new HashtableServerImpl(location, myId, configParams, useCaching, nodeClassNameMap, nodeClassMap, recovery);

        if (recordStats) HashtableNode.enableStats();

        HashtableNode.setNumServers(hostList.size());


        //synchronized lookup and graph loading logic
        boolean isCoordinator = false;
        if (myHostname.equals(coordinator))
            isCoordinator = true;

        DistributedSynchronizer distributedSync = null;

        int coordinatorPort = 33333;

        if (isCoordinator) {
            //Initialize RMI registry
            Registry localRegistryCoordinator = null;
            try {
                localRegistryCoordinator = LocateRegistry.createRegistry(coordinatorPort);
            } catch (RemoteException e) {

                System.out.println("Couldn't Create Registry");
                e.printStackTrace();
            }
            System.setSecurityManager(new RMISecurityManager());

            distributedSync = new DistributedSynchronizerImpl(hostList.size(), isCoordinator);
            //Creating a stub from DistributedSynchronizerImpl and exporting it with custom client and server socket configurations
            DistributedSynchronizer stub = (DistributedSynchronizer) UnicastRemoteObject.exportObject(distributedSync, 0, csf, ssf);
            localRegistryCoordinator.bind("DistributedSynchronizer", stub);
        }


        if (oldFileLoader == false) {
            if (isCoordinator) {
                distributedSync.synchronize(myHostname);
            } else {
                Thread.sleep(1000);
                String url = "//" + coordinator + ":" + coordinatorPort + "/DistributedSynchronizer";
                distributedSync = (DistributedSynchronizer) Naming.lookup(url);

                distributedSync.synchronize(myHostname);
            }

            if (isCoordinator) {
                distributedSync.reset();
            }
        }

        if (oldFileLoader == false) {
            if (!recovery) {
                System.out.println(myHostname + " started loading graph...");
                long startGraphLoad = System.currentTimeMillis();
                loadGraphData(args[3], HashtableNode, numNodes, null);        //Implementation specific method to be specified by caller
                long endGraphLoad = System.currentTimeMillis();
                System.out.println(myHostname + " completed graph load in " + (endGraphLoad - startGraphLoad) + " ms");
            }
        }

        //synchronized lookup and graph loading logic
        if (oldFileLoader == false) {
            if (isCoordinator) {
                distributedSync.synchronize(myHostname);
            } else {
                Thread.sleep(1000);
                distributedSync.synchronize(myHostname);
            }

            if (isCoordinator) {
                distributedSync.reset();
            }
        }


        //global socket factory configuration. Effective for all RMI connections
        RMISocketFactory.setSocketFactory(new RMISocketFactory() {
            public Socket createSocket(String host, int port)
                    throws IOException {
                int timeout = 30000;
                Socket socket = new Socket();
                socket.setSoTimeout(timeout);
                socket.setSoLinger(false, 0);
                socket.connect(new InetSocketAddress(host, port), timeout);
                return socket;
            }

            public ServerSocket createServerSocket(int port)
                    throws IOException {
                return new ServerSocket(port);
            }
        });


        //Initialize RMI registry
        Registry localRegistry = null;
        try {
            localRegistry = LocateRegistry.createRegistry(1099);
        } catch (RemoteException e) {

            System.out.println("Couldn't Create Registry");
            e.printStackTrace();
        }
        System.setSecurityManager(new RMISecurityManager());

        //binding HashTableServer to RMI registry
        HashtableServer htStub = (HashtableServer) UnicastRemoteObject.exportObject(HashtableNode, 0, csf, ssf);
        localRegistry.bind("HashtableServer", htStub);

        System.out.println("Hashtable server started");
        Thread.sleep(10000);

        //configure this server with handles of other servers while others are also doing the same
        HashtableNode.configureServer(hostList);
        System.out.println("Server configure complete");


        // start thrift server
        String RPC_MECHANISM = "RMI";
        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

        if (RPC_MECHANISM.equals("THRIFT")) {
            StartServer thriftServer = new StartServer(HashtableNode);
            thriftServer.start();
            Thread.sleep(10000);
        }


        //storage system does not need to know about the useCaching flag since we have separate APIs
        StorageSystem dataStorage;
        if (useCaching) {
            dataStorage = new StorageSystem(hostList, HashtableNode, writeThroughCache, configParams, nodeClassNameMap, nodeClassMap);
        } else {
            dataStorage = new StorageSystem(hostList, HashtableNode, configParams, nodeClassNameMap, nodeClassMap);
        }
        if (recordStats) dataStorage.enableStats();


        this.numNodes = Integer.parseInt(args[4]);

        //if graph partition hashing scheme is used, creating the partition table
        if (configParams.has("Server.FileLoader") && configParams.has("Server.NODE_HASHING_SCHEME")) {
            String loaderType = configParams.getStringValue("Server.FileLoader").trim();
            String hashingScheme = configParams.getStringValue("Server.NODE_HASHING_SCHEME").trim();
            if (loaderType.equals("Bucketized") && hashingScheme.equals("GRAPH_PARTITION")) {
                List<Integer> hostArrList = new ArrayList<Integer>();

                try {
                    String input_filename = args[3] + "/" + "metadata";
                    BufferedReader input = new BufferedReader(new FileReader(input_filename));
                    String line = "";
                    int ctr = 0;

                    while ((line = input.readLine()) != null) {
                        if (line.charAt(0) == '#' || line.trim().equals("") || line.length() < 5)
                            continue;

                        StringTokenizer lineTok = new StringTokenizer(line, " ");
                        String hostKey = lineTok.nextToken();
                        String fromNode = lineTok.nextToken();
                        String toNode = lineTok.nextToken();
                        int hostid = Integer.parseInt(hostKey);
                        int lastNodeStored = Integer.parseInt(toNode);
                        hostArrList.add(hostid, lastNodeStored);
                    }
                    input.close();
                } catch (Exception e) {
                    e.printStackTrace();
                    System.err.println("ERROR :: Failed to read metadata file");
                }

                int hostArr[] = new int[hostArrList.size()];
                for (int i = 0; i < hostArrList.size(); i++) {
                    hostArr[i] = hostArrList.get(i);
                }
                dataStorage.partitionedGraphHostNodesTable = hostArr;
            }
        }


        //binding Local Workpool to the RMI registry
        LocalWorkpoolImpl workpool = new LocalWorkpoolImpl(args[2], numThreads, dataStorage.partitionedGraphHostNodesTable, recovery);
        LocalWorkpoolServer localWorkpoolStub = (LocalWorkpoolServer) UnicastRemoteObject.exportObject(workpool, 0, csf, ssf);
        localRegistry.bind("WorkpoolServer", localWorkpoolStub);

        Thread.sleep(10000);
        workpool.lookupPeers(hostList);

        this.workpool = workpool;
        this.dataStorage = dataStorage;


        if (oldFileLoader == true) {
            if (master) {
                if (!recovery) {
                    System.out.println(myHostname + " master started loading graph...");
                    long startGraphLoad = System.currentTimeMillis();
                    loadGraphData(args[3], null, numNodes, dataStorage);    //Implementation specific method to be declared by caller
                    long endGraphLoad = System.currentTimeMillis();
                    System.out.println(myHostname + " completed graph load in " + (endGraphLoad - startGraphLoad) + " ms");
                }
            }
        }


        if (master && checkpoint) {
            new Checkpointer(workpool, dataStorage).start();
        }

        if (failureDetectionEnable) {
            String failureServerHostname = configParams.getStringValue("FailureDetection.hostname");
            String failure_url = "//" + failureServerHostname + "/FailureDetectionServer";
            try {
                System.out.println(myHostname + " looking up failure detector server...");
                FailureDetector failureDetectionServer = (FailureDetector) Naming.lookup(failure_url);
                System.out.println(myHostname + " connected to failure detector server");
                Heartbeat heartbeat = new Heartbeat(myHostname, failureDetectionServer);
                heartbeat.start();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    public abstract void loadGraphData(String input_file, HashtableServer server, int numNodes, StorageSystem dataStorage);

    public Vector<String> createHostList(String myHostname) {
        this.master = false;
        boolean first = true;
        Vector<String> hostList = new Vector<String>();
        this.myId = 0;
        try {
            BufferedReader bufferedReader = new BufferedReader(new FileReader(hostlistFile));

            String host;
            int hostCount = 0;
            while ((host = bufferedReader.readLine()) != null) {
                host = host.trim();
                if (host.length() == 0) continue;
                if (first) {
                    coordinator = host;

                    if (host.equals(myHostname)) {
                        this.master = true;
                        System.out.println("Master machine");
                    }
                    first = false;
                }
                hostList.add(host);
                if (host.equals(myHostname)) {
                    this.myId = hostCount;
                }
                hostCount++;
            }
            bufferedReader.close();
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println(e.getMessage());
            System.out.println("Error reading config file");
            System.exit(1);
        }
        return hostList;
    }

    public void printToResultFile(String args[], int numThreads, String classname) throws Exception {
        int count = 0;

        FileWriter fw = new FileWriter("result.txt", true);
        BufferedWriter out = new BufferedWriter(fw);

        BufferedReader br = new BufferedReader(new FileReader(args[0]));
        String line = null;
        out.write("============================================================================\n");
        out.write("New Run: " + new Date().toString() + "\n");

        out.write(classname + " : " + " " + Arrays.toString(args) + "\n");
        out.write("Number of Machines used");
        while ((line = br.readLine()) != null) {
            count++;
        }
        out.write(" " + Integer.toString(count) + "\n");

        out.write("Number of threads:" + Integer.toString(numThreads) + "\n");

        out.write("Experiment settings description: \n");
        br = new BufferedReader(new FileReader(args[5]));
        line = null;
        while ((line = br.readLine()) != null)

        {
            out.write(line + "\n");
        }
        out.close();
        br.close();
    }

    public void readConfigParams(Config configParams) {
        this.useCaching = configParams.getBoolValue("Server.useCaching");
        this.recordStats = configParams.getBoolValue("StorageSystem.recordStats");
        this.useNodeData = configParams.getBoolValue("StorageSystem.useNodeData");
        this.writeThroughCache = configParams.getBoolValue("StorageSystem.WriteThroughCache");
        this.numThreads = configParams.getIntValue("Worker.WORKER_COUNT");
        this.canonicalHostNames = configParams.getBoolValue("Server.canonicalHostNames");
        this.recovery = configParams.getBoolValue("System.recovery");
        this.checkpoint = configParams.getBoolValue("System.checkpoint");
        this.failureDetectionEnable = configParams.getBoolValue("FailureDetection.enable");
        if (configParams.has("Server.FileLoader")) {
            String loaderType = configParams.getStringValue("Server.FileLoader").trim();
            if (!loaderType.equals("Distributed") && !loaderType.equals("Bucketized"))
                this.oldFileLoader = true;
            else
                this.oldFileLoader = false;
        }
    }

    public boolean isUseCaching() {
        return useCaching;
    }

    public void setUseCaching(boolean useCaching) {
        this.useCaching = useCaching;
    }

    public boolean isRecordStats() {
        return recordStats;
    }

    public void setRecordStats(boolean recordStats) {
        this.recordStats = recordStats;
    }

    public boolean isWriteThroughCache() {
        return writeThroughCache;
    }

    public void setWriteThroughCache(boolean writeThroughCache) {
        this.writeThroughCache = writeThroughCache;
    }

    public int getMyId() {
        return myId;
    }

    public void setMyId(int myId) {
        this.myId = myId;
    }

    public String getHostlistFile() {
        return hostlistFile;
    }

    public void setHostlistFile(String hostlistFile) {
        this.hostlistFile = hostlistFile;
    }

    public LocalWorkpoolImpl getWorkpool() {
        return workpool;
    }

    public void setWorkpool(LocalWorkpoolImpl workpool) {
        this.workpool = workpool;
    }

    public StorageSystem getDataStorage() {
        return dataStorage;
    }

    public void setDataStorage(StorageSystem dataStorage) {
        this.dataStorage = dataStorage;
    }

    public String getLocation() {
        return location;
    }

    public void setLocation(String location) {
        this.location = location;
    }

    public int getNumNodes() {
        return numNodes;
    }

    public void setNumNodes(int numNodes) {
        this.numNodes = numNodes;
    }

    public boolean isUseNodeData() {
        return useNodeData;
    }

    public void setUseNodeData(boolean useNodeData) {
        this.useNodeData = useNodeData;
    }

    public boolean isMaster() {
        return master;
    }

    public void setMaster(boolean master) {
        this.master = master;
    }

    public int getNumThreads() {
        return numThreads;
    }

    public void setNumThreads(int numThreads) {
        this.numThreads = numThreads;
    }

    class StartServer extends Thread {
        HashtableServerImpl hs = null;

        public StartServer(HashtableServerImpl hs) {
            this.hs = hs;
        }

        @Override
        public void run() {

            Config configParams = Config.getInstance();
            int thriftPort = configParams.getIntValue("System.HashtableServer_THRIFT_PORT");

            HashtableServer_thrift.Processor processor;
            processor = new HashtableServer_thrift.Processor(hs);
            TServerTransport serverTransport = null;
            try {
                serverTransport = new TServerSocket(thriftPort);
            } catch (TTransportException e) {
                System.out.println("Could not start Thrift HashtableServer: " + e.getMessage());
                e.printStackTrace();
            }
            TServer server = new TThreadPoolServer(new TThreadPoolServer.Args(serverTransport).processor(processor).protocolFactory(new TCompactProtocol.Factory()));
            System.out.println("Starting Thrift HashtableServer on port " + thriftPort);
            server.serve();

        }

    }

    private static class Heartbeat extends Thread {
        FailureDetector failureDetectionServer;
        long period = 1000;
        String clusterName;

        Heartbeat(String clusterName, FailureDetector detector) {
            this.failureDetectionServer = detector;
            this.clusterName = clusterName;
        }

        Heartbeat(String clusterName, FailureDetector detector, long period) {
            this(clusterName, detector);
            this.period = period;
        }

        @Override
        public void run() {
            while (true) {
                try {
                    Thread.sleep(period);
                    failureDetectionServer.reportHeartbeat(clusterName, true, true);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
    }


    private static class Checkpointer extends Thread {

        private List<CheckpointListener> checkpointable;

        private long checkpointRate = 5000; // 5 seconds
        private long lastCheckpoint = 0;
        private boolean systemBooting = true;

        private File saveDir;
        private File restoreDir;
        private File bakDir;

        private LocalWorkpoolImpl workpool;

        Checkpointer(LocalWorkpoolImpl workpool, StorageSystem storage) {
            this.workpool = workpool;

            this.saveDir = new File(CheckpointPathBuilder.getRootSavePath());
            this.restoreDir = new File(CheckpointPathBuilder.getRootRestorePath());
            this.bakDir = new File(CheckpointPathBuilder.getRootBakPath());

            saveDir.mkdirs();
            restoreDir.mkdirs();
            bakDir.mkdirs();

            this.workpool = workpool;
            checkpointable = new ArrayList<>();
            checkpointable.add(workpool);
            checkpointable.add(storage);

            if (workpool.configParams.has("System.checkpointFrequency")) {
                checkpointRate = workpool.configParams.getIntValue("System.checkpointFrequency");
            }
        }

        public Checkpointer(LocalWorkpoolImpl workpool, StorageSystem storage, long checkpointRate) {
            this(workpool, storage);
            this.checkpointRate = checkpointRate;
        }

        private void backupCheckpoints() {
            try {
                Runtime rt = Runtime.getRuntime();

                String[] removeBakFiles = {"rm", "-r", bakDir.getAbsolutePath()};
                Process rmBakProcess = rt.exec(removeBakFiles);
                rmBakProcess.waitFor();

                String[] backupRestore = {"mv", restoreDir.getAbsolutePath(), bakDir.getAbsolutePath()};
                Process bakRestoreProcess = rt.exec(backupRestore);
                bakRestoreProcess.waitFor();

                String[] backupSave = new String[]{"cp", "-R", saveDir.getAbsolutePath(), restoreDir.getAbsolutePath()};
                Process bakSaveProcess = rt.exec(backupSave);
                bakSaveProcess.waitFor();

            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        @Override
        public void run() {

            try {
                // Wait for jobs to actually start before checkpointing, otherwise, initial tasks get lost

                while (workpool.status != 2 && systemBooting) {
                    Thread.sleep(1000);
                }
                systemBooting = false;
                lastCheckpoint = System.currentTimeMillis();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

            while (true) {
                try {
                    WorkStatus ws = workpool.getWorkStatus();
                    if (ws.status == StatusType.WORK_TERMINATED) {
                        break;
                    }

                    if (ws.status == StatusType.QUIESCENT) {
                        for (CheckpointListener c : checkpointable) {
                            c.checkpoint();
                        }
                        backupCheckpoints();
                        workpool.start();
                        System.out.println("System is active");
                    }

                    long currentTime = System.currentTimeMillis();
                    if (currentTime - lastCheckpoint > checkpointRate) {
                        System.out.println("Putting System in quiescent state");
                        lastCheckpoint = currentTime;
                        workpool.pause();
                    }
                } catch (Exception ignored) {

                }
            }
        }
    }
}
