/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package TestProgs.util;

import beehive.util.*;
import beehive.validation.ValidationService;
import beehive.workpool.*;

import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RMISocketFactory;
import java.util.*;

import java.io.*;
import java.net.*;

abstract public class BeehiveComputeLoader {

    String hostlistFile;                // The name of the file containing all of the hostnames in the cluster
    LocalWorkpoolImpl workpool;
    StorageSystem dataStorage;
    String location;                    // The hostname of this server
    int numNodes;
    boolean useNodeData = false;
    boolean master;                     // Whether or not this server is the master
    int numThreads = 0;
    int myId;
    boolean useCaching = false;
    boolean recordStats = false;
    boolean writeThroughCache = false;
    boolean canonicalHostNames = true;
    boolean oldFileLoader = true;

    String coordinator = null;          // The hostname of the master server

    /**
     * This is run on each Beehive server in the cluster.
     *
     * @param args             Arguments provided at runtime. Includes the following in the corresponding order:
     *                         <ol>
     *                         <li> A file containing the hostnames of all of the Beehive Servers in the given cluster </li>
     *                         <li> The configuration file </li>
     *                         <li> The hostname of the server running the {@link ValidationService} </li>
     *                         <li> The file containing the graph that is to be loaded </li>
     *                         <li> The number of nodes in the graph that is to be loaded </li>
     *                         <li> A file to which the results of the Application will be written to </li>
     *                         </ol>
     * @param classname        The name of the Application that is being run
     * @param nodeClassNameMap A mapping from a unique byte identifier to the name of each subclass of {@link Node},
     *                         including {@code Node} itself, that is to be used in the Application.
     * @param nodeClassMap     A mapping from a unique byte identifier to the Class object of each subclass of {@link Node},
     *                         including {@code Node} itself, that is to be used in the Application. It is not
     *                         required that the byte identifier be identical to that in {@code classname}.
     * @throws Exception Thrown if the application could not connect to the cluster and/or validator
     */
    public void init(String args[], String classname, HashMap<Byte, String> nodeClassNameMap, HashMap<Byte, Class> nodeClassMap) throws Exception {

		/*Initializing Config Parameters*/
        String configParamsFile = args[1];
        Config configParams = new Config(configParamsFile);
        readConfigParams(configParams);

        System.out.println("Number of workers:" + numThreads);

		/*Initialize host data*/
        this.hostlistFile = args[0];

        String myHostname;
        if (this.canonicalHostNames) {
            myHostname = InetAddress.getLocalHost().getCanonicalHostName();
        } else {
            myHostname = InetAddress.getLocalHost().getHostName();

        }

        //System.out.println("My hostname is: " + myHostname );

        Vector<String> hostList = createHostList(myHostname);
        if (master) {
            printToResultFile(args, numThreads, classname);
        }

        //this.location = "//" + myHostname + "/HashtableServer";
        //System.out.println("My URL is: " + location + " my hostlist size is " + hostList.size()  );

        //global socket factory configuration. Effective for all RMI connections
        RMISocketFactory.setSocketFactory(new RMISocketFactory() {
            public Socket createSocket(String host, int port)
                    throws IOException {
                int timeout = 30000;
                Socket socket = new Socket();
                socket.setSoTimeout(timeout);
                socket.setSoLinger(false, 0);
                socket.connect(new InetSocketAddress(host, port), timeout);
                return socket;
            }

            public ServerSocket createServerSocket(int port)
                    throws IOException {
                return new ServerSocket(port);
            }
        });


        //Initialize RMI registry
        Registry localRegistry = null;
        try {
            localRegistry = LocateRegistry.createRegistry(1099);
        } catch (RemoteException e) {

            System.out.println("Couldn't Create Registry");
            e.printStackTrace();
        }
        System.setSecurityManager(new RMISecurityManager());


        // start thrift server
        String RPC_MECHANISM = "RMI";
        if (configParams.has("System.RPC_MECHANISM")) {
            RPC_MECHANISM = configParams.getStringValue("System.RPC_MECHANISM");
        }

//        if (RPC_MECHANISM.equals("THRIFT")) {
//            StartServer thriftServer = new StartServer(HashtableNode);
//            thriftServer.start();
//            Thread.sleep(10000);
//        }


/** we can pass null for HashtableNode object - it ref to local ahstable storage  */

        //storage system does not need to know about the useCaching flag since we have separate APIs
        StorageSystem dataStorage;
        if (useCaching) {
            dataStorage = new StorageSystem(hostList, writeThroughCache, configParams, nodeClassNameMap, nodeClassMap);
        } else {
            dataStorage = new StorageSystem(hostList, configParams, nodeClassNameMap, nodeClassMap);
        }
        if (recordStats) dataStorage.enableStats();


        this.numNodes = Integer.parseInt(args[4]);

        //if graph partition hashing scheme is used, creating the partition table
        if (configParams.has("Server.FileLoader") && configParams.has("Server.NODE_HASHING_SCHEME")) {
            String loaderType = configParams.getStringValue("Server.FileLoader").trim();
            String hashingScheme = configParams.getStringValue("Server.NODE_HASHING_SCHEME").trim();
            if (loaderType.equals("Bucketized") && hashingScheme.equals("GRAPH_PARTITION")) {
                List<Integer> hostArrList = new ArrayList<Integer>();

                try {
                    String input_filename = args[3] + "/" + "metadata";
                    BufferedReader input = new BufferedReader(new FileReader(input_filename));
                    String line = "";
                    int ctr = 0;

                    while ((line = input.readLine()) != null) {
                        if (line.charAt(0) == '#' || line.trim().equals("") || line.length() < 5)
                            continue;

                        StringTokenizer lineTok = new StringTokenizer(line, " ");
                        String hostKey = lineTok.nextToken();
                        String fromNode = lineTok.nextToken();
                        String toNode = lineTok.nextToken();
                        int hostid = Integer.parseInt(hostKey);
                        int lastNodeStored = Integer.parseInt(toNode);
                        hostArrList.add(hostid, lastNodeStored);
                    }
                    input.close();
                } catch (Exception e) {
                    e.printStackTrace();
                    System.err.println("ERROR :: Failed to read metadata file");
                }

                int hostArr[] = new int[hostArrList.size()];
                for (int i = 0; i < hostArrList.size(); i++) {
                    hostArr[i] = hostArrList.get(i);
                }
                dataStorage.partitionedGraphHostNodesTable = hostArr;
            }
        }


        /*  args[2] is the hostname for the validator   */
        LocalWorkpoolImpl workpool = new LocalWorkpoolImpl(args[2], numThreads, dataStorage.partitionedGraphHostNodesTable);
        localRegistry.bind("WorkpoolServer", workpool);
        Thread.sleep(10000);
        /*  no need to to lookup peers or pass it a vector containing  */
        /* otherwise pass a vercto contain gjust myHostname  -- call it clientHost  */
        workpool.lookupPeers(hostList);

        this.workpool = workpool;
        this.dataStorage = dataStorage;

    }


    /**
     * Constructs a list of the hostnames of its peers in the cluster.
     *
     * @param myHostname This server's hostname so that it can be excluded from the list of peer hostnames
     * @return The hostnames of its peers in the cluster.
     */
    public Vector<String> createHostList(String myHostname) {
        this.master = false;
        boolean first = true;
        Vector<String> hostList = new Vector<String>();
        this.myId = 0;
        try {
            BufferedReader bufferedReader = new BufferedReader(new FileReader(hostlistFile));

            String host;
            int hostCount = 0;
            while ((host = bufferedReader.readLine()) != null) {
                host = host.trim();
                if (host.length() == 0) continue;
                if (first) {
                    coordinator = new String(host);

                    if (host.equals(myHostname)) {
                        this.master = true;
                        System.out.println("Master machine");
                    }
                    first = false;
                }
                hostList.add(host);
                if (host.equals(myHostname)) {
                    this.myId = hostCount;
                }
                hostCount++;
            }
            bufferedReader.close();
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Error reading config file");
            System.exit(1);
        }
        return hostList;
    }

    public void printToResultFile(String args[], int numThreads, String classname) throws Exception {
        int count = 0;

        FileWriter fw = new FileWriter("result.txt", true);
        BufferedWriter out = new BufferedWriter(fw);

        BufferedReader br = new BufferedReader(new FileReader(args[0]));
        String line = null;
        out.write("============================================================================\n");
        out.write("New Run: " + new Date().toString() + "\n");

        out.write(classname + " : " + " " + Arrays.toString(args) + "\n");
        out.write("Number of Machines used");
        while ((line = br.readLine()) != null) {
            count++;
        }
        out.write(" " + Integer.toString(count) + "\n");

        out.write("Number of threads:" + Integer.toString(numThreads) + "\n");

        out.write("Experiment settings description: \n");
        br = new BufferedReader(new FileReader(args[5]));
        line = null;
        while ((line = br.readLine()) != null)

        {
            out.write(line + "\n");
        }
        out.close();
        br.close();
    }

    public void readConfigParams(Config configParams) {
        this.useCaching = configParams.getBoolValue("Server.useCaching");
        this.recordStats = configParams.getBoolValue("StorageSystem.recordStats");
        this.useNodeData = configParams.getBoolValue("StorageSystem.useNodeData");
        this.writeThroughCache = configParams.getBoolValue("StorageSystem.WriteThroughCache");
        this.numThreads = configParams.getIntValue("Worker.WORKER_COUNT");
        this.canonicalHostNames = configParams.getBoolValue("Server.canonicalHostNames");
        if (configParams.has("Server.FileLoader")) {
            String loaderType = configParams.getStringValue("Server.FileLoader").trim();
            if (!loaderType.equals("Distributed") && !loaderType.equals("Bucketized"))
                this.oldFileLoader = true;
            else
                this.oldFileLoader = false;
        }
    }

    public boolean isUseCaching() {
        return useCaching;
    }

    public void setUseCaching(boolean useCaching) {
        this.useCaching = useCaching;
    }

    public boolean isRecordStats() {
        return recordStats;
    }

    public void setRecordStats(boolean recordStats) {
        this.recordStats = recordStats;
    }

    public boolean isWriteThroughCache() {
        return writeThroughCache;
    }

    public void setWriteThroughCache(boolean writeThroughCache) {
        this.writeThroughCache = writeThroughCache;
    }

    public int getMyId() {
        return myId;
    }

    public void setMyId(int myId) {
        this.myId = myId;
    }

    public String getHostlistFile() {
        return hostlistFile;
    }

    public void setHostlistFile(String hostlistFile) {
        this.hostlistFile = hostlistFile;
    }

    public LocalWorkpoolImpl getWorkpool() {
        return workpool;
    }

    public void setWorkpool(LocalWorkpoolImpl workpool) {
        this.workpool = workpool;
    }

    public StorageSystem getDataStorage() {
        return dataStorage;
    }

    public void setDataStorage(StorageSystem dataStorage) {
        this.dataStorage = dataStorage;
    }

    public String getLocation() {
        return location;
    }

    public void setLocation(String location) {
        this.location = location;
    }

    public int getNumNodes() {
        return numNodes;
    }

    public void setNumNodes(int numNodes) {
        this.numNodes = numNodes;
    }

    public boolean isUseNodeData() {
        return useNodeData;
    }

    public void setUseNodeData(boolean useNodeData) {
        this.useNodeData = useNodeData;
    }

    public boolean isMaster() {
        return master;
    }

    public void setMaster(boolean master) {
        this.master = master;
    }

    public int getNumThreads() {
        return numThreads;
    }

}
