/*
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 * % Beehive Parallel Programming Framework
 * % © Regents of the University of Minnesota.
 * % This software is licensed under GPL v3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html) open source license.
 * %     For an alternative license please contact the Office for Technology Commercialization at the University of Minnesota.
 * %
 * % How to cite:
 * %     If you use this software for your work, please use the following citation:
 * %   Anand Tripathi, Vinit Padhye, Tara Sasank Sunkara, Jeremy Tucker, Bhagavathi Dhass Thirunavukarasu, Varun Pandey, Rahul R. Sharma,
 * %   "A Transactional Model for Parallel Programming of Graph Applications on Computing Clusters,"
 * %   2017 IEEE 10th International Conference on Cloud Computing (CLOUD), Honolulu, HI, USA, June 25-30, 2017.
 * %
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 */

package beehive.server;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.*;

import beehive.util.*;

import java.util.concurrent.ConcurrentHashMap;

public interface HashtableServer extends Remote {

    Date ping() throws RemoteException;
	
	/*get/put/update + bulk*/

	ResultObject put(String key, Object value) throws RemoteException;

	//put for write through cache
    ResultObject put(String key, Object value, int callingServerId) throws RemoteException;

	ResultObject get(String key, long storageSystemThreadId) throws RemoteException;

	Vector<ResultObject> getObjects(Vector<String> keyList) throws RemoteException;

	Vector<ResultObject> putObjects(ConcurrentHashMap<String, Node> objList) throws RemoteException;

	//Interface methods to test the reflection apis

	ResultObject getNodeProperties(NodeData n) throws RemoteException;

	Vector<ResultObject> getObjectsProperties(Vector<NodeData> keyList) throws RemoteException;

	ResultObject updateNodeProperties(NodeData n) throws RemoteException;

	Vector<ResultObject> updateNodePropertiesParallel(Vector<NodeData> objList) throws RemoteException;
	
	ResultObject exec(NodeData nd) throws RemoteException;
	
	Vector<ResultObject> execBulk(Vector<NodeData> keyList) throws RemoteException;

	//putObjects for write through cache
    Vector<ResultObject> putObjects(ConcurrentHashMap<String, Node> objList, int callerServerId) throws RemoteException;
	
	/*helper functions*/

	boolean contains(String key) throws RemoteException;

	Vector<Node> getAllNodes() throws RemoteException;

	Vector<String> getAllKeys() throws RemoteException;

	String getStats() throws RemoteException;
	
	void clearStatContainers() throws RemoteException;

	Node getNodeFromTable(String nodeId) throws RemoteException;
	
	void loadNodes(Vector<Node> nodes) throws RemoteException;
	
	int getServerId() throws RemoteException;

	int getNumServers() throws RemoteException;
	
	void setNumServers(int n) throws RemoteException;
	
	/*relocation related*/
	
	ResultObject relocate(int nodeId, String serverName) throws RemoteException;

	void locationChange(int nodeId, String serverName) throws RemoteException;

	ResultObject lock(int nodeId) throws RemoteException;

	void removeLocationTableEntry(String key) throws RemoteException;

	/* caching related*/
    void cacheInvalidate(String nodeID) throws RemoteException;

	Node getNodeForCaching(String nodeID, int cachingSite) throws RemoteException;

	ResultObject cachePut(String nodeID, Node node, int masterCopySite, long storageSystemThreadId) throws RemoteException;

	Vector<ResultObject> cachePut(Map<Integer, Vector<Node>> nodes, long storageSystemThreadId) throws RemoteException;

	Vector<ResultObject> getLocalNodes(Vector<String> nodeIds, long storageSystemThreadId) throws RemoteException;

	Vector<Node> getNodesForCaching(Vector<String> nodeIds, int callingSite) throws RemoteException;

	void writeThroughToCache(String paramString, Object value) throws RemoteException;

	void writeThroughToCacheBulk(ConcurrentHashMap<String, Node> map) throws RemoteException;

	Node getNodeFromCacheTable(String nodeId) throws RemoteException;

	Vector<Node> getCachedObjects(Vector<String> nodeIds) throws RemoteException;

	void cacheEvictMasterNotify(String cachingServerId, Hashtable<String, Long> evictedNodes) throws RemoteException;


    int acquireLocks(SortedSet<LockRequest> lockRequests) throws RemoteException;

    void releaseLock(SortedSet<LockRequest> lockRequests) throws RemoteException;

	void commitChanges(Set<Node> write, Set<NodeData> writeNd, Map<String, Long> read) throws RemoteException;

    void checkpoint() throws RemoteException;
}
